/**************************************************************************
 *  enumerator - a program to enumerate large implicit search spaces      *
 *  Copyright (C) 2007-2008  Eric Robinson                                *
 *                                                                        *
 *  This program is free software: you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation, either version 3 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU General Public License     *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 **************************************************************************/

#ifndef PATHINTERFACE_H_
#define PATHINTERFACE_H_

#include <stdlib.h>

#include "tivstand.h"
#include "tivmath.h"
#include "tivcheck.h"

typedef BYTE * SEARCH_PATH;
typedef BYTE   PATH_STEP;
UINT32 PATH_WORD = sizeof( UINT32 );

class PathInterface
{
	UINT32 MAX_BRANCH, STEP_MAX, BLOCK_SIZE;
	BYTE   *shifts;
	UINT8 STEP_MASK, STEPS_PER_BYTE, STEPS_PER_WORD;

public:
	void Initialize( UINT32 MaxBranch )
	{
		MAX_BRANCH = MaxBranch;
		
		if(      MAX_BRANCH <= 2   ) STEP_MAX = 2;
		else if( MAX_BRANCH <= 4   ) STEP_MAX = 4;
		else if( MAX_BRANCH <= 16  ) STEP_MAX = 16;
		else if( MAX_BRANCH <= 256 ) STEP_MAX = 256;
		else assert( 0 );  // Currently doesn't work with NUM_GENS > 256
		
		STEPS_PER_BYTE = BYTE_LENGTH / log( 2, STEP_MAX );
		STEPS_PER_WORD = PATH_WORD * STEPS_PER_BYTE;
		STEP_MASK = STEP_MAX - 1;
		
		malloc_check( shifts, BYTE *, STEPS_PER_BYTE );
		for( UINT32 i = 0; i < STEPS_PER_BYTE; i++ )
			shifts[i] = BYTE_LENGTH - ( i + 1 ) * BYTE_LENGTH / STEPS_PER_BYTE;
	}
	
   // Get a path value
   PATH_STEP GET_STEP( SEARCH_PATH path, UINT32 step )
   {
      assert( step < *( ( UINT32 * ) path ) );
      SEARCH_PATH addr = path + step / STEPS_PER_BYTE + sizeof( UINT32 );
      BYTE shift = shifts[step % STEPS_PER_BYTE ];
      return ( *addr & ( STEP_MASK << shift ) ) >> shift;
   }
   
   void SET_STEP( SEARCH_PATH path, UINT32 step, PATH_STEP val )
   {
      assert( step < *( ( UINT32 * ) path ) );
      assert( val < STEP_MAX );
      SEARCH_PATH addr = path + step / STEPS_PER_BYTE + sizeof( UINT32 );
      BYTE shift = shifts[step % STEPS_PER_BYTE ];
      *addr = ( *addr & ~( STEP_MASK << shift ) ) | ( val << shift );
   }

   UINT32 PATH_SIZE( UINT32 maxSteps )
   {
      return ( maxSteps / STEPS_PER_WORD + ( maxSteps % STEPS_PER_WORD ? 2 : 1 ) ) * PATH_WORD;
   }
   
   UINT32 GET_STEPS( SEARCH_PATH path )
   {
      return *( ( UINT32 * ) path );
   }
   
   UINT32 GET_SIZE( SEARCH_PATH path )
   {
      return PATH_SIZE( GET_STEPS( path ) );
   }
   
   void SET_STEPS( SEARCH_PATH path, UINT32 steps )
   {
      *( ( UINT32 * ) path ) = steps;
   }
   
   INT32 pathComp( SEARCH_PATH left, SEARCH_PATH right )
   {
      UINT32 minSteps = MIN( GET_STEPS( left ), GET_STEPS( right ) ),
             maxBlock = minSteps / STEPS_PER_BYTE;
      
      // Check up to the full block size for the shortest
      INT32 val = memcmp( left + SIZE32, right + SIZE32, maxBlock );
      if( val ) return val;
      
      // Check the remaining bytes
      for( UINT32 count = 0; count < minSteps % STEPS_PER_BYTE; count++ )
      {
         val = GET_STEP( left, minSteps - minSteps % STEPS_PER_BYTE + count ) - 
               GET_STEP( right, minSteps - minSteps % STEPS_PER_BYTE + count );
         if( val ) return val;
      }
      
      // Find the longest and set it greater than the other
      if( GET_STEPS( left ) > GET_STEPS( right ) ) return 1;
      else if( GET_STEPS( right ) > GET_STEPS( left ) ) return -1;
      else return 0;
   }
      
};

#endif /*PATHINTERFACE_H_*/
