/**************************************************************************
 *  enumerator - a program to enumerate large implicit search spaces      *
 *  Copyright (C) 2007-2008  Eric Robinson                                *
 *                                                                        *
 *  This program is free software: you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation, either version 3 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU General Public License     *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 **************************************************************************/

/*************************************************************************
 * Usage:  Create a new class similar to the NeuClusterMPI model.        *
 *   The arguments to Initialize are:        				 *
 *   void Initialize( UINT64         MemorySize,        		 *
 *                    UINT64         DiskSize,        			 *
 *                    const CHAR *   CommonDirectory,        		 *
 *                    const CHAR *   TempDirectory,        		 *
 *                    const CHAR *   FinalDirectory,        		 *
 *                    UINT32         NumberOfMachines )        		 *
 *  All sizes are in units of bytes.        				 *
 *  The actual NumberOfMachines will be determined by MPI.  The argument *
 *  here should be the correct number (preferred) or an underestimate.   *
 *  The code assumes a common directory for the input files, which       *
 *  may be shared (e.g. NFS), or replicated local files at the same path.*
 *  The local run results will also be stored in this directory under a  *
 *  unique filename.							 *
 ************************************************************************/

#include <stdio.h>
#include <semaphore.h>

#include "ArchitectureInterface.h"

class MPIArchitecture :public ArchitectureInterface
{
public:
   // Semaphores used
   sem_t mpiLock;
   int flag;
   int argc;
   char **argv;
   
   // Sends a message to a machine, will inform if it fails and when it succeeds
   void sendMessage( void *mess, INT32 lenAct, UINT32 toAct, UINT32 tagAct )
   {
      sem_wait( &mpiLock );
      if( MPI_Send( mess, lenAct, MPI_UNSIGNED_CHAR, toAct, tagAct, MPI_COMM_WORLD ) !=
          MPI_SUCCESS )
      {
         printf( "Failed to send message %d to %d\n", tagAct, toAct );
         assert( 0 ); // Error Sending
      }
      sem_post( &mpiLock );
   }
   
   void probeMessage( INT32 fromReq, INT32 tagReq, 
                      INT32 &lenAct, UINT32 &fromAct, UINT32 &tagAct )
   {
      MPI_Status status;
      for( int mess = 0; !mess; )
      {
         sem_wait( &mpiLock );
         if( MPI_Iprobe( fromReq, tagReq, MPI_COMM_WORLD, &mess, &status )!=MPI_SUCCESS )
         {
            printf( "Error probing for message %d from %d\n", tagReq, fromReq );
            assert( 0 ); // Error Probing
         }
         sem_post( &mpiLock );
         usleep( 1000 );
      }
      MPI_Get_count( &status,  MPI_UNSIGNED_CHAR, &lenAct );
      fromAct = status.MPI_SOURCE;
      tagAct = status.MPI_TAG;
   }
   
   void recvMessage ( void *mess, INT32 lenReq, INT32 fromReq, INT32 tagReq,
                                  INT32 &lenAct, UINT32 &fromAct, UINT32 &tagAct )
   {
      MPI_Status status;
      sem_wait( &mpiLock );
      if( MPI_Recv( mess, lenReq, MPI_UNSIGNED_CHAR, fromReq, tagReq, MPI_COMM_WORLD, &status ) != 
          MPI_SUCCESS )
      {
         printf( "Error receiving message %d from %d\n", tagReq, fromReq );
         assert( 0 );  // Error Receiving
      }
      sem_post( &mpiLock );
      MPI_Get_count( &status,  MPI_UNSIGNED_CHAR, &lenAct );
      fromAct = status.MPI_SOURCE;
      tagAct = status.MPI_TAG;
   }
   
   void initializeComm( UINT32 &myID, UINT32 &numComputers )
   {
      int id, comp;
      
      TAG_ANY = MPI_ANY_TAG;
      SOURCE_ANY = MPI_ANY_SOURCE;
      sem_init( &(mpiLock), 0, 1 );
      MPI_Initialized( &flag );
      if( !flag && MPI_Init(&argc, &argv)!=MPI_SUCCESS ) assert( 0 );
      MPI_Comm_size(MPI_COMM_WORLD, &comp); numComputers = comp;
      MPI_Comm_rank(MPI_COMM_WORLD, &id); myID = id;
   }
   
   void setArgs( int argc, char **argv )
   {
      this->argc = argc;
      this->argv = argv;
   }
   
   void finalizeComm( )
   {
      if( !flag ) MPI_Finalize( );
   }
};

class TeraClusterMPI :public  MPIArchitecture
{
public:
	TeraClusterMPI( int argc, char **argv )
	{
      setArgs( argc, argv );
		Initialize(	1000000000ll,
						200000000000ll,
						"/home/erobinso/TreeBuilder/",
						"/state/partition1/home/erobinso/TBS/tmp/",
						"/state/partition1/home/erobinso/TBS/final/",
						32 );
	}
};

class LocalhostMPI :public  MPIArchitecture
{
public:
	LocalhostMPI( int argc, char **argv )
	{
      setArgs( argc, argv );
		Initialize(	500000000ll,
						5000000000ll,
						"/tmp/enumerator-0.1.5/",
						"/tmp/enumerator-0.1.5/tmp/",
						"/tmp/enumerator-0.1.5/final/",
						3 );
	}
};

class NeuClusterMPI :public  MPIArchitecture
{
public:
	NeuClusterMPI( int argc, char **argv )
	{
      setArgs( argc, argv );
		Initialize(	500000000ll,
						5000000000ll,
						"/home/tivadar/TreeBuilder/",
						"/tmp/TreeBuilder/tmp/",
						"/tmp/TreeBuilder/final/",
						64 );
	}
};

class OpportunityClusterMPI : public MPIArchitecture
{
public:
   OpportunityClusterMPI( int argc, char **argv )
   {
      setArgs( argc, argv );
      Initialize( 1000000000ll,
                  10000000000ll,
                  "/home/erobinso/TreeBuilder/",
                  "/scratch_local/TreeBuilder/tmp/",
                  "/scratch_local/TreeBuilder/final/",
                  16 );
   }
};

