/**************************************************************************
 *  enumerator - a program to enumerate large implicit search spaces      *
 *  Copyright (C) 2007-2008  Eric Robinson                                *
 *                                                                        *
 *  This program is free software: you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation, either version 3 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU General Public License     *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 **************************************************************************/

#include "ProblemInterface.h"

#include <stdio.h>

#include "tivbits.h"

class VectorMatrixMultiplyGreasing: public ProblemInterface
{
private:
   PathInterface *pathInt;
	typedef BYTE BandType;
	typedef UINT64 ChunkType;
	UINT32 BandWidth;
	UINT32 ConfigWidth;
	UINT32 BandMax;
	UINT32 Dimension;
	const CHAR *mname1;
	const CHAR *mname2;
	const CHAR *vname;
	UINT8 avoid[2];
	BYTE *matrices[2];
	BYTE *root;
	
	CHAR getNext( FILE *is )
	{
		CHAR input;
		for( 	input = getc( is );
				input!=']' && input!='0' && input!='1' && input!='['; 
				input = getc( is ) );
		return input;
	}
	
	void getRoots( BYTE *roots )
	{
		memcpy( roots, root, StateSize );
	}
	
	void multiply( const BYTE *target, BYTE *destination, unsigned char id )
	{
	   UINT32 block;
	   ChunkType *source;

	   memset( destination, '\0', StateSize );
	   
	   for( block = 0; block < BandMax; block++ )
	   {
	      // get the right spot
	      source = ( ChunkType * )	( matrices[id] + block * ConfigWidth * StateSize + StateSize * 
	    		  					( ( BandType * )target )[block] );
	      for( UINT32 location = 0; location < StateSize / sizeof( ChunkType ); location++ )
	    	  *( (ChunkType *) destination + location ) ^= *( source + location );
	   }
	}
	
	void convert_matrix( char *filename, char *outputname )
	{
	   UINT32 bytes = 0;
	   
	   // Get the input and output files ready
	   FILE *is = fopen( filename, "r" );
	   FILE *os = fopen( outputname, "w" );
	   char current_out; char current_in;
	   unsigned long row, column, count;
	 
	   // Make sure there is some beginning to the matrix
	   if( getNext( is )!='[' ) assert( 0 );
	  
	   // Process the row
	   for( row = 0, current_out = 0; row < Dimension; row++, current_out = 0 )
	   {
	      // Make sure we're at the start of the row
	      if( getNext( is )!='[' ) assert( 0 );
	     
	      // Read each column
	      for( column = 0; column < Dimension; column++ )
	      {
	         // Get the current character and make sure it's valid
	         current_in = getNext( is );
	         if( current_in!='0' && current_in!='1' ) assert( 0 );
	       
	         // Set the current output
	         current_out |= ( current_in - '0' ) << ( BYTE_LENGTH  - ( column % BYTE_LENGTH ) - 1 );
	       
	         // Write the block if we've finished it
	         if( column % BYTE_LENGTH == BYTE_LENGTH - 1 )
	         {
	            if( ( current_in = fputc( current_out, os ) ) != current_out )
	               assert( 0 );
	            bytes++;
	            current_out = 0;
	         }
	      }
	      
	      // Makes sure we're at the end of the row
	      if( getNext( is )!=']' ) assert( 0 );
	     
	      // Write the last block we were on, and keep track of the column
	      if( column % BYTE_LENGTH != 0 )
	      {
	         if( ( current_in = fputc( current_out, os ) ) != current_out )
	            assert( 0 );
	         bytes++;
	         column = ( column / BYTE_LENGTH + 1 ) * BYTE_LENGTH;
	      }

	      // Round it so it hits SIG_SIZE
	      for(; StateSize > column / BYTE_LENGTH; column += BYTE_LENGTH )
	      {
	         bytes++;
	         if( ( current_in = fputc( '\0', os ) ) != '\0' )
	            assert( 0 );
	      }
	   }
	   
	   // Make sure we're at the end of the matrix
	   if( getNext( is )!=']' ) assert( 0 );
	  
	   // Add extra rows to get a perfectly square matrix
	   for(; StateSize > row / BYTE_LENGTH; row++ )
	      for( count = 0; count < StateSize; count++ )
	      {
	         bytes++;
	         if( ( current_in = fputc( '\0', os ) ) != '\0' )
	            assert( 0 );
	      }
	      
	   printf( "The final number of bytes is %d\n", bytes );

	   // Close the input and output files
	   fclose( is );
	   fclose( os );
	}

	void load_matrix( char *filename, unsigned long num )
	{
		BYTE *current;
		FILE *fp;
	 
		malloc_check( matrices[num], BYTE *, ConfigWidth * StateSize * BandMax );
		malloc_check( current, BYTE *, BYTE_LENGTH * StateSize * StateSize );
	   
		fopen_check( fp, filename, "rb" );
		fread_check( current, BYTE_LENGTH * StateSize * StateSize, fp );
		fclose( fp );   
	  
		memset( matrices[num], '\0', ConfigWidth * StateSize * BandMax );
		for( unsigned long x = 0; x < BandMax; x++ )
			for( unsigned long y = 0; y < ConfigWidth; y++ )
				sum_band( num, x, y, current + x * StateSize * BandWidth );
	         
		free( current );
	}

	void sum_band( UINT32 num, UINT32 band, UINT32 config, BYTE *row_start )
	{
		BYTE *destination = matrices[num] + band * ConfigWidth * StateSize + config * StateSize;
	  
		// Compute over all sets of columns
		for( UINT32 pos = 0; pos < StateSize; pos++ )
		{
			// Sum all the rows
			for( UINT32 row = 0; row < BandWidth; row++ )
			{
				if( ( config & ( 1 << ( BandWidth - row - 1 ) ) ) )
					*( destination + pos ) ^= *( row_start + StateSize * row + pos );
			}
		}
	}
	
   void loadFirst( const CHAR *filename, BYTE * location )
   {
      FILE *file;
      INT8 input;
      
      fopen_check( file, filename, "r" );
      memset( location, '\0', StateSize );
      assert( getNext( file )=='[' );
      for( UINT32 count = 0; count < Dimension; count++ )
      {
         if( ( input = getNext( file ) )=='1' ) BV_SET( location, count );
         else if( input!='0' ) assert( 0 );
      }
      if( getNext( file )!=']' ) assert( 0 );
      fclose( file );
   }
	
	
public:
	VectorMatrixMultiplyGreasing( const CHAR *group, ArchitectureInterface *arch )
	{
		UINT64 SearchSize;
	
		if( !strcmp( group, "Baby Monster" ) )
		{
			SearchSize = 14000000000ll;
			Dimension = 4370;
			mname1 = "m1.bab.bin"; avoid[0] = 2;
			mname2 = "m2.bab.bin"; avoid[1] = 3;
			vname = "mx.bab.txt";		
		}
		else if( !strcmp( group, "Fischer 23" ) ) 
		{
			SearchSize = 12000000000ll;
			Dimension = 728;
			mname1 = "m1.fi23.bin"; avoid[0] = 2;
			mname2 = "m2.fi23.bin"; avoid[1] = 3;
			vname = "mx.fi23.txt";			
		}
		else if( !strcmp( group, "Janko 4" ) )
		{
			SearchSize = 173067389ll;
			Dimension = 112;
			mname1 = "m1.j4.bin"; avoid[0] = 2;
			mname2 = "m2.j4.bin"; avoid[1] = 4;
			vname = "mx.j4.txt";			
		}
		else if( !strcmp( group, "McLaughlin" ) )
		{
			SearchSize = 150000000ll;
			Dimension = 22;
			mname1 = "m1.mcl.bin"; avoid[0] = 0;
			mname2 = "m2.mcl.bin"; avoid[1] = 0;
			vname = "mx.mcl.txt";			
		}
		else if( !strcmp( group, "Harada Norton" ) ) 
		{
			SearchSize = 1140000ll;
			Dimension = 760;
			mname1 = "m1.hn.bin"; avoid[0] = 2;
			mname2 = "m2.hn.bin"; avoid[1] = 3;
			vname = "mx.hn.txt";			
		}
		else
		{
			assert( 0 );	// Not a valid group
		}
		
		UINT32 CHUNK_BYTES = sizeof( ChunkType );
		UINT32 CHUNK_BITS = CHUNK_BYTES * BYTE_LENGTH;
		UINT32 StateSize = ( Dimension / CHUNK_BITS + ( Dimension % CHUNK_BITS ? 1 : 0 ) ) * CHUNK_BYTES;
		
		// Set the band width and configuration width
		BandWidth = BYTE_LENGTH * sizeof( BandType );
		ConfigWidth = power( 2, BandWidth );
		BandMax = ( BYTE_LENGTH * StateSize / BandWidth );

		// Assert they're valid given the State Size
		assert( BandWidth % 8 == 0 && StateSize % ( BandWidth / 8 ) == 0 );

		( ( ProblemInterface * ) this )->Initialize( SearchSize, StateSize, 2, 2 );
		this->Initialize( arch->CommonDirectory );
	}

	void convertMatrices( char *directory )
	{
	   CHAR outputname[1000];
	   CHAR filename[1000];

      sprintf( filename, "%sinput/%s.txt", directory, mname1 );
      sprintf( outputname, "%sinput/%s", directory, mname1 );
      convert_matrix( filename, outputname );
      sprintf( filename, "%sinput/%s.txt", directory, mname2 );
      sprintf( outputname, "%sinput/%s", directory, mname2 );
      convert_matrix( filename, outputname );
	}

	void loadMatrices( char *directory )
	{
	   CHAR outputname[1000];
	   CHAR filename[1000];
	   
	   // Load in and compute the matrices
	   sprintf( filename, "%sinput/%s", directory, mname1 );
	   load_matrix( filename, 0 );
	   sprintf( filename, "%sinput/%s", directory, mname2 );
	   load_matrix( filename, 1 );
	   
	   sprintf( filename, "%sinput/%s", directory, vname );
	   malloc_check( root, BYTE *, StateSize );
	   loadFirst( filename, root );
	}
	
	void Initialize( char *directory )
	{
	   CHAR outputname[1000];
	   CHAR filename[1000];
	   
	   pathInt = new PathInterface;
	   pathInt->Initialize( 2 );
	}
	
   void applyTargetGen(   const BYTE *state, 
                          const UINT32 genLabel, 
                          BYTE *child, 
                          BOOL *success,
                          const SEARCH_PATH path = NULL,
                          const UINT32 pathSize = 0 )
   {
      if( path && pathInt->GET_STEPS( path ) >= avoid[genLabel] - 1 )
      {
         *success = false;
         for( UINT32 count = 0; count < avoid[genLabel] - 1 && !*success; count++ )
         {
            if( pathInt->GET_STEP( path, pathInt->GET_STEPS( path ) - 1 - count ) != genLabel )
               *success = true;
         }
      }
      else *success = true;
      if( *success ) multiply( state, child, genLabel );
   }
};
