/**************************************************************************
 *  enumerator - a program to enumerate large implicit search spaces      *
 *  Copyright (C) 2007-2008  Eric Robinson                                *
 *                                                                        *
 *  This program is free software: you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation, either version 3 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU General Public License     *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 **************************************************************************/

#ifndef SEARCH_H_
#define SEARCH_H_

// Treebuilder
/* Version 1.0.0 has finished checks base on head block mods on an individual
 *    machine.  It also stores the final output as 64 bit compressed signature
 *    files as well as a corresponding file with the bitstring to that
 *    compressed signature. */
/* Version 1.0.1 addressed the issues with piping output to a file. */
/* Version 1.0.2 reduced the block size and printed when a node was done
*     with processing its queue. */
/* Version 1.1.0 fixed a couple of memory leak bugs, as well as introducing
 *    the storage on the checker end and the storing of the hash table
 *    upon completion of the algorithm. */
/* Version 1.1.1 fixed my laziness.  Instead of having to edit all the shell
 *    scripts and procgroups each time, I can just edit a central computers.txt
 *    file and have it generate them for me!  Also fixed a bug with the tail
 *    block not being initialized to 0 */
/* Version 1.1.2 adjusted the scripts to fix the local/common dir problem on
 *    architectures where there is no common directory */
/* Version 1.1.4 fixed the bug within tivbits which calculated the maximum
 *    number of bits */
/* Version 1.2.0 fixed the tivbits bug as well as reducing the block size.
 *    This version was tested and discovered approximately 1B points
 *    overnight.  */
/* Version 1.2.1 changed the asserts to perrors for file opens */
/* Version 1.2.2 added a longer sleep time back in */
/* Version 1.3.0 fixed the issue with multithreaded MPI calls, it also added
 *     locks around much of the data, even if just used for a simple read
 *     of a word when another process may be writing that word */
/* Version 1.4.0 added synchronization around MPI_Init using sockets */
/* Version 1.4.1 removed synchronization and attempted to generalize
 *     tests on a variety of groups. */
/* Version 1.4.2 reduced the block size to make the computation memory
 *     manageable. */
/* Version 1.4.4 printed many messages to avoid a lockup with sends and
 *      receives. */
/* Version 1.5.0 generalized the compute block routine so that it could
 *     be used with other groups, it added special functionality to test
 *     for duplicates in the first block. */
/* Version 1.5.1 reorganized the SET_BIT routine so that it would set bits
 *     in the same order that it received them */
/* Version 1.5.2 made the block size much smaller and reduced the length of the
 *     string max in the first block */
/* Version 1.5.6 added better error logging */
/* Version 1.6 combined 1.5.0 and 1.5.6 to remove the bugs, also added
 *     error checking with freads and writes */
/* Version 1.8 fixed a bug with the store blocks, resulted in a full solution
 *     to the Harada Norton group. */
/* Version 1.9 uses the new MPI, removing the receive manager and special logic 
 *     for processing messages to self.  It generalizes the child calls and 
 *     separates them into two files, a config and a function */
/* Version 2.0 can do compressed/uncompressed sigs, uses Isend where 
 *     appropriate, automatically computes the Compressed sig size, deletes
 *     maybe queue files once finished with them, and handles the base 3 
 *     multiplication. */
/* Version 2.1 checks and stores the first block, keeping the shortest paths. */
/* Version 3 removed the pipelining for the BFS, it also allows for branching 
 *     factors up to 256. */
/* Version 3.1 allows for a hash-size of 0, performing a straightforward BFS 
 *     rather than a probabilistic one.  It guarentees the path lengths found 
 *     are the shortest that exist. */
// Enumerator:
/* Version 0.1 performs only Tiered DD with full signatures and paths. */
/* Version 0.1.1 removed sizes from the final and comptuation queues and uses 
 *     only maxStep to determine this information.  It also introduced the 
 *     use of getrlimit to determine the maximum number of files available. */
/* Version 0.1.2 added Tiered DD without paths */
/* Version 0.1.3 added Tiered DD with compressed states */
/* Version 0.1.4 moved communications functions to the Architecture Interface */
/* Version 0.1.5 documented code for tiered DDD */
// VERSION: 0.1.5

#include "tivlog.h"

#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>

#include "SearchInterface.h"
#include "tivmath.h"
#include "tivvect.h"
#include "pqueue.h"


class GeneralTieredDDD: public SearchInterface
{
protected:
   ////////////////////////////////////////////////////////////////
   //
   // Structures
   //
   ////////////////////////////////////////////////////////////////

   // Information about the current usage of the Memory-resident BFS blocks
   typedef struct blockAssign
   {
      UINT32 proc;
      UINT32 store;
      UINT32 cur;
      UINT32 head;
      UINT32 tail;
   } BLOCK_ASSIGN;

   // Used to store a state and the generator that lead to that state
   typedef struct pathNode
   {
      BYTE *state;
      PATH_STEP step;
   } PATH_NODE;

   // Information about the DDD blocks
   typedef struct dddQueue
   {
      INT32 first;
      FILE *file;
      const CHAR *name;
      INT32 cur;
      vector <UINT32> counts;
      vector <UINT32> maxSteps;
   } DDD_QUEUE;
     
   // The memory-resident hash
   typedef struct hash
   {
      BITVECTOR  hash;
      UINT64     size;
   } HASH_DATA;
   
   // A memory-resident BFS queue block
   typedef struct block
   {
      BOOL          used;
      UINT32        length;
      UINT32        pathSize;
      UINT32        stepMax;
      BYTE          *states;
      SEARCH_PATH   paths;
   } BLOCK;

    // Information for an entry in a DDD block
    typedef struct ddd_entry
    {
       CHAR fname[1000];
       BOOL coll;
       FILE *file;
       BYTE *data;
       UINT32 size;
    } DDD_ENTRY;
    
    ////////////////////////////////////////////////////////////////
    //
    // Enumerated Types
    //
    ////////////////////////////////////////////////////////////////

    // Possible message types that can be sent
    enum MESS_TYPE
    {
       MESS_CHECKED = 100,
       MESS_READY,
       MESS_GO,
       MESS_CHECK,
       MESS_FIRST,
       
       MESS_MAX
    };

    // The types of memory-resident BFS blocks possible
    enum BLOCK_TYPE
    {
       BLOCK_HEAD = 0,  // The head block of the computation, generators applied
       BLOCK_TAIL,      // The tail block, stores results of head block gen apps
       BLOCK_STORE,     // If a write block isn't full, stores it temporarily
      
       BLOCK_FREE
    };

    ////////////////////////////////////////////////////////////////
    //
    // Data Members
    //
    ////////////////////////////////////////////////////////////////

    // Memory-resident BFS blocks
    BLOCK blocks[BLOCK_FREE];
    // Memory-resident imperfect Hash
    HASH_DATA hash;
    // Information about the DDD blocks
    DDD_QUEUE fins, cols;
    // Information about the current usage of the Memory-resident BFS blocks
    BLOCK_ASSIGN ass;
    // Whether or not to store paths with values
    BOOL paths;
    // Keeps track of the number of points discovered locally
    UINT64 myPointTotal;
    // The maximum depth to compute the first BFS block to
    UINT32 FirstBlockPathMax;
    // MPI Rank Information
    UINT32 myRank, maxRank;
    // The number of states to batch process
    UINT64 StatesPerBlock;
    // The number of bits per element in the imperfect hash
    FLOAT64 HashMultiple;
    // The interfaces used during the computation
    ProblemInterface *probInt;
    ArchitectureInterface *archInt;
    PathInterface *pathInt;
    static GeneralTieredDDD *searchInt;
    // The level to debug at
    UINT64 DebugLevel;
    // The max length in bytes of a DDD block
    UINT64 QueueMax;
    
    ////////////////////////////////////////////////////////////////
    //
    // Virtual Functionality
    //
    ////////////////////////////////////////////////////////////////
   
    // The size of a compressed state
    virtual UINT64 getCompStateSize( ) = 0;
    // The size of the compressed key
    virtual UINT64 getCompKeySize( ) = 0;
    // The average path size
    virtual UINT64 getPathAverageSize( ) = 0;
    // The size of an entry in the disk queues
    virtual UINT32 getDataEntrySize( UINT32 maxSteps ) = 0;
    // Find the hash location in the table
    virtual BYTE *getCompState( BYTE *comp, BYTE *state ) = 0;
    

    ////////////////////////////////////////////////////////////////
    //
    // Constructors
    //
    ////////////////////////////////////////////////////////////////
    
    // Set the maximum length of the first block to 64 and prints only DDD block messages
    GeneralTieredDDD( )
    {
       StatesPerBlock = 0;
       QueueMax = 0;
       HashMultiple = 0;
       FirstBlockPathMax = 64;
       DebugLevel = 1;
    }
    
    ////////////////////////////////////////////////////////////////
    //
    // Comparators and functional arguments
    //
    ////////////////////////////////////////////////////////////////
            
    // Compare two compressed signatures in a DDD entry
    static int compSigComp( const void *left, const void *right )
    {
       return memcmp( left, right, searchInt->getCompKeySize() );
    }

    // Compare two paths in a DDD entry
    static int treeComp( const void *left, const void *right )
    {
       assert( searchInt->paths );
       return searchInt->pathInt->pathComp(
             ((BYTE *)left) + searchInt->getCompStateSize(),
             ((BYTE *)right) + searchInt->getCompStateSize());
    }

    // Compare two paths in the DDD entry structure
    static int fileTreeComp( const DDD_ENTRY *left, const DDD_ENTRY *right )
    {
       assert( searchInt->paths );
       return searchInt->pathInt->pathComp( 
             left->data + searchInt->getCompStateSize(), 
             right->data + searchInt->getCompStateSize());
    }

    // Compare two compressed signatures in the DDD entry structure
    static int fileCompSigComp( const DDD_ENTRY *left, const DDD_ENTRY *right )
    {
       return memcmp( left->data, right->data, searchInt->getCompKeySize() );
    }
    
    // Advance a DDD entry structure to the next entry in the DDD block
    static bool fileNext( DDD_ENTRY *entry )
    {
       if( fread( entry->data, 1, entry->size, entry->file ) != entry->size )
       {
          free( entry->data );
          fclose( entry->file );
          entry->file = NULL;
          if( entry->coll ) unlink( entry->fname );
          return false;
       }   
       return true;
    }
    
    ////////////////////////////////////////////////////////////////
    //
    // General functionality
    //
    ////////////////////////////////////////////////////////////////

    // Determine the owning computer of a compressed state
    inline UINT8 getProcessor( BYTE *compState )
    {
       UINT8 result = 0;
       for( UINT32 count = 0; count < getCompKeySize(); count++ ) result ^= compState[count];
       return result;
    }

    // Retrieve a pointer to a target state in a BFS block
    inline BYTE *getSig( UINT8 block, UINT32 number )
    {
       return blocks[block].states + number * probInt->StateSize;
    }
    
    // Retrieve a pointer to a target path in a BFS block
    inline SEARCH_PATH getPath( UINT8 block, UINT32 number )
    {
       assert( paths );
       return blocks[block].paths + number * blocks[block].pathSize;
    }
    
    // Report that this machine is done with the current phase of processing
    void syncCall( UINT64 myValue )
    {
        archInt->sendMessage( &( myValue ), SIZE64, 0, MESS_READY );
    }

    // Wait for all machines to be done with the current phase of processing
    void syncWait( UINT64 *values )
    {
       UINT32 fromAct, tagAct;
       INT32 fromReq, lenReq, lenAct, tagReq;
    
       if( myRank==0 )
       {
          // Master waits to hear from everyone that they're done
          for( UINT32 total = 0; total < maxRank; total++ )
          {
             archInt->probeMessage( archInt->SOURCE_ANY, MESS_READY, lenAct, fromAct, tagAct );
             archInt->recvMessage( &values[fromAct], lenAct, fromAct, tagAct,
                                                     lenAct, fromAct, tagAct );
          }
          // Master tells everyone to continue
          for( UINT32 total = 0; total < maxRank; total++ )
             archInt->sendMessage( values, SIZE64 * maxRank, total, MESS_GO );
       }
    
       // Everyone waits till the master says everyone's ready
       archInt->probeMessage( 0, MESS_GO, lenAct, fromAct, tagAct );
       archInt->recvMessage( values, lenAct, fromAct, tagAct,
                                     lenAct, fromAct, tagAct );
    }
    
    // Get a block of a certain type and set it to another type
    inline UINT8 getBlock( )
    {
       UINT8 resultBlock;

       for( resultBlock = BLOCK_FREE; resultBlock == BLOCK_FREE; usleep( 1000 ) )
          for( resultBlock = 0; resultBlock < BLOCK_FREE; resultBlock++ )
             if( blocks[resultBlock].used == false ) break;
       
       blocks[resultBlock].used = true;
       return resultBlock;
    }
    
    // Finalizes a DDD block and starts the next block
    void advanceFile( DDD_QUEUE &queue )
    {
       if( !queue.file || ftell( queue.file ) > 0 )
       {
          CHAR filename[1000];
          if( queue.file )
          {
             fclose( queue.file );
             log_string( 1, ( "Writing %s block %d, size %d, step max %d\n",
                         queue.name,
                         queue.cur,
                         queue.counts[queue.cur],
                         queue.maxSteps[queue.cur] ) );
              sprintf( filename, "%s%s.%d.%d.info", 
                    archInt->FinalDirectory, queue.name, myRank, queue.cur );
              fopen_check( queue.file, filename, "w" );
              fwrite_check( &(queue.counts[queue.cur]), SIZE32, queue.file );
              fwrite_check( &(queue.maxSteps[queue.cur]), SIZE32, queue.file );
              fclose( queue.file );
          }
          queue.counts.push_back(0);
          queue.maxSteps.push_back(0);
          sprintf( filename, "%s%s.%d.%d", archInt->FinalDirectory, queue.name, myRank, ++queue.cur );
          fopen_check( queue.file, filename, "w" );
       }
    }

    ////////////////////////////////////////////////////////////////
    //
    // Initialization
    //
    ////////////////////////////////////////////////////////////////

    // Initializes the data structures and generates the initial data
    bool initialize( RUN_MODE run, int *pargc, char ***pargv )
    {
        pthread_t hashMan;
        CHAR filename[1000];
        
        // The size of each file in the final/collision queue
        QueueMax  = (probInt->BranchingFactor) * (probInt->SearchSize) * 
                    (getPathAverageSize() + getCompStateSize()) / 
                    (archInt->NumberOfMachines * archInt->NumberOfFiles);
        UINT64 BLOCK_SIZE = (archInt->MessageSize * archInt->NumberOfMachines) / 
                            (getPathAverageSize() + getCompStateSize());

        if( StatesPerBlock==0 ) StatesPerBlock = BLOCK_SIZE / probInt->StateSize;
        if( QueueMax < BLOCK_FREE * StatesPerBlock * probInt->StateSize )
              QueueMax = BLOCK_FREE * StatesPerBlock * probInt->StateSize;

        INT64  AVAIL_MEM =  archInt->MemorySize -
           // Memory used by the blocks/queue structure
           MAX( BLOCK_FREE * StatesPerBlock * probInt->StateSize, QueueMax ) -
           // Memory used by the paths for the blocks
           BLOCK_FREE * StatesPerBlock * getPathAverageSize() - 
           // Memory used on both sides for checking
           2 * (getPathAverageSize() + getCompStateSize()) * StatesPerBlock;

        if( HashMultiple==0 )
            HashMultiple = MAX( 0, 8.0 * AVAIL_MEM / probInt->SearchSize * archInt->NumberOfMachines );
     
        printf( "Computation block size: %lld, Final/Collision block size: %lld\n",
               StatesPerBlock, QueueMax );
        printf( "Hash multiple: %lf because of %ld available memory\n",
               HashMultiple, AVAIL_MEM );
        
        if( HashMultiple==0 )
        {
            printf( "Cannot run computation, %lld available memory and hash requires at least %lld\n",
                    AVAIL_MEM,
                    probInt->SearchSize / 8 );
            return false;
        }
        if( run==RUN_REPORT ) return true;
       
        // Clean up the files and create the working directory
        sprintf( filename, "rm -rf %s %s", archInt->TempDirectory, archInt->FinalDirectory ); 
        system( filename );
        sprintf( filename, "mkdir -p %s %s", archInt->TempDirectory, archInt->FinalDirectory );
        system( filename );
       
       // Start the communication program and initialize the log
       archInt->initializeComm( myRank, maxRank );
       sprintf( filename, "%soutput.%d.txt", archInt->CommonDirectory, myRank );
       init_log( filename, DebugLevel );
       log_string( 0, ( "COMM Initiated: %d of %d\n", myRank, maxRank ) );

       // Initialize all of our blocks to free
       UINT32 memsize = MAX( QueueMax, StatesPerBlock * probInt->StateSize * BLOCK_FREE );
       malloc_check( blocks[0].states, BYTE *, memsize );
       for( UINT8 b = 0; b < BLOCK_FREE; b++ )
       {
          blocks[b].states = blocks[0].states + b * StatesPerBlock * probInt->StateSize;
          blocks[b].pathSize = blocks[b].stepMax = 0;
          blocks[b].paths = NULL;
          blocks[b].used = false;
       }
       
       // Open the queueing files
       fins.cur = -1; fins.file = NULL; fins.first = 0; fins.name = "final";
       cols.cur = -1;  cols.file = NULL;  cols.first = 0;  cols.name = "collision";
       advanceFile( fins );
       advanceFile( cols );

       // Initialize the hash table
       if( pthread_create( &hashMan, NULL, runHashManager, this )!=0 ) assert( 0 );
       
       // Wait to receive the first block
       computeFirstBlock( );

       return true;
    }
    
    // Compute the first block on the first processor,
    // Partition it up and distribute it
    void computeFirstBlock( )
    {
       MPI_Status status;
       UINT32 head, proc, size;
       UINT32 fromAct, tagAct;
       INT32 lenReq, lenAct, fromReq, tagReq;
       
       if( myRank == 0 )
       {
          // Load the first matrix
          ass.cur = 1;
          probInt->getRoots( getSig( ass.cur, 0 ) );
          blocks[ass.cur].length = probInt->NumberOfRoots;
          
          // Get room for the paths
          if( paths )
          {
             realloc_check( blocks[ass.cur].paths,
                            SEARCH_PATH,
                            pathInt->PATH_SIZE( FirstBlockPathMax ) * StatesPerBlock );
             for( head = 0; head < probInt->NumberOfRoots; head++ )
                pathInt->SET_STEPS( getPath( ass.cur, 0 ), 0 );
          }
             
          // Compute the first block
          log_string( 0, ( "Beginning computation of first block\n" ) );
          if( paths ) blocks[ass.cur].pathSize = pathInt->PATH_SIZE( FirstBlockPathMax );
          head = computeBlock( ass.cur, ass.cur, 0, FirstBlockPathMax );
          if( paths )
          {
             blocks[ass.cur].stepMax = pathInt->GET_STEPS( getPath( ass.cur, blocks[ass.cur].length - 1 ) );
             while( pathInt->PATH_SIZE( blocks[ass.cur].stepMax ) < 
                    pathInt->PATH_SIZE( FirstBlockPathMax ) )
                blocks[ass.cur].stepMax++;
          }
          assert( blocks[ass.cur].length==StatesPerBlock ||
                  blocks[ass.cur].stepMax==FirstBlockPathMax );
          log_string( 0, ( "First block step max %d, %d items, %d new items\n",
                      blocks[ass.cur].stepMax, 
                      blocks[ass.cur].length,
                      blocks[ass.cur].length - head ) );
          
          // Check the old values in the block and report the results
          size = blocks[ass.cur].length;
          blocks[ass.cur].length = head;
          checkBlock( );
          log_string( 0, ( "After duplicate stripping of old values %d items %d new items\n",
                      blocks[ass.cur].length + size - head,
                      size - head ) );
          myPointTotal += blocks[ass.cur].length;
                      
          // Check the new values in the block and report the results
          blocks[ass.cur].length = size - head;
          if( paths )
             memmove( getPath( ass.cur, 0 ),
                      getPath( ass.cur, head ),
                      blocks[ass.cur].length * blocks[ass.cur].pathSize );
          memmove( getSig( ass.cur, 0 ),
                   getSig( ass.cur, head ),
                   blocks[ass.cur].length * probInt->StateSize );
          checkBlock( );
          log_string( 0, ( "After duplicate stripping of new values %d new items\n",
                      blocks[ass.cur].length ) );
          
          // Send the block to the rest of the processors
          for( proc = 0, head = 0, size = blocks[ass.cur].length / maxRank; 
               proc < maxRank; 
               proc++, head += size )
          {
             if( proc == maxRank - 1 ) size = blocks[ass.cur].length - head;
             archInt->sendMessage( &size, SIZE32, proc, MESS_FIRST );
             if( size )
             {
                if( paths )
                {
                   archInt->sendMessage( &blocks[ass.cur].stepMax, SIZE32, proc, MESS_FIRST );
                   archInt->sendMessage( getPath( ass.cur, head ), size * blocks[ass.cur].pathSize, proc, MESS_FIRST );
                }
                archInt->sendMessage( getSig( ass.cur, head ), size * probInt->StateSize, proc, MESS_FIRST );
             }
             else archInt->sendMessage( &size, SIZE32, proc, MESS_FIRST );
          }
          ass.cur = BLOCK_FREE;
       }

       // Wait for the block length
       archInt->probeMessage( 0, MESS_FIRST, lenAct, fromAct, tagAct );
       archInt->recvMessage( &blocks[0].length, lenAct, fromAct, tagAct,
                                                lenAct, fromAct, tagAct );
       log_string( 0, ( "Receiving first block of size %d\n", blocks[0].length ) );
       
       // If there's data for me, get it
       if( blocks[0].length )
       {
          ass.proc = 0;
          blocks[ass.proc].used = true;
          
          if( paths )
          {
             // Wait for the step max
             archInt->probeMessage( 0, MESS_FIRST, lenAct, fromAct, tagAct );
             archInt->recvMessage( &blocks[ass.proc].stepMax, lenAct, fromAct, tagAct,
                                                              lenAct, fromAct, tagAct );
       
             // Wait for the paths
             archInt->probeMessage( 0, MESS_FIRST, lenAct, fromAct, tagAct );
             blocks[ass.proc].pathSize = lenAct / blocks[ass.proc].length;
             realloc_check( blocks[ass.proc].paths, 
                            SEARCH_PATH, 
                            blocks[ass.proc].pathSize * StatesPerBlock );
             archInt->recvMessage( blocks[ass.proc].paths, lenAct, fromAct, tagAct,
                                                           lenAct, fromAct, tagAct );
          }
          
          // Wait for the incoming block
          archInt->probeMessage( 0, MESS_FIRST, lenAct, fromAct, tagAct );
          archInt->recvMessage( blocks[ass.proc].states, lenAct, fromAct, tagAct,
                                                         lenAct, fromAct, tagAct );
          myPointTotal += blocks[ass.proc].length;          
       }
    }

    // Completes the initialization
    void finishInitialize( UINT64 *totalPoints, UINT64 &localTotal, UINT64 &sumTotal )
    {
       UINT64 finTot = 0;
       
       // Print out the information
       log_string( 0, ( "INIT finished with %lld points locally\n", totalPoints[myRank] ) );
       for( UINT8 count = 0; count < maxRank; count++ )
           finTot += totalPoints[count];
       log_string( 0, ( "INIT finished with %lld points globally\n", finTot ) );
       localTotal += totalPoints[myRank];
       log_string( 0, ( "Discovered %lld points locally so far\n", localTotal ) );
       sumTotal += finTot;
       log_string( 0, ( "Discovered %lld points globally so far\n", sumTotal ) );
       myPointTotal = 0;
    }

    ////////////////////////////////////////////////////////////////
    //
    // Hash Manager and associated functionality
    //
    ////////////////////////////////////////////////////////////////

    // Thread to manage the hash, allocates space then waits for blocks to check
    static void *runHashManager( void *input )
    {
       ArchitectureInterface *archInt = searchInt->archInt;
       vector <BYTE> *data = new vector <BYTE> [searchInt->maxRank];
       UINT32 lenReq, fromAct, tagAct;
       INT32 fromReq, lenAct, tagReq;
       
       // Initialize the hash
       searchInt->hash.size =  ( UINT64 )( searchInt->HashMultiple * searchInt->probInt->SearchSize / searchInt->maxRank );
       if( searchInt->hash.size )
       {
          malloc_check( searchInt->hash.hash, BITVECTOR, BV_SIZE( searchInt->hash.size ) );
          memset( searchInt->hash.hash, 0, BV_SIZE( searchInt->hash.size ) );
          log_string( 0, ( "Hash obtained for %lld items\n", searchInt->hash.size ) );
       }
       
       // Wait for check requests
       while( true )
       {
          // Wait for the message
          archInt->probeMessage( archInt->SOURCE_ANY, MESS_CHECK, lenAct, fromAct, tagAct );
          log_string( 3, ( "Hashing block from %d\n", fromAct ) );
          
          // Resize the block and receive the message
          data[fromAct].resize( lenAct );
          archInt->recvMessage( &(data[fromAct][0]), lenAct, fromAct, tagAct,
                                                     lenAct, fromAct, tagAct );
          UINT32 numItems = *( ( UINT32 * ) &data[fromAct][0*SIZE32] );
          UINT32 maxStep  = *( ( UINT32 * ) &data[fromAct][1*SIZE32] );
          UINT32 itemSize = searchInt->getDataEntrySize( maxStep );
          
          // Check the block
          *( ( UINT32 * ) &(data[fromAct][1*SIZE32]) ) = 
             searchInt->checkHash( data[fromAct], itemSize, numItems, maxStep );
          
          // Send it back to where it came from
          if( !searchInt->hash.size ) archInt->sendMessage( &(data[fromAct][0]), SIZE32, fromAct, MESS_CHECKED );
          else archInt->sendMessage( &(data[fromAct][1*SIZE32]), numItems + SIZE32, fromAct, MESS_CHECKED );
       }
    }

    // Check an individual hash block
    UINT32 checkHash( vector <BYTE> &block, UINT32 size, UINT32 count, UINT32 maxStep )
    {
       UINT32 total = 0;
       
       // See if the maybe block has been filled
       if( ( cols.counts[cols.cur] + count ) * 
             MAX( getDataEntrySize( cols.maxSteps[cols.cur] ), size ) > QueueMax ) 
          advanceFile( cols );
       
       // See if the final block has been filled
       if( ( fins.counts[fins.cur] + count ) * 
             MAX( getDataEntrySize( fins.maxSteps[fins.cur] ), size ) > QueueMax ) 
          advanceFile( fins );
       
       if( maxStep > cols.maxSteps[cols.cur] )  cols.maxSteps[cols.cur] = maxStep;
       if( maxStep > fins.maxSteps[fins.cur] )  fins.maxSteps[fins.cur] = maxStep;
       
       for( UINT32 cur = 0; cur < count; cur++ )
       {
          if( HashCheck( cur, block, size ) )
          {
             // Write out the sig to the maybe block
             total++;
             cols.counts[cols.cur]++;
             fwrite_check( &(block[cur * size + 2*SIZE32]),
                           getCompStateSize() + 
                              ( paths ? pathInt->GET_SIZE( &(block[cur * size + 2*SIZE32 + getCompStateSize()])) : 0 ), 
                           cols.file );
             block[cur + 2*SIZE32] = 0;
          }
          else
          {
             HashSet( cur, block, size );
             fins.counts[fins.cur]++;
             fwrite_check( &(block[cur * size + 2*SIZE32]), 
                           getCompStateSize() + 
                              ( paths ? pathInt->GET_SIZE( &(block[cur * size + 2*SIZE32 + getCompStateSize()])) : 0 ), 
                           fins.file );
             block[cur + 2*SIZE32] = 1;
          }
       }
       log_string( 3, ( "Finished hasing block with %d/%d collisions\n", total, count ) );
       return count - total;
    }
    
    // Returns true if an entry in the DDD block has a hash collision
    inline bool HashCheck( UINT64 target, vector <BYTE> &block, UINT64 eleSize )
    {
        return !hash.size || 
               BV_IS_SET( hash.hash,  
                          getHash( &( block[ target * eleSize + 2 * SIZE32 ] ) ) % hash.size );
    }
    
    // Fills the location in the hash corresponding to the entry in the DDD block
    inline void HashSet( UINT64 target, vector <BYTE> &block, UINT64 eleSize )
    {
        if( hash.size )
            BV_SET( hash.hash,
                    getHash( &( block[ target * eleSize + 2 * SIZE32 ] ) ) % hash.size );
            
    }
    
    // Finds the hash value (UINT64) of a compressed state
    UINT64 getHash( BYTE *compState )
    {
       UINT64 result = 0;
       for( UINT32 count = 0; count < getCompKeySize() / SIZE64; count++ )
          result ^= ( ( UINT64 * )compState )[count];
       return result;
    }
                            
    ////////////////////////////////////////////////////////////////
    //
    // BFS functionality
    //
    ////////////////////////////////////////////////////////////////

    // Thread to manage the BFS, performs:
    // (1) Load disk BFS block from head of BFS queue
    // (2) Compute new tail BFS block from neighbors of head BFS block
    // (3) Check new tail BFS block for duplicates
    // (4) Write new tail BFS block to disk
    static void *runBFS( void *input )
    {
       UINT64 location = 0;
       log_string( 0, ( "Beginning distributed breadth-first search\n" ) );
       
       // While there is a head BFS block to load
       for( searchInt->loadBlock( ); searchInt->ass.proc != BLOCK_FREE; searchInt->loadBlock( ) )
       {
          // Compute the tail BFS block
          searchInt->processBlock( location );
          // Strip the tail BFS block ofduplicates
          searchInt->myPointTotal += searchInt->checkBlock( );
          // Store the tail BFS block
          searchInt->storeBlock( );
       }
       
       // Tell the master we're done
       searchInt->syncCall( searchInt->myPointTotal );
       return NULL;
    }
    
    // Performs (1): Loads the head BFS block
    // (1a) Load no block if the current head block if not finished
    // (1b) Load a block from disk otherwise if available
    // (1c) Load no block otherwise
    void loadBlock( )
    {
       // Performs (1b): Loads the block from disk
       if( ass.proc==BLOCK_FREE && ass.head < ass.tail )
       {
          CHAR filename[1000];
          FILE *file;
          ass.proc = getBlock( );
          
          // Read in the info for the block
          sprintf( filename, "%sBFSinfo.%d.%d", archInt->TempDirectory, myRank, ass.head );
          fopen_check( file, filename, "rb" );
          fread_check( &(blocks[ass.proc].length), SIZE32, file );
          if( paths )
          {
             fread_check( &(blocks[ass.proc].pathSize), SIZE32, file );
             fread_check( &(blocks[ass.proc].stepMax), SIZE32, file );
          }
          log_string( 2, ( "Reading block %d, length %d, path size %d, max step %d\n", 
                           ass.head, blocks[ass.proc].length, blocks[ass.proc].pathSize, blocks[ass.proc].stepMax ) );
          fclose( file );
          remove( filename );
          if( paths )
          {
             realloc_check( blocks[ass.proc].paths, 
                            SEARCH_PATH, 
                            StatesPerBlock * blocks[ass.proc].pathSize );
          }
          
          // Read in the sigs
          sprintf( filename, "%sBFSstates.%d.%d", archInt->TempDirectory, myRank, ass.head );
          fopen_check( file, filename, "rb" );
          fread_check( blocks[ass.proc].states, 
                       blocks[ass.proc].length * probInt->StateSize, 
                       file );
          fclose( file );
          remove( filename );
          
          // Read in the bitstrs
          if( paths )
          {
             sprintf( filename, "%sBFSpaths.%d.%d", archInt->TempDirectory, myRank, ass.head );
             fopen_check( file, filename, "rb" );
             fread_check( blocks[ass.proc].paths,
                          blocks[ass.proc].length * blocks[ass.proc].pathSize,
                          file );
             fclose( file );
             remove( filename );
          }
          
          // Update the head value
          log_string( 2, ( "Block of length %d read\n", blocks[ass.proc].length ) );
          ass.head++;
       }
       // Performs (1b): load the block from the block about to be stored
       else if( ass.proc==BLOCK_FREE && ass.store!=BLOCK_FREE )
       {
          ass.proc = ass.store;
          ass.store = BLOCK_FREE;
       }
    }

    // Performs (2): Computes the neighbors of the head block to a tail block
    void processBlock( UINT64 &location )
    {
       // Get a new block to use as tail
       ass.cur = getBlock( );
       blocks[ass.cur].length = 0;
       if( paths )
       {
          blocks[ass.cur].stepMax = blocks[ass.proc].stepMax + 1;
          blocks[ass.cur].pathSize = pathInt->PATH_SIZE( blocks[ass.cur].stepMax );
       }

       // Compute the tail block
       log_string( 2, ( "Processing block of size %d with pathSize %d\n",
                   blocks[ass.proc].length, blocks[ass.proc].pathSize ) );
       if( paths )
       {
          realloc_check( blocks[ass.cur].paths, 
                         SEARCH_PATH,
                         StatesPerBlock * blocks[ass.cur].pathSize );
       }
       location = computeBlock( ass.proc, ass.cur, location, blocks[ass.cur].stepMax );   
       log_string( 2, ( "Block processed, size %d -> %d\n",
                   blocks[ass.proc].length, blocks[ass.cur].length ) );
            
       // Check and see if we're done with the head block
       if( location == blocks[ass.proc].length )
       {
          blocks[ass.proc].used = false;
          location = 0;
          ass.proc = BLOCK_FREE;
       }
    }

    // Helps perform (2): Compute neighbors of a block
    UINT32 computeBlock( UINT8 head, UINT8 tail, UINT32 pos, UINT32 pathMax )
    {
       UINT32 headSize = blocks[head].pathSize, 
              tailSize = blocks[tail].pathSize,
              headSteps,
              tar;
       SEARCH_PATH tailPath, headPath;
       BOOL success;
       
       while( pos < blocks[head].length && blocks[tail].length < StatesPerBlock )
       {
          if( paths )
          {
             headPath = getPath( head, pos );
             if( ( headSteps = pathInt->GET_STEPS( headPath ) ) >= pathMax ) break;
          }
          
          for(  tar = 0; 
                tar < probInt->NumberOfGenerators && blocks[tail].length < StatesPerBlock; 
                tar++ )
          {
             probInt->applyTargetGen( getSig( head, pos ),
                                   tar,
                                   getSig( tail, blocks[tail].length ),
                                   &success,
                                   paths ? headPath : NULL,
                                   paths ? headSize : 0);
             if( success )
             {
                if( paths )
                {
                   tailPath = getPath( tail, blocks[tail].length );
                   memcpy( tailPath, headPath, MIN( headSize, tailSize ) );
                   pathInt->SET_STEPS( tailPath, headSteps + 1 );
                   pathInt->SET_STEP( tailPath, headSteps, tar );
                }
                blocks[tail].length++;
             }
          }
          
          if( tar==probInt->NumberOfGenerators ) pos++;
       }
       
       return pos;
    }

    // Performs (3): Removes duplicates from the block:
    // (3a) Partition the states to be sent to owning nodes
    // (3b) Send the data and retrieve the results
    // (3c) Removes the duplicates
    UINT64 checkBlock( )
    {
       // The compressed signatures and paths to be sent to each machine for each entry
       vector <BYTE> data;
       // The owning machine for each entry
       vector <UINT32> procs;
       // The number of entries owned by lowered numbered machines for each machine, a cumsum
       vector <UINT32> counts;

       log_string( 2, ( "Checking block of size %d\n", blocks[ass.cur].length ) );
       
       // Performs (3a): Partitions the states
       partitionData( ass.cur, data, procs, counts );
       // Performs (3b): Sends the data and retrieves results
       getHashResults( ass.cur, data, counts );
       // Performs (3c): Removes the duplicates
       if( !hash.size ) blocks[ass.cur].length = 0;
       else stripDuplicates( ass.cur, data, procs, counts );

       log_string( 2, ( "Block checked of size %d\n", blocks[ass.cur].length ) );
       return blocks[ass.cur].length;
    }

    // Performs (3a): Partitions the values to be sent
    // (3ai)   Count number of values to be sent to each machine
    // (3aii)  Prepare header information for each machine
    // (3aiii) Write (compressed) states for each machine
    void partitionData( UINT32 check, vector <BYTE> &data, vector <UINT32> &procs, vector <UINT32> &counts )
    {
       UINT32 count, prev, sumtotal;
       BYTE compSig[ getCompStateSize() ];
       
       // Emptying out the data
       counts.resize(maxRank);
       memset( &(counts[0]), 0, maxRank * SIZE32 );
       data.resize( 2*SIZE32 * maxRank + // Room for the headers
                    blocks[check].length * ( getCompStateSize() + blocks[check].pathSize ) + // Room for the data
                    1 ); // Extra padding at the end for pointer
       procs.resize( blocks[check].length );

       // Perform (3ai): Compute procs and find the number of entries for each processor
       for( count = 0; count < blocks[check].length; count++ )
       {
           getCompState( compSig, getSig( check, count ) );
           procs[count] = getProcessor( compSig ) % maxRank;
           counts[procs[count]]++;
       }
       
       // Perform (3aii): Compute counts and intialize the data block
       for( count = 0, prev = 0, sumtotal = 0; count < maxRank; count++ )
       {
           memcpy( &data[count * 2*SIZE32 + 0*SIZE32 + 
                         (prev+sumtotal) * (getCompStateSize() + blocks[check].pathSize)],
                   &counts[count],
                   SIZE32);
           memcpy( &data[count * 2*SIZE32 + 1*SIZE32 + 
                         (prev+sumtotal) * (getCompStateSize() + blocks[check].pathSize)],
                   &blocks[check].stepMax,
                   SIZE32);
           sumtotal += prev;
           prev = counts[count];
           counts[count] = sumtotal;
       }
       
       // Perform (3aiii): Compute the data block
       vector <BYTE *> pointers; pointers.resize( maxRank );
       for( count = 0; count < maxRank; count++ )
           pointers[count] = &data[count * 2*SIZE32 + 2*SIZE32 + 
                                   counts[count] * (getCompStateSize() + blocks[check].pathSize)];
       for( count = 0; count < blocks[check].length; count++ )
       {
           getCompState( pointers[procs[count]],
                       getSig(check, count) );
           if( paths )
              memcpy( pointers[procs[count]] + getCompStateSize(),
                      getPath( check, count ),
                      blocks[check].pathSize );
           pointers[procs[count]] += blocks[check].pathSize + getCompStateSize();
       }
    }

    // Perform (3b): Send the data and retrive results:
    // (3bi)  Send the data
    // (3bii) Retrieve results
    void getHashResults( UINT32 check, vector <BYTE> &data, vector <UINT32> &counts )
    {
       UINT32 fromAct, tagAct;
       INT32 fromReq, lenReq, lenAct, tagReq;
       BYTE *target = &data[0];
       
       // Perform (3bi): Send out the hashes
       for( UINT32 count = 0; count < maxRank; count++ )
       {
            archInt->sendMessage( target,
                                  *(((UINT32 *) target)+0) * getDataEntrySize(blocks[check].stepMax) + 2*SIZE32,
                                  count,
                                  MESS_CHECK );
            target += *(((UINT32 *) target)+0) * getDataEntrySize(blocks[check].stepMax) + 2*SIZE32;
       }

       // Perform (3bii): Wait for all the incoming hash check results to come in
       for( UINT32 count = 0; count < maxRank; count++ )
       {
          // Wait till for any check result and store it
          archInt->probeMessage( archInt->SOURCE_ANY, MESS_CHECKED, lenAct, fromAct, tagAct );
          archInt->recvMessage( &data[fromAct * 2*SIZE32 + 
                                      1*SIZE32 + 
                                      counts[fromAct] * getDataEntrySize(blocks[check].stepMax)],
                                lenAct, fromAct, tagAct,
                                lenAct, fromAct, tagAct );
       }
    }

    // Perform (3c) Removes the duplicates
    void stripDuplicates( UINT32 check, vector <BYTE> &data, vector <UINT32> &procs, vector <UINT32> &counts )
    {
        UINT32 head, tail, count;
        
        // Remove the duplicates
        vector <BYTE *> pointers; pointers.resize( maxRank );
        for( count = 0; count < maxRank; count++ )
             pointers[count] = &data[count * 2*SIZE32 + 2*SIZE32 + 
                                     counts[count] * (getCompStateSize() + blocks[check].pathSize)];
        for( head = 0, tail = 0; tail < blocks[check].length; tail++ )
        {
            if( *(pointers[procs[tail]]) )
            {
                if( head!=tail )
                {
                   memcpy( getSig( check, head ),
                           getSig( check, tail ),
                           probInt->StateSize );
                   if( paths )
                      memcpy( getPath( check, head ),
                              getPath( check, tail ),
                              blocks[check].pathSize );
                }
                head++;
            }
            pointers[procs[tail]]++;
        }
        blocks[check].length = head;
    }

    // Performs (4): Stores a tail block to disk:
    // (4a) Append the block to the current store block if possible
    // (4b) Otherwise:
    // (4bi)  Write the current store block to disk
    // (4bii) Set the block to be the current store block
    void storeBlock( )
    {
       if( blocks[ass.cur].length )
       {
          log_string( 2, ( "Writing block of size %d strlen %d\n",
                      blocks[ass.cur].length, blocks[ass.cur].pathSize ) );
               
          // Performs (4a): Appends to the current store block
          if( ass.store!=BLOCK_FREE && 
              blocks[ass.cur].pathSize == blocks[ass.store].pathSize &&
              blocks[ass.cur].length + blocks[ass.store].length <= StatesPerBlock )
          {
             // Add it to the store
             blocks[ass.store].stepMax = MAX( blocks[ass.store].stepMax,
                                              blocks[ass.cur].stepMax );
             memcpy( getSig( ass.store, blocks[ass.store].length ),
                     getSig( ass.cur, 0 ),
                     blocks[ass.cur].length * probInt->StateSize );
             if( paths )
                memcpy( getPath( ass.store,  blocks[ass.store].length ),
                        getPath( ass.cur, 0 ),
                        blocks[ass.cur].length * blocks[ass.cur].pathSize );
             blocks[ass.store].length += blocks[ass.cur].length;
             
             // Free up the check block
             blocks[ass.cur].used = false;
          }
          // Performs (4b): Swaps out the store block
          else
          {
             // Performs (4bi): Writes the current store to disk
             if( ass.store != BLOCK_FREE )
             {
                writeBlock( &(blocks[ass.store]), ass.tail );
                blocks[ass.store].used = false;
             }
                 
             // Performs (4bii): Change the current block to a store block
             ass.store = ass.cur;
             ass.cur = BLOCK_FREE;
          }
       }
       else
       {
          blocks[ass.cur].used = false;
          ass.cur = BLOCK_FREE;
       }
    }

    // Performs (4bi): Writes a block to disk to the tail of the BFS queue
    void writeBlock( BLOCK *block, UINT32 &tail )
    {
       CHAR filename[1000];
       FILE *file;
       
       // If there's nothing to write, ignore it
       if( block->length > 0 )
       {
          log_string( 2, 
             ( "Writing block %d of size %d, stepMax %d\n",
               tail, block->length, block->stepMax ) );
                
          // Write out the info
          sprintf( filename, "%sBFSinfo.%d.%d", archInt->TempDirectory, myRank, tail );
          fopen_check( file, filename, "wb" );
          fwrite_check( &(block->length), SIZE32, file );
          if( paths )
          {
             fwrite_check( &(block->pathSize), SIZE32, file );
             fwrite_check( &(block->stepMax), SIZE32, file );
          }
          fclose( file );
          
          // Write out the sigs
          sprintf( filename, "%sBFSstates.%d.%d", archInt->TempDirectory, myRank, tail );
          fopen_check( file, filename, "wb" );
          fwrite_check( block->states, block->length * probInt->StateSize, file );
          fclose( file );
          
          // Write out the bitstrs
          if( paths )
          {
             sprintf( filename, "%sBFSpaths.%d.%d", archInt->TempDirectory, myRank, tail );
             fopen_check( file, filename, "wb" );
             fwrite_check( block->paths, block->length * block->pathSize, file );
             fclose( file );
          }
          
          // Increment the tail block
          log_string( 2, ( "Block %d written\n", tail ) );
          tail++;
       }
    }

    // Finalize the BFS portion of the computation
    void finishBFS( UINT64 *totalPoints, UINT64 &localTotal, UINT64 &sumTotal )
    {
       UINT64 finTot = 0;
       
       // Close out the maybe and final blocks for the second phase
       if( ftell( cols.file ) ) advanceFile( cols );
       if( ftell( fins.file ) ) advanceFile( fins );

       // Print out the information
       log_string( 0, ( "BFS finished with %lld points locally\n", totalPoints[myRank] ) );
       for( UINT8 count = 0; count < maxRank; count++ )
           finTot += totalPoints[count];
       log_string( 0, ( "BFS finished with %lld points globally\n", finTot ) );
       localTotal += totalPoints[myRank];
       log_string( 0, ( "Discovered %lld points locally so far\n", localTotal ) );
       sumTotal += finTot;
       log_string( 0, ( "Discovered %lld points globally so far\n", sumTotal ) );
       myPointTotal = 0;
    }
    
    ////////////////////////////////////////////////////////////////
    //
    // DDD functionality
    //
    ////////////////////////////////////////////////////////////////

    // Thread to manage the DDD, performs:
    // (1) For each DDD block, sort and standardize:
    // (2) Strip duplicates, sort new final blocks, and write to disk
    // (3) Compute the new BFS blocks from new final entries
    // (4) Compressed: Sort the new final blocks by compressed state
    static void *runDDD( void *input )
    {
       UINT32 totPoints = 0;
       CHAR filename[1000];
       bool done = true;
       
       log_string( 0, ( "Starting Disk-based Duplicate Detection\n" ) );
       
       if( searchInt->fins.first < searchInt->fins.cur )
       {
          // Sort the final blocks
          searchInt->sortFiles( searchInt->fins );
          searchInt->fins.first = searchInt->fins.cur;
       }
         
       if( searchInt->cols.first < searchInt->cols.cur )
       {
          // Sort the collision blocks
          searchInt->sortFiles( searchInt->cols );

          // Eliminate duplicates in the collision queue
          searchInt->killDuplicates( );

          // Convert the new final blocks to computation blocks
          if( searchInt->fins.first < searchInt->fins.cur )
          {
             searchInt->finalToComp( );
          }
       }
       log_string( 0, ( "DDD finished\n" ) );
       searchInt->cols.first = searchInt->cols.cur;
       searchInt->fins.first = searchInt->fins.cur;
       searchInt->syncCall( searchInt->myPointTotal );
       return NULL;
    }

    // Performs (1) in the DDD: sorts and standardizes DDD blocks
    // (1a) Load the block and standardize the sizes
    // (1b) Sort the block by compressed signature 
    // (1c) Write the block back to disk
    void sortFiles( DDD_QUEUE &q )
    {
       CHAR filename[1000];
       UINT32 count, space;
       FILE *pfFile;

       log_string( 1, ( "Sorting %s DDD blocks %d-%d\n", q.name, q.first, q.cur-1 ) );
       for( UINT32 cur = q.first; cur < q.cur; cur++ )
       {
          // Performs (1a): Reorganize the block
          log_string( 2, ( "Reorganizing %s block %d\n", q.name, cur ) );
          UINT32 blockSize = getDataEntrySize( q.maxSteps[cur] );
          
          // Open the file and reorganize
          sprintf( filename, "%s%s.%d.%d", archInt->FinalDirectory, q.name, myRank, cur );
          fopen_check( pfFile, filename, "r" );
          for( count = 0; count < q.counts[cur]; count++ )
          {
             fread_check( blocks[0].states + count * blockSize, getCompStateSize(), pfFile );
             if( paths )
             {
                fread_check( blocks[0].states + count * blockSize + getCompStateSize(), SIZE32, pfFile );
                fread_check( blocks[0].states + count * blockSize + getCompStateSize() + SIZE32, 
                             pathInt->GET_SIZE( blocks[0].states + count * blockSize + getCompStateSize() ) - SIZE32, 
                             pfFile );
             }
          }
          assert( fread( &count, 1, SIZE32, pfFile )==0 );
          fclose( pfFile );
          
          // Performs (1b): Sort the block
          log_string( 2, ( "Sorting %s block %d\n", q.name, cur ) );
          qsort( blocks[0].states, q.counts[cur], blockSize, compSigComp );
          
          // Performs (1c): Rewrite the block
          log_string( 2, ( "Writing %s block %d back to disk\n", q.name, cur ) );
          fopen_check( pfFile, filename, "w" );
          fwrite_check( blocks[0].states, q.counts[cur] * blockSize, pfFile );
          fclose( pfFile );
       }
    }

    // Performs (2) in the DDD: Strips duplicates to find new final blocks
    // (2) Use a priority queue over:
    //        The sorted DDD final blocks
    //        The sorted DDD collision blocks
    // (2a) Strip duplicates from the collision blocks
    // (2b) Sort the new final blocks by:
    //        Full states: compressed state
    //        Compressed states: path ordering
    // (2c) Write the new final blocks to disk
    void killDuplicates( )
    {
       UINT32 entSteps = 0;
       UINT64 totPoints = 0;
       DDD_ENTRY ent;
       vector <DDD_ENTRY> colEnts, finEnts;
       BYTE prvCol[getCompStateSize()];
       memset( prvCol, '\0', getCompStateSize() );
       
       log_string( 1, ( "Removing duplicates in collision queue blocks %d-%d\n", cols.first, cols.cur-1 ) );
       
       // Finding the new block size, use the max of all
       for( UINT32 cur = cols.first; cur < cols.cur; cur++ )
          if( cols.maxSteps[cur] > entSteps )
             entSteps = cols.maxSteps[cur];
       UINT32 entSize = getDataEntrySize( entSteps );

       // Performs (2): Initialize all the collision pqueue entries
       for( UINT32 cur = cols.first; cur < cols.cur; cur++ )
       {
          sprintf( ent.fname, "%s%s.%d.%d", archInt->FinalDirectory, cols.name, myRank, cur );
          fopen_check( ent.file, ent.fname, "r" );
          malloc_check( ent.data, BYTE *, entSize );
          ent.size = getDataEntrySize( cols.maxSteps[cur] );
          ent.coll = true;
          colEnts.push_back( ent );
       }

       // Performs (2): Initialize all the final pqueue entries
       for( UINT32 cur = 0; cur < fins.cur; cur++ )
       {
          sprintf( ent.fname, "%s%s.%d.%d", archInt->FinalDirectory, fins.name, myRank, cur );
          fopen_check( ent.file, ent.fname, "r" );
          ent.size = getDataEntrySize( fins.maxSteps[cur] );
          malloc_check( ent.data, BYTE *, ent.size );
          ent.coll = false;
          finEnts.push_back( ent );
       }
       
       // Performs (2): Initialize the queues themselves
       UINT32 blockMax = QueueMax / entSize;
       mergeQueue <DDD_ENTRY> 
          mqFin( fileCompSigComp, fileNext, &(finEnts[0]), finEnts.size() ),
          mqCol( fileCompSigComp, fileNext, &(colEnts[0]), colEnts.size() );
          DDD_ENTRY *curFin = mqFin.pop(), *curCol = NULL;
       
       // While the collision queue isn't empty
       do
       {
          // While the current final block isn't full
          while( fins.counts[fins.cur] < blockMax && ( curCol = mqCol.pop() ) )
          {
             // Performs (2a): Determines if the value is new
             assert( compSigComp( prvCol, curCol->data ) <= 0 );
             if( compSigComp( curCol->data, prvCol )!=0 )
             {
                // Copy the current signature into the old signature
                memcpy( prvCol, curCol->data, getCompStateSize() );

                // Scan through the Final Queue till the final >= current
                while( curFin && compSigComp( curFin->data, curCol->data ) < 0 )
                   curFin = mqFin.pop( );
                
                // If there's no more finals left or they don't compare, add it
                if( !curFin || compSigComp( curFin->data, curCol->data )!=0 )
                {
                   memcpy( blocks[0].states + fins.counts[fins.cur]++ * entSize,
                           curCol->data, 
                           entSize );
                   myPointTotal++;
                }
             }
             totPoints++;
          }
          // Write out the block
          if( fins.counts[fins.cur] )
          {
             // Performs (2b) Sorts the new final blocks by:
             //        Full states: compressed state
             //        Compressed states: path ordering
             fins.maxSteps[fins.cur] = entSteps;
             if( getCompStateSize() < probInt->StateSize )
             {
                // Sort the block by its paths
                qsort( blocks[0].states, fins.counts[fins.cur], entSize, treeComp );
             }
             else
             {
                // Sort the block by its compressed signatures
                qsort( blocks[0].states, fins.counts[fins.cur], entSize, compSigComp );
             }
             // Performs (2c): Writes the new final block to disk
             fwrite_check( blocks[0].states, fins.counts[fins.cur]*entSize, fins.file );
             advanceFile( fins );
          }
       } while( curCol );
       
       // Close out the final queue files and free the data
       for( UINT32 count = 0; count < finEnts.size(); count++ )
       {
          if( finEnts[count].file )
          {
             free( finEnts[count].data );
             fclose( finEnts[count].file );
          }
       }
       
       // Report the points discovered
       log_string( 0, ( "Found %lld/%lld new points\n", myPointTotal, totPoints ) );  
    }

    // Performs (3) and (4): Compute the new BFS blocks from new final entries:
    // (3a) Use a priority queue over the new final blocks:
    //        Full states: sorted by compressed state
    //        Compressed states: sorted by path ordering
    // (3b) For each new entry, compute its full state:
    //        Full states: just copy
    //        Compressed states: compute the state from the path
    // (3c) Write the BFS blocks to disk
    // (4) Compressed: Sort the new final blocks by compressed state
    void finalToComp( )
    {
       vector <DDD_ENTRY> ents;
       DDD_ENTRY ent, *target;
       UINT64 totMults = 0, posMults = 0;
       
       // Performs (3a) Forms the priority queue
       for( UINT32 cur = fins.first; cur < fins.cur; cur++ )
       {
          assert( fins.maxSteps[cur]==fins.maxSteps[fins.first] );
          sprintf( ent.fname, "%s%s.%d.%d", archInt->FinalDirectory, fins.name, myRank, cur );
          fopen_check( ent.file, ent.fname, "r" );
          ent.size = getDataEntrySize( fins.maxSteps[cur] );
          malloc_check( ent.data, BYTE *, ent.size );
          ent.coll = false;
          ents.push_back( ent );
       }
       
       mergeQueue <DDD_ENTRY> 
          mq( ( getCompStateSize() < probInt->StateSize ? fileTreeComp : NULL ), 
              fileNext, 
              &(ents[0]), 
              ents.size() );

       BYTE compSig[getCompStateSize()];
       vector <PATH_NODE> nodes;
       nodes.resize( 1 );
       malloc_check( nodes[0].state, BYTE *, probInt->StateSize );
       probInt->getRoots( nodes[0].state );
       nodes[0].step = probInt->NumberOfGenerators;

       // Initialize the first block appropriately
       if( paths )
       {
          blocks[0].pathSize = pathInt->PATH_SIZE( fins.maxSteps[fins.first] );
          realloc_check( blocks[0].paths, SEARCH_PATH, blocks[0].pathSize * StatesPerBlock );
          blocks[0].stepMax = fins.maxSteps[fins.first];
       }
       blocks[0].length = 0;

       // Performs (3b): For each new entry
       log_string( 1, ( "Adding finals (%d-%d) to comp queue\n", fins.first, fins.cur-1 ) );
       while( target = mq.pop() )
       {
          if( paths )
             memcpy( getPath( 0, blocks[0].length ), 
                     target->data + getCompStateSize(), 
                     blocks[0].pathSize );
          
          // Perform (3b): Find its full state
          if( getCompStateSize() < probInt->StateSize )
          {
             UINT32 loc = addPathNode( getPath( 0, blocks[0].length ), nodes );
             assert( memcmp( target->data, 
                             getCompState( compSig, nodes[loc].state ),
                             getCompKeySize() ) == 0 );
             memcpy( getSig( 0, blocks[0].length ), nodes[loc].state, probInt->StateSize );
          }
          else
          {
             memcpy( getSig( 0, blocks[0].length ), target->data, probInt->StateSize );
          }
                  
          // Perform (3c) Write the BFS block to disk
          if( ++blocks[0].length==StatesPerBlock )
          {
             writeBlock( blocks, ass.tail );
             blocks[0].stepMax = fins.maxSteps[fins.first];
             blocks[0].length = 0;
          }
       }
       
       // Perform (3c): Write the BFS block to disk
       if( blocks[0].length ) writeBlock( blocks, ass.tail );
       
       // Perform (4): Sort the new final blocks by compressed state
       if( getCompStateSize() < probInt->StateSize )
       {
          // Clean out the state data allocated for the nodes
          for( UINT32 cur = 0; cur < nodes.size(); cur++ )
             free( nodes[cur].state );
          // Sort the blocks by compressed signature again
          for( UINT32 cur = fins.first; cur < fins.cur; cur++ )
          {
             log_string( 1, ( "Resorting finals (%d-%d) by compressed signature\n", fins.first, fins.cur-1 ) );
             sprintf( ent.fname, "%s%s.%d.%d", archInt->FinalDirectory, fins.name, myRank, cur );
             fopen_check( ent.file, ent.fname, "r" );
             fread_check( blocks[0].states, getDataEntrySize(fins.maxSteps[cur]) * fins.counts[cur], ent.file );
             fclose( ent.file );
             qsort( blocks[0].states, fins.counts[cur], getDataEntrySize(fins.maxSteps[cur]), compSigComp );
             fopen_check( ent.file, ent.fname, "w" );
             fwrite_check( blocks[0].states, getDataEntrySize(fins.maxSteps[cur]) * fins.counts[cur], ent.file);
             fclose( ent.file );
          }
       }
    }

    // Performs (3b) in the DDD
    // (3bi)  Finds the portion of the path that remains the same
    // (3bii) Performs remaining generator applications
    UINT32 addPathNode( SEARCH_PATH path, vector <PATH_NODE> &nodes )
    {
       PATH_NODE newPath = { NULL, probInt->NumberOfGenerators };
       UINT32 loc;
       UINT64 mults;
       BOOL success;
       
       // Perform (3bi): Find the longest identical prefix
       for( loc = 0; nodes[loc].step != probInt->NumberOfGenerators && 
                     loc < pathInt->GET_STEPS( path ); loc++ )
          if( nodes[loc].step != pathInt->GET_STEP( path, loc ) ) break;
       
       assert( loc < pathInt->GET_STEPS( path ) );
       assert( nodes[loc].step == probInt->NumberOfGenerators || 
               nodes[loc].step < pathInt->GET_STEP( path, loc ) );
       
       // Perform (3bii): Perform remaining generator applications
       for( ; loc < pathInt->GET_STEPS( path ); loc++ )
       {
          // Get additional path nodes if required
          if( nodes.size() <= loc+1 )
          {
             nodes.push_back( newPath );
             malloc_check( nodes[loc+1].state, BYTE *, probInt->StateSize );
          }
          // Perform the generator application and add it to the end of the path
          nodes[loc].step = pathInt->GET_STEP( path, loc );
          probInt->applyTargetGen( nodes[loc].state, nodes[loc].step, nodes[loc+1].state, &success );
          assert( success );
       }
       return loc;
    }

    // Completes the DDD
    BOOL finishDDD( UINT64 *totalPoints, UINT64 &localTotal, UINT64 &sumTotal )
    {
       UINT64 finTot = 0;
       BOOL finished = true;
       
       // Print out the information
       log_string( 0, ( "DDD finished with %lld points locally\n", totalPoints[myRank] ) );
       for( UINT8 count = 0; count < maxRank; count++ )
       {
           finTot += totalPoints[count];
           if( totalPoints[count] ) finished = false;
       }
       log_string( 0, ( "DDD finished with %lld points globally\n", finTot ) );
       localTotal += totalPoints[myRank];
       log_string( 0, ( "Discovered %lld points locally so far\n", localTotal ) );
       sumTotal += finTot;
       log_string( 0, ( "Discovered %lld points globally so far\n", sumTotal ) );
       myPointTotal = 0;
       return finished;
    }
    
public:
   // Sets the level of messages to be printed in the computation
   // Must be called before run
   void SetDebugLevel( UINT64 level )
   {
      this->DebugLevel = level;
   }
   
   // Sets the number of states in a BFS block
   // Must be called before run
   void SetStatesPerBlock( UINT64 states )
   {
      this->StatesPerBlock = states;
   }
   
   // Sets the number of hash entries per search entries
   // Must be called before run
   void SetHashMultiple( FLOAT64 multiple )
   {
      this->HashMultiple = multiple;
   }
    
   // Change the maximum number of steps the first path can be
   // Must be called before run
   void SetFirstBlockPathMax( UINT32 max )
   {
      FirstBlockPathMax = max;
   }

   // Runs the computation
   bool Run( ProblemInterface *problem, ArchitectureInterface *architecture, RUN_MODE run, int *pargc, char ***pargv )
   {
      pthread_t BFSMan, DDDMan;
      UINT32 numPasses = 0;
      UINT64 *totalPoints;
      UINT64 localTotal = 0, sumTotal = 0;
      BOOL finished = false;
      myPointTotal = 0;
      
      searchInt = this;
      archInt = architecture;
      probInt = problem;
      if( paths )
      {
         pathInt = new PathInterface( );
         pathInt->Initialize( problem->NumberOfGenerators );
      }
      else pathInt = NULL;
      ass.cur = ass.proc = ass.store = BLOCK_FREE;
      ass.head = ass.tail = 0;
      
      // Initialize the computation and compute the base of the tree
      totalPoints = new UINT64[maxRank];
      initialize( run, pargc, pargv );
      syncCall( myPointTotal );
      syncWait( totalPoints );
      finishInitialize( totalPoints, localTotal, sumTotal );
      
      do
      {
         // Perform the breadth-first search using a distributed imperfect hash
         if( pthread_create( &BFSMan, NULL, runBFS, NULL )!=0 ) assert( 0 );
         syncWait( totalPoints );
         finishBFS( totalPoints, localTotal, sumTotal );
         
         // Perform the disk-based duplicate detection
         if( pthread_create( &DDDMan, NULL, runDDD, NULL )!=0 ) assert( 0 );
         syncWait( totalPoints );
         finished = finishDDD( totalPoints, localTotal, sumTotal );
         
         // Update the current pass
         numPasses++;
      }
      while( !finished );
      
      // Close the open files and finalize the computation.
      log_string( 0, ( "Completed in %ld passes with %lld points locally, %lld globally\n", 
                  numPasses, localTotal, sumTotal ) );
      finalize( totalPoints );
      return 0;
   }
    
private:
   // Completes the computation
   void finalize( UINT64 *totalPoints )
   {
      UINT64 total = 0, count = 0;
      fclose( fins.file );
      fclose( cols.file );
      for( count = 0; count <= fins.cur; count++ )
         total += fins.counts[count];
      syncCall( total );
      syncWait( totalPoints );
      for( count = 0, total = 0; count < maxRank; count++ )
         total += totalPoints[count];       
      log_string( 0, ( "Discovered %lld points globally\n", total ) );
      syncCall( 0 );
      syncWait( totalPoints );
      close_log( );
      archInt->finalizeComm( );
   }
};

GeneralTieredDDD *GeneralTieredDDD::searchInt = NULL;

class TieredDDD : public GeneralTieredDDD
{
   UINT64 getCompStateSize( )
   {
      return probInt->StateSize;
   }
   
   UINT64 getCompKeySize( )
   {
      return probInt->KeySize;
   }
   
   BYTE *getCompState( BYTE *comp, BYTE *sig )
   {
      memcpy( comp, sig, getCompStateSize() );
      return comp;
   }
   
   UINT64 getPathAverageSize( )
   {
      return pathInt->PATH_SIZE(log(probInt->BranchingFactor, probInt->SearchSize));
   }

   UINT32 getDataEntrySize( UINT32 maxSteps )
   {
      return pathInt->PATH_SIZE( maxSteps ) + getCompStateSize();
   }

public:
   TieredDDD( ) : GeneralTieredDDD( )
   {
      paths = true;
   }
};

class TieredDDDNoPath : public GeneralTieredDDD
{
   UINT64 getCompStateSize( )
   {
      return probInt->StateSize;
   }
   
   UINT64 getCompKeySize( )
   {
      return probInt->KeySize;
   }
   
   BYTE *getCompState( BYTE *comp, BYTE *sig )
   {
      memcpy( comp, sig, getCompStateSize() );
      return comp;
   }
   
   UINT64 getPathAverageSize( )
   {
      return 0;
   }
   
   UINT32 getDataEntrySize( UINT32 maxSteps )
   {
      return getCompStateSize();
   }

public:
   TieredDDDNoPath( ) : GeneralTieredDDD( )
   {
      paths = false;
   }
};

class TieredDDDCompressed : public GeneralTieredDDD
{
   UINT32 CompSize;
   
   UINT64 getCompStateSize( )
   {
      if( !CompSize ) CompSize = ( ( 2 * log( 2, probInt->SearchSize ) / 24 + 1 ) * SIZE32 );
      return CompSize;
   }
   
   UINT64 getCompKeySize( )
   {
      if( !CompSize ) CompSize = ( ( 2 * log( 2, probInt->SearchSize ) / 24 + 1 ) * SIZE32 );
      return CompSize;
   }
   
   BYTE *getCompState( BYTE *comp, BYTE *sig )
   {
      memset( comp, '\0', getCompStateSize( ) );
      for( UINT32 count = 0; count < probInt->KeySize; count++ )
         comp[count%getCompStateSize( )] ^= sig[count];
      return comp;
   }
   
   UINT64 getPathAverageSize( )
   {
      return pathInt->PATH_SIZE(log(probInt->BranchingFactor, probInt->SearchSize));
   }

   UINT32 getDataEntrySize( UINT32 maxSteps )
   {
      return pathInt->PATH_SIZE( maxSteps ) + getCompStateSize();
   }

public:
   TieredDDDCompressed( ) : GeneralTieredDDD( )
   {
      paths = true;
      CompSize = 0;
   }
};

#endif /*SEARCH_H_*/

