/**************************************************************************
 *  enumerator - a program to enumerate large implicit search spaces      *
 *  Copyright (C) 2007-2008  Eric Robinson                                *
 *                                                                        *
 *  This program is free software: you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation, either version 3 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU General Public License     *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 **************************************************************************/

#ifndef TIVVECTOR
#define TIVVECTOR

#include <assert.h>
#include <iostream>
using namespace std;
#include "tivcheck.h"

template <class T> class vector  {
public:
   // Moving this here temporarily
   T *data;
   
   // Constructors and destructor

   // Default Constructor - Initialize with length 0 and no buffer
   vector( )
   {
      data = NULL;
      buffer = length = 0;
   }

   // Copy Constructor - Initialize given another vector
   vector(vector<T> const & source)
   {
      data = NULL;
      buffer = length = 0;
      *this = source;
   }

   // Initialize given a buffer size
   vector(UINT32 padding)
   {
      if( padding == 0 )
      {
         data = NULL;
         buffer = length = 0;
      }
      else
      {
         data = new T[padding]; assert( data );
         buffer = padding;
         length = 0;
      }
   }

   // Destructor - free the buffer
   ~vector()
   {
      if( data ) delete [] data;
      data = 0;
   }

   // access to elements via subscript
   T & operator [] (UINT32 index) const
   {
      assert( index < length );
      return data[index];
   }
   
   T & back ( ) const
   {
      assert( length );
      return data[length-1];
   }

   // assignment
   vector<T> & operator = (const vector<T> &right)
   {
      // Set length to 0 so we don't copy, we will overwrite anyways
      length = 0;
      reserve(right.buffer);
      length = right.length;

      // copy the elements
      for (UINT32 i = 0; i < length; i++)
         data[i] = right.data[i];

      return *this;
   }

   // length of vector
   UINT32   size() const
   {
      return length;
   }

   // holding capacity of the vector
   UINT32   capacity() const
   {
      // return the number of elements in the vector
      return buffer;
   }

   // dynamically change size of vector given a new size
   UINT32  resize(UINT32 numberOfElements)
   {
      if( numberOfElements==length )
      {
         return length;
      }
      else if( numberOfElements != 0 )
      {
         if( numberOfElements > buffer )
         {
            // first create the new data area
            T *newData;

            if( buffer == 0 ) buffer = 1;
            while( buffer < numberOfElements )
               buffer *= 2;
         
            newData = new T[buffer]; assert( newData );

            for ( UINT32 i = 0; i < numberOfElements && i < length; i++)
               newData[i] = data[i];

            // delete the old data buffer
            if( data ) delete [] data;

            data = newData;
         }

         // update the data member fields
         length = numberOfElements;
      }
      else
      {
         length = 0;
      }

      // return new size
      return length;
   }

   // dynamically change size given new size and an initializer
   UINT32  resize(UINT32 numberOfElements, const T &initialValue)
   {
      UINT32 oldSize = length;
      resize( numberOfElements );

      for( ; oldSize < numberOfElements; oldSize++ ) data[oldSize] = initialValue;

      // return new length
      return length;
   }

   // dynamically change space
   UINT32  reserve(UINT32 space)
   {
      assert( space>=length );

      if( space==buffer )
      {
         return buffer;
      }
      else if( space != 0 )
      {
         // first create the new data area
         T *newData;
         buffer = space;
      
        
         newData = new T[buffer]; assert( newData );

         for ( UINT32 i = 0; i < length; i++)
            newData[i] = data[i];

         // delete the old data buffer
         if( data ) delete [] data;

         data = newData;
      }
      else
      {
         buffer = 0;
      }

      // return new padding
      return buffer;
   }

   // add to end
   void push_back( const T &element )
   {
      resize( length + 1, element );
   }

    // add to end
   T pop_back( )
   {
      T retval = data[length - 1];
      resize( length - 1 );
      return retval;
   }

   // Operations for a sorted vector are below...

   // insert an item into a sorted vector
   UINT32 insert( const T &element, UINT32 (*index_fun)( const void * ) )
   {
      // Insert an element into a sorted vector
      UINT32 target = index_fun( &element );
      UINT32 location = find_location( target, index_fun, false );
      UINT32 i;

      if( ++length > buffer )
      {
         // first create the new data area
         T *newData;

         if( buffer == 0 ) buffer = 1;
         while( buffer < length )
            buffer *= 2;
      
         newData = new T[buffer]; assert( newData );

         // Insert in the correct place
         for (i = 0; i < location; i++)
            newData[i] = data[i];
         newData[i] = element;
         for(i++; i < length; i++)
            newData[i] = data[i-1];

         // delete the old data buffer
         if( data ) delete [] data;

         data = newData;
      }
      else
      {
         for(i = length - 1; i > location; i-- )
            data[i] = data[i-1];
         data[location] = element;
      }

      // return new size
      return length;
   }
   // remove an item from a sorted vector
   // void remove( UINT32 index, UINT32 (*index_fun)( const void * ) );
   // find an item in a sorted vector
   UINT32 find( UINT32 index, UINT32 (*index_fun)( const void * ) )
   {
      return find_location( index, index_fun, true );
   }

   // get an item from a sorted vector
   T & get( UINT32 index, UINT32 (*index_fun)( const void * ) )
   {
      UINT32 location = find_location( index, index_fun, true );
      assert( location!= length );
      return data[location];
   }

private:
   // Helper functions

   // Search for a given value to insert or return
   UINT32 find_location( UINT32 target, UINT32 (*index_fun)( const void * ), bool absolute )
   {
      UINT32 low = 1;
      UINT32 high = length;
      UINT32 mid, current;

      while (low <= high)
      {
         mid = (low + high) / 2;
         current = index_fun( &data[mid - 1] );

         if (current == target) return mid - 1;
         if (current <  target) low  = mid + 1;
         else                   high = mid - 1;
      }

      if( absolute ) return length;
      else return low - 1;
   }

   // data areas
   UINT32  length;
   UINT32  buffer;
};

template <class T> ostream & operator << (ostream & out, const vector <T> vec)
{
   UINT32 length = vec.length ();
   UINT32 i;

   for (i = 0; i < length; i++)
      out << vec [i] << ", ";

   out << "**END**" << endl;

   return out;
}

#endif
