#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "cdll.h"

/*
 * Interface for a circularly-doubly-linked-list package.
 *
 * These types carry no payload, so they are meant to be embedded in the data
 * structure proper.
 */

#define CDLL_HEADER_MARKER	(void*)0xf100dbad
#define CDLL_ELEMENT_MARKER	(void*)0xbadf100d

typedef struct ele_tag *ele_p;

/*
 * A circularly-doubly-linked-list header.
 */
typedef struct hdr_tag {
  ele_p	next;
  ele_p	prev;
  void	*hmark;
} hdr_t, *hdr_p;

/*
 * A circularly-doubly-linked-list element.
 */
typedef struct ele_tag {
  ele_p	next;
  ele_p	prev;
  void	*emark;
} ele_t;

void cdll_header_init( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  hdr->next = (ele_p)hdr;
  hdr->prev = (ele_p)hdr;
  hdr->hmark = CDLL_HEADER_MARKER;
}

void cdll_header_destroy( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );
  hdr->next = NULL;
  hdr->prev = NULL;
  hdr->hmark = NULL;
}

void cdll_element_init( cdll_element_p e ) {
  ele_p ele = (ele_p)e;
  ele->next = ele;
  ele->prev = ele;
  ele->emark = CDLL_ELEMENT_MARKER;
}

void cdll_element_destroy( cdll_element_p e ) {
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );
  ele->next = NULL;
  ele->prev = NULL;
  ele->emark = NULL;
}

int cdll_empty( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );
  return ( hdr->next == hdr->prev ) && ( hdr->next == (ele_p)hdr );
}

void cdll_insert_tail( cdll_header_p h, cdll_element_p e ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );

  ele_p p = hdr->prev;

  p->next = ele;
  ele->prev = p;

  hdr->prev = ele;
  ele->next = (ele_p)hdr;
}

void cdll_insert_head( cdll_header_p h, cdll_element_p e ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );

  ele_p n = hdr->next;

  n->prev = ele;
  ele->next = n;

  hdr->next = ele;
  ele->prev = (ele_p)hdr;
}

cdll_element_p cdll_remove( cdll_element_p e ) {
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );
  assert( ele->next != NULL );
  assert( ele->prev != NULL );

  ele_p n = ele->next;
  ele_p p = ele->prev;

  p->next = n;
  n->prev = p;

  ele->next = ele->prev = NULL;
  return e;
}

cdll_element_p cdll_dequeue_head( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );

  if ( cdll_empty( h ) ) {
    return (cdll_element_p)NULL;
  } else {
    return cdll_remove( (cdll_element_p)hdr->next );
  }
}

cdll_element_p cdll_dequeue_tail( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );

  if ( cdll_empty( h ) ) {
    return (cdll_element_p)NULL;
  } else {
    return cdll_remove( (cdll_element_p)hdr->prev );
  }
}

cdll_element_p cdll_head( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );

  return (cdll_element_p)hdr->next;
}

cdll_element_p cdll_tail( cdll_header_p h ) {
  hdr_p	hdr = (hdr_p)h;
  assert( hdr->hmark == CDLL_HEADER_MARKER );

  return (cdll_element_p)hdr->prev;
}

cdll_element_p cdll_next( cdll_element_p e ) {
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );

  return (cdll_element_p)ele->next;
}

cdll_element_p cdll_prev( cdll_element_p e ) {
  ele_p ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );

  return (cdll_element_p)ele->prev;
}

void cdll_insert_before( cdll_element_p e, cdll_element_p n ) {
  ele_p	ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );
  ele_p new = (ele_p)n;
  assert( new->emark == CDLL_ELEMENT_MARKER );

  ele_p p = ele->prev;

  p->next = new;
  new->prev = p;

  ele->prev = new;
  new->next = ele;
}

void cdll_insert_after( cdll_element_p e, cdll_element_p n ) {
  ele_p	ele = (ele_p)e;
  assert( ele->emark == CDLL_ELEMENT_MARKER );
  ele_p new = (ele_p)n;
  assert( new->emark == CDLL_ELEMENT_MARKER );

  ele_p nx = ele->next;

  nx->prev = new;
  new->next = nx;

  ele->next = new;
  new->prev = ele;
}

#ifdef STANDALONE
static void testInternalSizes( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  assert( sizeof(cdll_header_t) == sizeof(hdr_t) );
  assert( sizeof(cdll_element_t) == sizeof(ele_t) );

  printf("%s: end\n", __func__);
}

static void testHeaderInit( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr0 = { 0 };
  cdll_header_init( &hdr0 );
  assert( cdll_empty( &hdr0 ) );

  cdll_header_t hdr1 = { 0 };
  cdll_header_init( &hdr1 );
  assert( cdll_empty( &hdr1 ) );

  cdll_header_p hdrp = (cdll_header_p)malloc( sizeof(cdll_header_t) );
  cdll_header_init( hdrp );
  assert( cdll_empty( hdrp ) );
  free( hdrp );

  hdrp = (cdll_header_p)calloc( 1, sizeof(cdll_header_t) );
  cdll_header_init( hdrp );
  assert( cdll_empty( hdrp ) );
  free( hdrp );

  printf("%s: end\n", __func__);
}

static void testHeaderDestroy( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr1 = { 0 };
  cdll_header_init( &hdr1 );
  cdll_header_destroy( &hdr1 );

  cdll_header_p hdrp = (cdll_header_p)malloc( sizeof(cdll_header_t) );
  cdll_header_init( hdrp );
  cdll_header_destroy( hdrp );
  free( hdrp );

  printf("%s: end\n", __func__);
}

static void testElementInit( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_element_t ele0 = { 0 };
  cdll_element_init( &ele0 );
  assert( cdll_next( &ele0 ) == cdll_prev( &ele0 ) );
  assert( cdll_next( &ele0 ) == &ele0 );

  cdll_element_t ele1 = { 0 };
  cdll_element_init( &ele1 );
  assert( cdll_next( &ele1 ) == cdll_prev( &ele1 ) );
  assert( cdll_next( &ele1 ) == &ele1 );

  cdll_element_p elep = (cdll_element_p)malloc( sizeof(cdll_element_t) );
  cdll_element_init( elep );
  assert( cdll_next( elep ) == cdll_prev( elep ) );
  assert( cdll_next( elep ) == elep );
  free( elep );

  elep = (cdll_element_p)calloc( 1, sizeof(cdll_element_t) );
  cdll_element_init( elep );
  assert( cdll_next( elep ) == cdll_prev( elep ) );
  assert( cdll_next( elep ) == elep );
  free( elep );

  printf("%s: end\n", __func__);
}

static void testElementDestroy( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_element_t ele1 = { 0 };
  cdll_element_init( &ele1 );
  cdll_element_destroy( &ele1 );

  cdll_element_p elep = (cdll_element_p)malloc( sizeof(cdll_element_t) );
  cdll_element_init( elep );
  cdll_element_destroy( elep );
  free( elep );

  printf("%s: end\n", __func__);
}

static void testEmpty( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr = { 0 };
  cdll_header_init( &hdr );
  assert( cdll_empty( &hdr ) );
  cdll_element_t ele = { 0 };
  cdll_element_init( &ele );
  cdll_insert_head( &hdr, &ele );
  assert( !cdll_empty( &hdr ) );
  cdll_header_destroy( &hdr );
  cdll_element_destroy( &ele );

  printf("%s: end\n", __func__);
}

typedef struct item_tag {
  cdll_element_t	cdll;
  long			val;
} item_t, *item_p;

static item_p itemAlloc(void) {
  item_p i = (item_p)calloc( 1, sizeof(item_t) );
  assert( NULL != i );
  cdll_element_init( &i->cdll );
  return i;
}

static void itemFree( item_p i ) {
  cdll_element_destroy( &i->cdll );
  free( i );
}

static void testInsertDequeueStack( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr = { 0 };
  cdll_header_init( &hdr );
  assert( cdll_empty( &hdr ) );

  int i;
  for ( i = 1; i <= 100; i++ ) {
    item_p item = itemAlloc();
    item->val = i;
    cdll_insert_head( &hdr, &item->cdll );
    assert( !cdll_empty( &hdr ) );
    }

  for ( i = 100; i >= 1; i-- ) {
    item_p item = (item_p)cdll_dequeue_head( &hdr );
    assert( item->val == i );
    itemFree( item );
    }

  assert( cdll_empty( &hdr ) );

  for ( i = 1; i <= 100; i++ ) {
    item_p item = itemAlloc();
    item->val = i;
    cdll_insert_tail( &hdr, &item->cdll );
    assert( !cdll_empty( &hdr ) );
    }

  for ( i = 100; i >= 1; i-- ) {
    item_p item = (item_p)cdll_dequeue_tail( &hdr );
    assert( item->val == i );
    itemFree( item );
    }

  assert( cdll_empty( &hdr ) );
  cdll_header_destroy( &hdr );

  printf("%s: end\n", __func__);
}

static void testInsertDequeueFifo( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr = { 0 };
  cdll_header_init( &hdr );
  assert( cdll_empty( &hdr ) );

  int i;
  for ( i = 1; i <= 100; i++ ) {
    item_p item = itemAlloc();
    item->val = i;
    cdll_insert_head( &hdr, &item->cdll );
    assert( !cdll_empty( &hdr ) );
    }

  for ( i = 1; i <= 100; i++ ) {
    item_p item = (item_p)cdll_dequeue_tail( &hdr );
    assert( item->val == i );
    itemFree( item );
    }

  assert( cdll_empty( &hdr ) );

  for ( i = 1; i <= 100; i++ ) {
    item_p item = itemAlloc();
    item->val = i;
    cdll_insert_tail( &hdr, &item->cdll );
    assert( !cdll_empty( &hdr ) );
    }

  for ( i = 1; i <= 100; i++ ) {
    item_p item = (item_p)cdll_dequeue_head( &hdr );
    assert( item->val == i );
    itemFree( item );
    }

  assert( cdll_empty( &hdr ) );
  cdll_header_destroy( &hdr );

  printf("%s: end\n", __func__);
}

static void testWalking( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr = { 0 };
  cdll_header_init( &hdr );
  assert( cdll_empty( &hdr ) );

  item_p item;
  int i;
  for ( i = 1; i <= 100; i++ ) {
    item = itemAlloc();
    item->val = i;
    cdll_insert_head( &hdr, &item->cdll );
    assert( !cdll_empty( &hdr ) );
    }

  assert( !cdll_empty( &hdr ) );

  item = (item_p)cdll_head( &hdr );
  for ( i = 100; ( item != (item_p)&hdr ); i-- ) {
    assert( i >= 1 );
    assert( item->val == i );
    item = (item_p)cdll_next( &item->cdll );
  }

  item = (item_p)cdll_tail( &hdr );
  for ( i = 1; ( item != (item_p)&hdr ); i++ ) {
    assert( i <= 100 );
    assert( item->val == i );
    item = (item_p)cdll_prev( &item->cdll );
  }

  item = (item_p)cdll_head( &hdr );
  assert( (item_p)&hdr == (item_p)cdll_prev( &item->cdll ) );

  item = (item_p)cdll_tail( &hdr );
  assert( (item_p)&hdr == (item_p)cdll_next( &item->cdll ) );

  item = (item_p)cdll_tail( &hdr );
  for ( i = 1; !cdll_empty( &hdr ); i++ ) {
    assert( i <= 100 );
    assert( item->val == i );
    assert( item == (item_p)cdll_remove( &item->cdll ) );
    itemFree( item );
    item = (item_p)cdll_tail( &hdr );
    }
  assert( (item_p)&hdr == item );

  assert( cdll_empty( &hdr ) );
  cdll_header_destroy( &hdr );

  printf("%s: end\n", __func__);
}

static void testInsertBeforeAfter( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);

  cdll_header_t hdr = { 0 };
  cdll_header_init( &hdr );
  assert( cdll_empty( &hdr ) );

  item_p pivot = itemAlloc();
  pivot->val = 3;

  cdll_insert_head( &hdr, &pivot->cdll );

  item_p item = itemAlloc();
  item->val = 1;
  cdll_insert_before( &pivot->cdll, &item->cdll );

  item = itemAlloc();
  item->val = 5;
  cdll_insert_after( &pivot->cdll, &item->cdll );

  item = itemAlloc();
  item->val = 2;
  cdll_insert_before( &pivot->cdll, &item->cdll );

  item = itemAlloc();
  item->val = 4;
  cdll_insert_after( &pivot->cdll, &item->cdll );

  int i;
  item = (item_p)cdll_head( &hdr );
  for ( i = 1; ( item != (item_p)&hdr ); i++ ) {
    assert( item->val == i );
    item = (item_p)cdll_next( &item->cdll );
    }
  assert( 6 == i );

  item = (item_p)cdll_tail( &hdr );
  for ( i = 5; ( item != (item_p)&hdr ); i-- ) {
    assert( item->val == i );
    item = (item_p)cdll_prev( &item->cdll );
    }
  assert( 0 == i );

  for ( i = 1; i <= 5; i++ ) {
    item = (item_p)cdll_dequeue_head( &hdr );
    assert( item->val == i );
    itemFree( item );
    }

  assert( cdll_empty( &hdr ) );
  cdll_header_destroy( &hdr );

  printf("%s: end\n", __func__);
}

static void testTemplate( int argc, char *argv[] ) {
  printf("%s: begin\n", __func__);
  printf("%s: end\n", __func__);
}

int main( int argc, char *argv[] ) {
  testInternalSizes( argc, argv );
  testHeaderInit( argc, argv );
  testHeaderDestroy( argc, argv );
  testElementInit( argc, argv );
  testElementDestroy( argc, argv );
  testEmpty( argc, argv );
  testInsertDequeueStack( argc, argv );
  testInsertDequeueFifo( argc, argv );
  testWalking( argc, argv );
  testInsertBeforeAfter( argc, argv );
  testTemplate( argc, argv );
  return 0;
}
#endif
