 /**********************************************************************
  * MPINU				                               *
  * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include "mpi.h"
#include "mpiimpl.h"

#define NUM_SLAVES_TO_CREATE_AT_ONCE 5

typedef struct timed_system_tag {
  char			*cmd;
  pthread_mutex_t	mu;
  pthread_cond_t	cv;
  int			done;
} timed_system_t, *timed_system_p;

/*
 * Thread routine which performs the actual system() call.
 */
static void *do_system( void *arg ) {
  timed_system_p params = (timed_system_p)arg;

  assert( 0 == system( params->cmd ) );

  assert( 0 == pthread_mutex_lock( &params->mu ) );
  params->done = TRUE;
  assert( 0 == pthread_cond_signal( &params->cv ) );
  assert( 0 == pthread_mutex_unlock( &params->mu ) );
  return NULL;
}

static void exec_system_with_interrupt( char *ssh_cmd, char *hostname, char *command ) {
  char *argv[4], probe_string[MPINU_MAX_PROCESSOR_NAME+100];

  if ( 0 != strcmp("localhost", hostname) ) {

    int len = snprintf( probe_string, sizeof (probe_string),
	    "exec %s %s pwd \">\" /dev/null",
	    ssh_cmd, hostname );
    if (len > sizeof (probe_string)) {
	printf ("Error: buffer overflow, formatted string longer than "
		"buffer (%d > %d)\n", len, sizeof (probe_string));
	exit (1);
	}

    pthread_t		systhd;
    timed_system_t	params;
    params.cmd = probe_string;
    params.done = FALSE;
    assert( 0 == pthread_mutex_init( &params.mu, NULL ) );
    assert( 0 == pthread_cond_init( &params.cv, NULL ) );

    assert( 0 == pthread_create( &systhd, NULL, do_system, (void *)&params ) );

    struct timespec	timeout;
    timeout.tv_nsec = 0;
    timeout.tv_sec = time(NULL) + 15;

    assert( 0 == pthread_mutex_lock( &params.mu ) );
    while ( !params.done ) {
      int stat = pthread_cond_timedwait( &params.cv, &params.mu, &timeout );
      if ( stat == ETIMEDOUT ) {
	fprintf(stderr, "*** MPINU:  Slave %s didn't reply to `%s'.\n",
            hostname, ssh_cmd);
	fflush( stderr );
	exit( 1 );
      } else if ( stat != 0 ) {
	fprintf( stderr, "pthread_cond_timedwait() failed with %d (%s)\n",
	    stat, strerror( stat ) );
	exit( stat );
      }
    }
    assert( 0 == pthread_mutex_unlock( &params.mu ) );

    assert( 0 == pthread_join( systhd, NULL ) );
  }

  argv[0] = "sh";
  argv[1] = "-c";
  argv[2] = command;
  argv[3] = NULL;
  /* It works to call alarm() here, if desired. */
  execv("/bin/sh", argv);
  exit(127);
}

int MPINU_set_and_exec_cmds(char *hostname, char *port, int argc, char *argv[], char *outfile)
{
  char cmd[256], args[256], *args_ptr;
  char *ssh_cmd;
  int i, fd, args_len;
  pid_t pid;
  FILE *ssh_script;

  ssh_script = fopen(outfile, "w");
  fprintf(ssh_script, "#!/bin/sh\n");
  if ( (ssh_cmd = getenv("SSH")) == NULL ) ssh_cmd = "ssh";
  if ( getenv("CALLBACK_HOST") != NULL ) hostname = getenv("CALLBACK_HOST");

  args[0] = '\0';
  args_ptr = args;
  args_len = sizeof (args);
  for (i=1; i < argc; i++) {
      int len = snprintf (args_ptr, args_len, "%s ", argv[i]);
      if (len > args_len) {
	  fprintf (stderr, "warning: args not all copied!\n");
	  break;
	  }
      args_ptr += len;
      args_len -= len;
      }
  for(i=1; i < PG_ARRAY_SIZE && MPINU_pg_array[i].processor != NULL; i++)
  {
    /* If localhost, avoid overhead and portability issues of local loop */
    /* Strictly speaking, should be snprintf(cmd, sizeof(cmd), "cd; exec ...");
       but then: localhost 1 - >slave.out   creates slave.out in home dir. */
    if ( 0 == strcmp("localhost", MPINU_pg_array[i].processor) )
	snprintf(cmd, sizeof (cmd), "exec %s %s %s %s -p4amslave",
	      MPINU_pg_array[i].process, args, hostname, port);
    /* Assumes user shell on remote processor accepts "exec" */
    else snprintf(cmd, sizeof (cmd), "exec %s %s exec %s %s %s %s -p4amslave", ssh_cmd,
		 MPINU_pg_array[i].processor, MPINU_pg_array[i].process,
		 args, hostname, port);
    fprintf(ssh_script, "%s &\n", cmd);
    if ( i % NUM_SLAVES_TO_CREATE_AT_ONCE == 0 )
      fprintf(ssh_script, "sleep(1)\n", cmd);
    fflush(ssh_script);
#ifdef DEBUG
    printf("cmd:  %s\n",cmd);
#endif
    CALL_CHK( pid = fork, () );
    if (pid == (pid_t)0) {  /* if child process */

      fclose(ssh_script);
      /* Could have used "ssh -n" to redirect stdin, but position of "-n"
         in "ssh" varies according to different dialects. */
#ifndef STDIN_FILENO
#define STDIN_FILENO 0 /* Should be standard in most dialects -- just in case */
#endif
      close(STDIN_FILENO);
      CALL_CHK( fd = open, ("/dev/null", O_RDONLY) );
      if ( fd != STDIN_FILENO ) { /* fd should be stdin, but check to be safe */
        CALL_CHK( dup2, (fd, STDIN_FILENO) );
        close( fd );
      }

#ifdef DEBUG
      printf ("[i=%d] ssh_cmd = %s, processor = %s, cmd = %s\n", i, ssh_cmd, MPINU_pg_array[i].processor, cmd);
#endif
      exec_system_with_interrupt(ssh_cmd, MPINU_pg_array[i].processor, cmd);
      exit(127); /* child shouldn't reach here */
    }
    if ( i % NUM_SLAVES_TO_CREATE_AT_ONCE == 0 )
      sleep(1);
    cmd[0] = '\0';
  }
  fclose(ssh_script);
  if ( i >= PG_ARRAY_SIZE )
    printf("MPINU: WARNING: more slaves than PG_ARRAY_SIZE(%d); not using all\n",
           PG_ARRAY_SIZE);
  return i - 1; /* return number of slaves */
}
