 /**********************************************************************
  * MPINU				                               *
  * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

/* Collective communication implemented by blocking send and receive
   for point-to-point communication.  For an Ethernet-style LAN, it
   would be difficult to do better without significant extra complication.
   These routines use MPI_COLL_COMM_TAG, which is not available to end-user
   programs.  */

#include "mpi.h"
#include "mpiimpl.h"


/* function used directly by MPI_Reduce and MPI_Scan */
/* this function applies op to sendbuf and recvbuf, and the result of
   the operation based on the datatype is written back on recvbuf */
void exec_op ( sendbuf, recvbuf, count, datatype, op )
     void *sendbuf;
     void *recvbuf;
     int count;
     MPI_Datatype datatype;
     MPI_Op op;
{
  int i;

  switch (op)
    {

    case MPI_MIN:
      switch (datatype)
	{
	case MPI_INT:
	  for ( i=0; i < count; i++ )
	    {
	      if (((int*)sendbuf)[i] < ((int*)recvbuf)[i] )
		((int*)recvbuf)[i] = ((int*)sendbuf)[i];
	    }
	  break;

	case  MPI_UNSIGNED:
	  for ( i=0; i < count; i++ )
            {
              if (((unsigned int*)sendbuf)[i] < ((unsigned int*)recvbuf)[i] )
                ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i];
            }
	  break;

	case MPI_SHORT:
	  for ( i=0; i < count; i++ )
            {
              if (((short*)sendbuf)[i] < ((short*)recvbuf)[i])
                ((short*)recvbuf)[i] = ((short*)sendbuf)[i];
            }
	  break;

	case MPI_UNSIGNED_SHORT:
	  for ( i=0; i < count; i++ )
            {
              if (((unsigned short*)sendbuf)[i] < ((unsigned short*)recvbuf)[i])
                ((unsigned short*)recvbuf)[i] = ((unsigned short*)sendbuf)[i];
            }
	  break;

	case MPI_LONG:
          for ( i=0; i < count; i++ )
            {
              if (((long*)sendbuf)[i] < ((long*)recvbuf)[i])
                ((long*)recvbuf)[i] = ((long*)sendbuf)[i];
            }
	  break;

	case MPI_UNSIGNED_LONG:
	  for ( i=0; i < count; i++ )
            {
              if (((unsigned long*)sendbuf)[i] < ((unsigned long*)recvbuf)[i])
                ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i];
            }
	  break;

	case MPI_FLOAT:
	  for ( i=0; i < count; i++ )
	    {
	      if ( ((float*)sendbuf)[i] < ((float*)recvbuf)[i] )
		((float*)recvbuf)[i] = ((float*)sendbuf)[i];
	    }
	  break;

	case MPI_DOUBLE:
	  for ( i=0; i < count; i++ )
            {
              if ( ((double*)sendbuf)[i] < ((double*)recvbuf)[i] )
                ((double*)recvbuf)[i] = ((double*)sendbuf)[i];
            }
	  break;

	case MPI_LONG_DOUBLE:
	  for ( i=0; i < count; i++ )
            {
              if ( ((long double*)sendbuf)[i] < ((long double*)recvbuf)[i] )
                ((long double*)recvbuf)[i] = ((long double*)sendbuf)[i];
            }
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
	}
      break;

    case MPI_MAX:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            {
              if (((int*)sendbuf)[i] > ((int*)recvbuf)[i] )
                ((int*)recvbuf)[i] = ((int*)sendbuf)[i];
            }
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            {
              if (((unsigned int*)sendbuf)[i] > ((unsigned int*)recvbuf)[i]
		  )
                ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i];
            }

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            {
              if (((short*)sendbuf)[i] > ((short*)recvbuf)[i])
                ((short*)recvbuf)[i] = ((short*)sendbuf)[i];
            }
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            {
              if (((unsigned short*)sendbuf)[i] > ((unsigned
          short*)recvbuf)[i])
                ((unsigned short*)recvbuf)[i] = ((unsigned
          short*)sendbuf)[i];
            }
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            {
              if (((long*)sendbuf)[i] > ((long*)recvbuf)[i])
                ((long*)recvbuf)[i] = ((long*)sendbuf)[i];
            }
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            {
              if (((unsigned long*)sendbuf)[i] > ((unsigned
          long*)recvbuf)[i])
                ((unsigned long*)recvbuf)[i] = ((unsigned
          long*)sendbuf)[i];
            }
	  break;

        case MPI_FLOAT:
          for ( i=0; i < count; i++ )
            {
              if ( ((float*)sendbuf)[i] > ((float*)recvbuf)[i] )
                ((float*)recvbuf)[i] = ((float*)sendbuf)[i];
            }
	  break;

        case MPI_DOUBLE:
          for ( i=0; i < count; i++ )
            {
              if ( ((double*)sendbuf)[i] > ((double*)recvbuf)[i] )
                ((double*)recvbuf)[i] = ((double*)sendbuf)[i];
            }
	  break;

        case MPI_LONG_DOUBLE:
          for ( i=0; i < count; i++ )
            {
              if ( ((long double*)sendbuf)[i] >
		   ((long double*)recvbuf)[i] )
                ((long double*)recvbuf)[i] = ((long double*)sendbuf)[i];
            }
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_PROD:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
                ((int*)recvbuf)[i] = ((int*)sendbuf)[i] * ((int*)recvbuf)[i];
	  break;

	case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
	    ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] *
	      ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
	    ((short*)recvbuf)[i] = ((short*)sendbuf)[i] * ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
                ((unsigned short*)recvbuf)[i] = ((unsigned
						  short*)sendbuf)[i] *
		  ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
	    ((long*)recvbuf)[i] = ((long*)sendbuf)[i] * ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
	    ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] *
	      ((unsigned long*)recvbuf)[i];
	  break;

        case MPI_FLOAT:
          for ( i=0; i < count; i++ )
                ((float*)recvbuf)[i] = ((float*)sendbuf)[i] * ((float*)recvbuf)[i];
	  break;

        case MPI_DOUBLE:
          for ( i=0; i < count; i++ )
	    ((double*)recvbuf)[i] = ((double*)sendbuf)[i] * ((double*)recvbuf)[i];
	  break;

        case MPI_LONG_DOUBLE:
          for ( i=0; i < count; i++ )
	    ((long double*)recvbuf)[i] = ((long double*)sendbuf)[i] *
	      ((long double*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_SUM :
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i<count; i++ )
	    ((int*)recvbuf)[i] = ((int*)sendbuf)[i] + ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] +
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] +
          ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
	    ((unsigned short*)recvbuf)[i] = ((unsigned
					      short*)sendbuf)[i] +
	      ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] +
          ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] +
              ((unsigned long*)recvbuf)[i];
	  break;

        case MPI_FLOAT:
          for ( i=0; i < count; i++ )
	    ((float*)recvbuf)[i] = ((float*)sendbuf)[i] + ((float*)recvbuf)[i];
	  break;

        case MPI_DOUBLE:
          for ( i=0; i < count; i++ )
            ((double*)recvbuf)[i] = ((double*)sendbuf)[i] +
          ((double*)recvbuf)[i];
	  break;

        case MPI_LONG_DOUBLE:
          for ( i=0; i < count; i++ )
            ((long double*)recvbuf)[i] = ((long double*)sendbuf)[i] +
              ((long double*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_LAND:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            ((int*)recvbuf)[i] = ((int*)sendbuf)[i] && ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] &&
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] &&
	      ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            ((unsigned short*)recvbuf)[i] = ((unsigned
                                              short*)sendbuf)[i] &&
              ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] &&
	      ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] &&
              ((unsigned long*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
	}
      break;

    case MPI_LOR:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            ((int*)recvbuf)[i] = ((int*)sendbuf)[i] || ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] ||
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] ||
	      ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            ((unsigned short*)recvbuf)[i] = ((unsigned
                                              short*)sendbuf)[i] ||
              ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] ||
	      ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] ||
              ((unsigned long*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_LXOR:
      /* we reduce each element in recvbuf and sendbuf to its logical
      value before we do a bitwise xor; that is if element is not equal to
      0 then, it is true(equal to 1) */
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
	    {
	      if ( ((int*)recvbuf)[i] != 0 )
		((int*)recvbuf)[i] = 1;
	      if ( ((int*)sendbuf)[i] != 0 )
		((int*)sendbuf)[i] = 1;
	      ((int*)recvbuf)[i] =  ((int*)sendbuf)[i] ^
		((int*)recvbuf)[i];
	    }
          break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
	    {
	      if ( ((unsigned int*)recvbuf)[i] != 0 )
		((unsigned int*)recvbuf)[i] = 1;
	      if ( ((unsigned int*)sendbuf)[i] != 0 )
		((unsigned int*)sendbuf)[i] = 1;

	      ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] ^
		((unsigned int*)recvbuf)[i];
	    }
          break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            {
	      if ( ((short*)recvbuf)[i] != 0 )
		((short*)recvbuf)[i] = 1;
	      if ( ((short*)sendbuf)[i] != 0 )
		((short*)sendbuf)[i] = 1;
	      ((short*)recvbuf)[i] =  ((short*)sendbuf)[i] ^
		((short*)recvbuf)[i];
            }
          break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            {
	      if ( ((unsigned short*)recvbuf)[i] != 0 )
                ((unsigned short*)recvbuf)[i] = 1;
              if ( ((unsigned short*)sendbuf)[i] != 0 )
                ((unsigned short*)sendbuf)[i] = 1;
              ((unsigned short*)recvbuf)[i] =  ((unsigned short*)sendbuf)[i] ^
                ((unsigned short*)recvbuf)[i];
            }
          break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            {
	      if ( ((long*)recvbuf)[i] != 0 )
                ((long*)recvbuf)[i] = 1;
              if ( ((long*)sendbuf)[i] != 0 )
                ((long*)sendbuf)[i] = 1;
              ((long*)recvbuf)[i] =  ((long*)sendbuf)[i] ^
                ((long*)recvbuf)[i];
            }
          break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            {
	      if ( ((unsigned long*)recvbuf)[i] != 0 )
                ((unsigned long*)recvbuf)[i] = 1;
              if ( ((unsigned long*)sendbuf)[i] != 0 )
                ((unsigned long*)sendbuf)[i] = 1;
              ((unsigned long*)recvbuf)[i] =  ((unsigned
						 long*)sendbuf)[i] ^
                ((unsigned long*)recvbuf)[i];

            }
          break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_BAND:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            ((int*)recvbuf)[i] = ((int*)sendbuf)[i] & ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] &
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] &
              ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            ((unsigned short*)recvbuf)[i] = ((unsigned
                                              short*)sendbuf)[i] &
              ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] &
              ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] &
              ((unsigned long*)recvbuf)[i];
	  break;

	case MPI_CHAR:
	  for ( i=0; i < count; i++ )
            ((char*)recvbuf)[i] = ((char*)sendbuf)[i] &
	      ((char*)recvbuf)[i];
	  break;

	case MPI_UNSIGNED_CHAR:
	case MPI_BYTE:
	  for ( i=0; i < count; i++ )
            ((unsigned char*)recvbuf)[i] = ((unsigned char*)sendbuf)[i] &
	      ((unsigned char*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_BOR:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            ((int*)recvbuf)[i] = ((int*)sendbuf)[i] | ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] |
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] |
              ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            ((unsigned short*)recvbuf)[i] = ((unsigned
                                              short*)sendbuf)[i] |
              ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] |
              ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] |
              ((unsigned long*)recvbuf)[i];
	  break;

	case MPI_CHAR:
	  for ( i=0; i < count; i++ )
            ((char*)recvbuf)[i] = ((char*)sendbuf)[i] |
	      ((char*)recvbuf)[i];
	  break;

	case MPI_UNSIGNED_CHAR:
	case MPI_BYTE:
	  for ( i=0; i < count; i++ )
            ((unsigned char*)recvbuf)[i] = ((unsigned char*)sendbuf)[i] |
	      ((unsigned char*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_BXOR:
      switch (datatype)
        {
        case MPI_INT:
          for ( i=0; i < count; i++ )
            ((int*)recvbuf)[i] = ((int*)sendbuf)[i] ^ ((int*)recvbuf)[i];
	  break;

        case  MPI_UNSIGNED:
          for ( i=0; i < count; i++ )
            ((unsigned int*)recvbuf)[i] = ((unsigned int*)sendbuf)[i] ^
              ((unsigned int*)recvbuf)[i];
	  break;

        case MPI_SHORT:
          for ( i=0; i < count; i++ )
            ((short*)recvbuf)[i] = ((short*)sendbuf)[i] ^
              ((short*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_SHORT:
          for ( i=0; i < count; i++ )
            ((unsigned short*)recvbuf)[i] = ((unsigned
                                              short*)sendbuf)[i] ^
              ((unsigned short*)recvbuf)[i];
	  break;

        case MPI_LONG:
          for ( i=0; i < count; i++ )
            ((long*)recvbuf)[i] = ((long*)sendbuf)[i] ^
              ((long*)recvbuf)[i];
	  break;

        case MPI_UNSIGNED_LONG:
          for ( i=0; i < count; i++ )
            ((unsigned long*)recvbuf)[i] = ((unsigned long*)sendbuf)[i] ^
              ((unsigned long*)recvbuf)[i];
	  break;

	case MPI_CHAR:
	  for ( i=0; i < count; i++ )
            ((char*)recvbuf)[i] = ((char*)sendbuf)[i] ^
	      ((char*)recvbuf)[i];
	  break;

	case MPI_UNSIGNED_CHAR:
	case MPI_BYTE:
	  for ( i=0; i < count; i++ )
            ((unsigned char*)recvbuf)[i] = ((unsigned char*)sendbuf)[i] ^
	      ((unsigned char*)recvbuf)[i];
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
        }
      break;

    case MPI_MAXLOC:
      switch (datatype)
	{
	case MPI_INT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct int_int*)sendbuf)[i].value > ((struct int_int*)recvbuf)[i].value)
		{
		  ((struct int_int*)recvbuf)[i].value = ((struct
							  int_int*)sendbuf)[i].value;
		  ((struct int_int*)recvbuf)[i].index = ((struct
							  int_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_SHORT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct short_int*)sendbuf)[i].value > ((struct short_int*)recvbuf)[i].value)
		{
		  ((struct short_int*)recvbuf)[i].value = ((struct
							    short_int*)sendbuf)[i].value;
		  ((struct short_int*)recvbuf)[i].index = ((struct
							    short_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_LONG_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct long_int*)sendbuf)[i].value > ((struct long_int*)recvbuf)[i].value)
		{
		  ((struct long_int*)recvbuf)[i].value = ((struct
							   long_int*)sendbuf)[i].value;
		  ((struct long_int*)recvbuf)[i].index = ((struct
							   long_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_FLOAT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct float_int*)sendbuf)[i].value > ((struct float_int*)recvbuf)[i].value)
		{
		  ((struct float_int*)recvbuf)[i].value = ((struct
							    float_int*)sendbuf)[i].value;
		  ((struct float_int*)recvbuf)[i].index = ((struct
							    float_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_DOUBLE_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct double_int*)sendbuf)[i].value > ((struct double_int*)recvbuf)[i].value)
		{
		  ((struct double_int*)recvbuf)[i].value = ((struct
							    double_int*)sendbuf)[i].value;
		  ((struct double_int*)recvbuf)[i].index = ((struct
							    double_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_LONG_DOUBLE_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct long_double_int*)sendbuf)[i].value > ((struct long_double_int*)recvbuf)[i].value)
		{
		  ((struct long_double_int*)recvbuf)[i].value = ((struct
							    long_double_int*)sendbuf)[i].value;
		  ((struct long_double_int*)recvbuf)[i].index = ((struct
							    long_double_int*)sendbuf)[i].index;
		}
            }
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
	}
      break;

    case MPI_MINLOC:
       switch (datatype)
	{
	case MPI_INT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct int_int*)sendbuf)[i].value < ((struct int_int*)recvbuf)[i].value)
		{
		  ((struct int_int*)recvbuf)[i].value = ((struct
							    int_int*)sendbuf)[i].value;
		  ((struct int_int*)recvbuf)[i].index = ((struct
							    int_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_SHORT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct short_int*)sendbuf)[i].value < ((struct short_int*)recvbuf)[i].value)
		{
		  ((struct short_int*)recvbuf)[i].value = ((struct
							    short_int*)sendbuf)[i].value;
		  ((struct short_int*)recvbuf)[i].index = ((struct
							    short_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_LONG_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct long_int*)sendbuf)[i].value < ((struct long_int*)recvbuf)[i].value)
		{
		  ((struct long_int*)recvbuf)[i].value = ((struct
							    long_int*)sendbuf)[i].value;
		  ((struct long_int*)recvbuf)[i].index = ((struct
							    long_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_FLOAT_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct float_int*)sendbuf)[i].value < ((struct float_int*)recvbuf)[i].value)
		{
		  ((struct float_int*)recvbuf)[i].value = ((struct
							    float_int*)sendbuf)[i].value;
		  ((struct float_int*)recvbuf)[i].index = ((struct
							    float_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_DOUBLE_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct double_int*)sendbuf)[i].value < ((struct double_int*)recvbuf)[i].value)
		{
		  ((struct double_int*)recvbuf)[i].value = ((struct
							    double_int*)sendbuf)[i].value;
		  ((struct double_int*)recvbuf)[i].index = ((struct
							    double_int*)sendbuf)[i].index;
		}
            }
	  break;

	case MPI_LONG_DOUBLE_INT:
	  for ( i=0; i < count; i++ )
            {
              if (((struct long_double_int*)sendbuf)[i].value < ((struct long_double_int*)recvbuf)[i].value)
		{
		  ((struct long_double_int*)recvbuf)[i].value = ((struct
							    long_double_int*)sendbuf)[i].value;
		  ((struct long_double_int*)recvbuf)[i].index = ((struct
							    long_double_int*)sendbuf)[i].index;
		}
            }
	  break;

	default:
	  fprintf(stderr, "MPI_MIN: invalid data type (%#lx).\n", datatype);
	  exit(1);
	}
       break;

    default:
      fprintf(stderr, "exec_op: invalid operation (%#lx).\n", op);
      exit(1);
    }
}
