#include "cache.h"

/*
 * Define a key-value pair element that is stored in the cache.
 */
typedef struct key_val_tag *key_val_p;
typedef struct key_val_tag {
    buf_key_t	key;
    buf_obj_t	val;
    key_val_p	next;
    } key_val_t;

typedef struct cache_tag {
    key_val_p	list;
    buf_key_t	lastKey;
    int (*cmp)(buf_key_t, buf_key_t);
    } cache_t, *cache_p;

/*
 * Initialize a cache buffer, with the routine to be called to compare objects
 * and keys. Returns zero on success, non-zero on failure.
 */
int buf_init (buf_p newCache, int (*cmp)(buf_key_t, buf_key_t))
    {
    cache_p	cache = (cache_p)calloc (1, sizeof (cache_t));
    if (NULL == cache)
	return ENOMEM;
    cache->cmp = cmp;
    *newCache = (buf_t)cache;
    return 0;
    }

/*
 * Associate a key with an object. Returns zero on success, non-zero on failure.
 */
int buf_put (buf_t cache, buf_key_t key, buf_obj_t obj)
    {
    key_val_p	pair = (key_val_p)malloc (sizeof (key_val_t));
    if (NULL == pair)
	return ENOMEM;
    pair->key = key;
    pair->val = obj;

    cache_p	iCache = cache;
    pair->next = iCache->list;
    iCache->list = pair;
    return 0;
    }

/*
 * Lookup a kay in the cache; returns TRUE if found, FALSE otherwise.
 */
int buf_lookup (buf_t cache, buf_key_t key)
    {
    cache_p	iCache = cache;
    key_val_p	ptr = iCache->list;

    while (ptr) {
	if (*(iCache->cmp) (key, ptr->key))
	    return TRUE;
	}
    return FALSE;
    }

/*
 * Pull a buffer from the cache matching the given key.
 *
 * If key is found and object is returned, then object must be the oldest
 * cache element of the cache having that key.
 */
int buf_get (buf_t cache, buf_key_t key, buf_obj_p obj)
    {
    cache_p	iCache = cache;
    key_val_p	ptr = iCache->list;
    key_val_p	pptr = NULL;

    while (ptr) {
	if (*(iCache->cmp) (key, ptr->key)) {
	    *obj = ptr->obj;
	    if (NULL == pptr)
		iCache->list = ptr->next;
	    else
		pptr->next = ptr->next;
	    ptr->next = NULL;
	    free (ptr);
	    return TRUE;
	    }
	pptr = ptr;
	ptr = ptr->next;
	}
    return 0;
    }

/*
 * Tear down a buffer cache, releasing any and all resources used to maintain
 * the cache.
 *
 * Return zero on success, non-zero on failure.
 */
int buf_destroy (buf_t cache, void (*freeObj)(buf_obj_t))
    {
    cache_p	iCache = cache;
    key_val_p	ptr = iCache->list;
    key_val_p	pptr;

    iCache->list = NULL;
    while (ptr) {
	(*freeObj)(ptr->obj);
	pptr = ptr;
	ptr = ptr->next;
	pptr->next = NULL;
	free (ptr);
	}
    return 0;
    }
