 /**********************************************************************
  * MPINU				                               *
  * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include "mpi.h"
#include "mpiimpl.h"

/* Record the version number of this particular build. */
static long long MPINU_version = MPINU_VERSION;

/* index into MPINU_pg_array is same as process rank in MPI_Comm_world */
pg_struct_t MPINU_pg_array[PG_ARRAY_SIZE];

int MPINU_myrank = -1;
int MPINU_num_slaves = -1;
static int ssh_slaves = -1; /* num orig slaves, not counting MPI_Spawn2() */
int MPINU_my_list_sd = -1;
fd_set MPINU_fdset;
int MPINU_max_sd = -1;
int MPINU_is_spawn2 = 0; /* Set inside mpi_spawn2 before unexec, then reset */
		/* Used by slave.c to determine if slave from mpi_spawn2 */
int MPINU_is_initialized = 0;
int MPINU_thread_level = 0;
pthread_t MPINU_thread_main = (pthread_t)NULL;
pthread_mutex_t MPINU_global = PTHREAD_MUTEX_INITIALIZER;
int MPINU_self_pipe = -1;
void *MPINU_anysrc_waitlist = NULL;

static const int datatype_size[MPI_MAX_DATATYPE+1] = {-1,
			     sizeof(char), sizeof(unsigned char),
			     sizeof(char),
			     sizeof(short), sizeof(unsigned short),
			     sizeof(int),
			     sizeof(unsigned), sizeof(long),
			     sizeof(unsigned long),
			     sizeof(float), sizeof(double),
			     sizeof(long double), sizeof(long long),
			     sizeof(struct short_int),
			     sizeof(struct int_int),
			     sizeof(struct long_int),
			     sizeof(struct float_int),
			     sizeof(struct double_int),
			     sizeof(struct long_double_int),
			     -1};

int MPINU_datatype_valid( MPI_Datatype datatype ) {
  return ( datatype > MPI_MIN_DATATYPE ) && ( datatype < MPI_MAX_DATATYPE );
}

int MPINU_datatype_size( MPI_Datatype datatype ) {
  assert( MPINU_datatype_valid( datatype ) );
  return datatype_size[(int)datatype];
}

int MPINU_msg_size( int count, MPI_Datatype datatype ) {
  assert( MPINU_datatype_valid( datatype ) );
  return count * datatype_size[(int)datatype];
}

static sho_t	sho;

static void null_handler( int signum, void *arg ) {
  /*
   * Now that this handler is called from a thread, not in an interrupt
   * context, we can raise the signal to let any installed handler take
   * action.
   */
  int i;
  for ( i = 0; i <= MPINU_num_slaves; i++ ) {
    MPINU_dump_cache( MPINU_pg_array[i].cache, MPINU_myrank, i );
  }
  raise( signum );
}

/*
 * Initialize the receive buffer data structures for each slave and the
 * master.
 */
static void init_recv_cache( void )
    {
    int i;
    for ( i = 0; i <= MPINU_num_slaves; i++ ) {
 	MPINU_pg_array[i].cache = MPINU_cache_init(
	    MPINU_pg_array[i].cache );
	}
    }

/*
 * Tear down the receive buffer data structures for each slave and the master.
 */
static void finalize_recv_cache( void )
    {
    int i;
    for ( i = 0; i <= MPINU_num_slaves; i++ ) {
 	MPINU_pg_array[i].cache = MPINU_cache_destroy(
		MPINU_pg_array[i].cache );
	}
    }

/*
 * Initialize the waitlist data structures for each slave and the
 * master.
 */
static void init_waitlists( void )
    {
    int i;
    for ( i = 0; i <= MPINU_num_slaves; i++ ) {
 	MPINU_pg_array[i].waitlist = MPINU_waitlist_init(
	    MPINU_pg_array[i].waitlist );
	}
    MPINU_anysrc_waitlist = MPINU_waitlist_init(
	MPINU_anysrc_waitlist );
    }

/*
 * Tear down the data structures for each slave and the master.
 */
static void finalize_waitlists( void )
    {
    int i;
    for ( i = 0; i <= MPINU_num_slaves; i++ ) {
 	MPINU_pg_array[i].waitlist = MPINU_waitlist_destroy(
		MPINU_pg_array[i].waitlist );
	}
    MPINU_anysrc_waitlist = MPINU_waitlist_destroy(
	MPINU_anysrc_waitlist );
    }

/*
 * Initialize the receive statistics data structures for each slave and the
 * master.
 */
static void init_recv_stats( void ) {
  int i;
  for ( i = 0; i <= MPINU_num_slaves; i++ )
    memset( &MPINU_pg_array[i].stats, 0, sizeof(MPINU_pg_array[i].stats) );
}

/*
 * Tear down the receive statistics data structures for each slave and the
 * master.
 */
static void finalize_recv_stats( void ) {
#ifdef DEBUG
  int i;
  for ( i = 0; i <= MPINU_num_slaves; i++ ) {
    pg_stats_p	st = &MPINU_pg_array[i].stats;
    if ( st->find_attempts )
      fprintf( stderr, "Rank %d: find_attempts[%d]: %d\n", MPINU_myrank, i, st->find_attempts );
    if ( st->find_blocking_attempts )
      fprintf( stderr, "Rank %d: find_blocking_attempts[%d]: %d\n", MPINU_myrank, i, st->find_blocking_attempts );
    if ( st->found_in_cache_first_try )
      fprintf( stderr, "Rank %d: found_in_cache_first_try[%d]: %d\n", MPINU_myrank, i, st->found_in_cache_first_try );
    if ( st->general_wake )
      fprintf( stderr, "Rank %d: general_wake[%d]: %d\n", MPINU_myrank, i, st->general_wake );
    if ( st->general_wakeup )
      fprintf( stderr, "Rank %d: general_wakeup[%d]: %d\n", MPINU_myrank, i, st->general_wakeup );
    if ( st->match_any_wakeup )
      fprintf( stderr, "Rank %d: match_any_wakeup[%d]: %d\n", MPINU_myrank, i, st->match_any_wakeup );
    if ( st->match_wakeup )
      fprintf( stderr, "Rank %d: match_wakeup[%d]: %d\n", MPINU_myrank, i, st->match_wakeup );
    if ( st->max_reader_on_attempts )
      fprintf( stderr, "Rank %d: max_reader_on_attempts[%d]: %d\n", MPINU_myrank, i, st->max_reader_on_attempts );
    if ( st->max_single_blocking_attempts )
      fprintf( stderr, "Rank %d: max_single_blocking_attempts[%d]: %d\n", MPINU_myrank, i, st->max_single_blocking_attempts );
    if ( st->max_spurious_wakeups )
      fprintf( stderr, "Rank %d: max_spurious_wakeups[%d]: %d\n", MPINU_myrank, i, st->max_spurious_wakeups );
    if ( st->not_found )
      fprintf( stderr, "Rank %d: not_found[%d]: %d\n", MPINU_myrank, i, st->not_found );
    if ( st->reader_on_first_try )
      fprintf( stderr, "Rank %d: reader_on_first_try[%d]: %d\n", MPINU_myrank, i, st->reader_on_first_try );
    if ( st->reader_any_wakeup )
      fprintf( stderr, "Rank %d: reader_any_wakeup[%d]: %d\n", MPINU_myrank, i, st->reader_any_wakeup );
    if ( st->reader_wakeup )
      fprintf( stderr, "Rank %d: reader_wakeup[%d]: %d\n", MPINU_myrank, i, st->reader_wakeup );
    if ( st->spurious_wakeups )
      fprintf( stderr, "Rank %d: spurious_wakeups[%d]: %d\n", MPINU_myrank, i, st->spurious_wakeups );
  }
#endif
}

pthread_cond_t *MPINU_rank_recv_cv_get( int rank ) {
  return &MPINU_pg_array[rank].recv_cond;
}

pthread_mutex_t *MPINU_rank_recv_mu_get( int rank ) {
  return &MPINU_pg_array[rank].recv_mutex;
}

pthread_mutex_t *MPINU_rank_send_mu_get( int rank ) {
  return &MPINU_pg_array[rank].send_mutex;
}

pthread_mutex_t *MPINU_rank_sock_mu_get( int rank ) {
  return &MPINU_pg_array[rank].sock_mutex;
}

void *MPINU_rank_cache_get( int rank ) {
  return MPINU_pg_array[rank].cache;
}

void *MPINU_rank_waitlist_get( int rank ) {
  return MPINU_pg_array[rank].waitlist;
}

pg_stats_p MPINU_rank_stats_get( int rank ) {
  return &MPINU_pg_array[rank].stats;
}

int MPINU_rank_reader_equal( int rank, pthread_t reader ) {
  return pthread_equal( MPINU_pg_array[rank].reader, reader );
}

void MPINU_rank_reader_set( int rank, pthread_t reader ) {
  MPINU_pg_array[rank].reader = reader;
}

int MPINU_rank_anysrc_tag_get(int rank) {
  return MPINU_pg_array[rank].anysrc_tag;
}

void MPINU_rank_anysrc_tag_set(int rank, int tag) {
  MPINU_pg_array[rank].anysrc_tag = tag;
}

int MPI_Init_thread( int *argc_ptr, char ***argv_ptr, int required,
    int *provided ) {
  char *p4pg_file = "procgroup";
  int i, p4pg_flag = 0;

#if (MPI_ANY_SOURCE != -1) || (MPI_ANY_TAG != -1) || (MPINU_NOT_A_TAG != -2) || (MPINU_INIT_TAG != -3) || (MPINU_COLL_COMM_TAG != -19)
  printf( "mpi: MPI_Init_thread: inconsistent C constant macros\n" );
  exit( 1 );
#endif

#ifdef DEBUG
  system( "hostname" );
  printf( "entering MPI_Init\n" );
  printf( " argc: %d\n", *argc_ptr );
  printf( "last argv: %s\n", (*argv_ptr)[*argc_ptr - 1] );
  fflush( stdout );
#endif

  if ( sizeof(INT) != 4 ) {
    printf( "sizeof(INT) != 4:  re-define INT in mpiimpl.h\n" );
    exit( 1 );
  }

  if ( MPINU_is_initialized ) {
    printf( "MPI_Init:  can't call MPI_Init after it's already started.\n" );
    exit( 1 );
  }

  /*
   * Give them single threaded behavior of they ask for it, otherwise they get
   * threads full bore!
   */
  switch ( required ) {
    case MPI_THREAD_SINGLE:
      MPINU_thread_level = MPI_THREAD_SINGLE;
      break;
    case MPI_THREAD_FUNNELED:
    case MPI_THREAD_SERIALIZED:
    case MPI_THREAD_MULTIPLE:
      MPINU_thread_level = MPI_THREAD_MULTIPLE;
      break;
    default:
      return MPI_FAIL;
  }

  FD_ZERO( &MPINU_fdset ); /* initialize, MPINU_new_listener() will modify */

  for( i = 0; i < PG_ARRAY_SIZE; i++ ) {
    /* initialize all entries */
    MPINU_pg_array[i].processor = NULL;
    MPINU_pg_array[i].num_threads = NULL;
    MPINU_pg_array[i].process = NULL;
    assert( 0 == pthread_mutex_init( &MPINU_pg_array[i].sock_mutex, NULL ) );
    MPINU_rank_socket_init( i );
    assert( 0 == pthread_mutex_init( &MPINU_pg_array[i].send_mutex, NULL ) );
#ifdef SENDRECV_RINGBUFF
    MPINU_pg_array[i].send_ops_idx = 0;
    memset( MPINU_pg_array[i].sendring, 0,
	sizeof(MPINU_pg_array[i].sendring)  );
#endif
    assert( 0 == pthread_mutex_init( &MPINU_pg_array[i].recv_mutex, NULL ) );
    assert( 0 == pthread_cond_init( &MPINU_pg_array[i].recv_cond, NULL ) );
#ifdef SENDRECV_RINGBUFF
    MPINU_pg_array[i].recv_ops_idx = 0;
    memset( MPINU_pg_array[i].recvring, 0,
	sizeof(MPINU_pg_array[i].recvring) );
#endif
    MPINU_pg_array[i].reader = NO_READER;	/* initialized below */
    MPINU_pg_array[i].anysrc_tag = MPINU_NOT_A_TAG;
    MPINU_pg_array[i].cache = NULL;		/* initialized below */
    MPINU_pg_array[i].waitlist = NULL;		/* initialized below */
  }

  if ( 0 != strcmp( (*argv_ptr)[*argc_ptr - 1], "-p4amslave" )
       && ! MPINU_is_spawn2 ) {
    for( i = 0; i < *argc_ptr; i++ ) {
      if ( p4pg_flag )
	(*argv_ptr)[i] = (*argv_ptr)[i+2];
      else if ( ! strcmp( (*argv_ptr)[i], "-p4pg" ) ) {
        p4pg_file = (*argv_ptr)[i+1];
        i--;
        *argc_ptr -= 2;
        p4pg_flag = 1;
      }
    }
    if ( p4pg_flag ) (*argv_ptr)[*argc_ptr] = NULL;
#ifdef DEBUG
    printf( "p4pg_file: %s\n", p4pg_file );fflush( stdout );
#endif
    { struct stat buf;
      if ( 0 != stat( p4pg_file, &buf ) || ! buf.st_mode & S_IFREG
           || ! buf.st_mode & S_IRUSR ) {
        fprintf( stderr, "*** MPINU:  can't read procgroup file:  %s\n",
                 p4pg_file );
        if (  0 != strcmp( p4pg_file, "procgroup" ) ) {
	  fprintf( stderr,
		   "  Either create \"progroup\" in current directory:\n"
		   "    %s/procgroup\n",
		   getcwd( NULL, 256 ) );
	  fprintf( stderr,
		   "  or else add a command line arg:"
                      "  -p4pg  ABSOLUTE_PATH_OF_PROCGROUP_FILE\n");
        }
	return MPI_FAIL; 
      }
    }
    MPINU_mpi_master( p4pg_file, *argc_ptr, *argv_ptr );
    ssh_slaves = MPINU_num_slaves;
  }
  else {  /* else slave */
#ifdef DEBUG
    printf( "calling MPINU_mpi_slave(%s, %s): %d\n",
	   (*argv_ptr)[*argc_ptr - 3], (*argv_ptr)[*argc_ptr - 2] );
    fflush( stdout );
#endif
    if ( MPINU_is_spawn2 ) MPINU_mpi_slave( NULL, NULL );
    else {
      MPINU_mpi_slave( (*argv_ptr)[*argc_ptr - 3],
	  (*argv_ptr)[*argc_ptr - 2] );

      // Change command line arg:  -p4amslave => -p4rank:XX
      snprintf( (*argv_ptr)[*argc_ptr - 1],
	        strlen ((*argv_ptr)[*argc_ptr - 1]) + 1,
	        "-p4rank:%02d", MPINU_myrank%100 );

      *argc_ptr -= 3;
    }
  }
  MPINU_is_initialized = 1;

  /*
   * Initialize the signal handling object facility. This also takes care of
   * setting up the various signal handling functions.
   */
  sho_init ();

  /*
   * Allocate and activate a signal handler for SIGINT, which will simply
   * raise() SIGINT in a separate thread rather than having it run the handler
   * a random thread picked by the OS.
   */
  sho = sho_alloc( SIGINT, null_handler, NULL );
  assert( NULL != sho );
  assert( 0 == sho_activate( sho ) );

  init_recv_cache();
  init_waitlists();
  init_recv_stats();

#if 1
  // Send message now to each slave, to initialize all-to-all connections
  // Replace this, when allowing connections to be set up on demand works.
  // To do that, have to add logic for case when two slaves each try
  //   to connect to each other at same time.  Can have rule like
  //   lower slave connects, higher slave accepts.  See file, TODO
  if ( MPINU_myrank > 0 ) {
    int buf = 42;
    MPI_Status status;
    // Lower rank connects to higher rank; Higher rank process calls accept()
    // MPINU_Int_Send() will force all calls to connect()
    //printf( "buf: %x\n", &buf );
    if ( MPINU_myrank < MPINU_num_slaves )
      for ( i = MPINU_myrank+1; i <= MPINU_num_slaves; i++ )
        MPINU_Int_Send( &buf, 1, MPI_INT, i, MPINU_INIT_TAG, MPI_COMM_WORLD );
    // MPINU_Int_Recv() will force all calls to accept()
    if ( MPINU_myrank > 1 )
      for ( i = 1; i < MPINU_myrank; i++ ) {
	int count;
	int rbuf;
        MPINU_Int_Recv(&rbuf, 1, MPI_INT, i, MPINU_INIT_TAG, MPI_COMM_WORLD,
	    &status);
	assert( status.MPI_SOURCE == i );
	assert( status.MPI_TAG == MPINU_INIT_TAG );
	MPI_Get_count( &status, MPI_INT, &count );
	assert( 1 == count );
	assert( buf == rbuf );
        MPINU_Int_Send( &buf, 1, MPI_INT, i, MPINU_INIT_TAG, MPI_COMM_WORLD );
      }
    // Confirm to connector that accept() is complete
    if ( MPINU_myrank < MPINU_num_slaves )
      for ( i = MPINU_myrank+1; i <= MPINU_num_slaves; i++ ) {
	int count;
	int rbuf;
        MPINU_Int_Recv(&rbuf, 1, MPI_INT, i, MPINU_INIT_TAG, MPI_COMM_WORLD,
	    &status);
	assert( status.MPI_SOURCE == i );
	assert( status.MPI_TAG == MPINU_INIT_TAG );
	MPI_Get_count( &status, MPI_INT, &count );
	assert( 1 == count );
	assert( buf == rbuf );
      }
  }
#endif
  MPINU_thread_main = pthread_self();
  // This implementation always provides the same level of support, whatever
  // that currently is.
  *provided = MPINU_thread_level;
  return MPI_SUCCESS;
}

int MPI_Is_thread_main( int *flag )
{
  if ( MPINU_thread_level == MPI_THREAD_MULTIPLE )
    *flag = pthread_equal( MPINU_thread_main, pthread_self() );
  else
    *flag = TRUE;
  return MPI_SUCCESS;
}

int MPI_Query_thread( int *provided )
{ *provided = MPINU_thread_level;
  return MPI_SUCCESS;
}

int MPI_Init( int *argc_ptr, char ***argv_ptr )
{ int dummy;
  return MPI_Init_thread( argc_ptr, argv_ptr, MPI_THREAD_SINGLE, &dummy );
}

int MPI_Initialized( int *flag )
{ *flag = MPINU_is_initialized;
  return MPI_SUCCESS;
}

int MPI_Finalize( void )
{ int i, statusp;
  MPI_Status status;

  /*
   * Perform an internal barrier to make sure all the processes enter
   * MPI_Finalize() before completing. The key here is that as we receive a
   * message from another node, we close the connection and deregister it so
   * that the MPI_ANY_SOURCE receive operations will not consider looking at
   * that source.
   *
   * FIX-ME: We need to deal with ANY_SOURCE checking our own internal socket.
   * So remove it first before we begin.
   */
  i = MPINU_rank_socket_get( MPINU_myrank );
  MPINU_rank_socket_set( MPINU_myrank, PG_NOSOCKET );
  close( i );

  for ( i = 0; i <= MPINU_num_slaves; i++ ) {
    if ( i == MPINU_myrank )
      continue;

    int val = MPINU_myrank | 0xbe1eef00;
    MPINU_Int_Send( &val, 1, MPI_INT, i, MPINU_FINAL_TAG,
	MPI_COMM_WORLD );
  }

  /*
   * FIX-ME: For all slaves, no assertions should kill the process from this
   * point forward. Not sure how to accomplish that. PAJP.
   */
  MPINU_send_thread_exit();

  for ( i = 0; i <= MPINU_num_slaves; i++ ) {
    if ( i == MPINU_myrank )
      continue;

    int r, cnt;
    MPI_Status status;

    MPINU_Int_Recv( &r, 1, MPI_INT, MPI_ANY_SOURCE, MPINU_FINAL_TAG,
	MPI_COMM_WORLD, &status );

    assert( status.MPI_SOURCE >= 0
	&& status.MPI_SOURCE <= MPINU_num_slaves
	&& status.MPI_SOURCE != MPINU_myrank );
    assert( MPINU_FINAL_TAG == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &cnt );
    assert( 1 == cnt );
#ifdef DEBUG
    fprintf( stderr, "rank: %d, i: %d, val: %#x\n", MPINU_myrank, i, r );
#endif
    r &= ~0xbe1eef00;
    assert( r == status.MPI_SOURCE );

    int s = MPINU_rank_socket_get( r );
    if ( s != PG_NOSOCKET ) {
      MPINU_rank_socket_set( r, PG_NOSOCKET );
      MPINU_dump_sockopt( s );
      if ( close( s ) )
	perror( "MPINU: MPI_Finalize: close" ); // but don't exit
    }
  }

  MPINU_is_initialized = 0;
  for ( i = 0; i <= MPINU_num_slaves; i++ ) {
    int s = MPINU_rank_socket_get( i );
    if ( s != PG_NOSOCKET ) {
      MPINU_dump_sockopt( s );
      if ( close( s ) )
	perror( "MPINU: MPI_Finalize: close" ); // but don't exit
    }
    if ( i > 0 ) {
      free( MPINU_pg_array[i].processor );
      free( MPINU_pg_array[i].num_threads );
      free( MPINU_pg_array[i].process );
      }
    assert( 0 == pthread_mutex_destroy( &MPINU_pg_array[i].sock_mutex ) );
    assert( 0 == pthread_mutex_destroy( &MPINU_pg_array[i].send_mutex ) );
    assert( 0 == pthread_mutex_destroy( &MPINU_pg_array[i].recv_mutex ) );
    assert( 0 == pthread_cond_destroy( &MPINU_pg_array[i].recv_cond ) );
    MPINU_pg_array[i].reader = NO_READER;
    }

  finalize_recv_stats();
  finalize_waitlists();
  finalize_recv_cache();
  assert( 0 == sho_deactivate( sho ) );
  sho_free( sho );
  sho_finalize();

  if ( close( MPINU_my_list_sd ) )
    perror( "MPINU: MPI_Finalize: close" ); // but don't exit
  if ( MPINU_myrank == 0 ) /* master waits for slaves to finish */
    for ( i = 1; i <= ssh_slaves; i++ )
      CALL_CHK( wait, (&statusp) );
  return MPI_SUCCESS;
}

/* Fill in all three on exit */
int MPINU_new_listener( int *sd, struct sockaddr_in *sin, int *port )
{ char host[256];		/* hostname for this process      */
  struct hostent *hp;           /* ptr to host info structure     */
  socklen_t socklen;
  int i;

  /* use the default protocol 6 = IPPROTO_TCP */
  *sd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
  MPINU_dump_sockopt( *sd );
  MPINU_setsockopt( *sd );
  /* Add new socket descriptor to set of all socket descriptors */
  FD_SET( *sd, &MPINU_fdset );
  if ( *sd > MPINU_max_sd )
    MPINU_max_sd = *sd;
#ifdef DEBUG
  printf( "MPINU_new_listener: socket: %d\n", *sd );fflush( stdout );
#endif

  CALL_CHK( gethostname, (host, sizeof(host)) );
  hp = gethostbyname( host );
  if ( NULL == hp ) {
    fprintf( stderr, "%s: gethostbyname(\"%s\") failed [%d](\"%s\")\n",
	__func__, host, errno, strerror( errno ) );
    abort();
  }

#if 0
  memset( (char *)sin, (char)0, sizeof(*sin) );
#else
  for ( i = 0; i < (signed)sizeof(*sin); i++ )
    ((char *)sin)[i] = 0;
#endif

  /* Is this next line (h_addr) needed? */
  memcpy( (char *)&(sin->sin_addr), hp->h_addr, hp->h_length );
  sin->sin_port = htons( 0 );
  sin->sin_family = hp->h_addrtype;
  /* Since INADDR_ANY is 0, zeroing out sin accomplished this, anyway. */
  sin->sin_addr.s_addr = INADDR_ANY;

#if 0
  printf( "sin_family (%d) and AF_INET (%d) should compare.\n",
      sin->sin_family, AF_INET );
  printf( "master: hp->h_name: %s\n", hp->h_name );
#endif

  do {
    sin->sin_addr.s_addr = INADDR_ANY;
    CALL_CHK( bind, (*sd, (struct sockaddr *)sin, sizeof(*sin)) );
    /* CALL_CHK sets errno = 0; O/S sometimes assigns a stale address */
  } while( errno == EADDRINUSE );

  /* i acts as dummy for sin_len below */
  socklen = sizeof(struct sockaddr_in);
  CALL_CHK( getsockname, (*sd, (struct sockaddr *)sin, &socklen) );
  *port = ntohs( sin->sin_port );
#ifdef DEBUG
  printf( "MPINU_new_listener: port: %d == %d; sin_len: %d\n",
	 ntohs( sin->sin_port ), *port, i ); fflush( stdout );
#endif
  if ( sin->sin_port == 0 ) {
    printf( "LISTENER FAILED TO GET NEW PORT!!\n" );
    exit( 1 );
  }

  /* Under SunOS 4.1, SOMAXCONN = 5; Could have slaves re-try after timeout */
  CALL_CHK( listen, (*sd, SOMAXCONN) );
  /* On Solaris 2.6, getsockname (and maybe bind?) zero out sin->sin_addr
    (converting it to localhost for efficiency?); So copy it back in */
  memcpy( (char *)&(sin->sin_addr), hp->h_addr, hp->h_length );
  return 0;
}
