/* This is the private portion of the MPI include file */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>  /* Needed for struct timeval for select */
#include <time.h>    /* Needed for struct timespec for nanosleep */
#include <netdb.h>
#include <netinet/in.h>
#ifdef __hpux
# include <arpa/inet.h>
#endif
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h> /* Needed with fcntl.h for Sys V */
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <sys/times.h>
#if defined(__APPLE__)
# include <sys/syslimits.h>
#endif
#include "signal-handling.h"

// Default is HAVE_PTHREAD.  Need to   #define HAVE_PTHREAD 0   to disable
#ifndef HAVE_PTHREAD
# define HAVE_PTHREAD 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif

#ifndef MAX
#define MAX(_a_,_b_)	((_a_) >= (_b_) ? (_a_) : (_b_))
#endif

#if HAVE_PTHREAD
# include <pthread.h>
# if defined(__APPLE__)
#  include <semaphore.h> /* See note in BUG for why this isn't used. */
# else
#  include "sem-pthread.h"
# endif
#else
# define pthread_mutex_t int /* any pthread_mutex_t won't be used anyway. */
# define PTHREAD_MUTEX_INITIALIZER 0
# define pthread_mutex_lock(x) 0 /* success return value */
# define pthread_mutex_unlock(x) 0 /* success return value */
# define sem_t int /* any sem_t won't be used anyway. */
# define sem_init(x,y,z) 0 /* success return value */
# define sem_wait(x) 0 /* success return value */
# define sem_post(x) 0 /* success return value */
#endif

/* Want to disguise `"' because ANSI C won't do substitution inside quotes.
 * #ifdef __STDC__
 *  #define quote(arg) #arg
 * #else
 *   #define quote(arg) "arg"
 * #endif
 * THEN USE:    perror(quote(function))  */
#define ANSIC ( -1 < (unsigned char) 1 )
#define PERROR(__function) if (ANSIC) \
    {fprintf(stderr,"%s:%d:%s(): ",__FILE__, __LINE__, __func__); perror(#__function);} \
          else perror("function")
/*
#if ANSIC
#define PERROR(function) fprintf(stderr,__FILE__ "(line %d): ", __LINE__); \
			 perror(#function)
#else
#define PERROR(function) perror("function")
#endif
*/

#ifndef assert_perror
/* We use assert_perror with system calls.  Never set NDEBUG. */
# ifdef NDEBUG
#  error "NDEBUG WAS SET.  IT SHOULD NOT BE SET."
# endif
# define assert_perror(x) \
     { if ( (long int)(x) == 0L ) { \
       char str[1000]; \
       snprintf( str, sizeof (str), "%s:%s:%d", __FILE__, __func__, __LINE__ ); \
       perror( str ); \
       fprintf( stderr, "  " ); \
       assert( #x == NULL ); /* Force assert to trigger */ } }
#endif

/* Re-write this based on select_block_sigs() in sendrecv.c */
#define CALL_CHK(__function, __args) \
  while (1) { \
    /* On EINTR (interrupted by signal), just try again */ \
    /* On EAGAIN, kernel wants us to just try again */ \
    /* On ECHILD (error no child in wait), ignore error */ \
    /* On EADDRINUSE (bind <= stale address), caller fixes and tries again */ \
    /* On ECONNRESET || EPIPE (broken connection), caller can return MPI_FAIL */ \
    errno = 0; \
    if ( (__function __args) == -1 ) { \
      /* Remove first if, and fall through, when have confidence it works */ \
      /* printf("errno: %d; ",errno); */ \
      if ( (errno == EADDRINUSE) ) { \
        PERROR( __function ); errno = EADDRINUSE; break; } \
      if ( (errno == ECHILD) || (errno == EADDRINUSE) ) break; \
      if ( (errno == ECONNRESET) || (errno == EPIPE) ) break; \
      if ( errno == EINTR ) continue; \
      if ( errno == EAGAIN ) continue; \
      PERROR( __function ); exit( 1 ); \
    } \
    else break; /* successful system call */ \
  }

#define RCV_EAT_HDR 1
#define RCV_DONT_EAT_HDR 0
#define RCV_BLOCKING 0
#define RCV_NOT_BLOCKING 1

#define MPINU_RECV	0
#define MPINU_PROBE	1

extern int MPINU_num_slaves;
extern int MPINU_myrank;
extern int MPINU_my_list_sd;
extern fd_set MPINU_fdset;
extern int MPINU_max_sd;
extern int MPINU_is_spawn2;
extern int MPINU_is_initialized;
extern int MPINU_thread_level;
extern pthread_t MPINU_thread_main;
extern pthread_mutex_t MPINU_global;
extern int MPINU_self_pipe;

typedef struct float_int { float value;
                 int index;
               } float_int;

typedef struct double_int { double value;
                 int index;
               } double_int;

typedef struct long_int { long value;
                 int index;
               } long_int;

typedef struct short_int { short value;
                 int index;
               } short_int;

typedef struct int_int { int value;
                 int index;
               } int_int;

typedef struct long_double_int { long double value;
                 int index;
               } long_double_int;

/*
 * Not-a-tag: used to mark a tag field as not having a valid value.
 */
#define MPINU_NOT_A_TAG			(MPI_ANY_TAG - 1)

/*
 * Initialization sequence tag.
 */
#define MPINU_INIT_TAG			(MPINU_NOT_A_TAG - 1)

/*
 * Finalization sequence tag
 */
#define MPINU_FINAL_TAG			(MPINU_INIT_TAG - 1)

/*
 * Collective Communication tags
 */
#define MPINU_ALLGATHERV_TAG		(MPINU_FINAL_TAG - 1)
#define MPINU_ALLGATHER_TAG		(MPINU_FINAL_TAG - 2)
#define MPINU_ALLREDUCE_TAG		(MPINU_FINAL_TAG - 3)
#define MPINU_ALLTOALLV_TAG		(MPINU_FINAL_TAG - 4)
#define MPINU_ALLTOALL_TAG		(MPINU_FINAL_TAG - 5)
#define MPINU_BARRIER_TAG		(MPINU_FINAL_TAG - 6)
#define MPINU_BCAST_TAG			(MPINU_FINAL_TAG - 7)
#define MPINU_GATHERV_TAG		(MPINU_FINAL_TAG - 8)
#define MPINU_GATHER_TAG		(MPINU_FINAL_TAG - 9)
#define MPINU_REDUCE_SCATTER_TAG	(MPINU_FINAL_TAG - 10)
#define MPINU_REDUCE_TAG		(MPINU_FINAL_TAG - 11)
#define MPINU_SCAN_TAG			(MPINU_FINAL_TAG - 12)
#define MPINU_SCATTERV_TAG		(MPINU_FINAL_TAG - 13)
#define MPINU_SCATTER_TAG		(MPINU_FINAL_TAG - 14)
// Must be last.
#define MPINU_COLL_COMM_TAG		(MPINU_FINAL_TAG - 15)

#define MPINU_MAX_PROCESSOR_NAME 256
#define PG_ARRAY_SIZE  1000
#define PROCGROUP_LEN 10000
#define PG_NOSOCKET -1 /* must not coincide with socket descriptor int */

#ifdef SENDRECV_RINGBUFF
# define RINGSIZE	40
typedef enum {
  OP_BAD	   	= 0,
  OP_RECV_NONBLOCKING	= 0xf100d0,
  OP_RECV_BLOCKING    	= 0xf100d1,
  OP_SEND_BLOCKING    	= 0xf100d2,
  OP_RECV_MSG_HDR_CACHE	= 0xf100d3
} optype_t;

typedef struct op_tag {
  optype_t	type;
  ssize_t	result;
  int		flags;
  int		errnov;
} op_t, *op_p;

# define RINGBUFF_RECV_ENTER(_s_,_t_,_r_,_f_,_e_) { \
    int _src_ = (_s_); \
    int _idx_ = MPINU_pg_array[_src_].recv_ops_idx; \
    MPINU_pg_array[_src_].recvring[_idx_].type = (_t_); \
    MPINU_pg_array[_src_].recvring[_idx_].result = (_r_); \
    MPINU_pg_array[_src_].recvring[_idx_].flags = (_f_); \
    MPINU_pg_array[_src_].recvring[_idx_].errnov = (_e_); \
    MPINU_pg_array[_src_].recv_ops_idx++; \
    if (MPINU_pg_array[_src_].recv_ops_idx >= RINGSIZE) \
      MPINU_pg_array[_src_].recv_ops_idx = 0; \
    }
# define RINGBUFF_SEND_ENTER(_s_,_t_,_r_,_f_,_e_) { \
    int _src_ = (_s_); \
    int _idx_ = MPINU_pg_array[_src_].send_ops_idx; \
    MPINU_pg_array[_src_].sendring[_idx_].type = (_t_); \
    MPINU_pg_array[_src_].sendring[_idx_].result = (_r_); \
    MPINU_pg_array[_src_].sendring[_idx_].flags = (_f_); \
    MPINU_pg_array[_src_].sendring[_idx_].errnov = (_e_); \
    MPINU_pg_array[_src_].send_ops_idx++; \
    if (MPINU_pg_array[_src_].send_ops_idx >= RINGSIZE) \
      MPINU_pg_array[_src_].send_ops_idx = 0; \
    }
#else
# define RINGBUFF_RECV_ENTER(_s_,_t_,_r_,_f_,_e_)
# define RINGBUFF_SEND_ENTER(_s_,_t_,_r_,_f_,_e_)
#endif

typedef struct pg_stats_tag {
  int	find_attempts;
  int	find_blocking_attempts;
  int	found_in_cache_first_try;
  int	general_wake;
  int	general_wakeup;
  int	match_wakeup;
  int	match_anytag_wakeup;
  int	match_anysrc_wakeup;
  int	match_anysrc_anytag_wakeup;
  int	max_reader_on_attempts;
  int	max_single_blocking_attempts;
  int	max_spurious_wakeups;
  int	not_found;
  int	reader_on_first_try;
  int	reader_wakeup;
  int	reader_anytag_wakeup;
  int	reader_anysrc_wakeup;
  int	reader_anysrc_anytag_wakeup;
  int	spurious_wakeups;
  int	waiters_gtr_events;
  int	waiters_gtr_events_anytag;
  int	wakeup_noop;
  int	reader_waiters_gtr_events;
  int	reader_waiters_gtr_events_anytag;
  int	reader_wakeup_noop;
} pg_stats_t, *pg_stats_p;

typedef struct pg_struct_tag {
  /*
   * These three fields do not need a mutex to synchronize their access since
   * they are only written by the MPINU_main_thread and read by all other
   * threads.
   */
  char *processor;
  char *num_threads;
  char *process;
  /*
   * The sock_mutex protects access to the "sd" and "listener_addr" fields.
   */
  pthread_mutex_t sock_mutex;
  /*
   * socket descriptor
   */
  int sd;
  /*
   * listener port to create new sockets
   */
  struct sockaddr_in listener_addr;
  /*
   * The send mutex is locked by any thread wishing to transmit data to a
   * destination rank different from MPINU_myrank. It does not protect any
   * fields in the structure.
   */
  pthread_mutex_t send_mutex;
#ifdef SENDRECV_RINGBUFF
  /*
   * The send mutex needs to be held to access this ring buffer.
   */
  unsigned int send_ops_idx;
  op_t sendring[RINGSIZE];
#endif
  /*
   * The receive mutex is locked by any thread wishing to receive data from a
   * source rank, including MPINU_myrank. It is also locked when a thread
   * wants to "send" a message to itself. It protects thread "reader" field
   * and the "cache".
   */
  pthread_mutex_t recv_mutex;
  pthread_cond_t recv_cond;
#ifdef SENDRECV_RINGBUFF
  /*
   * The receive mutex needs to be held to access this ring buffer.
   */
  unsigned int recv_ops_idx;
  op_t recvring[RINGSIZE];
#endif
#define NO_READER	(pthread_t)NULL
  pthread_t reader;
  int anysrc_tag;
  void *cache;
  void *waitlist;
  pg_stats_t stats;
} pg_struct_t, *pg_struct_p; 

/* index into MPINU_pg_array is same as process rank in MPI_Comm_world */
extern pg_struct_t MPINU_pg_array[];

/* waitlist of ANY_SOURCE operations */
extern void *MPINU_anysrc_waitlist;
extern pg_stats_t MPINU_anysrc_waitlist_stats;

/* INT must be 32-bit quantity.  Re-define to long if necessary. */
#if (1 << 31) == 0
#define INT long
#else
#define INT int
#endif

typedef struct init_msg_tag { /* Initial message from master to slave */
  INT len; /* For consistency check;  This should equal sizeof(init_msg) */
  INT rank;
  INT num_slaves;
} init_msg_t, *init_msg_p;

typedef struct msg_hdr_tag {
  INT tag;
  INT size;
  INT rank;
} msg_hdr_t, *msg_hdr_p;

/*
 * Internal versions of MPI_Recv and MPI_Send, used by the collective
 * communication subsystem, among others.
 */
extern int MPINU_Int_Recv(void *buf, int count, MPI_Datatype datatype,
    int source, int tag, MPI_Comm comm, MPI_Status *status);
extern int MPINU_Int_Send(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm);

/*
 * Socket option routines.
 */
extern void MPINU_dump_sockopt(int sd);
extern void MPINU_setsockopt(int sd);

/*
 * Non-blocking send entry points.
 */
extern void MPINU_send_thread_exit(void);
extern ssize_t MPINU_send_nonblock(int s, const void *buf, size_t len,
    int flags);

/*
 * Cache management entry points.
 */
extern void *MPINU_cache_init(void *cache);
extern void *MPINU_cache_destroy(void *cache);
extern void MPINU_enter_msg_in_cache(void *cache, void *buf_hdr,
    size_t buf_hdr_len, void *buf_body, size_t buf_body_len);
extern int MPINU_find_msg_in_cache(void *cache, int tag, int peek_msg,
    msg_hdr_p buf_hdr, void *buf_body, size_t buf_body_len);
extern void MPINU_dump_cache (void *cache, int myrank, int rank);

/*
 * Wait-list entry points.
 */
extern void *MPINU_waitlist_init(void *wl);
extern void *MPINU_waitlist_destroy(void *wl);
extern void MPINU_waitlist_block(void *wl, pthread_mutex_t *mu, int tag,
    int type, pg_stats_p stats);
extern void MPINU_waitlist_block_init(void *wl, int tag, int type,
    pg_stats_p stats);
extern void MPINU_waitlist_block_wait(void *wl, pthread_mutex_t *mu, int tag,
    int type, pg_stats_p stats);
extern void MPINU_waitlist_block_destroy(void *wl, int tag, int type,
    pg_stats_p stats);
extern void MPINU_waitlist_register(void *wl, int tag, void *awl,
    pthread_mutex_t *amu, pg_stats_p stats, pg_stats_p astats);
extern void MPINU_waitlist_unregister(void *wl, int tag, void *awl,
    pthread_mutex_t *amu, pg_stats_p stats, pg_stats_p astats);
extern void MPINU_waitlist_wakeup(void *wl, int tag, pg_stats_p stats);
extern void MPINU_waitlist_wakeup_reader(void *wl, pg_stats_p stats);

/*
 * Rank getters and setters
 */
extern int MPINU_rank_socket_get(int rank);
extern void MPINU_rank_socket_set(int rank, int sd);
extern void MPINU_rank_socket_init(int rank);
extern pthread_cond_t *MPINU_rank_recv_cv_get(int rank);
extern pthread_mutex_t *MPINU_rank_recv_mu_get(int rank);
extern pthread_mutex_t *MPINU_rank_send_mu_get(int rank);
extern pthread_mutex_t *MPINU_rank_sock_mu_get(int rank);
extern void *MPINU_rank_cache_get(int rank);
extern void *MPINU_rank_waitlist_get(int rank);
extern pg_stats_p MPINU_rank_stats_get(int rank);
extern void MPINU_rank_reader_set(int rank, pthread_t reader);
extern int MPINU_rank_reader_equal(int rank, pthread_t reader);
extern int MPINU_rank_anysrc_tag_get(int rank);
extern void MPINU_rank_anysrc_tag_set(int rank, int tag);

/*
 * Wrappers for recv() and send() system calls.
 */
extern ssize_t MPINU_recvall(int s, void *buf, size_t len, int flags);
extern ssize_t MPINU_sendall(int s, void *buf, size_t len, int flags);

/*
 * Miscellaneous routines.
 */
extern void MPINU_mpi_master(char *p4pg_file, int argc, char *argv[]);
extern void MPINU_mpi_slave(char *master_host, char *port);
extern int MPINU_parse();
extern int MPINU_new_listener();
extern int MPINU_set_and_exec_cmds();
extern int MPINU_rank_from_socket(int s);
extern void MPINU_create_self_pipe (int rank);
extern int MPINU_tag_match(int pattern, int tag);
extern int MPINU_datatype_valid(MPI_Datatype datatype);
extern int MPINU_datatype_size(MPI_Datatype datatype);
extern int MPINU_msg_size(int count, MPI_Datatype datatype);
