 /**********************************************************************
  * MPINU				                               *
  * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

/*
 * Exports four interfaces to the cache:
 *
 *	MPINU_cache_init
 *	MPINU_cache_destroy
 *	MPINU_enter_msg_in_cache
 *	MPINU_find_msg_in_cache
 *
 * Note that these interfaces are not a true cache encapsulation, as knowledge
 * of MPI definitions and constructs, particularly of MPI_ANY_TAG and the
 * message header itself, are needed. The internal routines that implement
 * cache operations are independent of MPI, except for the assertions. This
 * code moves us a step closer to an independent cache implementation, where
 * different cache modules could be plugged in for experimentation.
 */

#include "mpi.h"
#include "mpiimpl.h"

#define BUF_DEQUEUE_PEEK 1
#define BUF_DEQUEUE_PULL 0

typedef struct buf_queue_tag {
  void *buf;
  size_t end;
  int curs;
  int tag;
} buf_queue_t, *buf_queue_p;

static buf_queue_p buf_init(buf_queue_p cache);
static void *buf_destroy(buf_queue_p cache);
static void buf_reset(buf_queue_p cache, int tag);
static int buf_avail(buf_queue_p cache, size_t len);
static void buf_enqueue(buf_queue_p cache, void *buf, size_t len);
static void buf_dequeue(buf_queue_p cache, void *buf, size_t buf_len,
    size_t len, int peek_msg);
static void *buf_peek(buf_queue_p cache, size_t len);
static void buf_skip(buf_queue_p cache, size_t len);

//======================================================================

static buf_queue_p buf_init( buf_queue_p cache ) 
    {
    if (!cache) {
	/*
	 * Allocate a buffer queue structure since one does not already exist.
	 * Make sure the buffer pointer is cleared so that the code bellow
	 * does not think this is a re-initialization.
	 */
	assert_perror( cache = (buf_queue_p)malloc( sizeof( buf_queue_t ) ) );
	cache->buf = NULL;
	}
    if (cache->buf) {
	/*
	 * Apparently this routine has been called to reinitialize the caches.
	 * So free the previous buffer and set to NULL; a new one will be
	 * allocated when something is ready to be received (or sent to self).
	 */
	free (cache->buf);
	cache->buf = NULL;
	}
    /*
     * Set "end" to zero since we have nothing in the buffer since we have no
     * buffer yet.
     */
    cache->end = 0;
    /*
     * Set the "curs"or to zero as well, since there is nothing in the buffer.
     */
    cache->curs = 0;
    /*
     * Note that here the tag is set to zero, but zero is a valid tag. It
     * doesn't matter because the cursor is also zero, so that tag value is
     * not material until cursor is advanced.
     */
    cache->tag = 0;

    return cache;
    }

static void *buf_destroy( buf_queue_p cache )
    {
    if (cache) {
	if (cache->buf) {
	    free (cache->buf);
	    cache->buf = NULL;
	    }
	free (cache);
	cache = NULL;
	}
    return cache;
    }

static void buf_reset(buf_queue_p cache, int tag)
    {
    if ( NULL == cache )
	return;

    /*
     * We want to set the cursor to the beginning of the cache only if the
     * given tag does not match the recorded tag; that is the indication that
     * a new search is starting. If the tags are the same, there is no point
     * in resetting the cursor as nothing in the cache prior to the current
     * point would match.
     */

    if ( 0 == cache->curs ) {
	if ( cache->tag != tag ) {
	    /*
	     * The cursor is already at the beginning of the cache, so we want
	     * to record the new tag since it is different from the recorded
	     * tag.
	     *
	     * NOTE: Since the cursor is zero, the actual value of the tag is
	     * ignored; once the cursor is moved forward, however, then the
	     * cursor will mark the point before which no headers have this
	     * recorded tag.
	     */
	    cache->tag = tag;
	    }
	else {
	    /*
	     * There is nothing to do, as the cache is already in the proper
	     * state.
	     */
	    }
	}
    else if ( cache->tag != tag ) {
	/*
	 * The given tag does not match the recorded tag. Reset the cursor to
	 * the beginning of the buffer since we don't know if there are any
	 * headers which match the given tag.
	 */
	cache->curs = 0;

	/*
	 * NOTE: Since the cursor is now zero, the actual value of the tag is
	 * ignored; once the cursor is moved forward, however, then the cursor
	 * will mark the point before which no headers have this recorded tag.
	 */
	cache->tag = tag;
	}
    else {
	/*
	 * Nothing to do: the cursor is not at the beginning, and the given
	 * tag matches the recorded tag.
	 */
	}
    }

static int buf_avail( buf_queue_p cache, size_t len ) {
    if ( NULL == cache )
	return FALSE;
    return cache->curs + len <= cache->end;
}

static void buf_enqueue( buf_queue_p cache, void *buf, size_t len ) {
    assert_perror( cache->buf
	    = (buf_queue_p)realloc( cache->buf, cache->end + len ) );
    assert_perror( memcpy( ((char *)cache->buf) + cache->end, buf, len ) );
    cache->end += len;
}

static void buf_dequeue( buf_queue_p cache, void *buf, size_t buf_len,
			 size_t len, int peek_msg ) {
    int i, j;
    char *buf_queue = cache->buf;
    assert( cache->curs + len <= cache->end );
    size_t copy_len = len <= buf_len ? len : buf_len;
    memcpy( buf, ((char *)cache->buf) + cache->curs, copy_len );
    if ( !peek_msg ) {
      for ( i = cache->curs, j = cache->curs+len; 
            j < cache->end; i++, j++ )
        buf_queue[i] = buf_queue[j];
      assert( i + len == cache->end );
      cache->end = i;
      if ( cache->end == 0 ) {
        free( cache->buf );
	cache->buf = NULL;
	}
    }
}

static void *buf_peek( buf_queue_p cache, size_t len ) {
    assert( cache->curs + len <= cache->end );
    return (void *)(((char *)cache->buf) + cache->curs);
}

static void buf_skip( buf_queue_p cache, size_t len ) {
    cache->curs += len;
    assert( cache->curs <= cache->end );
}

#ifndef STANDALONE
void MPINU_enter_msg_in_cache( void *icache, void *buf_hdr,
			       size_t buf_hdr_len, void *buf_body,
			       size_t buf_body_len ) {
  buf_queue_p cache = (buf_queue_p)icache;
  buf_enqueue( cache, buf_hdr, buf_hdr_len ); /* Store hdr */
  if ( buf_body_len > 0 ) {
    assert( NULL != buf_body );
    buf_enqueue( cache, buf_body, buf_body_len ); /* Store body */
  }
  return;
}
#endif

#ifndef STANDALONE
int MPINU_find_msg_in_cache( void *icache, int tag, int peek_msg,
			     msg_hdr_p buf_hdr, void *buf_body,
			     size_t buf_body_len ) {
  buf_queue_p cache = (buf_queue_p)icache;
  buf_reset( cache, tag );
  size_t len = sizeof (*buf_hdr);
  int msg_found = FALSE;
  while ( !msg_found && buf_avail( cache, len ) ) {
    msg_found = MPINU_tag_match(
	tag, ntohl( ((msg_hdr_p)buf_peek( cache, len ))->tag ) );
    if ( msg_found ) {
      buf_dequeue( cache, buf_hdr, len, len, peek_msg );
      if (!peek_msg) {
	int size = ntohl( buf_hdr->size );
	buf_dequeue( cache, buf_body, buf_body_len, size, 0 );
      }
    }
    else {
      size_t skip_len = len
	+ ntohl( ((msg_hdr_p)buf_peek( cache, len ))->size );
      buf_skip( cache, skip_len ); /* Skip header & body */
    }
  }
  if ( !msg_found )
    assert( !buf_avail( cache, len ) );
  return msg_found;
}
#endif

#ifndef STANDALONE
void MPINU_dump_cache ( void *icache, int myrank, int rank ) {
# ifdef DEBUG
  buf_queue_p cache = (buf_queue_p)icache;
  pthread_mutex_t *mu = MPINU_rank_recv_mu_get( rank );
  assert( 0 == pthread_mutex_lock( mu ) );
  fprintf( stderr, "[%d]Dumping cache(%d): %#lx (curs: %d, end: %d)\n",
      myrank, rank, icache, cache->curs, cache->end );
  cache->curs = 0;
  cache->tag = 0;
  int len = sizeof (msg_hdr_t);
  while ( buf_avail( cache, len ) ) {
    msg_hdr_t buf_hdr;
    memcpy( &buf_hdr, ((char *)cache->buf) + cache->curs, len );
    int size = ntohl( buf_hdr.size );
    fprintf( stderr, "[%d]  tag: %#x, size: %d, rank: %d\n",
	myrank, ntohl( buf_hdr.tag ), size, ntohl( buf_hdr.rank ) );
    buf_skip( cache, len + size );
  }
  cache->curs = 0;
  assert( 0 == pthread_mutex_unlock( mu ) );
# endif
}
#endif

#ifndef STANDALONE
void *MPINU_cache_init(void *cache) {
  return (void *)buf_init( (buf_queue_p)cache );
}
#endif

#ifndef STANDALONE
void *MPINU_cache_destroy(void *cache) {
  return (void *)buf_destroy( (buf_queue_p)cache );
}
#endif


//=============================================================

#ifdef STANDALONE
int main( int argc, char *argv[] ) {
  char buf[6];
  char pbuf[6];
  char *ptr;
  buf_queue_p cache = buf_init( NULL );

  assert( cache );

  buf_enqueue( cache, "klmno", 5 );
  buf_skip( cache, 5 );
  buf_enqueue( cache, "pqrst", 5 );
  buf_enqueue( cache, "abcde", 5 );
  buf_enqueue( cache, "fghij", 5 );
  buf_skip( cache, 5 );

  buf_dequeue( cache, buf, 5, 5, BUF_DEQUEUE_PULL );
  buf[5] = '\0';
  printf(buf);		// abcde
  assert( 0 == memcmp( "abcde", buf, 5 ) );

  memcpy (buf, buf_peek( cache, 5 ), 5);
  buf[5] = '\0';
  printf(buf);		// fghij
  assert( 0 == memcmp( "fghij", buf, 5 ) );

  buf_dequeue( cache, pbuf, 5, 5, BUF_DEQUEUE_PEEK );
  assert( 0 == memcmp( pbuf, buf, 5 ) );

  buf_dequeue( cache, buf, 5, 5, BUF_DEQUEUE_PULL );
  assert( 0 == memcmp( "fghij", buf, 5 ) );

  buf_reset( cache, 5 );

  buf_dequeue( cache, buf, 5, 5, BUF_DEQUEUE_PULL );
  buf[5] = '\0';
  printf(buf);		// klmno
  assert( 0 == memcmp( "klmno", buf, 5 ) );

  buf_dequeue( cache, buf, 5, 5, BUF_DEQUEUE_PULL );
  buf[5] = '\0';
  printf("%s", buf);	// pqrst
  assert( 0 == memcmp( "pqrst", buf, 5 ) );

  assert( !buf_avail( cache, 5 ) );

  buf_enqueue( cache, "uvwxyz", 6 );
  memset( buf, '\0', 6 );
  buf_dequeue( cache, buf, 5, 6, BUF_DEQUEUE_PULL );
  assert( '\0' == buf[5] );
  printf("%s", buf);	// uvwxy
  assert( 0 == memcmp( "uvwxy", buf, 5 ) );

  buf_enqueue( cache, "z", 1 );
  memset( buf, '\0', 6 );
  buf_dequeue( cache, buf, 5, 1, BUF_DEQUEUE_PULL );
  assert( '\0' == buf[5] );
  assert( '\0' == buf[4] );
  assert( '\0' == buf[3] );
  assert( '\0' == buf[2] );
  assert( '\0' == buf[1] );
  printf("%s\n", buf);	// z
  assert( 0 == memcmp( "z", buf, 1 ) );

  assert( NULL == buf_destroy( cache ) );
  return 0;
}
#endif
