 /**********************************************************************
  * MPINU                                                              *
  * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

/*
 * MPI_Send calls MPINU_send_nonblock(), which in turn calls MPINU_sendall().
 * MPI_Recv, MPI_Probe and MPI_Iprobe all call msg_read()
 *   msg_read() calls msg_read_setup(), msg_read_select()
 *                    msg_read_get_buf(), and msg_read_post_check().
 *   msg_read_get_buf() calls recv_msg_with_cache(), which is
 *                    a wrapper for MPINU_recvall(), which is a wrapper
 *                    for recv().
 */
#include "mpi.h"
#include "mpiimpl.h"

// The maximum number of tags that the self-pipe will hold.
#define MAX_SELF_TAGS	128

pg_stats_t MPINU_anysrc_waitlist_stats = { 0 };
static pthread_mutex_t anysrc_waitlist_mu = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t anysrc_reader_mu = PTHREAD_MUTEX_INITIALIZER;
static pthread_t anysrc_reader = NO_READER;

// ============================================================
// Utilities for checking if socket is dead, or declaring it dead

static int socket_dead( int dest, int fd ) {
  MPINU_rank_socket_set( dest, PG_NOSOCKET );
  return 0;
}

static int is_socket_connected( int socket ) {
  socklen_t mpi_tmp_namelen = 0;
  struct sockaddr mpi_tmp_name;
  int retval = getpeername( socket, &mpi_tmp_name, &mpi_tmp_namelen );
  if ( retval == -1 ) {
    fprintf( stderr, "errno: %d\n", errno );
    PERROR( getpeername );
  }
  return 1 + getpeername( socket, &mpi_tmp_name, &mpi_tmp_namelen );
}

// ============================================================
// Utilities for dealing with socket options

void MPINU_dump_sockopt( int sd ) {
#if DEBUG
  int			val;
  int			val_len;
  struct timeval	timeout;
  struct linger		linger;

  val = 0;
  val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_KEEPALIVE,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_KEEPALIVE: %d\n",
      MPINU_myrank, sd, val );
  linger.l_onoff = 0;
  linger.l_linger = 0;
  val_len = sizeof(linger);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_LINGER,
	  (void *)&linger, &val_len) );
  assert( sizeof(linger) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_LINGER: %d[%d]\n",
      MPINU_myrank, sd, linger.l_onoff, linger.l_linger );
  val = 0;
  val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_SNDBUF,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_SNDBUF: %d\n",
      MPINU_myrank, sd, val );
  val = 0;
  val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_RCVBUF,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_RCVBUF: %d\n",
      MPINU_myrank, sd, val );
  val = 0;
  val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_SNDLOWAT,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_SNDLOWAT: %d\n",
      MPINU_myrank, sd, val );
  val = 0;
  val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_RCVLOWAT,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_RCVLOWAT: %d\n",
      MPINU_myrank, sd, val );

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  val_len = sizeof(timeout);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_SNDTIMEO,
	  (void *)&timeout, &val_len) );
  assert( sizeof(timeout) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_SNDTIMEO: (%d,%d)\n",
      MPINU_myrank, sd, timeout.tv_sec, timeout.tv_usec );

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  val_len = sizeof(timeout);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_RCVTIMEO,
	  (void *)&timeout, &val_len) );
  assert( sizeof(timeout) == val_len );
  fprintf( stderr, "rank %d, socket: %d, SO_RCVTIMEO: (%d,%d)\n",
      MPINU_myrank, sd, timeout.tv_sec, timeout.tv_usec );
#endif
}

void MPINU_setsockopt( int sd ) {
  int val = 1;
  CALL_CHK( setsockopt, (sd, SOL_SOCKET, SO_KEEPALIVE,
	  (void *)&val, sizeof(val)) );

  val = 0;
  socklen_t val_len = sizeof(val);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_KEEPALIVE,
	  (void *)&val, &val_len) );
  assert( sizeof(val) == val_len );
  assert( val != 0 );

  struct linger	linger;
  linger.l_onoff = 1;
  linger.l_linger = 60*10;	// Ten minutes
  CALL_CHK( setsockopt, (sd, SOL_SOCKET, SO_LINGER,
	  (void *)&linger, sizeof(linger)) );

  linger.l_onoff = 0;
  linger.l_linger = 0;
  val_len = sizeof(linger);
  CALL_CHK( getsockopt, (sd, SOL_SOCKET, SO_LINGER,
	  (void *)&linger, &val_len) );
  assert( sizeof(linger) == val_len );
  assert( linger.l_onoff != 0 );
  assert( 60*10 == linger.l_linger );

#ifdef MPINU_USE_SO_RCVTIMEO_SOCKOPT
  struct timeval timeout;
  timeout.tv_sec = 2;
  timeout.tv_usec = 0;
  CALL_CHK( setsockopt, (sd, SOL_SOCKET, SO_RCVTIMEO,
	  (void *)&timeout, sizeof(timeout)) );
#endif
}

// ============================================================
// list_accept_new_socket, list_connect_new_socket
// Each process has listener socket.  Any other process can call it
//  to set up the initial connection, or to create a new one if the
//  old one dies.

static void list_accept_new_socket( int list_sd ) {
  struct sockaddr_in new_sin;
  int ns, source;
  socklen_t fromlen = sizeof(new_sin);
  char buf[sizeof(INT)];

#ifdef DEBUG
  printf( "Rank(%d): accepting new socket\n", MPINU_myrank );
  fflush( stdout );
#endif
  CALL_CHK( ns = accept, (list_sd, (struct sockaddr *)&new_sin, &fromlen) );
  MPINU_dump_sockopt( ns );
  MPINU_setsockopt( ns );
  CALL_CHK( MPINU_recvall, (ns, buf, sizeof(INT), 0) );
  source = ntohl(*((INT *)buf));

  if ( source < 0 || source > MPINU_num_slaves ) {
    printf( "list_accept_new_socket:  Bad sender rank received: %d\n",
	source );
    fflush( stdout );
  } else {
    pthread_mutex_t *mu = MPINU_rank_sock_mu_get( source );
    assert( 0 == pthread_mutex_lock( mu ) );
    assert( PG_NOSOCKET == MPINU_rank_socket_get( source ) );
#ifdef DEBUG
    printf( "accepted socket (%d) from %d\n", ns, source );
    fflush( stdout );
#endif
    MPINU_rank_socket_set( source, ns );
    assert( 0 == pthread_mutex_unlock( mu ) );
  }
}

static int list_connect_new_socket( int myrank, int dest_rank ) {
  int sd;
  char buf[sizeof(INT)];

#ifdef DEBUG
  printf( "connecting new socket from %d to %d\n", myrank, dest_rank );
  fflush( stdout );
#endif
  *((INT *)buf) = htonl(myrank);
  pthread_mutex_t *mu = MPINU_rank_sock_mu_get( dest_rank );
  assert( 0 == pthread_mutex_lock( mu ) );
  CALL_CHK( sd = socket, (AF_INET, SOCK_STREAM, IPPROTO_TCP) );
  MPINU_dump_sockopt( sd );
  MPINU_setsockopt( sd );
  while ( -1 == connect( sd,
		  (struct sockaddr *)&MPINU_pg_array[dest_rank].listener_addr,
		  sizeof(struct sockaddr_in) ) ) {
    if ( errno == EALREADY || errno == EISCONN ) break;
    if ( errno != EINTR ) {
      assert( 0 == pthread_mutex_unlock( mu ) );
      return -1; /* return failure */
    }
  }
  CALL_CHK( MPINU_sendall, (sd, buf, sizeof(INT), 0) );

#ifdef DEBUG
  printf( "connected new socket(%d) from %d to %d\n", sd, myrank, dest_rank );
  fflush( stdout );
#endif
  MPINU_rank_socket_set( dest_rank, sd );
  assert( 0 == pthread_mutex_unlock( mu ) );
  return sd ;
}

/*
 * Create a self-pipe for use by threads "sending" to self so that ANY_SOURCE
 * threads can "block" in select() and be woken up when a message from another
 * thread is available. The master and each slave need to create these pipes.
 */
void MPINU_create_self_pipe( int rank ) {
  /*
   * Setup self pipe so that ANY_SOURCE readers will know when a thread sends
   * data to MPINU_myrank.
   */
  int selfFds[2];
  // Create the pipe.
  int ret_stat = pipe( selfFds );
  if ( ret_stat < 0 ) {
    perror( "pipe" );
    abort();
  }
  // Make the read side non-blocking.
  ret_stat = fcntl( selfFds[0], F_SETFL, O_NONBLOCK );
  if ( ret_stat < 0 ) {
    perror( "fcntl" );
    abort();
  }  
  // file descriptor where data will be written.
  MPINU_self_pipe = selfFds[1];
  // file descriptor where data will be read
  MPINU_rank_socket_set( rank, selfFds[0] );
}

/* ============================================================ */
/* This blocks certain signals while inside select()		*/
/* Note that a signal in the middle of a system call that is	*/
/*   caught and returns, may either restart the signal or else	*/
/*   exit the system call.  In particular, SIGINT appears to	*/
/*   cause select to return with some of readfds being set.	*/
/* Note also that recv() is called only after msg_read()	*/
/*   guarantees that a message is available, or else after a	*/
/*   select() inside msg_read() that guarantees a message	*/
/*   is available.  So, recv() should never block.		*/

/* When this works, eliminate re-start after EINTR in mpiimpl.h */
/* SIGINT may have user-installed signal handler.  Call it after
 *   catching the signal.  If we're still here after that, go
 *   go back and re-start the system call.  This can be done
 *   generically with a single MPINU signal handler, since it
 *   can check what signal it caught.  But we must first save all
 *   old signal handlers in an array, to raise them or restore them
 *   later.
 */
/* When this works, do the same for SIGCONT, SIGSTOP, SIGTSTP (^Z)
 *  SIGPIPE ("broken pipe" => broken socket), if those have other
 *  signal handlers that should be called;
 * I've been told that SO_KEEPALIVE causes a SIGPIPE if no
 *  activity within a certain time
 */

/* Under POSIX, the system calls used here that are interruptible (EINTR)
 *  are:	 connect(), recv(), send(), select()
 */
static int select_block_sigs( int n, fd_set *readfds, fd_set *writefds,
			      fd_set *exceptfds, struct timeval *timeout ) {
  int answer, cnt = 0;
  fd_set origreadfds;
  fd_set origwritefds;
  fd_set origexceptfds;

  /* save fdsets */
  if ( readfds != NULL ) {
    memcpy( (char *)&origreadfds, (char *)readfds, sizeof(*readfds) );
    cnt++;
  }
  if ( writefds != NULL ) {
    memcpy( (char *)&origwritefds, (char *)writefds, sizeof(*writefds) );
    cnt++;
  }
  if ( exceptfds != NULL ) {
    memcpy( (char *)&origexceptfds, (char *)exceptfds, sizeof(*exceptfds) );
    cnt++;
  }
  assert( cnt > 0 );

  /* If a blocked signal arrives during select, it returns -1 with errno ==
   * EINTR. In principle, if a blocked signal arrives before then, for which
   * there is no signal handler, then we block forever here. Something similar
   * happens if a signal arrives after select and before sigprocmask restores
   * old mask. This could be fixed by installing a signal handler that writes
   * the signal number to a pipe, and allows us to read and propagate the
   * signal later. But so far, this hasn't been a problem. pselect() was
   * created to solve this, but Linux pselect() is not yet native, and so
   * still has the race condition.
   */

  while ( 1 ) {
    errno = 0;
    answer = select( n, readfds, writefds, exceptfds, timeout );

    if ( answer != -1 )
      break; /* successful system call */

    /* restore fdsets on error from select */
    memcpy( (char *)readfds, (char *)&origreadfds, sizeof(origreadfds) );
    if ( writefds != NULL )
      memcpy( (char *)writefds, (char *)&origwritefds,
	  sizeof(origwritefds) );
    if ( exceptfds != NULL )
      memcpy( (char *)exceptfds, (char *)&origexceptfds,
	      sizeof(origexceptfds) );

    /* Remove first if, and fall through, when have confidence it works */
    /* printf( "errno: %d; ", errno ); */
    if ( errno == EADDRINUSE ) {
      PERROR( select ); errno = EADDRINUSE; break; }

    /* On EINTR (interrupted by signal), just try again */
    /* On EAGAIN, kernel wants us to just try again */
    /* On ECHILD (error no child in wait), ignore error */
    /* On EADDRINUSE (bind <= stale address), caller fixes and tries again */
    /* On ECONNRESET || EPIPE (broken connection), caller can
       return MPI_FAIL */
    if ( ( errno == ECHILD ) || ( errno == EADDRINUSE ) ) break;
    if ( ( errno == ECONNRESET ) || ( errno == EPIPE ) )  break;
    if ( errno == EINTR ) continue;
    if ( errno == EAGAIN ) continue;
    PERROR( select ); abort();
  }

  return answer;
}

/*
 * Define the number of times MPINU_recv_nonblocking(), MPINU_recvall() and
 * MPINU_sendall() will attempt their operations in the face of interrupts and
 * zero receives/sends.
 */
#define TRIES 100

/*
 * MPINU_recvall()
 *
 * Wrapper for the recv() system call. Receive the specified number of bytes
 * from the provided socket storing them in the given buffer. It will handle
 * partial receives, and keep retrying in the face of interrupts and zero
 * receives. It will also handle non-blocking sockets by using select() to
 * wait until the socket has data to be received.
 */
ssize_t MPINU_recvall( int s, void *buf, size_t len, int flags ) {
    int result = 0; /* if while loop not executed, len was 0 */
    int received = 0;
    int timeout = TRIES;
    int intr = 0;
    int zero_recvd = 0;
    int peek_not_recvd = 0;

#ifdef LOG_STRING
    if ( len == sizeof(msg_hdr_t) )
      log_string( 4, "MPINU_recvall(rank %d): msg_hdr\n", MPINU_myrank );
    else
      log_string( 4, "MPINU_recvall(rank %d): msg body(coming): size: %d\n",
	  MPINU_myrank, len );
#endif
    assert( s >= 0 );
    assert( len > 0 );
    while ( received < len && timeout > 0 ) {
      result = recv( s, (char *)buf + received, len - received, flags );
      if ( ( flags & MSG_PEEK ) && ( result != len ) ) {
	if ( result > 0 ) {
	  peek_not_recvd++;
	  result = 0;
	}
      }
      if ( result > 0 ) {
	received += result;
      }
      else if ( 0 == result ) {
	timeout--;
	zero_recvd++;
      }
      else if ( ( -1 == result ) && ( EINTR == errno ) ) {
	timeout--;
	intr++;
      }
      else {
	abort();
	break;
      }
    }
    if ( 0 == timeout ) {
      int i;
      if ( ( 0 == received )
	  && ( peek_not_recvd > ( TRIES - 5 ) )
	  && ( TRIES > 6 ) ) {
	fprintf( stderr, "MPINU_recvall(rank %d, len %d, flags %s): many attempts to peek(%d), "
	    "but only partial buffer receved;\n", MPINU_myrank, len, (flags & MSG_PEEK ? "MSG_PEEK" : "<none>"), peek_not_recvd );
      } else if ( ( 0 == received )
	  && ( zero_recvd > ( TRIES - 5 ) )
	  && ( TRIES > 6 ) ) {
	fprintf( stderr, "MPINU_recvall(rank %d, len %d, flags %s): no receive(%d);"
	    " peer performed orderly shutdown.\n", MPINU_myrank, len, (flags & MSG_PEEK ? "MSG_PEEK" : "<none>"), zero_recvd );
	for ( i = 0; i <= MPINU_num_slaves; i++ ) {
	  void *cache = MPINU_rank_cache_get( i );
	  MPINU_dump_cache( cache, MPINU_myrank, i );
	}
      } else {
	fprintf( stderr, "MPINU_recvall(rank %d, len %d, flags %s): %d tries, %d interrupts, "
		"%d zeros, %d peeks;  Only received %d bytes out of %d.\n",
	    MPINU_myrank, len, (flags & MSG_PEEK ? "MSG_PEEK" : "<none>"), TRIES, intr, zero_recvd, peek_not_recvd,
	    received, len );
      }
      int rank = MPINU_rank_from_socket( s );
      fprintf( stderr, " receiving from rank %d:\n", rank );
      errno = ECONNRESET; // On Linux, recv can return 0,
			  // instead of -1 and setting errno = ECONNRESET
#ifdef SENDRECV_RINGBUFF
      if ( rank >= 0 )
	RINGBUFF_RECV_ENTER( rank, OP_RECV_BLOCKING, -1, flags, errno );
#endif
      return -1;
    }
    if ( result < 0 ) {
      int i;
      fprintf( stderr, "Rank %d receiving from rank %d: ",
	  MPINU_myrank, MPINU_rank_from_socket( s ) );
      perror( "MPINU_recvall" );
    }
    else if ( received == len )
      result = len;

#ifdef LOG_STRING
    if ( len == sizeof(msg_hdr_t) )
      log_string( 4, "MPINU_recvall(rank %d): msg_hdr:"
		    " size: %d, tag: %d, rank: %d\n",
		  MPINU_myrank,
		  ntohl( ((msg_hdr_p)buf)->size ),
		  ntohl( ((msg_hdr_p)buf)->tag ),
		  ntohl( ((msg_hdr_p)buf)->rank ) );
    else
      log_string( 4, "MPINU_recvall(rank %d): msg body(arrived): size: %d\n",
		 MPINU_myrank, len );
#endif
#ifdef SENDRECV_RINGBUFF
    int rank = MPINU_rank_from_socket( s );
    if ( rank >= 0 )
      RINGBUFF_RECV_ENTER( rank, OP_RECV_BLOCKING, result, flags, errno );
#endif
    return result;
}

/*
 * MPINU_sendall()
 *
 * Wrapper for the send() system call. Send the specified number of bytes from
 * the given buffer over the provided socket. It will handle partial sends,
 * and keep retrying in the face of interrupts and zero sends. It will also
 * handle non-blocking sockets by using select() to wait until the socket is
 * ready for transmission.
 */
ssize_t MPINU_sendall( int s, void *buf, size_t len, int flags ) {
  int result = 0; /* if while loop not executed, len was 0 */
  int sent = 0;
  int timeout = TRIES;
  int intr = 0;
  int zero_sent = 0;
  assert( s >= 0 );
  while ( sent < len && timeout > 0 ) {
    result = send( s, (char *)buf + sent, len - sent, flags );
    if ( result > 0 ) {
      sent += result;
    }
    else if ( 0 == result ) {
      timeout--;
      zero_sent++;
    }
    else if ( result == -1 && errno == EINTR ) {
      timeout--;
      intr++;
    }
    else
      break;
  }
  if ( 0 == timeout ) {
    if ( 0 == sent && zero_sent > TRIES - 5 && TRIES > 6 )
      fprintf( stderr, "MPINU_sendall(rank %d): nothing sent after %d tries\n",
	  MPINU_myrank, TRIES );
    else
      fprintf( stderr, "MPINU_sendall(rank %d): %d tries, %d interrupts, %d zeros;"
	  "	Only sent %d bytes out of %d.\n",
	  MPINU_myrank, TRIES, intr, zero_sent, sent, len );
    errno = ECONNRESET;
    return -1;
  }
  assert( timeout > 0 && ( sent == len || result == -1 ) );
  if ( result == -1 ) {
    int i;
    fprintf( stderr, "MPINU_sendall: Rank %d sending to rank %d: ",
	MPINU_myrank, MPINU_rank_from_socket( s ) );
    perror( "MPINU_sendall" );
  }
  if ( result >= 0 && sent == len )
    result = sent;
#ifdef SENDRECV_RINGBUFF
  int rank = MPINU_rank_from_socket( s );
  if ( rank >= 0 )
    RINGBUFF_SEND_ENTER( rank, OP_SEND_BLOCKING, result, flags, errno );
#endif
  return result;
}

/*
 * Helper routine for writing a tag to the self pipe.
 */
static void write_self_pipe( int tag ) {
  int bytes_to_write = sizeof(tag);
  char *tag_ptr = (char *)&tag;
  do {
    int ret_stat = write( MPINU_self_pipe, tag_ptr, bytes_to_write );
    if ( ret_stat < 0 ) {
      if ( errno != EWOULDBLOCK ) {
	perror( "write(self) failed" );
	abort();
      }
    } else {
      // FIX-ME: If we get here, this means we can't fit the size of an
      // integer into the pipe. This might cause an assertion on the read side
      // when an integral number of tags is not read.
      bytes_to_write -= ret_stat;
      tag_ptr += ret_stat;
    }
  } while ( bytes_to_write );
}

// ==============================================================
// MPI_Send (thread-safe, checks for dead sockets)

static int msg_send( void *buf, int count, size_t len, int dest,
	      int tag, MPI_Comm comm ) {
  msg_hdr_t buf_hdr;
  ssize_t num_bytes;
  sigset_t orig;
  pthread_mutex_t *mu;
  sho_block_async_signals( &orig );
  if ( dest == MPINU_myrank ) {
    mu = MPINU_rank_recv_mu_get( MPINU_myrank );
    void *waitlist = MPINU_rank_waitlist_get( MPINU_myrank );
    void *cache = MPINU_rank_cache_get( MPINU_myrank );
    pg_stats_p stats = MPINU_rank_stats_get( MPINU_myrank );
    assert( 0 == pthread_mutex_lock( mu ) );
    msg_hdr_t hdr_buf;
    hdr_buf.size = htonl( len );
    hdr_buf.tag = htonl( tag );
    hdr_buf.rank = htonl( MPINU_myrank );
    MPINU_enter_msg_in_cache( cache, &hdr_buf, sizeof( hdr_buf ), buf, len );
    /*
     * Wake up any ANY_SOURCE readers.
     *
     * Note that only ANY_SOURCE readers mark themselves as "reader" for this
     * source, as there are is no socket to read from; this the thread calling
     * msg_send() is placing the message directly into the cache.
     */
    if ( !MPINU_rank_reader_equal( MPINU_myrank, NO_READER ) ) {
      assert( !MPINU_rank_reader_equal( MPINU_myrank, pthread_self() ) );

      /*
       * The ANY_SOURCE reader set the rank's anysrc_tag field to the tag it
       * is looking for. So don't bother writing to the self pipe if the
       * message tag doesn't match.
       */
      if ( MPINU_tag_match( MPINU_rank_anysrc_tag_get( MPINU_myrank ), tag ) ) {
	write_self_pipe( tag );
      }
    }
    /*
     * Wake up any local readers as well.
     */
    MPINU_waitlist_wakeup( waitlist, tag, stats );
    assert( 0 == pthread_mutex_unlock( mu ) );
    sho_restore_async_signals( &orig );
    return MPI_SUCCESS;
  }
  int s = MPINU_rank_socket_get( dest );
  if ( s == PG_NOSOCKET ) {
    assert( dest != MPINU_myrank );
    s = list_connect_new_socket( MPINU_myrank, dest );
    if ( -1 == s ) {
      sho_restore_async_signals( &orig );
      return MPI_FAIL; /* Return failure if can't connect to dest */
    }
  }
  if ( ! is_socket_connected( s ) ) {
    socket_dead( dest, s );
    sho_restore_async_signals( &orig );
    return MPI_FAIL;
  }
  buf_hdr.tag = htonl( tag );
  buf_hdr.size = htonl( len );
  buf_hdr.rank = htonl( MPINU_myrank );

  // Don't separate send of message header and message
  mu = MPINU_rank_send_mu_get( dest );
  assert( 0 == pthread_mutex_lock( mu ) );
  num_bytes = MPINU_send_nonblock( s, (char *)&buf_hdr, sizeof(buf_hdr), 0 );
  if ( ( num_bytes == -1 ) || ( !is_socket_connected( s ) ) ) {
    socket_dead( dest, s );
    assert( 0 == pthread_mutex_unlock( mu ) );
    sho_restore_async_signals( &orig );
    return MPI_FAIL;
  }
  assert( num_bytes == sizeof(buf_hdr) );

  num_bytes = MPINU_send_nonblock( s, (char *)buf, len, 0 );

  if ( ( num_bytes == -1 ) || ( ! is_socket_connected(s) ) ) {
    socket_dead( dest, s );
    assert( 0 == pthread_mutex_unlock( mu ) );
    sho_restore_async_signals( &orig );
    return MPI_FAIL;
  }
  assert( 0 == pthread_mutex_unlock( mu ) );
  assert( num_bytes == len );

  sho_restore_async_signals( &orig );
  return MPI_SUCCESS;
}

int MPI_Send( void *buf, int count, MPI_Datatype datatype, int dest,
	      int tag, MPI_Comm comm ) {
  if ( dest > MPINU_num_slaves || dest < 0 || count < 0
      || !MPINU_datatype_valid( datatype ) || ( tag < 0 ) ) {
    printf( "MPI_Send( .., %d, %d, %d, %d, ..): invalid argument\n",
	count, datatype, dest, tag );
    fflush( stdout );
    abort();
  }
  size_t len = MPINU_msg_size( count, datatype );
#ifdef DEBUG
  printf( "%s: source: %d, dest:  %d, count: %d, size: %d, tag: %#x\n",
      __func__, MPINU_myrank, dest, count, len, tag );
  fflush( stdout );
#endif
  return msg_send( buf, count, len, dest, tag, comm );
}

int MPINU_Int_Send( void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm ) {
  if ( dest > MPINU_num_slaves || dest < 0 || count < 0
      || !MPINU_datatype_valid( datatype )
      || tag > MPINU_INIT_TAG || tag < MPINU_COLL_COMM_TAG ) {
    printf( "MPINU_Int_Send( .., %d, %d, %d, %d, ..): invalid argument\n",
	count, datatype, dest, tag );
    fflush( stdout );
    abort();
  }
  size_t len = MPINU_msg_size( count, datatype );
#ifdef DEBUG
  printf( "%s: source: %d, dest: %d, count: %d, size: %d, tag: %#x\n",
      __func__, MPINU_myrank, dest, count, len, tag );
  fflush( stdout );
#endif
  return msg_send( buf, count, len, dest, tag, comm );
}

/*
 * Given a tag pattern, match again a provided tag. TRUE if a match, FALSE
 * otherwise.
 */
int MPINU_tag_match( int pattern, int tag ) {
  int ret_val;

  assert( tag != MPI_ANY_TAG );
  if ( MPI_ANY_TAG == pattern ) {
    /*
     * MPI_ANY_TAG cannot match internal tags.
     */
    ret_val = ( tag >= 0 );
  } else {
    ret_val = ( pattern == tag );
  }
  return ret_val;
}

/*
 * Receive one message, placing a message that doesn't match the expected tag
 * into the cache for all receive operations: Iprobe(), Probe() and Recv().
 * This routine should only block in MPINU_recvall() waiting for the message
 * body (or possibly the header as well, but unlikely), allowing the caller to
 * block in select() for the next header if it is not in the socket yet.
 */
static ssize_t recv_msg_with_cache( int source, int tag,
				    msg_hdr_p buf_hdr, size_t buf_hdr_len,
				    void *buf_body, size_t buf_body_len,
				    int eat_hdr, int blocking ) {
  assert( source != -1 );
  assert( source != MPINU_myrank );
  int s = MPINU_rank_socket_get(source);
#ifdef DEBUG
  printf( "%s(rank %d): receiving on socket %d out of MPINU_max_sd %d\n",
      __func__, MPINU_myrank, s, MPINU_max_sd );
  fflush( stdout );
#endif
  assert( sizeof(msg_hdr_t) == buf_hdr_len );
  int flags;
  assert( ( RCV_BLOCKING == blocking )
      || ( ( RCV_DONT_EAT_HDR == eat_hdr )
	  && ( RCV_NOT_BLOCKING == blocking ) ) );
  if ( RCV_DONT_EAT_HDR == eat_hdr )
    flags = MSG_PEEK;
  else
    flags = 0;
  size_t rlen;

  assert( MPINU_rank_reader_equal( source, pthread_self() ) );

  rlen = MPINU_recvall( s, buf_hdr, buf_hdr_len, flags );
  if ( rlen < 0 ) {
    /*
     * If the received failed, drop to the bottom of this routine and return
     * to the caller.
     */
    RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
    return rlen;
  }
  assert( rlen == buf_hdr_len );
  if ( MPINU_tag_match( tag, ntohl( buf_hdr->tag ) ) ) {
    /*
     * We found a matching tag; the buffer has been pulled from the socket in
     * the case of eat_hdr == RCV_EAT_HDR, or left on the socket otherwise. In
     * either case, it is in the right place. If eat_hdr == RCV_EAT_HDR then
     * we need to pull the body off as well.
     */
    if ( RCV_EAT_HDR == eat_hdr ) {
      size_t size = ntohl( buf_hdr->size );
      if ( size > 0 ) {
	/*
	 * Receive what can fit in the provided buffer.
	 */
	size_t rsize = size <= buf_body_len ? size : buf_body_len;
	if ( rsize > 0 ) {
	  CALL_CHK( rlen = MPINU_recvall, (s, (char *)buf_body, rsize, 0 ) );
	}
	if ( size > buf_body_len ) {
	  /*
	   * Now there is still stuff on the wire, so pull it all off and
	   * throw it away.
	   */
	  size_t remainder = size - buf_body_len;
	  do {
	    char trash[1000];
	    size_t tlen;
	    rsize = remainder <= sizeof(trash) ? remainder : sizeof(trash);
	    CALL_CHK( tlen = MPINU_recvall, (s, trash, rsize, 0 ) );
	    remainder -= tlen;
	  } while ( remainder > 0 );
	  assert( 0 == remainder );
	}
      }
    }
    RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
    return rlen;
  }

  /*
   * We have a message header, but it does not match.
   */
  assert( !MPINU_tag_match( tag, ntohl( buf_hdr->tag ) ) );

  if ( RCV_DONT_EAT_HDR == eat_hdr ) {
    /*
     * [I]Probe(RCV_DONT_EAT_HDR, RCV_BLOCKING|RCV_NOT_BLOCKING): we need to
     * pull the message header off the socket so that it can be placed into
     * the cache below. We want to empty the socket so that if the caller
     * decides to block, it can block in select() and not here.
     */
    CALL_CHK( rlen = MPINU_recvall, ( s, buf_hdr, buf_hdr_len, 0 ) );
    if ( rlen < 0 ) {
      /*
       * If the received failed, notify the caller.
       */
      RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
      return rlen;
    }
    assert( rlen == buf_hdr_len );
  }

  assert( !MPINU_tag_match( tag, ntohl( buf_hdr->tag ) ) );

  pthread_mutex_t *mu = MPINU_rank_recv_mu_get( source );
  void *waitlist = MPINU_rank_waitlist_get( source );
  void *cache = MPINU_rank_cache_get( source );
  pg_stats_p stats = MPINU_rank_stats_get( source );

  /*
   * At this point, we have a header that needs to be placed in the cache.
   *
   * If the body of the message is zero, just place the header in the cache
   * and leave.
   */
  size_t body_len = ntohl( buf_hdr->size );
  if ( 0 == body_len ) {
    assert( 0 == pthread_mutex_lock( mu ) );
    MPINU_enter_msg_in_cache( cache, buf_hdr, buf_hdr_len, NULL, 0 );
    if ( MPINU_tag_match( MPINU_rank_anysrc_tag_get( source ),
	    ntohl( buf_hdr->tag ) ) ) {
      write_self_pipe( tag );
    }
    MPINU_waitlist_wakeup( waitlist, ntohl( buf_hdr->tag ), stats );
    RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
    assert( 0 == pthread_mutex_unlock( mu ) );

    /*
     * We have successfully placed this message in the cache, tell the caller
     * we did not find a match.
     */
    return 0;
  }

  /*
   * Pull the body off the wire and then place them both in the cache.
   */
  assert( body_len > 0 );
# define STATIC_SIZE 1000
  char body_buf_array[STATIC_SIZE];
  void *body_buf;
  if ( body_len > STATIC_SIZE ) {
    body_buf = malloc( body_len );
    assert( body_buf );
  } else {
    body_buf = body_buf_array;
  }
  CALL_CHK( rlen = MPINU_recvall, ( s, body_buf, body_len, 0 ) );
  if (rlen < 0) {
    /*
     * If the received failed, notify the caller.
     */
    RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
    return rlen;
  }
  assert( rlen == body_len );
  assert( 0 == pthread_mutex_lock( mu ) );
  MPINU_enter_msg_in_cache( cache, buf_hdr, buf_hdr_len, body_buf, body_len );
  if ( MPINU_tag_match( MPINU_rank_anysrc_tag_get( source ),
	  ntohl( buf_hdr->tag ) ) ) {
    write_self_pipe( tag );
  }
  MPINU_waitlist_wakeup( waitlist, ntohl( buf_hdr->tag ), stats );
  assert( 0 == pthread_mutex_unlock( mu ) );
  if ( body_len > STATIC_SIZE )
    free( body_buf );

  /*
   * We have successfully placed this message in the cache, tell the caller we
   * did not find a match.
   */
  rlen = 0;

  RINGBUFF_RECV_ENTER(source, OP_RECV_MSG_HDR_CACHE, rlen, flags, errno);
  return rlen;
}


// ==============================================================
// msg_read and helper functions

#define NO_MSG_WAITING -1
#define READ_ERROR -2

static void msg_read_setup( int blocking, struct timeval **timeout,
			    struct timeval *zerotime ) {
  /*
   * Pick the proper timeout for select. If this is a non-blocking operation
   * (Iprobe), make sure select does not block. Otherwise, this thread as the
   * only reader, can block for a Probe or Recv operation.
   */
  if ( RCV_BLOCKING == blocking ) {
    /* NULL timeout means means wait forever in select (block) */
    *timeout = NULL;
  } else {
    /* poll and return */
    zerotime->tv_sec = 0;
    zerotime->tv_usec = 0;
    *timeout = zerotime;
  }
}

static int msg_read_select_any( int max_sd, fd_set *fd_readset_ptr,
				int blocking ) {
  int num_sd;

#ifdef DEBUG
  printf( "entering %s(%d)\n", __func__, MPINU_myrank );
  fflush( stdout );
#endif

  /*
   * Setup the timeout for the select call below based on the blocking mode.
   */
  struct timeval *timeout, zerotime;
  msg_read_setup( blocking, &timeout, &zerotime );

  /* if no socket, hopefully listener will accept new connection */
  FD_SET( MPINU_my_list_sd, fd_readset_ptr );
  max_sd = MAX( MPINU_my_list_sd, max_sd ) + 1;
  assert( max_sd <= MPINU_max_sd + 1 );

  /*
   * Use select to find active port.
   *
   * FIX-ME: We need to synchronize which thread is listening for new
   * connections. Only one at a time, though.
   *
   * num_sd is number of socket descriptors with available messages
   */
  CALL_CHK( num_sd = select_block_sigs, ( max_sd, fd_readset_ptr,
	  NULL, NULL, timeout) );
  if ( ( num_sd > 0 ) && ( FD_ISSET( MPINU_my_list_sd, fd_readset_ptr ) ) ) {
    FD_CLR( MPINU_my_list_sd, fd_readset_ptr );
    /* A msg on this socket is a request to open a new socket. */
    list_accept_new_socket( MPINU_my_list_sd );
    // Account for checking the listener socket.
    num_sd--;
  }

#ifdef DEBUG
  printf( "exited %s(%d)\n"
      "timeout == NULL: %d; num_sd: %d; blocking: %d\n",
      __func__, MPINU_myrank, timeout == NULL, num_sd, blocking );
  fflush( stdout );
#endif
  return num_sd;
}

static int msg_read_select( int source, int blocking ) {
  assert( source != MPI_ANY_SOURCE );

#ifdef DEBUG
  printf( "entering %s(%d)\n", __func__, MPINU_myrank );
  fflush( stdout );
  { struct sockaddr_in sin;
    int i = sizeof(struct sockaddr_in);
    CALL_CHK( getsockname, (MPINU_my_list_sd, (struct sockaddr *)&sin, &i) );
    printf( "CHECKING(%d) listener port w/ sd(%d): sin.sin_port: %d\n",
	MPINU_myrank, MPINU_my_list_sd, sin.sin_port );
    fflush( stdout );
  }
#endif

  /*
   * Setup the timeout for the select call below based on the blocking mode.
   */
  struct timeval *timeout, zerotime;
  msg_read_setup( blocking, &timeout, &zerotime );

  /*
   * Translate the source to a file descriptor for use with select(). Add it
   * to the readset, and record the maximum file descriptor number in the set.
   */
  int num_sd, max_sd;
  int fd = MPINU_rank_socket_get( source );

  fd_set fd_readset;
  FD_ZERO( &fd_readset );
  if ( fd != PG_NOSOCKET ) {
    FD_SET( fd, &fd_readset );
    max_sd = fd;
  } else {
    max_sd = 0;
  }

  /*
   * Use select to find active port; if performing a blocking operation, then
   * this thread also needs to check the listener port
   *
   * FIX-ME: We need to synchronize which thread is listening for new
   * connections. Only one at a time, though.
   *
   * num_sd is number of socket descriptors with available messages
   */

  /* if no socket, hopefully listener will accept new connection */
  FD_SET( MPINU_my_list_sd, &fd_readset );
  max_sd = MAX( MPINU_my_list_sd, max_sd ) + 1;
  assert( max_sd <= MPINU_max_sd + 1 );

  CALL_CHK( num_sd = select_block_sigs, ( max_sd, &fd_readset,
	  NULL, NULL, timeout ) );
  if ( ( num_sd > 0 ) && ( FD_ISSET( MPINU_my_list_sd, &fd_readset ) ) ) {
    FD_CLR( MPINU_my_list_sd, &fd_readset );
    /* A msg on this socket is a request to open a new socket. */
    list_accept_new_socket( MPINU_my_list_sd );
    // Account for checking the listener socket.
    num_sd--;
  }
  if ( num_sd > 0 ) {
    assert( FD_ISSET( fd, &fd_readset ) );
  }

#ifdef DEBUG
  printf( "exiting %s(%d)\n"
      "timeout == NULL: %d; num_sd: %d; blocking: %d\n",
      __func__, MPINU_myrank, timeout == NULL, num_sd, blocking );
  fflush( stdout );
#endif
  return num_sd;
}

static int msg_read_get_buf( int source, int tag, msg_hdr_p buf_hdr,
			     void *buf_body, size_t buf_body_len, int eat_hdr,
			     int blocking ) {
  if ( source == MPINU_myrank ) {
    // Read up to MAX_SELF_TAGS at a time.
    int tags[MAX_SELF_TAGS];
    int p = MPINU_rank_socket_get( source );
    char *tag_ptr = (char *)tags;
    int bytes_remaining = sizeof(tags);

    // Drain the pipe, MAX_SELF_TAGS at a time.
    for (;;) {
      int bytes_read = read( p, tag_ptr, bytes_remaining );
      if ( bytes_read < 0 ) {
	if ( errno == EWOULDBLOCK ) {
	  // Nothing left to read, we're done. Assert that we have read a
	  // multiple number of tags.
	  int ts = sizeof(tags[0]);
	  assert( bytes_remaining == ( ( bytes_remaining / ts ) * ts ) );
	  break;
	}
	perror( "read" );
	abort();
      } else {
	bytes_remaining -= bytes_read;
	tag_ptr += bytes_read;
      }
    }
    // We never actually find anything here. The data is always in the cache.
    return FALSE;
  }

  int num_bytes = recv_msg_with_cache( source, tag, buf_hdr,
      sizeof(*buf_hdr), buf_body, buf_body_len, eat_hdr, blocking );

  int ret_val;
  if ( num_bytes <= 0 ) {
    if ( num_bytes < 0 && source <= MPINU_num_slaves ) {
      if ( list_connect_new_socket( MPINU_myrank, source ) == -1 ) {
	socket_dead( source, MPINU_rank_socket_get( source ) );
      }
    }
    // failed to receive; return FALSE to let the caller try again
    ret_val = FALSE;
  } else {
    // Confirm that the source for that socket is the same as source in
    // buf_hdr
    assert( source <= MPINU_num_slaves );
    assert( source == ntohl( buf_hdr->rank ) );
    assert( MPINU_tag_match( tag, ntohl( buf_hdr->tag ) ) );
    ret_val = TRUE;
  }
  return ret_val;
}

/*
 * Pull the data out of a message header into an MPI_Status structure.
 */
static int hdr_to_status( msg_hdr_p buf_hdr, MPI_Status *status ) {
  status->MPI_TAG = ntohl( buf_hdr->tag );
  int size = ntohl( buf_hdr->size );
  status->mpi_size = size;
  status->MPI_SOURCE = ntohl( buf_hdr->rank );
  return size;
}

/*
 * Acquire readership on behalf of the current thread for the given source.
 */
static int reader_tryset( int source ) {
  int reader;
  pthread_t self = pthread_self();

  if ( MPINU_rank_reader_equal( source, NO_READER  ) ) {
    /*
     * No reader present, so this thread is it.
     */
    MPINU_rank_reader_set( source, self );
    reader = TRUE;
  } else {
    /*
     * Another thread is the reader.
     */
    assert( !MPINU_rank_reader_equal( source, self ) );
    reader = FALSE;
  }
  return reader;
}

static int reader_fd_tryset( int source, int *max_sd, fd_set *readfds ) {
  int set;
  if ( MPINU_rank_reader_equal( source, NO_READER  ) ) {
    /*
     * Become the reader for this source.
     */
    int fd = MPINU_rank_socket_get( source );
    if ( fd != PG_NOSOCKET ) {
      /*
       * Only if this source has a socket do we want to play the game.
       */
      MPINU_rank_reader_set( source, pthread_self() );
      FD_SET( fd, readfds );
      *max_sd = MAX( fd, *max_sd );
      assert( *max_sd <= MPINU_max_sd );
      set = TRUE;
    } else {
      set = FALSE;
    }
  } else {
    set = FALSE;
  }
  return set;
}

void reader_unset( int rank ) {
  assert( MPINU_rank_reader_equal( rank, pthread_self() ) );
  MPINU_rank_reader_set( rank, NO_READER );
}

/*
 * Acquire ANY_SOURCE readership on behalf of the current thread.
 */
static int become_anysrc_reader( void ) {
  int reader;
  pthread_t self = pthread_self();

  assert( 0 == pthread_mutex_lock( &anysrc_reader_mu ));
  if ( NO_READER == anysrc_reader ) {
    /*
     * We are now the one and only reader for ANY_SOURCE operations. All
     * other threads performing ANY_SOURCE operations must block on the
     * ANY_SOURCE waitlist if they can't find anything in the cache.
     */
    anysrc_reader = self;
    reader = TRUE;
  } else {
    assert( anysrc_reader != self );
    reader = FALSE;
  }
  assert( 0 == pthread_mutex_unlock( &anysrc_reader_mu ));

  return reader;
}

static void release_anysrc_reader( void ) {
  assert( 0 == pthread_mutex_lock( &anysrc_reader_mu ) );
  assert( pthread_self() == anysrc_reader );
  anysrc_reader = NO_READER;
  /*
   * Pick a waiter and wake it up to become the ANY_SOURCE reader.
   */
  MPINU_waitlist_wakeup_reader( MPINU_anysrc_waitlist,
      &MPINU_anysrc_waitlist_stats );
  assert( 0 == pthread_mutex_unlock( &anysrc_reader_mu ) );
}

/*
 * Handle message receive operations against self.
 */
static int msg_read_self( msg_hdr_p buf_hdr, void *buf, size_t buf_len,
			  int tag, int blocking, int eat_hdr ) {
  int type = RCV_DONT_EAT_HDR == eat_hdr ? MPINU_PROBE : MPINU_RECV;
  int found, source = MPINU_myrank;
  pthread_mutex_t *mu = MPINU_rank_recv_mu_get( source );
  pg_stats_p stats = MPINU_rank_stats_get( source );
  void *waitlist = MPINU_rank_waitlist_get( source );
  void *cache = MPINU_rank_cache_get( source );

  /*
   * The source is the same as for the current executing thread; there is no
   * socket to check. In a single threaded environment, we return
   * NO_MSG_WAITING since the application is coded incorrectly.
   *
   * But, in a threaded environment, it is possible for another thread to
   * enter something in the cache for this source. In that case, this
   * operation will block waiting for something to be entered into the cache.
   */
  assert( 0 == pthread_mutex_lock( mu ) );

  assert( !MPINU_rank_reader_equal( source, pthread_self() ) );

  if ( MPI_THREAD_MULTIPLE == MPINU_thread_level ) {
    if ( RCV_NOT_BLOCKING == blocking ) {
      found = MPINU_find_msg_in_cache( cache, tag, RCV_DONT_EAT_HDR == eat_hdr,
	  buf_hdr, buf, buf_len );
      /*
       * Keep track of some stats.
       */
      stats->find_attempts++;
      if ( !found )
	stats->not_found++;
    } else {
      assert( RCV_BLOCKING == blocking );
      int attempts = 0;
      do {
	found = MPINU_find_msg_in_cache( cache, tag,
	    RCV_DONT_EAT_HDR == eat_hdr, buf_hdr, buf, buf_len );
	attempts++;
	if ( found ) {
	  /*
	   * Keep track of some stats.
	   */
	  stats->find_attempts += attempts;
	  if ( 1 == attempts )
	    stats->found_in_cache_first_try++;
	  if ( RCV_BLOCKING == blocking ) {
	    stats->find_blocking_attempts += attempts;
	    stats->max_single_blocking_attempts = MAX(
		stats->max_single_blocking_attempts,
		attempts );
	  }
	} else {
	  /*
	   * There was already a reader, and this is a Probe or Recv
	   * operation. So block this thread until it can become the reader or
	   * another thread places a message with a matching tag in the cache.
	   */
	  MPINU_waitlist_block( waitlist, mu, tag, type, stats );
	}
      } while ( !found );
    }
  } else {
    assert( MPI_THREAD_SINGLE == MPINU_thread_level );
    found = MPINU_find_msg_in_cache( cache, tag, RCV_DONT_EAT_HDR == eat_hdr,
	buf_hdr, buf, buf_len );
    /*
     * Keep track of some stats.
     */
    stats->find_attempts++;
    if ( !found )
      stats->not_found++;
  }

  assert( 0 == pthread_mutex_unlock( mu ) );
  return found;
}

/*
 * Release any readships the current thread may have acquired across all known
 * sources.
 */
static void release_readerships( char *readership ) {
  int src;
  for ( src = MPINU_num_slaves; src >= 0; src-- ) {
    if ( readership[src] != 0 ) {
      /*
       * This thread is the reader for this source, release it.
       */
      pthread_mutex_t *mu = MPINU_rank_recv_mu_get( src );
      void *waitlist = MPINU_rank_waitlist_get( src );
      pg_stats_p stats = MPINU_rank_stats_get( src );
      assert( 0 == pthread_mutex_lock( mu ) );
      if ( readership[src] > 0 ) {
	reader_unset( src );
	if ( src == MPINU_myrank ) {
	  // ANY_SOURCE reader for myrank; remove the recorded tag.
	  MPINU_rank_anysrc_tag_set( src, MPINU_NOT_A_TAG );
	}
	MPINU_waitlist_wakeup_reader( waitlist, stats );
      } else {
	assert( readership[src] < 0 );
	/*
	 * This thread recorded a tag for this source, remove it.
	 */
	MPINU_rank_anysrc_tag_set( src, MPINU_NOT_A_TAG );
      }
      assert( 0 == pthread_mutex_unlock( mu ) );
      readership[src] = 0;
    }
  }
}

/*
 * Handle non-blocking message receive operations (MPI_Iprobe) for ANY_SOURCE.
 */
static int msg_read_anysrc_nonblk( msg_hdr_p buf_hdr, int tag ) {
  char *readership = (char *)malloc( MPINU_num_slaves + 1 );
  assert( readership );

  int found = FALSE, skipped;
  do {
    skipped = FALSE;
    int readerships = 0, max_sd = 0;
    fd_set readfds;

    FD_ZERO( &readfds );
    memset( readership, 0, MPINU_num_slaves + 1 );

    int src;
    for ( src = 0; src <= MPINU_num_slaves && !found; src++ ) {
      pthread_mutex_t *mu = MPINU_rank_recv_mu_get( src );
      int stat = pthread_mutex_trylock( mu );
      if ( EBUSY == stat ) {
        /*
         * Could not acquire the lock immediately, don't bother on this pass,
         * but make sure we don't block in select either.
         */
        skipped = TRUE;
      } else {
        /*
         * We acquired the lock, now we can check the cache.
         */
        assert( 0 == stat );
	void *cache = MPINU_rank_cache_get( src );
        found = MPINU_find_msg_in_cache( cache, tag, TRUE, buf_hdr, NULL, 0 );
        if ( !found ) {
	  assert( !readership[src] );
	  /*
	   * Try to become the reader for this source since the thread
	   * could not find a matching tag in this source's cache and this
	   * thread is the designated ANY_SOURCE reader.
	   */
	  if ( reader_fd_tryset( src, &max_sd, &readfds ) ) {
	    readership[src]++;
	    readerships++;
	  } else {
	    assert( !MPINU_rank_reader_equal( src, pthread_self() ) );
	  }
	}

	/*
	 * We are done with this source, release the mutex.
	 */
	assert( 0 == pthread_mutex_unlock( mu ) );
      }
    }

    if ( found ) {
      /*
       * Wow, we found a match.
       */

      /*
       * Oh gees, we need to go back to every source and release our
       * readerships.
       */
      release_readerships( readership );
      readerships = 0;
    } else if ( readerships > 0 ) {
      /*
       * NOTE WELL: If this thread was unable to acquire any readerships, it
       * will leave having been satisfied that it could not find anything in
       * the caches.
       */

      // Assert that the maximum file descriptor number in the read is not
      // beyond what the library wide value is.
      assert( max_sd <= MPINU_max_sd );

      /*
       * Check all the sockets for which we are a reader to see if anything is
       * there to be processed.
       */
      int num_sd = msg_read_select_any( max_sd, &readfds, RCV_NOT_BLOCKING );
      if ( -1 == num_sd ) {
	release_readerships( readership );
	readerships = 0;
	skipped = FALSE;	// We want to exit the loop.
      } else {
	// Assert that at most the number of ready sockets is equal to the
	// number of readerships we have.
	assert( num_sd <= readerships );

        for ( src = 0; ( src <= MPINU_num_slaves ) && !found; src++ ) {
          int fd = MPINU_rank_socket_get( src );
          if ( fd == PG_NOSOCKET )
            continue;
          if ( fd > MPINU_max_sd ) {
            printf( "Unknown socket descriptor from select: %d MAX_SD:%d.\n",
		fd, MPINU_max_sd );
            fflush( stdout );
            abort();
          }
          if ( FD_ISSET( fd, &readfds ) ) {
	    assert( fd <= max_sd );
            assert( MPINU_rank_reader_equal( src, pthread_self() ) );
	    num_sd--;
            found = msg_read_get_buf( src, tag, buf_hdr, NULL, 0,
		RCV_DONT_EAT_HDR, RCV_NOT_BLOCKING );
          }
        }
	// Assert that each file descriptor was processed.
	assert( found || 0 == num_sd );

        /*
         * Whether or not we found anything on any of the sockets, we need to
         * release all our readerships.
         */
	release_readerships( readership );
	readerships = 0;
      }
    }
    assert( 0 == readerships );
  } while ( !found && skipped );

  free( readership );
  readership = NULL;
  return found;
}

/*
 * Return 1 for "found", 0 for "not found" and -1 for "error".
 */
static int msg_read_anysrc_reader( msg_hdr_p buf_hdr, void *buf, size_t buf_len,
				   int tag, int eat_hdr ) {
  int error = FALSE;
  int found = FALSE;
  int readerships = 0, marks = 0;
  char *readership = (char *)malloc( MPINU_num_slaves + 1 );
  assert( readership );
  memset( readership, 0, MPINU_num_slaves + 1 );

  int max_sd = 0;
  fd_set readfds;
  FD_ZERO( &readfds );

  /*
   * Check every cache for the tag. If we can't get the lock for the source,
   * remember that so that we won't block in select indefinitely (we'll just
   * come back and try the lock again). If we can get the lock, and its not
   * in the cache, become the reader for this source; if there is already a
   * reader, then remember that so we won't block in select indefinitely (as
   * above). If it is in the cache, then make sure we drop readership on all
   * the previous sources
   */
  int dontblock = 0;

  int src;
  for ( src = 0; src <= MPINU_num_slaves && !found; src++ ) {
    pthread_mutex_t *mu = MPINU_rank_recv_mu_get( src );
    int stat = pthread_mutex_trylock( mu );
    if ( EBUSY == stat ) {
      /*
       * Could not acquire the lock immediately, don't bother on this pass,
       * but make sure we don't block in select either.
       */
      dontblock++;
    } else {
      /*
       * We acquired the lock, now we can check the cache.
       */
      assert( 0 == stat );
      void *cache = MPINU_rank_cache_get( src );
      found = MPINU_find_msg_in_cache( cache, tag,
	  RCV_DONT_EAT_HDR == eat_hdr, buf_hdr, buf, buf_len );
      if ( !found ) {
	assert( !readership[src] );
	/*
	 * Try to become the reader for this source since the thread
	 * could not find a matching tag in this source's cache and this
	 * thread is the designated ANY_SOURCE reader.
	 */
	if ( reader_fd_tryset( src, &max_sd, &readfds ) ) {
	  if ( src == MPINU_myrank ) {
	    // ANY_SOURCE reader for myrank; record the tag we are looking for
	    // so that only the proper messages wake us up.
	    MPINU_rank_anysrc_tag_set( src, tag );
	  }
	  readership[src]++;
	  readerships++;
	} else {
	  assert( !MPINU_rank_reader_equal( src, pthread_self() ) );
	  /*
	   * This thread is the ANY_SOURCE reader, but it is not the reader
	   * for this source. So just mark the source as having an ANY_SOURCE
	   * reader looking for the given tag.
	   */
	  readership[src]--;
	  MPINU_rank_anysrc_tag_set( src, tag );
	  marks++;
	}
      }

      /*
       * We are done with this source, release the mutex.
       */
      assert( 0 == pthread_mutex_unlock( mu ) );
    }
  }

  if ( readerships > 0 ) {
    if ( !found ) {
      /*
       * NOTE WELL: If we are the reader, but were unable to acquire any
       * readerships, we'll keep spinning until we do. FIX-ME.
       */

      // Assert that the maximum file descriptor number in the read is not
      // beyond what the library wide value is.
      assert( max_sd <= MPINU_max_sd );

      /*
       * Check all the sockets for which we are a reader to see if anything is
       * there to be processed. Block only if we are the reader for all
       * sockets.
       *
       * FIX-ME: we need to allow a thread which is the reader on only a few
       * of the sockets to block, having registered itself as a waiter for
       * others.
       */
      int num_sd = msg_read_select_any( max_sd, &readfds,
	  dontblock ? RCV_NOT_BLOCKING : RCV_BLOCKING );

      if ( num_sd < 0 ) {
	/*
	 * An error was encountered while select()ing on the socket list.
	 */
	error = TRUE;
      } else if ( num_sd > 0 ) {
	/*
	 * Assert that at most the number of ready sockets is equal to the
	 * number of readerships we have.
	 */
	assert( num_sd <= readerships );

	for ( src = 0; ( src <= MPINU_num_slaves ) && !found; src++ ) {
	  int fd = MPINU_rank_socket_get( src );
	  if ( fd == PG_NOSOCKET )
	    continue;
	  if ( fd > MPINU_max_sd ) {
	    printf( "Unknown socket descriptor from select: %d MAX_SD:%d.\n",
		fd, MPINU_max_sd );
	    fflush( stdout );
	    abort();
	  }
	  if ( FD_ISSET( fd, &readfds ) ) {
	    assert( fd <= max_sd );
	    assert( MPINU_rank_reader_equal( src, pthread_self() ) );
	    num_sd--;
	    found = msg_read_get_buf( src, tag, buf_hdr, buf, buf_len, eat_hdr,
		RCV_BLOCKING );
	  }
	}
	// Assert that each file descriptor was processed.
	assert( found || 0 == num_sd );
      }
    }

    /*
     * Whether or not we found anything, we need to release our readerhips,
     * as we want to defer processing to threads that are dedicated to a
     * given socket.
     *
     * FIX-ME: Is this good? Having one thread read for many sockets does
     * not scale, but neither do many threads in a system. Ideally, there
     * should be one or two I/O threads per CPU overlapping some
     * computational thread. Multiple sources might better served using a
     * dedicated thread for reading ...
     */
    release_readerships( readership );
    readerships = 0;
    marks = 0;
  }

  free( readership );
  readership = NULL;

  return error ? -1 : found ? 1 : 0;
}

static int msg_read_anysrc_waiter( msg_hdr_p buf_hdr, void *buf, size_t buf_len,
				   int tag, int eat_hdr ) {
  int type = RCV_DONT_EAT_HDR == eat_hdr ? MPINU_PROBE : MPINU_RECV;
  int found = FALSE;
  char *registered = (char *)malloc( MPINU_num_slaves + 1 );
  assert( registered );
  memset( registered, 0, MPINU_num_slaves + 1 );
  int registered_cnt = 0;

  /*
   * Check every cache for the tag. If we can't get the lock for the source,
   * remember that so that we won't block in select indefinitely (we'll just
   * come back and try the lock again). If we can get the lock, and its not
   * in the cache, become the reader for this source; if there is already a
   * reader, then remember that so we won't block in select indefinitely (as
   * above). If it is in the cache, then make sure we drop readership on all
   * the previous sources
   */
  int dontblock = 0;

  int src;
  for ( src = 0; src <= MPINU_num_slaves && !found; src++ ) {
    pthread_mutex_t *mu = MPINU_rank_recv_mu_get( src );
    int stat = pthread_mutex_trylock( mu );
    if ( EBUSY == stat ) {
      /*
       * Could not acquire the lock immediately, don't bother on this pass,
       * but make sure we don't block in select either.
       */
      dontblock++;
    } else {
      /*
       * We acquired the lock, now we can check the cache.
       */
      assert( 0 == stat );
      void *cache = MPINU_rank_cache_get( src );
      found = MPINU_find_msg_in_cache( cache, tag,
	  RCV_DONT_EAT_HDR == eat_hdr, buf_hdr, buf, buf_len );
      if ( !found && !dontblock && !registered[src] ) {
	if ( 0 == registered_cnt ) {
	  /*
	   * This thread is not the reader, and it is able to block. So
	   * initialize a blocking operation on the ANY_SOURCE waitlist
	   * for the given tag. That way, any packets arriving after we
	   * registered with a source, will be recorded on the ANY_SOURCE
	   * waitlist so that this thread will know not to block, but to
	   * recheck all the caches.
	   */
	  assert( 0 == pthread_mutex_lock( &anysrc_reader_mu ) );
	  MPINU_waitlist_block_init( MPINU_anysrc_waitlist, tag, type,
	      &MPINU_anysrc_waitlist_stats );
	  assert( 0 == pthread_mutex_unlock( &anysrc_reader_mu ) );
	}

	/*
	 * Register an ANY_SOURCE waiter on this source for the given tag.
	 */
	void *waitlist = MPINU_rank_waitlist_get( src );
	pg_stats_p stats = MPINU_rank_stats_get( src );
	MPINU_waitlist_register( waitlist, tag, MPINU_anysrc_waitlist,
	    &anysrc_reader_mu, stats, &MPINU_anysrc_waitlist_stats );
	registered[src]++;
	registered_cnt++;
      }

      /*
       * We are done with this source, release the mutex.
       */
      assert( 0 == pthread_mutex_unlock( mu ) );
    }
  }

  if ( registered_cnt > 0 ) {
    if ( !found && !dontblock ) {
      assert( 0 == pthread_mutex_lock( &anysrc_reader_mu ) );
      /*
       * Place ourselves on the ANY_SOURCE waitlist for the specified tag.
       */
      MPINU_waitlist_block_wait( MPINU_anysrc_waitlist, &anysrc_reader_mu, tag,
	  type, &MPINU_anysrc_waitlist_stats );
      assert( 0 == pthread_mutex_unlock( &anysrc_reader_mu ) );
    }

    int src;
    for ( src = 0; src <= MPINU_num_slaves; src++ ) {
      if ( registered[src] ) {
	pthread_mutex_t *mu = MPINU_rank_recv_mu_get( src );
	void *waitlist = MPINU_rank_waitlist_get( src );
	pg_stats_p stats = MPINU_rank_stats_get( src );
	assert( 0 == pthread_mutex_lock( mu ) );
	MPINU_waitlist_unregister( waitlist, tag, MPINU_anysrc_waitlist,
	    &anysrc_reader_mu, stats, &MPINU_anysrc_waitlist_stats );
	if ( found && MPINU_rank_reader_equal( src, NO_READER ) ) {
	  /*
	   * This thread is leaving, so make there is a reader for this
	   * source, if possible.
	   */
	  MPINU_waitlist_wakeup_reader( waitlist, stats );
	}
	assert( 0 == pthread_mutex_unlock( mu ) );
	registered[src]--;
	registered_cnt--;
      }
    }
    assert( 0 == pthread_mutex_lock( &anysrc_reader_mu ) );
    MPINU_waitlist_block_destroy( MPINU_anysrc_waitlist, tag, type,
	&MPINU_anysrc_waitlist_stats );
    if ( found && pthread_equal( anysrc_reader, NO_READER ) ) {
      /*
       * Pick a waiter and wake it up to become the ANY_SOURCE reader since we
       * might have been the targeted reader before we found a packet.
       */
      MPINU_waitlist_wakeup_reader( MPINU_anysrc_waitlist,
	  &MPINU_anysrc_waitlist_stats );
    }
    assert( 0 == pthread_mutex_unlock( &anysrc_reader_mu ) );
    assert( 0 == registered_cnt );
  }

  free( registered );
  registered = NULL;

  return found;
}

/*
 * Handle blocking message receive operations (MPI_Recv / MPI_Probe) for
 * ANY_SOURCE.
 *
 * The goal here is to make sure that support for ANY_SOURCE operations do not
 * negatively impact the performance of normal operations. So if there is a
 * trade-off to be made, the performance of ANY_SOURCE operations will suffer
 * instead of normal operations (at least a little bit).
 *
 * The way ANY_SOURCE operations work is a bit complex due to the nature and
 * semantics of ANY_SOURCE. A thread wishing to receive or [i]probe a message
 * from any source first visits every cache for every source. As it checks
 * each cache, if it does not find a matching message, it marks the associated
 * waitlist's tag entry (creating one if one does not exist) to record that
 * there is an ANY_SOURCE waiter for that tag.
 *
 * Upon finding that there are no matching messages in the cache, the thread
 * will wait on whatever sockets it has become the reader. 
 */
static int msg_read_anysrc( msg_hdr_p buf_hdr, void *buf, size_t buf_len,
			    int tag, int eat_hdr ) {
  int ret_val = 0;
  do {
    if ( become_anysrc_reader() ) {
      ret_val = msg_read_anysrc_reader( buf_hdr, buf, buf_len, tag, eat_hdr );
      release_anysrc_reader();
    } else {
      ret_val = msg_read_anysrc_waiter( buf_hdr, buf, buf_len, tag, eat_hdr );      
    }
  } while ( 0 == ret_val );

  return ret_val;
}

/*
 * Read from a single source
 */
static int msg_read_single( msg_hdr_p buf_hdr, void *buf, size_t buf_len,
			    int source, int tag, int blocking, int eat_hdr ) {
  int type = RCV_DONT_EAT_HDR == eat_hdr ? MPINU_PROBE : MPINU_RECV;
  int found;
  int reader = FALSE;
  int attempts = 0;
  pthread_mutex_t *mu = MPINU_rank_recv_mu_get( source );
  pg_stats_p stats = MPINU_rank_stats_get( source );
  void *waitlist = MPINU_rank_waitlist_get( source );
  void *cache = MPINU_rank_cache_get( source );

  assert( 0 == pthread_mutex_lock( mu ) );

  do {
    found = MPINU_find_msg_in_cache( cache, tag, RCV_DONT_EAT_HDR == eat_hdr,
	buf_hdr, buf, buf_len );
    attempts++;
    if ( found ) {
      /*
       * Keep track of some stats.
       */
      stats->find_attempts += attempts;
      if ( 1 == attempts )
	stats->found_in_cache_first_try++;
      if ( RCV_BLOCKING == blocking ) {
	stats->find_blocking_attempts += attempts;
	stats->max_single_blocking_attempts = MAX(
	    stats->max_single_blocking_attempts,
	    attempts );
      }
    } else {
      /*
       * This thread could not find a match in the cache; it will now try to
       * become the reader for this source.
       */
      reader = reader_tryset( source );
      if ( reader ) {
	/*
	 * Keep track of some stats.
	 */
	stats->find_attempts += attempts;
	if ( 1 == attempts )
	  stats->reader_on_first_try++;
	if ( RCV_BLOCKING == blocking ) {
	  stats->find_blocking_attempts += attempts;
	  stats->max_single_blocking_attempts = MAX(
	      stats->max_single_blocking_attempts,
	      attempts );
	  stats->max_reader_on_attempts = MAX(
	      stats->max_reader_on_attempts,
	      attempts );
	}
      } else {
	if ( RCV_NOT_BLOCKING == blocking ) {
	  /*
	   * Keep track of some stats.
	   */
	  stats->not_found++;
	} else {
	  /*
	   * There was already a reader, and this is a Probe or Recv
	   * operation. So block this thread until it can become the reader or
	   * another thread places a message with a matching tag in the cache.
	   */
	  MPINU_waitlist_block( waitlist, mu, tag, type, stats );
	}
      }
    }
  } while ( !found && !reader && RCV_BLOCKING == blocking );

  if ( found ) {
    /*
     * This thread found an entry in the cache, it should not have also become
     * the reader. Perform this check while still holding the lock.
     */
    assert( !reader );
    assert( !MPINU_rank_reader_equal( source, pthread_self() ) );
  } else if ( reader ) {
    assert( MPINU_rank_reader_equal( source, pthread_self() ) );
  }

  if ( found || ( !reader && RCV_NOT_BLOCKING == blocking ) ) {
    /*
     * This thread is leaving, so make there is a reader for this source, if
     * possible.
     */
    if ( MPINU_rank_reader_equal( source, NO_READER ) ) {
      MPINU_waitlist_wakeup_reader( waitlist, stats );
    }
  }

  assert( 0 == pthread_mutex_unlock( mu ) );

  if ( found || ( !reader && RCV_NOT_BLOCKING == blocking ) ) {
    /*
     * Either this thread found a matching message, or it did not find one for
     * a non-blocking operation (Iprobe) and did not become the reader; in
     * both cases just have it return that result.
     */
    return found;
  }

  /*
   * This thread must be a reader at this point; actuals asserted above under
   * the lock.
   */
  assert( reader );

  /*
   * At this point the thread is the reader for this source. For Iprobe,
   * select() with no wait, and if a message is there, peeking at it and only
   * return it on a match. For Probe and Recv, loop until a match is found,
   * entering messages into the cache when they don't match.
   */
  do {
    /*
     * FIX-ME: Could block in select(), so need to register the fact that we
     * are waiting for a tag, and have another thread wake us up.
     */
    int num_sd = msg_read_select( source, blocking );

    if ( -1 == num_sd ) {
      break;
    } else if ( num_sd > 0 ) {
      assert( 1 == num_sd );
      found = msg_read_get_buf( source, tag, buf_hdr, buf, buf_len, eat_hdr,
	  blocking );
    }
  } while ( RCV_BLOCKING == blocking && !found );

  assert( 0 == pthread_mutex_lock( mu ) );
  reader_unset( source );
  /*
   * Pick a waiter and wake it up to become the reader.
   */
  MPINU_waitlist_wakeup_reader( waitlist, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  return found;
}

static void msg_read_post_check( int eat_hdr, int source, int tag,
				 int size, MPI_Datatype datatype,
				 int count, MPI_Status *status ) {
  assert( size != -1 );
  assert( status->MPI_SOURCE != MPI_ANY_SOURCE );
  if ( !MPINU_tag_match( tag, status->MPI_TAG ) ) {
    printf( "%s: MPINU(rank %d): Msg received with tag (%d) from source (%d)\n"
	"		    incompatible with requested tag (%d).\n",
	__func__, MPINU_myrank, status->MPI_TAG, status->MPI_SOURCE, tag );
    fflush( stdout );
    abort();
  }
  if ( RCV_EAT_HDR == eat_hdr ) {
    if ( count < 0 || !MPINU_datatype_valid( datatype ) ) {
      printf( "%s: invalid count or datatype argument\n", __func__ );
      fflush( stdout );
      abort();
    }
  }
  if ( ( source != status->MPI_SOURCE ) && ( source != MPI_ANY_SOURCE ) ) {
    printf( "%s:  inconsistent source,\n"
	"  expecting source(%d) and received "
	"from source(%d) on proc. %d\n",
	__func__, source, status->MPI_SOURCE, MPINU_myrank );
    fflush( stdout );
    abort();
  }
}

static void msg_read_pre_check( int eat_hdr, MPI_Datatype dt, int count ) {
  if ( RCV_EAT_HDR == eat_hdr ) {
    if ( count < 0 || !MPINU_datatype_valid( dt ) ) {
      printf( "%s: invalid count or datatype argument\n", __func__ );
      fflush( stdout );
      abort();
    }
  }
}

/* returns size (in bytes) of data that follows, or read, or NO_MSG_WAITING
   if no msg */
/*  source:  on entry, MPI_Rank
 *  blocking: on entry, RCV_NOT_BLOCKING means Iprobe (zero timeout)
 *			RCV_BLOCKING means Probe or Recv (NULL timeout)
 *  eat_hdr: on entry, RCV_EAT_HDR means Recv
 *		       RCV_DONT_EAT_HDR means Probe or Iprobe
 *  void	*buf: buffer filled for Recv
 *  MPI_Status	*status:  set on exit
 */
static int msg_read( void *buf, int count, MPI_Datatype datatype, int source,
		     int tag, MPI_Comm comm, MPI_Status *status, int blocking,
		     int eat_hdr ) {
  int size, found;
  msg_hdr_t buf_hdr;

#ifdef DEBUG
  printf( "Entering %s(): MPINU_myrank: %d, source: %d, tag: %#x\n",
      __func__, MPINU_myrank, source, tag );
  fflush( stdout );
#endif

  msg_read_pre_check( eat_hdr, datatype, count );
  size_t buf_len;
  if ( count > 0 ) {
    buf_len = MPINU_msg_size( count, datatype );
  } else {
    buf_len = 0;
  }

  if ( source == MPINU_myrank ) {
    found = msg_read_self( &buf_hdr, buf, buf_len, tag, blocking, eat_hdr );
  } else if ( MPI_ANY_SOURCE == source && RCV_NOT_BLOCKING == blocking ) {
    found = msg_read_anysrc_nonblk( &buf_hdr, tag );
  } else if ( MPI_ANY_SOURCE == source && RCV_BLOCKING == blocking ) {
    found = msg_read_anysrc( &buf_hdr, buf, buf_len, tag, eat_hdr );
  } else {
    found = msg_read_single( &buf_hdr, buf, buf_len, source, tag, blocking,
	eat_hdr );
  }

  if ( found >= 1 ) {
    assert( found == 1 );
    /*
     * Iprobe, Probe or Recv actually found a message to return.
     */
    size = hdr_to_status ( &buf_hdr, status );
    msg_read_post_check( eat_hdr, source, tag, size, datatype, count, status );
  } else if ( 0 == found ) {
    /*
     * Iprobe with no message waiting matching the specified tag.
     */
    size = NO_MSG_WAITING;
  } else {
    assert( found == -1 );
    size = READ_ERROR;
  }

#ifdef DEBUG
  if ( size <= NO_MSG_WAITING ) {
    printf( "%s(): source: %d, dest: %d, size: %d, tag: %#x\n",
	__func__, source, MPINU_myrank, size, tag );
  } else {
    printf( "%s(): source: %d, dest: %d, size: %d, tag: %#x\n",
	__func__, status->MPI_SOURCE, MPINU_myrank, size, status->MPI_TAG );
  }
  fflush( stdout );
#endif

  return size;
}

// ==============================================================
// MPI_Recv, MPI_Probe, etc. (thread-safe, checks for dead sockets)

int MPI_Recv( void *buf, int count, MPI_Datatype datatype, int source,
	     int tag, MPI_Comm comm, MPI_Status *status ) {
  int size, i;
  MPI_Status dummy_status;

  sigset_t orig;
  if ( source > MPINU_num_slaves || source < MPI_ANY_SOURCE || count < 0
      || !MPINU_datatype_valid( datatype )
      || tag < MPI_ANY_TAG || comm != MPI_COMM_WORLD ) {
    printf( "%s( .., %d, %d, %d, %d, ..): invalid argument\n",
	__func__, count, datatype, source, tag );
    fflush( stdout );
    abort();
  }
  if (status == MPI_STATUS_IGNORE)
    status = &dummy_status;
  // After initialization, _all_ receives must go through this mutex.
  // Don't separate receive of message header and message
#ifdef DEBUG
  printf( "%s(MPINU_myrank=%d)\n", __func__, MPINU_myrank ); fflush( stdout );
#endif
  sho_block_async_signals( &orig );
  size = msg_read( buf, count, datatype, source, tag, comm, status,
      RCV_BLOCKING, RCV_EAT_HDR );
  if ( size <= NO_MSG_WAITING ) {
    printf( "%s(%d,%d):  msg_read claims no message waiting.\n",
	__func__, MPINU_myrank, source );
    fflush( stdout );
    abort();
  }
  sho_restore_async_signals( &orig );
  return MPI_SUCCESS;
}

int MPINU_Int_Recv( void *buf, int count, MPI_Datatype datatype, int source,
    int tag, MPI_Comm comm, MPI_Status *status ) {
  int size, i;
  MPI_Status dummy_status;

  sigset_t orig;
  if ( source > MPINU_num_slaves || source < MPI_ANY_SOURCE || count < 0
      || !MPINU_datatype_valid( datatype )
      || tag > MPINU_INIT_TAG || tag < MPINU_COLL_COMM_TAG
      || comm != MPI_COMM_WORLD ) {
    printf( "%s( .., %d, %d, %d, %d, ..): invalid argument\n",
	__func__, count, datatype, source, tag );
    fflush( stdout );
    abort();
  }
  if (status == MPI_STATUS_IGNORE)
    status = &dummy_status;
  // After initialization, _all_ receives must go through this mutex.
  // Don't separate receive of message header and message
#ifdef DEBUG
  printf( "%s(MPINU_myrank=%d)\n", __func__, MPINU_myrank ); fflush( stdout );
#endif
  sho_block_async_signals( &orig );
  size = msg_read( buf, count, datatype, source, tag,
		   comm, status, RCV_BLOCKING, RCV_EAT_HDR );
  if ( size <= NO_MSG_WAITING ) {
    printf( "%s(%d,%d): msg_read claims no message waiting.\n",
	__func__, MPINU_myrank, source );
    fflush( stdout );
    abort();
  }
  sho_restore_async_signals( &orig );
  return MPI_SUCCESS;
}

int MPI_Probe( int source, int tag, MPI_Comm comm, MPI_Status *status ) {
  sigset_t orig;
  sho_block_async_signals( &orig );
#ifdef DEBUG
  printf( "%s(MPINU_myrank=%d)\n", __func__, MPINU_myrank ); fflush( stdout );
#endif
  MPI_Status dummy_status;
  if ( status == MPI_STATUS_IGNORE )
    status = &dummy_status;
  int size = msg_read( NULL, 0, MPI_MIN_DATATYPE, source, tag, comm,
		       status, RCV_BLOCKING, RCV_DONT_EAT_HDR );
  if ( size <= NO_MSG_WAITING ) {
    printf( "%s(%d):  msg_read claims no message waiting.\n", __func__,
	MPINU_myrank );
    fflush( stdout );
    abort();
  }
  sho_restore_async_signals( &orig );
  return MPI_SUCCESS;
}

int MPI_Iprobe( int source, int tag, MPI_Comm comm, int *flag,
	        MPI_Status *status ) {
  sigset_t orig;
  sho_block_async_signals( &orig );
#ifdef DEBUG
  printf( "%s (MPINU_myrank=%d)\n", MPINU_myrank ); fflush( stdout );
#endif
  MPI_Status dummy_status;
  if ( status == MPI_STATUS_IGNORE )
    status = &dummy_status;
  int size = msg_read( NULL, 0, MPI_MIN_DATATYPE, source, tag, comm,
		       status, RCV_NOT_BLOCKING, RCV_DONT_EAT_HDR );
  if ( size <= READ_ERROR ) {
    printf( "%s(%d):  msg_read error.\n", __func__, MPINU_myrank );
    fflush( stdout );
    abort();
  }
  /* flag updated by msg_read, can be checked by caller */
  *flag = ( size >= 0 );
  sho_restore_async_signals( &orig );
  return MPI_SUCCESS;
}

int MPI_Get_count( MPI_Status *status, MPI_Datatype datatype, int *count ) {
  if ( !MPINU_datatype_valid( datatype ) )
    return MPI_FAIL;

  *count = status->mpi_size / MPINU_datatype_size( datatype );
#ifdef DEBUG
  printf( "%s:	count: %d\n", __func__, *count ); fflush( stdout );
#endif
  return MPI_SUCCESS;
}
