/**********************************************************************
 * MPINU				                               *
 * Copyright (c) 2004-2005 Gene Cooperman <gene@ccs.neu.edu>          *
 *                                                                    *
 * This library is free software; you can redistribute it and/or      *
 * modify it under the terms of the GNU Lesser General Public         *
 * License as published by the Free Software Foundation; either       *
 * version 2.1 of the License, or (at your option) any later version. *
 *                                                                    *
 * This library is distributed in the hope that it will be useful,    *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 * Lesser General Public License for more details.                    *
 *                                                                    *
 * You should have received a copy of the GNU Lesser General Public   *
 * License along with this library (see file COPYING); if not, write  *
 * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
 * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
 * <gene@ccs.neu.edu>.                                                *
 **********************************************************************/

#include "mpi.h"
#include "mpiimpl.h"

/*
 * Signal handlers and threads don't mix well. We avoid the very difficult
 * problems the arise trying to debug threaded programs using signal handlers
 * by create a facility that uses a sigwait thread to service objects that
 * describe actions to be taken when a signal occurs.
 */

/*
 * We begin by defining the internal data of the object, represented by an
 * opaque handle to the outside world.
 */

typedef struct isho_tag {
  struct isho_tag	*next;
  int			signal_number;
  void			(*func)(int, void *);
  void			*arg;
} isho_t, *isho_p;

typedef enum states_tag {
  SHOTHD_STARTING = -1,		// sigwait thread starting
  SHOTHD_WAITING = 1,		// in sigwait() call
  SHOTHD_PROCESSING = 2,	// processing objects
  SHOTHD_STOPPING = 3		// sigwait thread stopping
} states_t;

/*
 * Thread handle for the sigwait thread.
 */
static pthread_t	sigwait_thd = (pthread_t)NULL;

/*
 * Mutex protecting the array of actions for each signal.
 */
static pthread_mutex_t	sigwait_mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * State of all sigwait thread.
 */
static states_t		sigwait_state = SHOTHD_STARTING;

/*
 * The array of signal objects.
 */
#define	MAX_SIG		64
static isho_p		sigwait_objects[MAX_SIG] = { 0 };

/*
 * Set all the bits for asynchronous signals in the given signal set.
 */
static void set_async_signals (sigset_t *set)
{
  // Fill the signal set to cover all signals
  sigfillset (set);

  // Remove the ones we know are synchronous
  sigdelset (set, SIGBUS);
#ifdef SIGEMT
  sigdelset (set, SIGEMT);
#endif
  sigdelset (set, SIGFPE);
  sigdelset (set, SIGILL);
  sigdelset (set, SIGIOT);
  sigdelset (set, SIGPIPE);
  sigdelset (set, SIGSEGV);
  sigdelset (set, SIGSYS);
  sigdelset (set, SIGTRAP);
}

/*
 * Thread routine for the sigwaiter thread.
 */
static void *sigwait_rtn (void *arg)
{
  int		status;
  int		sig;
  sigset_t	set;

  /*
   * Create a set of signals that don't include the synchronous ones. We'll
   * use this set of signal in the call to sigwait() below.
   */
  set_async_signals (&set);

  status = pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL);
  if (0 != status)
    exit (status);

  status = pthread_mutex_lock (&sigwait_mutex);
  if (0 != status)
    exit (status);

  assert (SHOTHD_STARTING == sigwait_state);
  sigwait_state = SHOTHD_WAITING;

  status = pthread_mutex_unlock (&sigwait_mutex);
  if (0 != status)
    exit (status);

  while (TRUE) {
    status = sigwait (&set, &sig);
    if (0 != status)
      exit (status);

    isho_p pCur = sigwait_objects[sig];
    if (pCur) {
      status = pthread_mutex_lock (&sigwait_mutex);
      if (0 != status)
	exit (status);

      // Re-read to make sure we get the proper value now that we hold
      // the lock. Note that we hold the lock while processing the whole
      // list so that we don't have to do anything fancy to deal with
      // concurrent inserts or removes.
      pCur = sigwait_objects[sig];
      while (pCur) {
	assert (sig == pCur->signal_number);
	(*pCur->func)(sig, pCur->arg);
	pCur = pCur->next;
      }

      status = pthread_mutex_unlock (&sigwait_mutex);
      if (0 != status)
	exit (status);
    }
  }
}

/*
 * Block all asynchronous signals for the calling thread. We assume the POSIX
 * signal model here: signals are blocked per-thread, handlers are
 * per-process.
 */
void sho_block_async_signals (sigset_t *save) {
  sigset_t	set;

  // Fill in the set of asynchronous signals.
  set_async_signals (&set);

  // Now block them for this thread
  int status = pthread_sigmask (SIG_BLOCK, &set, save);
  if (status != 0) {
    fprintf (stderr, "Unable to block signals for thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }
}

/*
 * Restore the given signal set to the thread.
 */
void sho_restore_async_signals (sigset_t *set) {
  int status = pthread_sigmask (SIG_SETMASK, set, NULL);
  if (status != 0) {
    fprintf (stderr, "Unable to restore async signals for thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }
}

/*
 * Initialize the signal handling thread. 
 */
int sho_init (void)
{
  /*
   * Eliminate broken socket exits. SO_KEEPALIVE can generate SIGPIPE.
   */
  signal (SIGPIPE, SIG_IGN);

  /*
   * Create the sigwait() thread.
   */
  int status = pthread_create (&sigwait_thd, NULL, sigwait_rtn, NULL);
  if (status != 0) {
    fprintf (stderr, "Unable to create sigwait() thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }

  return status;
}

int sho_finalize (void)
{
#if 0
  // Post a cancel against the thread in case it is in sigwait.
  int status = pthread_cancel (sigwait_thd);
  if (status != 0) {
    fprintf (stderr, "Unable to cancel sigwait() thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }

  /*
   * Send it an asynchronous signal (FIX-ME: which one that the user won't
   * have overriden)
   */
  status = pthread_kill (sigwait_thd, SIGUSR1);
  if (status != 0) {
    fprintf (stderr, "Unable to kill sigwait() thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }

  // Make sure it is gone before returning.
  status = pthread_join (sigwait_thd, NULL);
  if (status != 0) {
    fprintf (stderr, "Unable to join with sigwait() thread, (%d) %s!\n",
	status, strerror (status));
    exit (status);
  }
#endif

  return 0;
}

/*
 * Allocate a sho, given a signal number, a function and an argument for that
 * function.
 */
sho_t sho_alloc (int sn, void (*f)(int, void *), void *a)
{
  assert ((sn <= MAX_SIG) && (sn >= 1));

  isho_p	sho = (isho_p)malloc (sizeof (isho_t));
  if (NULL != sho) {
    sho->next = NULL;
    sho->signal_number = sn;
    sho->func = f;
    sho->arg = a;
  }
  return sho;
}

/*
 * Activate the sho so that it will start handling signals.
 */
int sho_activate (sho_t osho)
{
  int		status, sn;
  isho_p	sho = (isho_p)osho;

  sn = sho->signal_number;
  assert (NULL == sho->next);
  assert ((sn <= MAX_SIG) && (sn >= 1));

  status = pthread_mutex_lock (&sigwait_mutex);
  if (0 == status) {
    /*
     * Insert the object into the array based on its signal number.
     */
    sho->next = sigwait_objects[sn];
    sigwait_objects[sn] = sho;

    pthread_mutex_unlock (&sigwait_mutex);
  }

  return status;
}

/*
 * Deactivate the sho so that signals won't be handled any longer.
 */
int sho_deactivate (sho_t osho)
{
  int		status = ENOENT, sn;
  isho_p	sho = (isho_p)osho;

  sn = sho->signal_number;
  assert ((sn <= MAX_SIG) && (sn >= 1));

  status = pthread_mutex_lock (&sigwait_mutex);
  if (0 == status) {
    isho_p	pCur, pPrv;
    for (pCur = sigwait_objects[sho->signal_number], pPrv = NULL;
	 pCur != NULL;
	 pPrv = pCur, pCur = pCur->next) {
      if (pCur == sho) {
	// Found it.
	if (pPrv == NULL) {
	  // Remove it from the array pointer.
	  sigwait_objects[sho->signal_number] = pCur->next;
	}
	else {
	  // Remove it from between two elements or at the end of
	  // the array.
	  pPrv->next = pCur->next;
	}
	// Return success
	status = 0;
      }
    }
    pthread_mutex_unlock (&sigwait_mutex);
  }

  return status;
}

/*
 * Free the sho and any memory it might have allocated to it.
 */
void sho_free (sho_t osho)
{
  isho_p	sho = (isho_p)osho;

  if (NULL != sho) {
    assert (sho->next == NULL);
    free (sho);
  }
}
