#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "mpi.h"

typedef enum { WORKTAG1, WORKTAG2, WORKTAG3, BARRIERTAG } work_type_t;

#define WORKVAL	0x00abad00

#define WORKLEN 500000

#define POISON	0x8badb10d

/*
 * We do a barrier the hard way. Everyone sends a message of 400 KB (something
 * larger than the network buffer), and then they receive it. This requires
 * the use of MPI_Bsend (or possibly MPI_Isend) to guarantee no deadlock. But
 * if the macro TOPC is not defined, MPI_Send also works.
 */
static void barrier( int myrank, int maxrank ) {
  int		rank, i;
  MPI_Status	status;
  int		*work = (int *)calloc( WORKLEN, sizeof(int) );
  assert( work );
  for ( i = 0; i < WORKLEN; i++ ) {
    work[i] = POISON;
  }

  for ( rank = 0; rank < maxrank; rank++ ) {
    if ( rank != myrank )
      /* This will turn into either MPI_Bsend or MPI_Isend
       * in the future. */
      MPI_Send( work, WORKLEN, MPI_INT, rank, BARRIERTAG, MPI_COMM_WORLD );
  }
  for ( rank = 0; rank < maxrank; rank++ )
    if ( rank != myrank ) {
      memset( work, 0, WORKLEN * sizeof(int) );
      MPI_Recv( work, WORKLEN, MPI_INT, rank, MPI_ANY_TAG, MPI_COMM_WORLD,
	  &status );
      assert( rank == status.MPI_SOURCE );
      if ( status.MPI_TAG != BARRIERTAG ) {
	fprintf( stderr, "MPI: Application bug:  application message"
	    " being sent across barrier from Rank %d to Rank %d\n",
	    rank, myrank );
      }
      assert( BARRIERTAG == status.MPI_TAG );
      int count;
      MPI_Get_count( &status, MPI_INT, &count );
      assert( WORKLEN == count );
      for ( i = 0; i < WORKLEN; i++ ) {
	assert( POISON == work[i] );
      }
    }

  /* Could do MPI_Testall here, if we were using MPI_Isend */
  free( work );
}

/*
 * Name:	hello_mpi2
 *
 * Purpose:	A) On Rank 0, verify that messages can be sent to self and
 *		   properly received by Iprobe, Probe and Recv.
 *		B) On Rank 1, verify that messages can be received out of
 *		   order.
 *		C) Verify an all-to-all barrier with sends followed by
 *		   receives works as expected.
 *		D) Verify that two large messages can be first sent from each
 *		   node to the other node, and then received by each node.
 *
 * Slaves:	1, as additional slaves will only participate in the barrier.
 */
int main( int argc, char *argv[] ) {
  int		myrank, maxrank, work_send, work_recv, count;
  MPI_Status	status;

  MPI_Init( &argc, &argv );		/* initialize MPI */
  // always use this process rank, 0 thru N-1
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  // always use this #processes in application = N
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  /* Test ability to send to self. */

  printf( "Sending messages to self.\n" ); fflush( stdout );

  work_send = WORKVAL;
  MPI_Send( &work_send, 1, MPI_INT, myrank, WORKTAG1, MPI_COMM_WORLD );
  MPI_Send( &work_send, 1, MPI_INT, myrank, WORKTAG1, MPI_COMM_WORLD );

  int flag = 0;
  MPI_Iprobe( myrank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
  assert( 1 == flag );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  MPI_Probe( myrank, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  work_recv = 0;
  MPI_Recv( &work_recv, 1, MPI_INT, myrank, WORKTAG1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
  assert( work_recv == work_send );

  flag = 0;
  MPI_Iprobe( myrank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
  assert( 1 == flag );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  MPI_Probe( myrank, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  work_recv = 0;
  MPI_Recv( &work_recv, 1, MPI_INT, myrank, WORKTAG1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( WORKTAG1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
  assert( work_recv == work_send );

  MPI_Iprobe( myrank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
  assert( 0 == flag );

  /* Test out-of-order receives when tags arrive out of order. */
  if ( 0 == myrank ) { /* Send in reverse order */
    printf( "Sending 3 worktags and receiving out of order.\n" );
    fflush( stdout );

    work_send = 1001;
    MPI_Send( &work_send, 1, MPI_INT, 1, WORKTAG1, MPI_COMM_WORLD );
    work_send = 1002;
    MPI_Send( &work_send, 1, MPI_INT, 1, WORKTAG2, MPI_COMM_WORLD );
    work_send = 1003;
    MPI_Send( &work_send, 1, MPI_INT, 1, WORKTAG3, MPI_COMM_WORLD );
  } else if ( 1 == myrank ) { /* Can slave receive in correct order? */
    MPI_Recv( &work_send, 1, MPI_INT, 0, WORKTAG3, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( WORKTAG3 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 1003 == work_send );

    MPI_Recv( &work_send, 1, MPI_INT, 0, WORKTAG1, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( WORKTAG1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 1001 == work_send );

    MPI_Recv( &work_send, 1, MPI_INT, 0, WORKTAG2, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( WORKTAG2 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 1002 == work_send );
  }

  /*
   * Test nonblocking sends.
   * All-to-all sends followed by all-to-all recv.
   */
  if ( 0 == myrank ) {
    printf( "Testing nonblocking sends via all-to-all barrier.\n" );
    fflush( stdout );
  }
  barrier( myrank, maxrank );

  /*
   * Test nonblocking sends of very large messages.
   *   This forces a deadly embrace unless one of the two processes
   *   can send two large messages, and begin a receive, while the
   *   other process has not yet begun any receive.
   * However, it appears that some O/S (e.g. Linux 2.6.12 ??) have
   *   a trick by which even if the master blocks on MPI_Send, when the
   *   slave enters MPI_Send, this somehow blocks the master.
   */
  int	*workarray = (int *)calloc( WORKLEN, sizeof(int) );
  assert( workarray );
  int i;
  for ( i = 0; i < WORKLEN; i++ ) {
    workarray[i] = POISON;
  }

  if ( 0 == myrank ) {
    printf( "Sending two large messages before a receive at other end.\n" );
    fflush( stdout );

    MPI_Send( workarray, WORKLEN, MPI_INT, 1, WORKTAG1,	MPI_COMM_WORLD );
    MPI_Send( workarray, WORKLEN, MPI_INT, 1, WORKTAG1,	MPI_COMM_WORLD );

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 1, WORKTAG1,	MPI_COMM_WORLD,
	&status );
    assert( 1 == status.MPI_SOURCE );
    assert( WORKTAG1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( POISON == workarray[i] );
    }

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 1, WORKTAG1,	MPI_COMM_WORLD,
	&status );
    assert( 1 == status.MPI_SOURCE );
    assert( WORKTAG1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( POISON == workarray[i] );
    }
  } else if ( 1 == myrank ) {
    MPI_Send( workarray, WORKLEN, MPI_INT, 0, WORKTAG1,	MPI_COMM_WORLD );
    MPI_Send( workarray, WORKLEN, MPI_INT, 0, WORKTAG1,	MPI_COMM_WORLD );

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 0, WORKTAG1,	MPI_COMM_WORLD,
	&status );
    assert( 0 == status.MPI_SOURCE );
    assert( WORKTAG1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( POISON == workarray[i] );
    }

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 0, WORKTAG1,	MPI_COMM_WORLD,
	&status );
    assert( 0 == status.MPI_SOURCE );
    assert( WORKTAG1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( POISON == workarray[i] );
    }
  }
  free( workarray );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
