#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"

typedef enum { WORKTAG1, WORKTAG2, WORKTAG3, BARRIERTAG } work_type_t;

#define WORKLEN 500000

/*
 * We do a barrier the hard way. Everyone sends a message of 400 KB (something
 * larger than the network buffer), and then they receive it. This requires
 * the use of MPI_Bsend (or possibly MPI_Isend) to guarantee no deadlock. But
 * if the macro TOPC is not defined, MPI_Send also works.
 */
static void barrier( int myrank, int maxrank ) {
  int		rank;
  MPI_Status	status;
  int		*work = (int *)malloc( sizeof(int) * WORKLEN );
  assert( work );

  for ( rank = 0; rank < maxrank; rank++ ) {
    if ( rank != myrank )
      MPI_Send( work, WORKLEN, MPI_INT, rank, BARRIERTAG, MPI_COMM_WORLD );
  }
  for ( rank = 0; rank < maxrank; rank++ )
    if ( rank != myrank ) {
      MPI_Recv( work, WORKLEN, MPI_INT, rank, MPI_ANY_TAG, MPI_COMM_WORLD,
	  &status );
      assert( rank == status.MPI_SOURCE );
      if ( status.MPI_TAG != BARRIERTAG ) {
	fprintf( stderr, "MPI: Application bug:  application message"
	    " being sent across barrier from Rank %d to Rank %d\n",
	    rank, myrank );
      }
      assert( BARRIERTAG == status.MPI_TAG );
      int count;
      MPI_Get_count( &status, MPI_INT, &count );
      assert( WORKLEN == count );
    }
  free( work );
}

static void testRecvRank1(void) {
  int		work;
  MPI_Status	status;

  puts( "1 Sending 0 T1, and 0 T3." );
  fflush( stdout );

  work = 1001;
  MPI_Send( &work, 1, MPI_INT, 0, WORKTAG1, MPI_COMM_WORLD );
  work = 1003;
  MPI_Send( &work, 1, MPI_INT, 0, WORKTAG3, MPI_COMM_WORLD );

  // Now on rank 0, its rank 1 cache will have work 1001 in
  // it, as it will have only received the work 1003 so far.
}

static void testRecvRank0(void) {
  int		work;
  MPI_Status	status;

  MPI_Recv( &work, 1, MPI_INT, 1, WORKTAG3, MPI_COMM_WORLD, &status );
  assert( 1003 == work );
  // At this point on rank 0, the cache for rank 1 will have work 1001
  // and 1003 in it.

  puts( "0 Sending 0 T2." );
  fflush( stdout );

  work = 1002;
  MPI_Send( &work, 1, MPI_INT, 0, WORKTAG2, MPI_COMM_WORLD );
  // At this point on rank 0, the cache for rank 0 will have work
  // 1002, and cursor for that cache will be at the beginning of the
  // buffer.

  MPI_Recv( &work, 1, MPI_INT, MPI_ANY_SOURCE, WORKTAG1, MPI_COMM_WORLD,
      &status );
  assert( 1001 == work );
  // This above receive should not crash; but at a certain point in time,
  // the ANY source selection would see that the rank 0 cache has something
  // in it, but since it did not check the tag, it would try to read from
  // the network and crash.

  MPI_Recv( &work, 1, MPI_INT, 0, WORKTAG2, MPI_COMM_WORLD, &status );
  assert( 1002 == work );
}

/*
 * Name:	hello_mpi3
 *
 * Purpose:	Test that the cache for rank 0 is not consulted on rank 0 when
 *		an ANY_SOURCE receive for a specific TAG occurs.
 *
 * Slaves:	> 1, so that some MPI internal setup communication must have
 *		occurred.
 */
int main( int argc, char *argv[] ) {
  int myrank, maxrank;

  MPI_Init( &argc, &argv );	/* initialize MPI */
  // always use this process rank, 0 thru N-1
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  // always use this #processes in application = N
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  if ( 1 == myrank ) {
    testRecvRank1();
  }
  else if ( 0 == myrank ) {
    testRecvRank0();
  }

  barrier( myrank, maxrank );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
