#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"

typedef enum { WORKTAG1, WORKTAG2, WORKTAG3, BARRIERTAG } work_type_t;
typedef enum { WITH_IPROBE, WITHOUT_IPROBE } barrier_type_t;

#define WORKLEN 500000

/*
 * We do a barrier the hard way. Everyone sends a message of 400 KB (something
 * larger than the network buffer), and then they receive it. This requires
 * the use of MPI_Bsend (or possibly MPI_Isend) to guarantee no deadlock. But
 * if the macro TOPC is not defined, MPI_Send also works.
 */
static void barrier( int myrank, int maxrank, barrier_type_t type ) {
  int		rank;
  MPI_Status	status;
  int		*work = (int *)malloc( sizeof(int) * WORKLEN );
  assert( work );

  for ( rank = 0; rank < maxrank; rank++ ) {
    if ( rank != myrank )
      MPI_Send( work, WORKLEN, MPI_INT, rank, BARRIERTAG, MPI_COMM_WORLD );
  }
  for ( rank = 0; rank < maxrank; rank++ )
    if ( rank != myrank ) {
      int count;
      if ( WITH_IPROBE == type ) {
	int flag = 0;
	MPI_Iprobe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
	while ( 0==flag ) {
	  sleep( 1 );
	  MPI_Iprobe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
	}
	assert( flag != 0 );
	assert( rank == status.MPI_SOURCE );
	if ( status.MPI_TAG != BARRIERTAG ) {
	  fprintf( stderr, "MPI: Application bug:  application message"
	      " being sent across barrier from Rank %d to Rank %d\n",
	      rank, myrank );
	}
	assert( BARRIERTAG == status.MPI_TAG );
	MPI_Get_count( &status, MPI_INT, &count );
	assert( WORKLEN == count );
      }
      else
	assert( WITHOUT_IPROBE == type );
      MPI_Recv( work, WORKLEN, MPI_INT, rank, MPI_ANY_TAG, MPI_COMM_WORLD,
	  &status );
      assert( rank == status.MPI_SOURCE );
      if ( status.MPI_TAG != BARRIERTAG ) {
	fprintf( stderr, "MPI: Application bug:  application message"
	    " being sent across barrier from Rank %d to Rank %d\n",
	    rank, myrank );
      }
      assert( BARRIERTAG == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( WORKLEN == count );
    }

  free( work );
}

/*
 * Name:	hello_mpi4
 *
 * Purpose:	Verify an all-to-all barrier with sends followed by
 *		receives works as expected, with and without Iprobes
 *
 * Slaves:	> 1, enough to have sufficient participants in the barriers
 */
int main( int argc, char *argv[] ) {
  int		myrank, maxrank, work;
  MPI_Status	status;

  MPI_Init( &argc, &argv );		/* initialize MPI */
  // always use this process rank, 0 thru N-1
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  // always use this #processes in application = N
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );


  /*
   * Test barrier without Iprobes first, since that is what the other
   * tests use.
   */
  if ( 0 == myrank ) {
    printf( "Barrier without Iprobes\n" );
    fflush( stdout );
  }
  barrier( myrank, maxrank, WITHOUT_IPROBE );

  /* Test barrier with Iprobes. */
  if ( 0 == myrank ) {
    printf( "Barrier with Iprobes\n" );
    fflush( stdout );
  }
  barrier( myrank, maxrank, WITH_IPROBE );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
