#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"

typedef enum { WORKTAG1, WORKTAG2, WORKTAG3, WORKTAG4, BARRIERTAG, DIETAG } work_type_t;

#define WORKLEN 500000

/*
 * We do a barrier the hard way. Everyone sends a message of 400 KB (something
 * larger than the network buffer), and then they receive it. This requires
 * the use of MPI_Bsend (or possibly MPI_Isend) to guarantee no deadlock. But
 * if the macro TOPC is not defined, MPI_Send also works.
 */
static void barrier( int myrank, int maxrank ) {
  int		rank;
  MPI_Status	status;
  int		*work = (int *)malloc( sizeof(int) * WORKLEN );
  assert( work );

  for ( rank = 0; rank < maxrank; rank++ ) {
    if ( rank != myrank )
      MPI_Send( work, WORKLEN, MPI_INT, rank, BARRIERTAG, MPI_COMM_WORLD );
  }
  for ( rank = 0; rank < maxrank; rank++ )
    if ( rank != myrank ) {
      MPI_Recv( work, WORKLEN, MPI_INT, rank, MPI_ANY_TAG, MPI_COMM_WORLD,
	  &status );
      assert( rank == status.MPI_SOURCE );
      if ( status.MPI_TAG != BARRIERTAG ) {
	fprintf( stderr, "MPI: Application bug:  application message"
	    " being sent across barrier from Rank %d to Rank %d\n",
	    rank, myrank );
      }
      assert( BARRIERTAG == status.MPI_TAG );
      int count;
      MPI_Get_count( &status, MPI_INT, &count );
      assert( WORKLEN == count );
    }
  free( work );
}

/*
 * Name:	hello_mpi5a
 *
 * Purpose:	Test the use of MPI_ANY_SOURCE with Iprobe().
 *
 * Slaves:	2, as additional slaves will only participate in the barrier.
 */
int main( int argc, char *argv[] ) {
  int		myrank, maxrank, rank, work;
  MPI_Status	status;

  MPI_Init( &argc, &argv );		/* initialize MPI */

  /*
   * Always use this; process rank, 0 thru N-1
   */
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );

  /*
   * Always use this; #processes in application = N
   */
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  switch ( myrank ) {
    case 1: {
      puts( "Rank 1 Sending Rank 0 T1 and T2." ); fflush( stdout );
      work = 1001;
      MPI_Send( &work, 1, MPI_INT, 0, WORKTAG1, MPI_COMM_WORLD );
      work = 1003;
      MPI_Send( &work, 1, MPI_INT, 0, WORKTAG2, MPI_COMM_WORLD );
      // Now on rank 0, its rank 1 cache will have work 1001 in
      // it, as it will have only received the work 1003 so far.

      /*
       * Now have all the other ranks block waiting for Rank 0 to tell them to
       * finish up.
       */
      MPI_Recv( &work, 1, MPI_INT, 0, DIETAG, MPI_COMM_WORLD, &status );
      assert( 0 == work );
      break;
    }
    case 2: {
      puts( "Rank 2 Sending Rank 0 T4 and T3." ); fflush( stdout );
      work = 1005;
      MPI_Send( &work, 1, MPI_INT, 0, WORKTAG4, MPI_COMM_WORLD );
      work = 1004;
      MPI_Send( &work, 1, MPI_INT, 0, WORKTAG3, MPI_COMM_WORLD );
      // Now on rank 0, its rank 2 cache will have work 1005 in
      // it, as it will have only received the work 1004 so far.

      /*
       * Now have all the other ranks block waiting for Rank 0 to tell them to
       * finish up.
       */
      MPI_Recv( &work, 1, MPI_INT, 0, DIETAG, MPI_COMM_WORLD, &status );
      assert( 0 == work );
      break;
    }
    case 0: {
      puts( "Rank 0 Iprobing for Rank 1's T1." ); fflush( stdout );

      // Nothing in the cache at all at this point.
      int flag = 0, count;
      do {
	MPI_Iprobe( MPI_ANY_SOURCE, WORKTAG1, MPI_COMM_WORLD, &flag, &status );
      }
      while ( 0 == flag );
      assert( 1 == status.MPI_SOURCE );
      assert( WORKTAG1 == status.MPI_TAG );
      MPI_Get_count ( &status, MPI_INT, &count );
      assert( 1 == count );
      // Still nothing in the cache as the Iprobe() should have only
      // found the message header in the socket.

      puts( "Rank 0 Iprobing for Rank 2's T4." ); fflush( stdout );

      flag = 0;
      do {
	MPI_Iprobe( MPI_ANY_SOURCE, WORKTAG4, MPI_COMM_WORLD, &flag, &status );
      }
      while ( 0 == flag );
      assert( 2 == status.MPI_SOURCE );
      assert( WORKTAG4 == status.MPI_TAG );
      MPI_Get_count ( &status, MPI_INT, &count );
      assert( 1 == count );
      // Still nothing in the cache as the Iprobe() should have only
      // found the message header in the socket.

      // Iprobe() T1 and T4 again, just to make sure we can find them on
      // the socket.

      puts( "Rank 0 Iprobing for Rank 1's T1, and Rank 2's T4." );
      fflush( stdout );

      MPI_Iprobe( MPI_ANY_SOURCE, WORKTAG1, MPI_COMM_WORLD, &flag, &status );
      assert( 1 == status.MPI_SOURCE );
      assert( WORKTAG1 == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );

      MPI_Iprobe( MPI_ANY_SOURCE, WORKTAG4, MPI_COMM_WORLD, &flag, &status );
      assert( 2 == status.MPI_SOURCE );
      assert( WORKTAG4 == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );

      // Consume all the messages
      puts( "Rank 0 Consume Rank 1's T2." ); fflush( stdout );
      MPI_Recv( &work, 1, MPI_INT, 1, WORKTAG2, MPI_COMM_WORLD, &status );
      assert( 1 == status.MPI_SOURCE );
      assert( WORKTAG2 == status.MPI_TAG );
      assert( 1003 == work );
      puts( "Rank 0 Consume Rank 2's T3." ); fflush( stdout );
      MPI_Recv( &work, 1, MPI_INT, 2, WORKTAG3, MPI_COMM_WORLD, &status );
      assert( 2 == status.MPI_SOURCE );
      assert( WORKTAG3 == status.MPI_TAG );
      assert( 1004 == work );
      puts( "Rank 0 Consume Rank 1's T1." ); fflush( stdout );
      MPI_Recv( &work, 1, MPI_INT, 1, WORKTAG1, MPI_COMM_WORLD, &status );
      assert( 1 == status.MPI_SOURCE );
      assert( WORKTAG1 == status.MPI_TAG );
      assert( 1001 == work );
      puts( "Rank 0 Consume Rank 2's T4." ); fflush( stdout );
      MPI_Recv( &work, 1, MPI_INT, 2, WORKTAG4, MPI_COMM_WORLD, &status );
      assert( 2 == status.MPI_SOURCE );
      assert( WORKTAG4 == status.MPI_TAG );
      assert( 1005 == work );

      // We are done, tell the rank and file to exit.
      for ( rank = 1; rank < maxrank; rank++ ) {
	printf( "Rank 0 kill Rank %d.\n", rank ); fflush( stdout );
	int zero = 0;
	MPI_Send( &zero, 1, MPI_INT, rank, DIETAG, MPI_COMM_WORLD );
      }
      break;
    }
    default:
      // Any other rank does nothing but the barrier.
      break;
  }

  // All ranks participate in the barrier.
  barrier( myrank, maxrank );

  printf( "Rank %d finalizing...\n", myrank ); fflush( stdout );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
