#include <stdio.h>
#include <assert.h>
#include "mpi.h"

/*
 * The work tag will have "lead" in the upper 16 bits, and the sending rank in
 * the next lower 8 bits, and a unique tag in the lowest 8 bits.
 */
#define TAG_WORK		0x1ead0000

/*
 * There will be TAG_COUNT number of messages sent from all slaves to the
 * master, each with its own tag.
 */
#define TAG_MAX			6

/*
 * Each of the TAG_MAX tags sent by the slaves takes the following form.
 */
#define MakeTag(_r_,_id_)	(TAG_WORK | ((_r_) << 8) | (_id_))

static void sendRankX( int myrank ) {
  int		i;
  int		tag;
  MPI_Status	status;

  /*
   * Send TAG_MAX distinct tags to rank 0. Rank 0 will end up placing some in
   * the cache, and leaving others on the socket at various times.
   */
  for ( i = 0; i < TAG_MAX; i++ ) {
    printf( "%d Sending 0 T%d\n", myrank, i );
    fflush( stdout );

    tag = MakeTag(myrank,i);
    MPI_Send( &i, 1, MPI_INT, 0, tag, MPI_COMM_WORLD );

    /*
     * Now have all the other ranks block waiting for Rank 0 to tell them to
     * finish up.
     */
    int next_i;
    MPI_Recv( &next_i, 1, MPI_INT, 0, TAG_WORK, MPI_COMM_WORLD, &status );
    assert( i+1 == next_i );
  }
}

static void verifyCache( int tag_in_socket, int maxrank ) {
  int		flag, count, tag, rank;
  MPI_Status	status;

  // Now verify that if we use MPI_ANY_SOURCE, we should be able to Iprobe()
  // for all the tags in the cache and the socket for a given rank, and not
  // find the tags that are still on the wire.
  for ( tag = 0; tag <= tag_in_socket; tag++ ) {
    for ( rank = 1; rank < maxrank; rank++ ) {
      flag = 0;
      MPI_Iprobe( MPI_ANY_SOURCE, MakeTag(rank,tag), MPI_COMM_WORLD, &flag,
	  &status );
      assert( 1 == flag );
      assert( status.MPI_SOURCE == rank );
      assert( status.MPI_TAG == MakeTag(rank,tag) );
      MPI_Get_count(&status, MPI_INT, &count );
      assert( 1 == count );
    }
  }

  // Iprobe() one beyond those sent just to be sure we don't find something
  // they did not send.
  for ( tag = (tag_in_socket+1); tag <= TAG_MAX; tag++ ) {
    for ( rank = 1; rank < maxrank; rank++ ) {
      flag = 1;
      MPI_Iprobe( MPI_ANY_SOURCE, MakeTag(rank,tag), MPI_COMM_WORLD, &flag,
	  &status );
      assert( 0 == flag );
    }
  }

  // Verify that we should be able to do the operation as above using Iprobe()
  // for all the tags in the cache and the socket for a given rank, and not
  // find the tags that have not been sent.
  int src;
  for ( src = 1; src < maxrank; src++ ) {
    for ( tag = 0; tag <= tag_in_socket; tag++ ) {
      for ( rank = 1; rank < maxrank; rank++ ) {
	flag = 0;
	MPI_Iprobe( src, MakeTag(rank,tag), MPI_COMM_WORLD, &flag, &status );
	if ( src == rank ) {
	  assert( 1 == flag );
	  assert( status.MPI_SOURCE == rank );
	  assert( status.MPI_TAG == MakeTag(rank,tag) );
	  MPI_Get_count( &status, MPI_INT, &count );
	  assert( 1 == count );
	}
	else
	  assert( 0 == flag );
      }
    }

    // Iprobe() one beyond those sent just to be sure we don't find something
    // we did not send.
    for ( tag = (tag_in_socket+1); tag <= TAG_MAX; tag++ ) {
      for ( rank = 1; rank < maxrank; rank++ ) {
	flag = 1;
	MPI_Iprobe( src, MakeTag(rank,tag), MPI_COMM_WORLD, &flag, &status );
	assert( 0 == flag );
      }
    }
  }
}

static void testCacheRank0( int maxrank ) {
  int		work, rank, tag, flag, count;
  MPI_Status	status;

  // Iprobe rank 0's cache to make sure it is empty, and it's cache tag
  // should be set to MPI_ANY_TAG.
  MPI_Iprobe( 0, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
  assert( 0 == flag );

  /*
   * First, Iprobe all the other ranks to verify their first message sent.
   */
  for ( rank = 1; rank < maxrank; rank++ ) {
    flag = 0;
    do {
      MPI_Iprobe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
    }
    while ( 0 == flag );

    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    // The cache is empty, with cache tag MPI_ANY_TAG
  }

  // At this point, the cache should not have anything in it, and the cache
  // tag should be MPI_ANY_TAG, from above. Repeated Iprobes on each of the
  // ranks should only return the first message sent.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int j;
    for ( j = 0; j < 1000; j++ ) {
      // Repeatedly, find the first message on the socket
      flag = 0;
      MPI_Iprobe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
      assert( 1 == flag );
      assert( status.MPI_SOURCE == rank );
      assert( status.MPI_TAG == MakeTag(rank,0) );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );
    }
  }

  // Now Iprobe for all the other possible tags. They should not have been
  // sent yet because we have not told the slaves to do so. These Iprobes
  // will have moved the first message from the socket for each source and
  // placed it into the respective caches.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int k;
    for ( k = 1; k < TAG_MAX; k++ ) {
      flag = 1;
      MPI_Iprobe( rank, MakeTag(rank,k), MPI_COMM_WORLD, &flag, &status );
      assert( 0 == flag );
    }

    // Note that the cache at this point has MakeTag(rank,0) in it, but
    // the cache tag will have been set to MakeTag(rank,(TAG_MAX-1))
  }

  // Now, a Probe on the first message should find it, in the cache. Verify
  // that no other messages exist.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,0), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    int k;
    for ( k = 1; k < TAG_MAX; k++ ) {
      flag = 1;
      MPI_Iprobe( rank, MakeTag(rank,k), MPI_COMM_WORLD, &flag, &status );
      assert( 0 == flag );
    }

    // Note that the cache at this point has MakeTag(rank,0) in it, but
    // the cache tag will have been set to MakeTag(rank,(TAG_MAX-1))
  }

  // Tell the slaves to send their next message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int one = 1;
    MPI_Send( &one, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }

  // Now, a Probe on the second message should leave the second message in
  // the socket, and should not have disturbed the first message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,1), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,1) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache cursor should be at the end of the cache, with tag
    // MakeTag(rank,1).

    // Verify with Iprobe(), that the first two messages are still
    // findable, but the third is not.
    flag = 0;
    MPI_Iprobe( rank, MakeTag(rank,0), MPI_COMM_WORLD, &flag, &status );
    assert( 1 == flag );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    // The cache cursor should be at the beginning of the cache.
    flag = 0;
    MPI_Iprobe( rank, MakeTag(rank,1), MPI_COMM_WORLD, &flag, &status );
    assert( 1 == flag );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,1) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache cursor should be at the end of the cache, having found
    // MakeTag(rank,1) in the socket still.

    int k;
    for ( k = 2; k < TAG_MAX; k++ ) {
      flag = 1;
      MPI_Iprobe( rank, MakeTag(rank,k), MPI_COMM_WORLD, &flag, &status );
      assert( 0 == flag );
    }

    // Make sure a Probe() will still find the first message.
    MPI_Probe( rank, MakeTag(rank,0), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // Make sure a Probe() for ANY tag will also find the first message.
    MPI_Probe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // Make sure a second Probe() for ANY tag still finds the first
    // message.
    MPI_Probe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // Make sure Iprobe() for ANY tag behaves the same way, too.
    flag = 0;
    MPI_Iprobe( rank, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status );
    assert( 1 == flag );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,0) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
  }

  verifyCache( 1, maxrank );

  // Tell the slaves to send their next message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int two = 2;
    MPI_Send( &two, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }

  // Now move the second message into the cache by Probe()ing the third.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,2), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,2) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache at this point should have tags 0 & 1 in it, with tag 2
    // sitting in the socket still, and the cache tag will have been set
    // to MakeTag(rank,2)
  }

  verifyCache( 2, maxrank );

  // Tell the slaves to send their next message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int three = 3;
    MPI_Send( &three, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }

  // Now move the third message into the cache by Probe()ing the fourth.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,3), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,3) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache at this point should have tags 0, 1 & 2 in it, with tag 3
    // sitting in the socket still, and the cache tag will have been set
    // to MakeTag(rank,3)
  }

  verifyCache( 3, maxrank );

  // Tell the slaves to send their next message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int four = 4;
    MPI_Send( &four, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }

  // Now move the fourth message into the cache by Probe()ing the fifth.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,4), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,4) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache at this point should have tags 0, 1, 2 & 3 in it, with
    // tag 4 sitting in the socket still, and the cache tag will have been
    // set to MakeTag(rank,4)
  }

  verifyCache( 4, maxrank );

  // Tell the slaves to send their last message.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int five = 5;
    MPI_Send( &five, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }

  // Now move the fifth message into the cache by Probe()ing the sixth.
  for ( rank = 1; rank < maxrank; rank++ ) {
    MPI_Probe( rank, MakeTag(rank,5), MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    assert( status.MPI_TAG == MakeTag(rank,5) );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    // The cache at this point should have tags 0, 1, 2, 3 & 4 in it, with
    // tag 5 sitting in the socket still, and the cache tag will have been
    // set to MakeTag(rank,5)
  }

  verifyCache( 5, maxrank );

  // Now Iprobe() the cache in reverse order of tags to make sure we can
  // still find everything in the proper cache, and not find tags
  // improperly.
  int src;
  for ( src = 1; src < maxrank; src++ ) {
    for ( tag = TAG_MAX-1; tag >= 0; tag-- ) {
      for ( rank = 1; rank < maxrank; rank++ ) {
	flag = 0;
	MPI_Iprobe(src, MakeTag(rank,tag), MPI_COMM_WORLD, &flag,
	    &status);
	if ( src == rank ) {
	  assert( 1 == flag );
	  assert( status.MPI_SOURCE == rank );
	  assert( status.MPI_TAG == MakeTag(rank,tag) );
	  MPI_Get_count( &status, MPI_INT, &count );
	  assert( 1 == count );
	}
	else
	  assert( 0 == flag );
      }
    }
  }

  // Probe() all tags in order as well.
  for ( tag = 0; tag < TAG_MAX; tag++ ) {
    for ( rank = 1; rank < maxrank; rank++ ) {
      MPI_Probe( rank, MakeTag(rank,tag), MPI_COMM_WORLD, &status );
      assert( status.MPI_SOURCE == rank );
      assert( status.MPI_TAG == MakeTag(rank,tag) );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );
    }
  }

  // Now verify that when we use MPI_ANY_SOURCE, we should be able to
  // Probe() for all the tags in the cache and the socket for a given rank.
  for ( tag = 0; tag < TAG_MAX; tag++ ) {
    for ( rank = 1; rank < maxrank; rank++ ) {
      MPI_Probe( MPI_ANY_SOURCE, MakeTag(rank,tag), MPI_COMM_WORLD, &status );
      assert( status.MPI_SOURCE == rank );
      assert( status.MPI_TAG == MakeTag(rank,tag) );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );
    }
  }

  // Now pull everything out of the cache in reverse order.
  for ( tag = TAG_MAX-1; tag >= 0; tag-- ) {
    for ( rank = 1; rank < maxrank; rank++ ) {
      int val;
      MPI_Recv( &val, 1, MPI_INT, rank, MakeTag(rank,tag), MPI_COMM_WORLD,
	  &status );
      assert( status.MPI_SOURCE == rank );
      assert( status.MPI_TAG == MakeTag(rank,tag) );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 1 == count );
      assert( val == tag );
    }
  }

  // We are done, tell the rank and file to exit.
  for ( rank = 1; rank < maxrank; rank++ ) {
    int six = 6;
    MPI_Send( &six, 1, MPI_INT, rank, TAG_WORK, MPI_COMM_WORLD );
  }
}

/*
 * Name:	hello_mpi6
 *
 * Purpose:	Step through the reception of 6 messages, performing various
 *		Probes and Iprobes on the individual sources to verify that
 *		each message is handled as expected.
 *
 * Slaves:	>= 1, as additional slaves will help verify source management.
 */
int main( int argc, char *argv[] ) {
  int myrank, maxrank;

  MPI_Init( &argc, &argv );	/* initialize MPI */
  // always use this process rank, 0 thru N-1
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  // always use this #processes in application = N
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  /*
   * Perform a number of tests of the cache and the socket buffer for a bunch
   * of tags sent from different ranks to rank 0.
   */
  if ( 0 == myrank )
    testCacheRank0( maxrank );
  else
    sendRankX( myrank );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
