#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#define WORKTAG1	0x00beef00
#define WORKTAG2	0x00feed00
#define WORKTAG3	0x00f00d00
#define BARRIERTAG	0x001eaf00

#define WORKVAL		0x00abad00

#define WORKLEN		500000

#define Poison(_r_,_i_)	(0x80000000 | ((_r_) << 16) | (_i_))

typedef struct envi_tag {
  int	myrank;
  int	maxrank;
  int	myid;
} envi_t, *envi_p;

/*
 * We do a barrier the hard way. Everyone sends a message of 400 KB (something
 * larger than the network buffer), and then they receive it. This requires
 * the use of MPI_Bsend (or possibly MPI_Isend) to guarantee no deadlock. But
 * if the macro TOPC is not defined, MPI_Send also works.
 */
static void barrier( int myrank, int maxrank, int myid ) {
  int		rank, i;
  MPI_Status	status;
  int		*work = (int *)calloc( WORKLEN, sizeof(int) );
  assert( work );
  for ( i = 0; i < WORKLEN; i++ ) {
    work[i] = Poison(myrank, myid);
  }

#if DEBUG == 1
  printf( "%s:(%#x,%#x)(%d,%#lx) maxrank(%#x)\n",
      __func__, myrank, myid, getpid(), pthread_self(), maxrank );
  fflush( stdout );
#endif

  for ( rank = 0; rank < maxrank; rank++ ) {
    if ( rank != myrank ) {
#if DEBUG == 1
      printf( "%s:(%#x,%#x)(%d,%#lx) MPI_Send(%#x)\n",
	  __func__, myrank, myid, getpid(), pthread_self(), rank );
      fflush( stdout );
#endif
      MPI_Send( work, WORKLEN, MPI_INT, rank, BARRIERTAG | myid,
	  MPI_COMM_WORLD );
    }
  }
  for ( rank = 0; rank < maxrank; rank++ )
    if ( rank != myrank ) {
#if DEBUG == 1
      printf( "%s:(%#x,%#x)(%d,%#lx) MPI_Recv(%#x)\n",
	  __func__, myrank, myid, getpid(), pthread_self(), rank );
      fflush( stdout );
#endif
      memset( work, 0, WORKLEN * sizeof(int) );
      MPI_Recv( work, WORKLEN, MPI_INT, rank, BARRIERTAG | myid, MPI_COMM_WORLD,
	  &status );
      assert( rank == status.MPI_SOURCE );
      if ( status.MPI_TAG != (BARRIERTAG | myid) ) {
	fprintf( stderr, "%s:(%#x,%#x)(%d,%#lx) MPI: Application bug:  "
	    "application message being sent across barrier from Rank %d "
	    "(%#lx, expected %#lx)\n",
	    __func__, myrank, myid, getpid(), pthread_self(), rank,
	    status.MPI_TAG, BARRIERTAG | myid);
      }
      assert( (BARRIERTAG | myid) == status.MPI_TAG );
      int count;
      MPI_Get_count( &status, MPI_INT, &count );
      assert( WORKLEN == count );
      for ( i = 0; i < WORKLEN; i++ ) {
	assert( Poison(rank, myid) == work[i] );
      }
    }
  free( work );
}

/*
 * Name:	th_func1
 *
 * Purpose:	Using multiple threads, perform a test much like hello_mpi2:
 *
 *		A) On Rank 0, verify that messages can be sent to self and
 *		   properly received by Iprobe, Probe and Recv.
 *		B) On Rank 1, verify that messages can be received out of
 *		   order.
 *		C) Verify an all-to-all barrier with sends followed by
 *		   receives works as expected.
 *		D) Verify that two large messages can be first sent from each
 *		   node to the other node, and then received by each node.
 *
 * Slaves:	1, as additional slaves will only participate in the barrier.
 */
static void *testit ( void *arg ) {
  pthread_t	self = pthread_self();
  envi_p	e = (envi_p)arg;
  int		myid = e->myid;
  int		myrank = e->myrank;
  int		maxrank = e->maxrank;
  int		work_send;
  int		work_recv;
  int		count;
  MPI_Status	status;

  /*
   * Setup various tags for this thread id.
   */
  int		worktag1 = WORKTAG1 | myid;
  int		worktag2 = WORKTAG2 | myid;
  int		worktag3 = WORKTAG3 | myid;

  // We don't need it anymore.
  free( e );

  /* Test ability to send to self. */

  printf( "%s: (%#x,%#x)(%d,%#lx) Sending messages to self.\n",
      __func__, myrank, myid, getpid(), self );
  fflush( stdout );

  work_send = WORKVAL | myid;
  MPI_Send( &work_send, 1, MPI_INT, myrank, worktag1, MPI_COMM_WORLD );
  MPI_Send( &work_send, 1, MPI_INT, myrank, worktag1, MPI_COMM_WORLD );

  int flag = 0;
  MPI_Iprobe( myrank, worktag1, MPI_COMM_WORLD, &flag, &status );
  assert( 1 == flag );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  MPI_Probe( myrank, worktag1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  work_recv = 0;
  MPI_Recv( &work_recv, 1, MPI_INT, myrank, worktag1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
  assert( work_recv == work_send );

  flag = 0;
  MPI_Iprobe( myrank, worktag1, MPI_COMM_WORLD, &flag, &status );
  assert( 1 == flag );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  MPI_Probe( myrank, worktag1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );

  work_recv = 0;
  MPI_Recv( &work_recv, 1, MPI_INT, myrank, worktag1, MPI_COMM_WORLD, &status );
  assert( myrank == status.MPI_SOURCE );
  assert( worktag1 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
  assert( work_recv == work_send );

  MPI_Iprobe( myrank, worktag1, MPI_COMM_WORLD, &flag, &status );
  assert( 0 == flag );

  /* Test out-of-order receives when tags arrive out of order. */
  if ( 0 == myrank ) { /* Send in reverse order */
    printf( "%s: (%#x,%#x)(%d,%#lx) Sending 3 worktags"
	" and receiving out of order.\n",
	__func__, myrank, myid, getpid(), self );
    fflush( stdout );

    work_send = 0x00100100 | myid;
    MPI_Send( &work_send, 1, MPI_INT, 1, worktag1, MPI_COMM_WORLD );
    work_send = 0x00100200 | myid;
    MPI_Send( &work_send, 1, MPI_INT, 1, worktag2, MPI_COMM_WORLD );
    work_send = 0x00100300 | myid;
    MPI_Send( &work_send, 1, MPI_INT, 1, worktag3, MPI_COMM_WORLD );
  }
  else if ( 1 == myrank ) { /* Can slave receive in correct order? */
    MPI_Recv( &work_send, 1, MPI_INT, 0, worktag3, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( worktag3 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 0x00100300 | myid == work_send );

    MPI_Recv( &work_send, 1, MPI_INT, 0, worktag1, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( worktag1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 0x00100100 | myid == work_send );

    MPI_Recv( &work_send, 1, MPI_INT, 0, worktag2, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    assert( worktag2 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );
    assert( 0x00100200 | myid == work_send );
  }

  /*
   * Test nonblocking sends.
   * All-to-all sends followed by all-to-all recv.
   */
  if ( 0 == myrank ) {
    printf( "%s: (%#x,%#x)(%d,%#lx) Testing nonblocking sends"
	" via all-to-all barrier.\n",
	__func__, myrank, myid, getpid(), self );
    fflush( stdout );
  }
  barrier( myrank, maxrank, myid );

  /*
   * Test nonblocking sends of very large messages.
   *   This forces a deadly embrace unless one of the two processes
   *   can send two large messages, and begin a receive, while the
   *   other process has not yet begun any receive.
   * However, it appears that some O/S (e.g. Linux 2.6.12 ??) have
   *   a trick by which even if the master blocks on MPI_Send, when the
   *   slave enters MPI_Send, this somehow blocks the master.
   */
  int	*workarray = (int *)calloc( WORKLEN, sizeof(int) );
  assert( workarray );
  int i;
  for ( i = 0; i < WORKLEN; i++ ) {
    workarray[i] = Poison(myrank, myid);
  }

  if ( 0 == myrank ) {
    printf( "%s: (%#x,%#x)(%d,%#lx) "
	"Sending two large messages before a receive at other end.\n",
	__func__, myrank, myid, getpid(), self );
    fflush( stdout );

    MPI_Send( workarray, WORKLEN, MPI_INT, 1, worktag1,	MPI_COMM_WORLD );
    MPI_Send( workarray, WORKLEN, MPI_INT, 1, worktag1,	MPI_COMM_WORLD );

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 1, worktag1,	MPI_COMM_WORLD,
	&status );
    assert( 1 == status.MPI_SOURCE );
    assert( worktag1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( Poison(1, myid) == workarray[i] );
    }

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 1, worktag1,	MPI_COMM_WORLD,
	&status );
    assert( 1 == status.MPI_SOURCE );
    assert( worktag1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( Poison(1, myid) == workarray[i] );
    }
  } else if ( 1 == myrank ) {
    MPI_Send( workarray, WORKLEN, MPI_INT, 0, worktag1,	MPI_COMM_WORLD );
    MPI_Send( workarray, WORKLEN, MPI_INT, 0, worktag1,	MPI_COMM_WORLD );

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 0, worktag1,	MPI_COMM_WORLD,
	&status );
    assert( 0 == status.MPI_SOURCE );
    assert( worktag1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( Poison(0, myid) == workarray[i] );
    }

    memset( workarray, 0, WORKLEN * sizeof(int) );
    MPI_Recv( workarray, WORKLEN, MPI_INT, 0, worktag1,	MPI_COMM_WORLD,
	&status );
    assert( 0 == status.MPI_SOURCE );
    assert( worktag1 == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( WORKLEN == count );
    for ( i = 0; i < WORKLEN; i++ ) {
      assert( Poison(0, myid) == workarray[i] );
    }
  }
  free( workarray );

#if DEBUG == 1
  printf( "%s: (%#x,%#x)(%d,%#lx) Finished.\n",
      __func__, myrank, myid, getpid(), self );
  fflush( stdout );
#endif
  return NULL;
}

int main( int argc, char *argv[] ) {
  int		myrank, maxrank, status, i, provided;
  pthread_t	self = pthread_self();
  int		numthreads = 10;
  pthread_t	threads[10];
  envi_p	e;

  /* initialize MPI */
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );
  /* always use this process rank, 0 thru N-1 */
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  /* always use this #processes in application = N */
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  for ( i = 1; i < numthreads; i++ ) {
    e = (envi_p)malloc( sizeof(envi_t) );
    e->myrank = myrank;
    e->maxrank = maxrank;
    e->myid = i;
    status = pthread_create( &threads[i], NULL, testit, (void *)e );
    if ( status != 0 ) {
      fprintf( stderr, "%s:(%d,%#lx) pthread_create(%#x) failed with (%d)%s\n",
	  __func__, getpid(), self, i, status, strerror( status ) );
      return status;
    }
  }

  e = (envi_p) malloc( sizeof(envi_t) );
  e->myrank = myrank;
  e->maxrank = maxrank;
  e->myid = i;
  testit( (void *)e );

  for ( i = 1; i < numthreads; i++ ) {
    status = pthread_join( threads[i], NULL );
    if ( status != 0 ) {
      fprintf( stderr,
	  "%s:(%d,%#lx) pthread_join([%#x]%#lx) failed with (%d)%s\n",
	  __func__, getpid(), self, i, threads[i], status, strerror( status ) );
      return status;
    }
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
