#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#define WORKTAG1	0x00beef00

#define WORKVAL		0x00abad00

typedef struct envi_tag {
  int	myrank;
  int	myid;
  int	numthreads;
} envi_t, *envi_p;

static pthread_mutex_t	remaining_sends_mu = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t remaining_sends_cv = PTHREAD_COND_INITIALIZER;
static int remaining_sends = 0;
static pthread_mutex_t	sends_processed_mu = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t sends_processed_cv = PTHREAD_COND_INITIALIZER;
static int sends_processed = 0;

/*
 * Name:	th_func3
 *
 * Purpose:	Verify two kinds of send/receive operations to "self" using a
 *		receive on ANY_SOURCE
 *
 *		1. Multiple Sends are issued before a Recv is issued
 *		2. A Recv is issued before multiple Sends
 *
 * Slaves:	None required, but using multiple slaves would allow for
 *		additional randomness in thread scheduling.
 */
static void *testit( void *arg ) {
  pthread_t	self = pthread_self();
  envi_p	e = (envi_p)arg;
  int		myid = e->myid;
  int		myrank = e->myrank;
  int		numthreads = e->numthreads;
  int		work_send = WORKVAL | myid;
  int		work_recv;
  int		work_tag = WORKTAG1 | myid;
  MPI_Status	status;

  // We don't need it anymore.
  free( e );

  /*
   * T1-(N-1): send to self
   * TN: recv from ANY_SOURCE (1-(N-1) times)
   * T1: recv from ANY_SOURCE (1)
   * T2: send to self (1)
   */

  if ( myid != numthreads ) {
    // Perform a send to the "self", myrank.
    MPI_Send( &work_send, 1, MPI_INT, myrank, work_tag, MPI_COMM_WORLD );

    // Tell receive thread (N) we have sent the data.
    assert( 0 == pthread_mutex_lock( &remaining_sends_mu ) );
    assert( remaining_sends > 0 );
    remaining_sends--;
    assert( 0 == pthread_cond_signal( &remaining_sends_cv ) );
    assert( 0 == pthread_mutex_unlock( &remaining_sends_mu ) );

    // wait for receive thread (N) to tell us we can send again.
    assert( 0 == pthread_mutex_lock( &sends_processed_mu ) );
    while ( sends_processed < (numthreads - 1) ) {
      assert( 0 == pthread_cond_wait( &sends_processed_cv, &sends_processed_mu ) );
    }
    assert( 0 == pthread_mutex_unlock( &sends_processed_mu ) );

    // Wait five seconds for the receive thread (N) to start waiting on an
    // MPI_ANY_SOURCE, and then blast it with another send to myrank, which
    // the receive thread (N) should pick up.
    sleep( 5 );

    // Perform another send to "self", myrank.
    MPI_Send( &work_send, 1, MPI_INT, myrank, work_tag, MPI_COMM_WORLD );
  } else {
    // Allocate an array to store a count of messages received from different
    // threads based on their assigned "id".
    int *tags = (int *)malloc( sizeof(int) * numthreads );
    assert( NULL != tags );
    memset( tags, 0, sizeof(int) * numthreads );

    printf( "Rank %d, waiting for other threads to send...\n", myrank );
    fflush( stdout );

    // wait for other threads to send to myrank
    assert( 0 == pthread_mutex_lock( &remaining_sends_mu ) );
    while ( remaining_sends > 0) {
      assert( 0 == pthread_cond_wait( &remaining_sends_cv, &remaining_sends_mu ) );
    }
    assert( 0 == pthread_mutex_unlock( &remaining_sends_mu ) );

    printf( "Rank %d, receiving from other threads...\n", myrank );
    fflush( stdout );

    // Now that all of the other threads have told us that they sent their
    // data, and it should all be in the cache, we should be able to receive
    // it all, and only find one message from each thread.
    int i;
    for ( i = 1; i < numthreads; i++ ) {
      int cnt, id;
      work_recv = 0;
      MPI_Recv( &work_recv, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG,
	  MPI_COMM_WORLD, &status );
      assert( myrank == status.MPI_SOURCE );
      MPI_Get_count( &status, MPI_INT, &cnt );
      assert( 1 == cnt );

      id = work_recv & ~WORKVAL;
      assert( ( id >= 1 ) && ( id < numthreads ) );
      tags[id]++;
    }

    // Success is defined to be the reception of only one message from each
    // thread. Note that the thread IDs run from 1 to numthreads-1, not from
    // zero.
    assert( 0 == tags[0] );
    for ( i = 1; i < numthreads; i++ ) {
      assert( 1 == tags[i] );
    }

    // Re-initialize the array.
    memset( tags, 0, sizeof(int) * numthreads );

    // Tell receive thread we have sent the data.
    assert( 0 == pthread_mutex_lock( &sends_processed_mu ) );
    assert( 0 == sends_processed );
    sends_processed = ( numthreads - 1 );
    assert( 0 == pthread_cond_broadcast( &sends_processed_cv ) );
    assert( 0 == pthread_mutex_unlock( &sends_processed_mu ) );

    printf( "Rank %d, receiving from other threads again...\n", myrank );
    fflush( stdout );

    // We should be able to receive a message from the other threads, even
    // though this thread started the receive first (the other threads sleep
    // for five seconds before sending).
    for ( i = 1; i < numthreads; i++ ) {
      int cnt, id;

      work_recv = 0;
      MPI_Recv( &work_recv, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG,
	  MPI_COMM_WORLD, &status );
      assert( myrank == status.MPI_SOURCE );
      assert( WORKTAG1 == ( status.MPI_TAG & ~0xff ) );
      MPI_Get_count( &status, MPI_INT, &cnt );
      assert( 1 == cnt );
      assert( WORKVAL == ( work_recv & ~0xff ) );
      id = status.MPI_TAG & ~WORKTAG1;
      assert( ( id >= 1 ) && ( id < numthreads ) );
      assert( id == ( work_recv & ~WORKVAL ) );

      tags[id]++;
    }

    // Success is defined to be the reception of only one message from each
    // thread. Note that the thread IDs run from 1 to numthreads-1, not from
    // zero.
    assert( 0 == tags[0] );
    for ( i = 1; i < numthreads; i++ ) {
      assert( 1 == tags[i] );
    }
  }
  return NULL;
}

int main( int argc, char *argv[] ) {
  int		myrank, status, i, provided;
  pthread_t	self = pthread_self();
  int	 	numthreads = 10;
  pthread_t	*threads;
  envi_p	e;

  /* initialize MPI */
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );
  /* always use this process rank, 0 thru N-1 */
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );

  threads = (pthread_t *)calloc( numthreads, sizeof(pthread_t) );
  assert( NULL != threads );

  // There are "numthreads" participating, including this thread, and all but
  // one will be sending data.
  remaining_sends = numthreads - 1;

  for ( i = 1; i < numthreads; i++ ) {
    e = (envi_p)malloc( sizeof(envi_t) );
    assert( NULL != e );
    e->myrank = myrank;
    e->myid = i;
    e->numthreads = numthreads;
    status = pthread_create( &threads[i], NULL, testit, (void *)e );
    if ( status != 0 ) {
      fprintf( stderr, "%s:(%d,%#lx) pthread_create(%#x) failed with (%d)%s\n",
	  __func__, getpid(), self, i, status, strerror (status) );
      return status;
    }
  }

  e = (envi_p) malloc( sizeof (envi_t) );
  assert( NULL != e );
  e->myrank = myrank;
  e->myid = i;
  e->numthreads = numthreads;
  testit( (void *)e );

  for ( i = 1; i < numthreads; i++ ) {
    status = pthread_join( threads[i], NULL );
    if ( status != 0 ) {
      fprintf( stderr, "%s:(%d,%#lx) pthread_join([%#x]%#lx) failed with (%d)%s\n",
	  __func__, getpid(), self, i, threads[i], status, strerror (status) );
      return status;
    }
  }

  free( threads );

#ifdef DEBUG
  fprintf( stderr, "Rank %d, calling MPI_Finalize()\n", myrank);
#endif
  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
