#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#define	MAX_BUF_SIZE	100
#define WORK_SIZE	(MAX_BUF_SIZE * sizeof(int))
#define MAX_THREADS	15

static void *master_thread( void *arg ) {
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD,	/* always use this */
      &myrank );			/* process rank, 0 thru N-1 */

  int tag = (int)arg;
  int *work = malloc( WORK_SIZE );
  assert( work );

#ifdef DEBUG
  int mypid = getpid();
  pthread_t self = pthread_self();
#endif

  MPI_Status status;
  int rsize;
  int i;
  for ( i = 0; i <= MAX_BUF_SIZE; i++ ) {
    memset( work, -1, WORK_SIZE );
    MPI_Recv( work, MAX_BUF_SIZE, MPI_INT, myrank, tag, MPI_COMM_WORLD,
	&status );

    assert( myrank == status.MPI_SOURCE );
    assert( tag == status.MPI_TAG );
    int cnt;
    MPI_Get_count( &status, MPI_INT, &cnt );
    assert( i == cnt );
    int pattern;
    memset( &pattern, i, sizeof(pattern) );
    int j;
    for ( j = 0; j < cnt; j++ ) {
#ifdef DEBUG
      if ( work[j] != pattern )
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 1, work[%d] = %#lx expected: %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j], pattern );
      else
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 1, work[%d] = %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j] );
      fflush( stdout );
#else
      assert( work[j] == pattern );
#endif
    }
    for ( j = cnt; j < MAX_BUF_SIZE; j++ ) {
#ifdef DEBUG
      if ( work[j] != -1 )
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 1, work[%d] = %#lx expected: %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j], pattern );
      else
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 1, work[%d] = %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j] );
      fflush( stdout );
#else
      assert( -1 == work[j] );
#endif
    }
  }
  for ( i = 0, rsize = MAX_BUF_SIZE; i <= MAX_BUF_SIZE; i++, rsize-- ) {
    memset( work, -1, WORK_SIZE );
    MPI_Recv( work, rsize, MPI_INT, myrank, tag, MPI_COMM_WORLD, &status );

    assert( myrank == status.MPI_SOURCE );
    assert( tag == status.MPI_TAG );
    int cnt;
    MPI_Get_count( &status, MPI_INT, &cnt );
    assert( i == cnt );
    int pattern;
    memset( &pattern, i, sizeof(pattern) );
    int j;
    int valid = cnt > rsize ? rsize : cnt;
    for ( j = 0; j < valid; j++ ) {
#ifdef DEBUG
      if ( work[j] != pattern )
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 2, work[%d] = %#lx expected: %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j], pattern );
      else
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 2, work[%d] = %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j] );
      fflush( stdout );
#else
      assert( work[j] == pattern );
#endif
    }
    for ( j = valid; j < MAX_BUF_SIZE; j++ ) {
#ifdef DEBUG
      if ( work[j] != -1 )
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 2, work[%d] = %#lx expected: %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j], pattern );
      else
	printf( "%s:(%#x,%#x)(%d,%#lx) pass 2, work[%d] = %#lx\n",
	    __func__, myrank, tag, mypid, self, j, work[j] );
      fflush( stdout );
#else
      assert( -1 == work[j] );
#endif
    }
  }
  return NULL;
}

static void *slave_thread( void *arg ) {
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD,	/* always use this */
      &myrank );			/* process rank, 0 thru N-1 */

  int tag = (int)arg;
  int *work = malloc( WORK_SIZE );
  assert( work );

  int i;
  for ( i = 0; i <= MAX_BUF_SIZE; i++ ) {
    memset( work, i, WORK_SIZE );
    MPI_Send( work, i, MPI_INT, myrank, tag, MPI_COMM_WORLD );
  }
  for ( i = 0; i <= MAX_BUF_SIZE; i++ ) {
    memset( work, i, WORK_SIZE );
    MPI_Send( work, i, MPI_INT, myrank, tag, MPI_COMM_WORLD );
  }
  free( work );
  return NULL;
}

/*
 * Name:	th_func5
 *
 * Purpose:	Verify that receive operations only read as much data as
 *		requested, and make sure short messages don't result in
 *		larger buffers being changed beyond the end of the provided
 *		data, and that provided buffers shorter than the received
 *		don't cause problems with buffer overflows or data reception.
 *
 *		This is almost the same as th_func4, but only deals with self
 *		operations.
 *
 * Slaves:	= 1, to verify send and receive
 */
int main( int argc, char *argv[] ) {
  int		i, provided;
  pthread_t	master_threads[MAX_THREADS];
  pthread_t	slave_threads[MAX_THREADS];

  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  for ( i = 0; i < MAX_THREADS; i++ ) {
    assert( 0 == pthread_create(
		&master_threads[i], NULL, master_thread, (void*)i ) );
    assert( 0 == pthread_create(
		&slave_threads[i], NULL, slave_thread, (void*)i ) );
  }

  for ( i = 0; i < MAX_THREADS; i++ ) {
    assert( 0 == pthread_join( master_threads[i], NULL ) );
    assert( 0 == pthread_join( slave_threads[i], NULL ) );
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
