#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#define CASES	6
#define THREADS	17

typedef struct op_tag {
  int	rank;
  int	id;
} op_t, *op_p;

static void doIprobe( int rank, int tag ) {
  MPI_Status status;
  int count, flag;

  for ( flag = 0; !flag; ) {
    MPI_Iprobe( rank, tag, MPI_COMM_WORLD, &flag, &status );
  }
  assert( 1 == flag );
  assert( status.MPI_SOURCE == rank );
  assert( status.MPI_TAG == tag );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
}

static void doProbe( int rank, int tag ) {
  MPI_Status status;
  int count;

  MPI_Probe( rank, tag, MPI_COMM_WORLD, &status );
  assert( status.MPI_SOURCE == rank );
  assert( status.MPI_TAG == tag );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 1 == count );
}

static void *testIprobeSlave( void *arg ) {
  op_p op = (op_p)arg;
  int rank = op->rank;
  int id = op->id;
  free( op );
  int thisCase;

  for ( thisCase = 1; thisCase <= CASES; thisCase++ ) {
    int tag = (thisCase * 10000) + (id * 100) + 1;
    MPI_Send( &tag, 1, MPI_INT, rank, tag, MPI_COMM_WORLD );
    tag++;
    MPI_Send( &tag, 1, MPI_INT, rank, tag, MPI_COMM_WORLD );
  }
  return NULL;
}

static void *testIprobeMaster( void *arg ) {
  op_p op = (op_p)arg;
  int rank = op->rank;
  int id = op->id;
  free( op );
  int thisCase;
  for ( thisCase = 1; thisCase <= CASES; thisCase++ ) {
#ifdef DEBUG
    printf( "Case %d\n", thisCase );
#endif
    int tag1 = (thisCase * 10000) + (id * 100) + 1;
    int tag2 = tag1 + 1;
    switch ( thisCase ) {
      case 1: {
	doProbe( rank, tag1 );
	doProbe( rank, tag2 );
	doIprobe( rank, tag2 );
	break;
      }
      case 2: {
	doIprobe( rank, tag1 );
	doIprobe( rank, tag2 );
	break;
      }
      case 3: {
	doIprobe( rank, tag1 );
	doProbe( rank, tag2 );
	break;
      }
      case 4: {
	doProbe( rank, tag2 );
	doIprobe( rank, tag2 );
	doProbe( rank, tag1 );
	break;
      }
      case 5: {
	doIprobe( rank, tag2 );
	doIprobe( rank, tag1 );
	break;
      }
      case 6: {
	doIprobe( rank, tag2 );
	doProbe( rank, tag1 );
	break;
      }
      default: {
	fprintf( stderr, "%s: invalid case encountered %d\n",
	    __func__, thisCase );
	abort();
      }
    }
  }
  return NULL;
}

/*
 * Name:	th_func6
 *
 * Purpose:	show that Iprobe is "best effort", meaning a busy loop using
 *		Iprobe() behaves the same as a Probe().
 *
 *		    Given: Send T1
 *			   Send T2
 *
 *		    The following sets should work:
 *
 *		    1. P  T1  2. Ip T1  3. Ip T1
 *		       P  T2     Ip T2     P  T2
 *		       Ip T2
 *		       -----     -----     -----
 *		    4. P  T2  5. Ip T2  6. Ip T2
 *		       Ip T2     Ip T1     P  T1
 *		       P  T1
 *
 *		Shows this for multiple threads, and for self operations.
 *
 * Slaves:	1
 */
int main( int argc, char *argv[] ) {
  int myrank, provided, i;

  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  // always use this process rank, 0 thru N-1
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );

  // Verify that Iprobe and Probe work the same on self.
  op_p op = (op_p)malloc( sizeof(op_t) );
  assert( op );
  op->rank = myrank;
  op->id = 0;
  pthread_t slave;
  assert( 0 == pthread_create( &slave, NULL, testIprobeSlave, (void*)op ) );
  op = (op_p)malloc( sizeof(op_t) );
  assert( op );
  op->rank = myrank;
  op->id = 0;
  testIprobeMaster( (void*)op );
  assert( 0 == pthread_join( slave, NULL ) );
  op = NULL;

  pthread_t threads[THREADS];
  if ( 1 == myrank ) {
    for ( i = 0; i < THREADS; i++ ) {
      op = (op_p)malloc( sizeof(op_t) );
      assert( op );
      op->rank = 0;
      op->id = i;
      assert( 0 == pthread_create(
		  &threads[i], NULL, testIprobeSlave, (void*)op ) );
    }
  }
  else {
    assert( 0 == myrank );
    for ( i = 0; i < THREADS; i++ ) {
      op = (op_p)malloc( sizeof(op_t) );
      assert( op );
      op->rank = 1;
      op->id = i;
      assert( 0 == pthread_create(
		  &threads[i], NULL, testIprobeMaster, (void*)op ) );
    }
  }

  for ( i = 0; i < THREADS; i++ ) {
    assert( 0 == pthread_join( threads[i], NULL ) );
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
