#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

static void *doRecv( void *arg ) {
  int rank = (int)arg;
  MPI_Status status;
  int work, cnt;

#ifdef DEBUG
  printf( "%s:(%#x)(%d,%#lx), begin.\n",
      __func__, rank, getpid(), pthread_self() );
  fflush( stdout );
#endif

  /*
   * Wait for a message.
   */
  MPI_Recv( &work, 1, MPI_INT, rank, 0, MPI_COMM_WORLD, &status );
  assert( rank == status.MPI_SOURCE );
  MPI_Get_count( &status, MPI_INT, &cnt );
  assert( 1 == cnt );
  assert( 0 == work );

#ifdef DEBUG
  printf( "%s:(%#x)(%d,%#lx), end.\n",
      __func__, rank, getpid(), pthread_self() );
  fflush( stdout );
#endif
  return NULL;
}

static void *recvr( void *arg ) {
  int work, cnt, maxrank = (int)arg;
  MPI_Status status;

#ifdef DEBUG
  printf( "%s:(%#x)(%d,%#lx), begin.\n",
      __func__, 0, getpid(), pthread_self() );
  fflush( stdout );
#endif

  /*
   * Wait for a message from each source.
   */
  char *checkbox = (char *)calloc( maxrank, sizeof(char) );
  assert( checkbox );
  int i;
  for ( i = 0; i < maxrank; i++ ) {
    MPI_Recv( &work, 1, MPI_INT, MPI_ANY_SOURCE, 1, MPI_COMM_WORLD, &status );
    assert( 0 == checkbox[status.MPI_SOURCE] );
    MPI_Get_count( &status, MPI_INT, &cnt );
    assert( 1 == cnt );
    assert( 1 == work );
    checkbox[status.MPI_SOURCE] = '1';
  }

#ifdef DEBUG
  printf( "%s:(%#x)(%d,%#lx), done.\n",
      __func__, 0, getpid(), pthread_self() );
  fflush( stdout );
#endif
}

static void master( int myrank ) {
  /*
   * Create a thread for all sources which will become the reader for those
   * sources.
   */
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );
  pthread_t *threads = (pthread_t*)calloc( maxrank+1, sizeof(pthread_t) );
  assert( threads );
  int i;
  for ( i = 0; i < maxrank; i++ ) {
    assert( 0 == pthread_create( &threads[i], NULL, doRecv, (void*)i ) );
    sched_yield();
  }

  // Wait for created threads to become readers.
  sleep( 1 );

  // Create a thread to perform the ANY_SOURCE MPI_Recv()
  pthread_t recvrth;
  assert( 0 == pthread_create( &recvrth, NULL, recvr, (void*)maxrank ) );

  // Wait for the recvr thread to establish itself.
  sleep( 1 );

  /*
   * Send to each source so that they will in turn send messages so that
   * the recvr() thread will complete
   */
  int work = 1;
  for ( i = 1; i < maxrank; i++ ) {
    MPI_Send( &work, 1, MPI_INT, i, 0, MPI_COMM_WORLD );
  }
  work = 1;
  MPI_Send( &work, 1, MPI_INT, 0, 1, MPI_COMM_WORLD );

  assert( 0 == pthread_join( recvrth, NULL ) );

  /*
   * Send to each source so that they will in turn send messages so that
   * doRecv() threads will complete
   */
  work = 0;
  for ( i = 0; i < maxrank; i++ ) {
    MPI_Send( &work, 1, MPI_INT, i, 0, MPI_COMM_WORLD );
  }

  for ( i = 0; i < maxrank; i++ ) {
    assert( 0 == pthread_join( threads[i], NULL ) );
  }

  free( threads );
}

/*
 * The purpose of the slave nodes are to feed the threads for each source in
 * the master. The slaves wait for a message from the master. The payload of
 * the message is the TAG value to use for the response.
 */
static void slave( int myrank ) {
  MPI_Status status;
  int work, cnt;
  int mypid = getpid();
  pthread_t self = pthread_self();

  do {
    /*
     * Wait for a message from the master telling us to quit.
     */
    MPI_Recv( &work, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status );
    assert( 0 == status.MPI_SOURCE );
    MPI_Get_count( &status, MPI_INT, &cnt );
    assert( 1 == cnt );

#ifdef DEBUG
    printf( "%s:(%#x,%#x)(%d,%#lx), processing tag: %#lx\n",
	__func__, myrank, work, mypid, self, work );
  fflush( stdout );
#endif

    /*
     * Send the payload back to the master using the payload value as the tag.
     */
    MPI_Send( &work, 1, MPI_INT, 0, work, MPI_COMM_WORLD );
  } while ( work != 0 );

#ifdef DEBUG
  printf( "%s:(%#x,%#x)(%d,%#lx), done.\n",
      __func__, myrank, work, mypid, self );
  fflush( stdout );
#endif
}

/*
 * Name:	th_func7
 *
 * Purpose:	Verify that when all sources already have reader threads
 *		(including self), that an ANY_SOURCE reader can still receive
 *		packets from all sources.
 *
 * Slaves:	> 1, so that there will be multiple sources to deal with at
 *		one time.
 */
int main( int argc, char *argv[] ) {
  int provided;
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  /*
   * Always use this process rank, 0 thru N-1
   */
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );

  if ( 0 == myrank ) {
    master( myrank );
  } else {
    slave( myrank );
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
