#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <pthread.h>
#if defined(__APPLE__)
# include <sys/sysctl.h>
#endif
#include "mpi.h"

static void *doRecv( void *arg ) {
  int rank = (int)arg;
  MPI_Status status;
  int work, cnt;

  /*
   * Wait for a message.
   */
  MPI_Recv( &work, 1, MPI_INT, rank, 1, MPI_COMM_WORLD, &status );
  assert( rank == status.MPI_SOURCE );
  MPI_Get_count( &status, MPI_INT, &cnt );
  assert( 1 == cnt );
  assert( 1 == work );

  return NULL;
}

static void *hiPrioRecv( void *arg ) {
  int work, cnt;
  MPI_Status status;
  MPI_Recv( &work, 1, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status );
  assert( 0 == status.MPI_SOURCE );
  MPI_Get_count( &status, MPI_INT, &cnt );
  assert( 1 == cnt );
  assert( 0 == work );

  return NULL;
}

static void master(void) {
  /*
   * Create a thread for all sources which will become the reader for those
   * sources.
   */
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );
  pthread_t *threads = (pthread_t*)calloc( maxrank+1, sizeof(pthread_t) );
  assert( threads );
  int i;
  for ( i = 0; i < maxrank; i++ ) {
    // Create all threads SCHED_FIFO, priority 1
    assert( 0 == pthread_create( &threads[i], NULL, doRecv, (void*)i ) );
    sched_yield();
  }

  // Wait for created threads to become readers.
  sleep( 1 );

  // Set the current thread to be SCHED_FIFO, priority 2, now that all the
  // readers are up and running.

  // Create a thread to perform the following MPI_Recv() at SCHED_FIFO,
  // priority 3. If MPI_Recv() busy waits for MPI_ANY_SOURCE, then this thread
  // won't get a chance to run.
  pthread_t recvr;
  assert( 0 == pthread_create( &recvr, NULL, hiPrioRecv, NULL ) );

  // Let the high priority receiver thread prove itself to be quiet.
  sleep ( 1 );

  // If the high priority receiver thread is not quiet, we'll be kept from
  // getting here, as this is a uni-processor and that thread is not yielding
  // the CPU.
  int work = 1;
  MPI_Send( &work, 1, MPI_INT, 0, 0, MPI_COMM_WORLD );

  /*
   * Send to each source so that they will in turn send messages so that
   * doRecv() threads will complete
   */
  work = 1;
  MPI_Send( &work, 1, MPI_INT, 0, 1, MPI_COMM_WORLD );
  work = 0;
  for ( i = 1; i < maxrank; i++ ) {
    MPI_Send( &work, 1, MPI_INT, i, 0, MPI_COMM_WORLD );
  }

  for ( i = 0; i < maxrank; i++ ) {
    assert( 0 == pthread_join( threads[i], NULL ) );
  }

  free( threads );
}

static void slave(void) {
  MPI_Status status;
  int work, cnt;

  /*
   * Wait for a message from the master telling us to quit.
   */
  MPI_Recv( &work, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status );
  assert( 0 == status.MPI_SOURCE );
  MPI_Get_count( &status, MPI_INT, &cnt );
  assert( 1 == cnt );
  assert( 0 == work );

  /*
   * Send tag 1 to the master so that doRecv() will complete.
   */
  work = 1;
  MPI_Send( &work, 1, MPI_INT, 0, 1, MPI_COMM_WORLD );
}

/*
 * Figure out the number of CPUs available on this platform.
 */
static long getNCPUs(void) {
  long ncpus;
#if defined(_SC_NPROCESSORS_ONLN)
  ncpus = sysconf( _SC_NPROCESSORS_ONLN );
  if ( ncpus < 0 ) {
    ncpus = 0;
  }
#elif defined(__APPLE__)
  int mib[2];
  mib[0] = CTL_HW;
  mib[1] = HW_NCPU;
  size_t len = sizeof(ncpus);
  int ret_val = sysctl( mib, 2, &ncpus, &len, NULL, 0 );
  if ( ret_val < 0 ) {
    ncpus = 0;
  }
#else
  ncpus = 0;
#endif
  return ncpus;
}

/*
 * Name:	th_func7
 *
 * Purpose:	Verify that when all sources already have reader threads
 *		(including self), that an ANY_SOURCE reader does not end up
 *		busy waiting.
 *
 * Slaves:	> 1, so that there will be multiple sources to deal with at
 *		one time.
 */
int main( int argc, char *argv[] ) {
  int provided;
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  /*
   * Always use this process rank, 0 thru N-1
   */
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  if (200112L == sysconf( _SC_THREAD_PRIORITY_SCHEDULING ) ) {
    long ncpus = getNCPUs();
    if ( 0 == ncpus ) {
      fprintf( stderr, "%s: can't determine CPU count: skipping\n",
	  basename( __FILE__ ) );
    } else if ( 1 == ncpus ) {
      if ( 0 == myrank ) {
	master();
      } else {
	slave();
      }
    } else {
      fprintf( stderr, "%s run on a multi-processor (%d): skipping\n",
	  basename( __FILE__ ), ncpus );
    }
  } else {
    fprintf( stderr, "%s run on a system that does not support POSIX Thread "
	"Execution Scheduling Option: skipping\n", basename( __FILE__ ) );
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
