#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#define LISTENERS		4
#define GATHERERS		4

#define INFO_RESP_TAG		0xd1
#define INFO_REQU_TAG		0xa1
#define INFO_KILL_TAG		0x42
#define SLAVE_DONE_TAG		0x90
#define SLAVE_FINISH_TAG	0x91
#define MakeTag(_t_,_i_)	(((_t_) << 16) | (_i_))

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif

typedef struct work_tag {
  int	dest;
  int	src;
  int	id;
} work_t, *work_p;

typedef struct resp_tag {
  int	dest;
  int	id;
  int	info;
} resp_t, *resp_p;

static int check( int *info, int maxrank ) {
  int i;
  for ( i = 0; i < maxrank; i++ ) {
    if ( -1 == info[i] )
      return FALSE;
  }
  return TRUE;
}

static void *gatherer( void *arg ) {
  unsigned int seed = 42 /* (int)pthread_self() */;

  int id = (int)arg;

  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  int *info = (int*)malloc( maxrank * sizeof(int) );
  assert( info );
  memset( info, -1, maxrank * sizeof(int) );

  /*
   * Randomly pick a source. If no information is available for it, randomly
   * ask another source for that information.
   */
  int done;
  for ( done = FALSE; !done; done = check( info, maxrank ) ) {
    int src = rand_r( &seed ) % maxrank;
    if ( -1 == info[src] ) {
      work_t work;
      work.dest = src;
      work.src = myrank;
      work.id = id;
      int rsrc = rand_r( &seed ) % maxrank;
      MPI_Send( &work, 3, MPI_INT, rsrc, INFO_REQU_TAG, MPI_COMM_WORLD );

      MPI_Status status;
      int count;
      resp_t resp;
      int tag = MakeTag( INFO_RESP_TAG, id );
      MPI_Recv( &resp, 3, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD,
	  &status );
      assert( tag == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 3 == count );
      assert( src == status.MPI_SOURCE );
      assert( src == resp.info );
      assert( id == resp.id );
      info[src] = resp.info;
    }
  }
  
  return NULL;
}

static void *listener( void *arg ) {
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  int done = FALSE;
  do {
    MPI_Status status;
    int count;
    // The work packet received contains three ints: the first describes the
    // source from which data is requested, the second int describes the
    // source requesting the data, the third is the ID of the thread
    // performing request.
    work_t work;

    // Receive an information request TAG from any source
    MPI_Recv( &work, 3, MPI_INT, MPI_ANY_SOURCE, INFO_REQU_TAG,
	MPI_COMM_WORLD, &status );
    assert( INFO_REQU_TAG == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 3 == count );

    if ( -1 == work.dest ) {
      done = TRUE;
    } else {
      /*
       * There are two cases: the work item will either be the ID of the
       * current node, myrank, or of another source to query.
       */
      if ( work.dest != myrank ) {
	assert( work.dest >= 0 && work.dest < maxrank );
	// Forward request on to the source specified by the work item,
	// sending a work item with -1 so that source will actually respond.
	MPI_Send( &work, 3, MPI_INT, work.dest, status.MPI_TAG,
	    MPI_COMM_WORLD );
      } else {
	// Respond with the information: this thread's rank.
	resp_t resp;
	resp.dest = work.src;
	resp.id = work.id;
	resp.info = myrank;
	int tag = MakeTag( INFO_RESP_TAG, resp.id );
	MPI_Send( &resp, 3, MPI_INT, resp.dest, tag, MPI_COMM_WORLD );
      }
    }
  } while ( !done );

  return NULL;
}

/*
 * Name:	th_func8
 *
 * Purpose:	A threaded Random-all-to-all test: each source has to goal of
 *		finding all of the "information" from all of its peers. To do
 *		this, each source has multiple threads that try to gather all
 *		the information about other sources, but asks a random source
 *		for information about another random source. Each source has a
 *		number of threads listening with ANY_SOURCE, which will then
 *		respond with that source's "information".
 *
 * Slaves:	> 1, so that there will be multiple sources to deal with at
 *		one time.
 */
int main( int argc, char *argv[] ) {
  int provided;
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  /*
   * Always use this process rank, 0 thru N-1
   */
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  pthread_t *listeners = (pthread_t *)calloc( LISTENERS, sizeof(pthread_t) );
  assert( listeners );
  pthread_t *gatherers = (pthread_t *)calloc( GATHERERS, sizeof(pthread_t) );
  assert( gatherers );

  /*
   * Create a bunch of listeners.
   */
  int i;
  for ( i = 0; i < LISTENERS; i++ ) {
    assert( 0 == pthread_create( &listeners[i], NULL, listener, NULL ) );
  }

  /*
   * Create the gatherers. When they have each filled up their information
   * array, they will exit.
   */
  for ( i = 0; i < GATHERERS; i++ ) {
    assert( 0 == pthread_create( &gatherers[i], NULL, gatherer, (void*)i ) );
  }

  /*
   * Now wait for all the gatherers to finish.
   */
  for ( i = 0; i < GATHERERS; i++ ) {
    assert( 0 == pthread_join( gatherers[i], NULL ) );
  }

  if ( 0 == myrank ) {
    /*
     * Now wait for all the slaves to finish
     */
    MPI_Status status;
    int count;
    for ( i = 1; i < maxrank; i++ ) {
      MPI_Recv( NULL, 0, MPI_INT, MPI_ANY_SOURCE, SLAVE_DONE_TAG,
	  MPI_COMM_WORLD, &status );
      assert( SLAVE_DONE_TAG == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 0 == count );
    }
    /*
     * Now that all the gatherers have finished, in this node and in all the
     * slaves, send out a set of kill messages, one for each listener, to all
     * source listeners.
     */
    work_t work;
    work.dest = -1;
    work.src = myrank;
    work.id = -1;
    for ( i = 0; i < maxrank; i++ ) {
      int j;
      for ( j = 0; j < LISTENERS; j++ ) {
	MPI_Send( &work, 3, MPI_INT, i, INFO_REQU_TAG, MPI_COMM_WORLD );
      }
    }

    /*
     * Now tell the slaves to proceed.
     */
    for ( i = 1; i < maxrank; i++ ) {
      MPI_Send( NULL, 0, MPI_INT, i, SLAVE_FINISH_TAG, MPI_COMM_WORLD );
    }
  } else {
    /*
     * Tell the master we are done, and then wait until the master tells us we
     * can proceed to exit.
     */
    MPI_Send( NULL, 0, MPI_INT, 0, SLAVE_DONE_TAG, MPI_COMM_WORLD );

    MPI_Status status;
    int count;
    MPI_Recv( NULL, 0, MPI_INT, 0, SLAVE_FINISH_TAG, MPI_COMM_WORLD, &status );
    assert( SLAVE_FINISH_TAG == status.MPI_TAG );
    assert( 0 == status.MPI_SOURCE );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 0 == count );
  }

  /*
   * Wait for all the listener's to run down, if they have not already.
   */
  for ( i = 0; i < LISTENERS; i++ ) {
    assert( 0 == pthread_join( listeners[i], NULL ) );
  }

  free( listeners );
  free( gatherers );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
