#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif

#define DIE_TAG	0x5000

#define THREADS	10

typedef void *(*prober_t)(void*);

static void doIprobe( int rank, int tag ) {
  MPI_Status status;
  int count, flag;

  for ( flag = 0; !flag; ) {
    MPI_Iprobe( rank, tag, MPI_COMM_WORLD, &flag, &status );
  }
  assert( 1 == flag );
  if ( MPI_ANY_SOURCE != rank ) {
    assert( status.MPI_SOURCE == rank );
  }
  if ( MPI_ANY_TAG != tag ) {
    assert( status.MPI_TAG == tag );
  }
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 0 == count );
}

static void doProbe( int rank, int tag ) {
  MPI_Status status;
  int count;

  MPI_Probe( rank, tag, MPI_COMM_WORLD, &status );
  if ( MPI_ANY_SOURCE != rank ) {
    assert( status.MPI_SOURCE == rank );
  }
  if ( MPI_ANY_TAG != tag ) {
    assert( status.MPI_TAG == tag );
  }
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 0 == count );
}

typedef struct op_tag {
  int	rank;
  int	tag;
} op_t, *op_p;

static void *prober( void *arg ) {
  op_p op = (op_p)arg;
  doProbe( op->rank, op->tag );
  free( op );
  return NULL;
}

static void *iprober( void *arg ) {
  op_p op = (op_p)arg;
  doIprobe( op->rank, op->tag );
  free( op );
  return NULL;
}

static void doTest( int rank, int prank, prober_t p1, prober_t p2 ) {
  printf( "%s(%d,%#x): %d, %d, %#lx, %#lx\n", __func__, getpid(),
      pthread_self(), rank, prank, p1, p2 );
  fflush( stdout );

  pthread_t threads[THREADS];

  int i;
  for ( i = 0; i < THREADS; i++ ) {
    op_p op = (op_p)malloc( sizeof(op_t) );
    assert( op );
    op->rank = prank;
    op->tag = 0;
    assert( 0 == pthread_create( &threads[i], NULL,
	    i & 1 == 0 ? p1 : p2, (void*)op ) );
    sched_yield();
  }
  MPI_Send( NULL, 0, MPI_INT, rank, 0, MPI_COMM_WORLD );
  for ( i = 0; i < THREADS; i++ ) {
    assert( 0 == pthread_join( threads[i], NULL ) );
  }
  MPI_Status status;
  int count;
  MPI_Recv( NULL, 0, MPI_INT, rank, 0, MPI_COMM_WORLD, &status );
  assert( status.MPI_SOURCE == rank );
  assert( 0 == status.MPI_TAG );
  MPI_Get_count( &status, MPI_INT, &count );
  assert( 0 == count );
}

static void echo( int rank ) {
  int done = FALSE;

  do {
    MPI_Status status;
    int count;
    MPI_Recv( NULL, 0, MPI_INT, rank, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    assert( status.MPI_SOURCE == rank );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 0 == count );

    done = ( DIE_TAG == status.MPI_TAG );
    if ( !done ) {
      MPI_Send( NULL, 0, MPI_INT, rank, status.MPI_TAG, MPI_COMM_WORLD );
    }
  } while ( !done );
}

/*
 * Name:	th_func9
 *
 * Purpose:	Demonstrate that when multiple threads are Probe-ing for the
 *		same tag they will all see the one message with that tag when
 *		it arrives. Perform this for all sources, both targetted
 *		sources and for ANY_SOURCE.
 *
 * Slaves:	> 1, so that there will be multiple sources to deal with at
 *		one time.
 */
int main( int argc, char *argv[] ) {
  int provided;
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  /*
   * Always use this process rank, 0 thru N-1
   */
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );


  /*
   * First, demonstrate that when a number of threads are Probe-ing on
   * MPINU_myrank for a given tag, they all see the one message that arrives.
   */


  /*
   * Try MPI_Probe() only.
   */
  doTest( myrank, myrank, prober, prober );

  /*
   * Try MPI_Iprobe() only.
   */
  doTest( myrank, myrank, iprober, iprober );

  /*
   * Try both MPI_Probe() and MPI_Iprobe().
   */
  doTest( myrank, myrank, prober, iprober );


  /*
   * Do it again for an ANY_SOURCE.
   */


  /*
   * Try MPI_Probe() only.
   */
  doTest( myrank, MPI_ANY_SOURCE, prober, prober );

  /*
   * Try MPI_Iprobe() only.
   */
  doTest( myrank, MPI_ANY_SOURCE, iprober, iprober );

  /*
   * Try both MPI_Probe() and MPI_Iprobe().
   */
  doTest( myrank, MPI_ANY_SOURCE, prober, iprober );


  if ( 0 == myrank ) {
    /*
     * Next, demonstrate that when a number of threads are Probe-ing on a source
     * other than MPINU_myrank for a given tag, they also all see the one
     * message that arrives.
     */


    /*
     * Try MPI_Probe() only.
     */
    doTest( 1, 1, prober, prober );

    /*
     * Try MPI_Iprobe() only.
     */
    doTest( 1, 1, iprober, iprober );

    /*
     * Try both MPI_Probe() and MPI_Iprobe().
     */
    doTest( 1, 1, prober, iprober );


    /*
     * Finally, demonstrate the previous scenario, but use ANY_SOURCE.
     */


    /*
     * Try MPI_Probe() only.
     */
    doTest( 1, MPI_ANY_SOURCE, prober, prober );

    /*
     * Try MPI_Iprobe() only.
     */
    doTest( 1, MPI_ANY_SOURCE, iprober, iprober );

    /*
     * Try both MPI_Probe() and MPI_Iprobe().
     */
    doTest( 1, MPI_ANY_SOURCE, prober, iprober );


    // Kill the slave.
    MPI_Send( NULL, 0, MPI_INT, 1, DIE_TAG, MPI_COMM_WORLD );
  } else {
    assert( 1 == myrank );
    echo( 0 );
  }

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
