#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "mpi.h"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif

#define ITERATIONS	1000
#define FREQ		500
#define MAX_CLIENTS	19
#define DIE		0x5000
#define PING_REQU_TAG	0x120000
#define PING_RESP_TAG	0x120001
#define CLIENT_REQU_TAG	0x130000
#define SERVER_RESP_TAG	0x130001
#define DONE_TAG	0x999999

static void *ping_server( void *arg ) {
  int done = FALSE;
  int ops = 0;
  int pid = getpid();
  pthread_t self = pthread_self();

  do {
    MPI_Status status;
    int count, work;
    MPI_Recv( &work, 1, MPI_INT, MPI_ANY_SOURCE, PING_REQU_TAG, MPI_COMM_WORLD,
	&status );
    assert( PING_REQU_TAG == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    done = ( DIE == work );
    if ( !done ) {
      ops++;
      if ( ( ops % FREQ ) == 0 ) {
	printf( "%s(%d,%#x): ops = %d\n", __func__, pid, self, ops );
      }
      MPI_Send( NULL, 0, MPI_INT, status.MPI_SOURCE, PING_RESP_TAG,
	  MPI_COMM_WORLD );
    }
  } while ( !done );

  return NULL;
}

static void *pinger( void *arg ) {
  int iterations = (int)arg;
  int pid = getpid();
  pthread_t self = pthread_self();
  int ops = 0;

  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  int i;
  for ( i = 0; i < iterations; i++ ) {
    int src;
    for ( src = 0; src < maxrank; src++ ) {
      int work = 0;
      MPI_Send( &work, 1, MPI_INT, src, PING_REQU_TAG, MPI_COMM_WORLD );

      MPI_Status status;
      int count;
      MPI_Recv( NULL, 0, MPI_INT, src, PING_RESP_TAG, MPI_COMM_WORLD, &status );
      assert( PING_RESP_TAG == status.MPI_TAG );
      assert( src == status.MPI_SOURCE );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 0 == count );

      ops++;
      if ( ( ops % FREQ ) == 0 ) {
	printf( "%s(%d,%#x): ops = %d\n", __func__, pid, self, ops );
      }
    }
  }

  return NULL;
}

static void *server( void *arg ) {
  int rank = (int)arg;
  int done = FALSE;
  int pid = getpid();
  pthread_t self = pthread_self();
  int ops = 0;

  do {
    MPI_Status status;
    int count, work;
    MPI_Recv( &work, 1, MPI_INT, MPI_ANY_SOURCE, CLIENT_REQU_TAG,
	MPI_COMM_WORLD, &status );
    assert( CLIENT_REQU_TAG == status.MPI_TAG );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 1 == count );

    done = ( DIE == work );
    if ( !done ) {
      ops++;
      if ( ( ops % FREQ ) == 0 ) {
	printf( "%s(%d,%#x): ops = %d\n", __func__, pid, self, ops );
      }
      MPI_Send( NULL, 0, MPI_INT, status.MPI_SOURCE, SERVER_RESP_TAG,
	  MPI_COMM_WORLD );
    }
  } while ( !done );

  return NULL;
}

static void *client( void *arg ) {
  int iterations = (int)arg;
  int pid = getpid();
  pthread_t self = pthread_self();
  int ops = 0;

  unsigned int seed = 42 /* pthread_self() */;

  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  int i;
  for ( i = 0; i < iterations; i++ ) {
    int src;
    do {
      src = rand_r( &seed ) % maxrank;
    } while ( src == myrank );

    int work = 0;
    MPI_Send( &work, 1, MPI_INT, src, CLIENT_REQU_TAG, MPI_COMM_WORLD );

    MPI_Status status;
    int count;
    MPI_Recv( NULL, 0, MPI_INT, src, SERVER_RESP_TAG, MPI_COMM_WORLD, &status );
    assert( SERVER_RESP_TAG == status.MPI_TAG );
    assert( src == status.MPI_SOURCE );
    MPI_Get_count( &status, MPI_INT, &count );
    assert( 0 == count );

    ops++;
    if ( ( ops % FREQ ) == 0 ) {
      printf( "%s(%d,%#x): ops = %d\n", __func__, pid, self, ops );
    }
  }

  return NULL;
}

/*
 * Name:	th_funcA
 *
 * Purpose:	Inter-mix ANY_SOURCE operations with targetted ones.
 *
 *		  1. each node has a ping thread and a ping server thread
 *		  2. each node has a server thread for every rank in system
 *		  3. each node has a random number of client threads
 *		  4. each node's ping thread sends a message to every other
 *		     node's pinger thread, which in turn responds that it is
 *		     alive
 *		  5. each client on a node asks a random node's server thread
 *		     to respond to a request with a reply
 *
 * Slaves:	> 1, so that there will be multiple sources to deal with at
 *		one time.
 */
int main( int argc, char *argv[] ) {
  int provided;
  MPI_Init_thread( &argc, &argv, MPI_THREAD_MULTIPLE, &provided );
  assert( MPI_THREAD_MULTIPLE == provided );

  /*
   * Always use this process rank, 0 thru N-1
   */
  int myrank;
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  int maxrank;
  MPI_Comm_size( MPI_COMM_WORLD, &maxrank );

  pthread_t ping_server_thread;
  assert( 0 == pthread_create( &ping_server_thread, NULL, ping_server, NULL ) );
  sched_yield();
  pthread_t pinger_thread;
  assert( 0 == pthread_create( &pinger_thread, NULL, pinger,
	  (void*)ITERATIONS ) );
  sched_yield();

  pthread_t *server_threads = (pthread_t*)malloc( maxrank * sizeof(pthread_t) );
  assert( server_threads );
  int src;
  for ( src = 0; src < maxrank; src++ ) {
    if ( src == myrank ) {
      memset( &server_threads[src], 0, sizeof(server_threads[src]) );
    } else {
      assert( 0 == pthread_create( &server_threads[src], NULL, server,
	      (void*)src ) );
      sched_yield();
    }
  }

  unsigned int seed = 42 /* (int)pthread_self() */;
  int client_cnt = rand_r( &seed ) % MAX_CLIENTS;

  pthread_t *client_threads = (pthread_t*)malloc(
      client_cnt * sizeof(pthread_t) );
  assert( client_threads );
  int i;
  for ( i = 0; i < client_cnt; i++ ) {
    assert( 0 == pthread_create( &client_threads[i], NULL, client,
	    (void*)ITERATIONS ) );
    sched_yield();
  }

  assert( 0 == pthread_join( pinger_thread, NULL ) );
  for ( i = 0; i < client_cnt; i++ ) {
    assert( 0 == pthread_join( client_threads[i], NULL ) );
  }
  free( client_threads );

  if ( 0 == myrank ) {
    for ( src = 1; src < maxrank; src++ ) {
      MPI_Status status;
      int count;
      MPI_Recv( NULL, 0, MPI_INT, MPI_ANY_SOURCE, DONE_TAG,
	  MPI_COMM_WORLD, &status );
      assert( DONE_TAG == status.MPI_TAG );
      MPI_Get_count( &status, MPI_INT, &count );
      assert( 0 == count );
    }
    for ( src = 0; src < maxrank; src++ ) {
      int work = DIE;
      for ( i = 1; i < maxrank; i++ ) {
	MPI_Send( &work, 1, MPI_INT, src, CLIENT_REQU_TAG, MPI_COMM_WORLD );
      }
      MPI_Send( &work, 1, MPI_INT, src, PING_REQU_TAG, MPI_COMM_WORLD );
    }
  } else {
    MPI_Send( NULL, 0, MPI_INT, 0, DONE_TAG, MPI_COMM_WORLD );
  }

  assert( 0 == pthread_join( ping_server_thread, NULL ) );

  for ( src = 0; src < maxrank; src++ ) {
    if ( src == myrank ) {
      continue;
    }
    assert( 0 == pthread_join( server_threads[src], NULL ) );
  }
  free( server_threads );

  MPI_Finalize();			/* cleanup MPI */
  return 0;
}
