 /**********************************************************************
  * MPINU				                               *
  * Copyright (c) 2004-2006 Gene Cooperman <gene@ccs.neu.edu>          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include "mpi.h"
#include "mpiimpl.h"
#include "cdll.h"

#define TAG_ENTRY_POISON	(tag_entry_p)0xbeedb1ed
#define TAG_AWL_POISON		(wait_list_p)0xbaadb1ad
#define TAG_AMU_POISON		(pthread_mutex_t *)0xb10db1ed
#define TAG_ASTATS_POISON	(pg_stats_p)0xb00db1ed

typedef struct wait_list_tag *wait_list_p;

typedef struct wait_tag {
  int			events;		// Number of "events" that have occurred
  int			waiters;	// Count of Probe threads waiting
  pthread_cond_t	cv;		// Associated mutex is the source's
					// receive mutex
} wait_t, *wait_p;

static void wait_init( wait_p w ) {
  w->events = 0;
  w->waiters = 0;
  assert( 0 == pthread_cond_init( &w->cv, NULL ) );
}

static void wait_destroy( wait_p w ) {
  w->events = -1;
  w->waiters = -1;
  assert( 0 == pthread_cond_destroy( &w->cv ) );
}

static int wait_signal( wait_p w ) {
  if ( w->waiters > w->events ) {
    // There are waiters for this tag, wake one.
    w->events++;
    assert( 0 == pthread_cond_signal( &w->cv ) );
    return TRUE;
  } else {
    return FALSE;
  }
}

static int wait_broadcast( wait_p w ) {
  if ( w->waiters > w->events ) {
    // There are waiters for this tag, wake them all.
    w->events = w->waiters;
    assert( 0 == pthread_cond_broadcast( &w->cv ) );
    return TRUE;
  } else {
    return FALSE;
  }
}

typedef struct tag_entry_tag {
  cdll_element_t	ce;		// CDLL element
  int			tag;
  wait_t		p;		// MPI_Probe waiters
  wait_t		r;		// MPI_Recv waiters
  int			awaiters;	// Count of ANY_SOURCE threads waiting
  wait_list_p		awl;		// ANY_SOURCE waitlist to use
  pthread_mutex_t	*amu;		// ANY_SOURCE waitlist mutex
  pg_stats_p		astats;		// ANY_SOURCE waitlist statistics
} tag_entry_t, *tag_entry_p;

typedef struct wait_list_tag {
  cdll_header_t		head;		// Linked list of sorted tag entries
  tag_entry_t		any;		// Keep a special entry for ANY_TAG.
} wait_list_t;

static void tag_entry_init( tag_entry_p tep, int tag ) {
  cdll_element_init( &tep->ce );
  tep->tag = tag;
  wait_init( &tep->p );
  wait_init( &tep->r );
  tep->awaiters = 0;
  tep->awl = NULL;
  tep->amu = NULL;
  tep->astats = NULL;
}

static void tag_entry_destroy( tag_entry_p tep ) {
  cdll_element_destroy( &tep->ce );
  tep->tag = -1;
  wait_destroy( &tep->p );
  wait_destroy( &tep->r );
  tep->awaiters = -1;
  tep->awl = TAG_AWL_POISON;
  tep->amu = TAG_AMU_POISON;
  tep->astats = TAG_ASTATS_POISON;
}

static tag_entry_p tag_entry_alloc( int tag ) {
  tag_entry_p tep = (tag_entry_p)malloc( sizeof(tag_entry_t) );
  tag_entry_init( tep, tag );
  return tep;
}

static void tag_entry_free( tag_entry_p tep ) {
  tag_entry_destroy( tep );
  free( tep );
}

static void free_tag_list( cdll_header_p header ) {
  /*
   * Run down the list of tags and free each one, and cut loose all the wait
   * entry structures. We don't have to worry about those because the caller
   * is assuring us it is okay to reinitialize.
   */
  while ( !cdll_empty( header ) ) {
    tag_entry_p curr = (tag_entry_p)cdll_dequeue_head( header );
    tag_entry_free( curr );
  }
}

/*
 * Initialize, or re-initialize, a wait list structure.
 */
void *MPINU_waitlist_init( void *wl ) {
  wait_list_p	wlp = (wait_list_p)wl;

  if ( NULL == wlp ) {
    wlp = malloc( sizeof(wait_list_t) );
    assert( wlp );
    memset( wlp, 0, sizeof(*wlp) );
    cdll_header_init( &wlp->head );
    tag_entry_init( &wlp->any, MPI_ANY_TAG );
  }

  /*
   * Free the tag list if anything is on it.
   */
  free_tag_list( &wlp->head );

  return wlp;
}

/*
 * Destroy a wait list structure, free the memory and returning a NULL
 * pointer.
 */
void *MPINU_waitlist_destroy( void *wl ) {
  wait_list_p	wlp = (wait_list_p)wl;

  if ( NULL != wlp ) {
    free_tag_list( &wlp->head );
    cdll_header_destroy( &wlp->head );
    tag_entry_destroy( &wlp->any );
    free( wlp );
    wlp = NULL;
  }

  return wlp;
}

/*
 * Find a matching tag entry, allocating one if a match is not found.
 */
static tag_entry_p find_alloc_match( wait_list_p wlp, int tag ) {
  tag_entry_p tep;
  if ( tag == MPI_ANY_TAG ) {
    /*
     * ANY_TAG waiters all used the same pre-allocated tag entry.
     */
    tep = &wlp->any;
  } else {
    /*
     * Look down the list of tag entries to find a matching tag. It is kept in
     * sorted in reverse order so that ANY_TAGs are at the end (we'll match
     * targeted waiters first).
     */
    tep = (tag_entry_p)cdll_head( &wlp->head );
    while ( ( tep != (tag_entry_p)&wlp->head ) && tep->tag < tag ) {
      tep = (tag_entry_p)cdll_next( &tep->ce );
    }
    if ( tep == (tag_entry_p)&wlp->head ) {
      tep = tag_entry_alloc( tag );
      cdll_insert_tail( &wlp->head, (cdll_element_p)tep );
    } else if ( tep->tag > tag ) {
      tag_entry_p ntep = tag_entry_alloc( tag );
      cdll_insert_before( &tep->ce, &ntep->ce );
      tep = ntep;
    }
  }
  return tep;
}

void MPINU_waitlist_block( void *wl, pthread_mutex_t *mu, int tag, int type,
    pg_stats_p stats ) {
  wait_list_p wlp = (wait_list_p)wl;

  tag_entry_p tep = find_alloc_match( wlp, tag );
  assert( tep->tag == tag );

  /*
   * Record another waiter on this tag.
   */
  wait_p w = ( MPINU_RECV == type ) ? &tep->r : & tep->p;
  w->waiters++;

  /*
   * Wait on the CV while there are no events.
   */
  while ( 0 == w->events ) {
    assert( 0 == pthread_cond_wait( &w->cv, mu ) );
  }

  /*
   * Remove this waiter from the tag, taking a posted event with it.
   */
  w->events--;
  assert( w->waiters > 0 );
  w->waiters--;

  /*
   * FIX-ME: When do we get rid of tag entries? For now, just leave them. A
   * program would have to go out of its way to use a different TAG for every
   * operation in order to leak memory.
   */

  /*
   * Fairness comes at the cost of performance. If the design here used a
   * mechanism which handed the thread woken up either the matching message or
   * the readership, it may be a very long time before the targeted thread
   * actually gets to run, while other threads might be in a better position
   * to take advantage of the state change.
   *
   * As such, upon waking up, this thread just enters the normal pool of
   * threads trying to find a match in the cache or becomes the reader for
   * this source.
   */
}

void MPINU_waitlist_block_init( void *wl, int tag, int type,
    pg_stats_p stats ) {
  wait_list_p wlp = (wait_list_p)wl;

  tag_entry_p tep = find_alloc_match( wlp, tag );
  assert( tep->tag == tag );

  /*
   * Record another waiter on this tag.
   */
  ( MPINU_RECV == type ) ? tep->r.waiters++ : tep->p.waiters++;

  /*
   * We have created a tag entry, or found an existing one, and marked it as
   * having another waiter. The caller will call block above to wait normally
   * for an event when it is ready.
   */
}

void MPINU_waitlist_block_destroy( void *wl, int tag, int type,
    pg_stats_p stats ) {
  wait_list_p wlp = (wait_list_p)wl;

  tag_entry_p tep = find_alloc_match( wlp, tag );
  assert( tep->tag == tag );

  /*
   * Unregister this waiter from the waitlist.
   */
  wait_p w = ( MPINU_RECV == type ) ? &tep->r : &tep->p;
  assert( w->waiters > 0 );
  w->waiters--;
}

void MPINU_waitlist_block_wait( void *wl, pthread_mutex_t *mu, int tag,
    int type, pg_stats_p stats ) {
  wait_list_p wlp = (wait_list_p)wl;

  tag_entry_p tep = find_alloc_match( wlp, tag );
  assert( tep->tag == tag );

  /*
   * The caller is assuring us that it already registered as a waiter by
   * calling MPINU_waitlist_block_register().
   */
  wait_p w = ( MPINU_RECV == type ) ? &tep->r : &tep->p;
  assert( w->waiters > 0 );

  /*
   * Wait on the CV while there are no events.
   */
  while ( 0 == w->events ) {
    assert( 0 == pthread_cond_wait( &w->cv, mu ) );
  }

  /*
   * Consume the posted event, but don't take a waiter yet, the caller will
   * have to do that.
   */
  w->events--;
  assert( w->waiters > 0 );
}

void MPINU_waitlist_register( void *wl, int tag, void *awl, pthread_mutex_t *amu,
    pg_stats_p stats, pg_stats_p astats ) {
  wait_list_p wlp = (wait_list_p)wl;
  wait_list_p awlp = (wait_list_p)awl;

  tag_entry_p tep = find_alloc_match( wlp, tag );
  assert( tep->tag == tag );

  /*
   * Record another ANY_SOURCE waiter on this tag.
   */
  assert( tep->awaiters >= 0 );
  if ( 0 == tep->awaiters ) {
    tep->awl = awlp;
    tep->amu = amu;
    tep->astats = astats;
  } else {
    assert( tep->awl == awlp );
    assert( tep->amu == amu );
    assert( tep->astats == astats );
  }
  tep->awaiters++;
}

void MPINU_waitlist_unregister( void *wl, int tag, void *awl, pthread_mutex_t *amu,
    pg_stats_p stats, pg_stats_p astats ) {
  wait_list_p wlp = (wait_list_p)wl;
  wait_list_p awlp = (wait_list_p)awl;
  tag_entry_p tep;
  if ( tag == MPI_ANY_TAG ) {
    /*
     * ANY_TAG waiters all used the same pre-allocated tag entry.
     */
    tep = &wlp->any;
  } else {
    /*
     * Look down the list of tag entries to find a matching tag. It is kept in
     * sorted in reverse order so that ANY_TAGs are at the end (we'll match
     * targeted waiters first).
     */
    tep = (tag_entry_p)cdll_head( &wlp->head );
    while ( tep != (tag_entry_p)&wlp->head && tep->tag < tag ) {
      tep = (tag_entry_p)cdll_next( &tep->ce );
    }
    if ( tep == (tag_entry_p)&wlp->head ) {
      /*
       * Could not find a tag entry to unregister. Who would have unregistered
       * the caller?
       */
      assert( tep != (tag_entry_p)&wlp->head );
    } else if ( tep->tag > tag ) {
      /*
       * Could not find a matching tag entry to unregister. Who would have
       * unregistered the caller?
       */
      assert( tep->tag == tag );
    }
  }
  assert( tep->tag == tag );

  /*
   * Decrement the count of ANY_SOURCE waiters on this tag.
   */
  assert( tep->awaiters > 0 );
  assert( tep->awl == awlp );
  assert( tep->amu == amu );
  assert( tep->astats == astats );
  tep->awaiters--;
  if ( 0 == tep->awaiters ) {
    /*
     * Last one out closes the lights.
     */
    tep->awl = NULL;
    tep->amu = NULL;
    tep->astats = NULL;
  }
}

#ifdef STANDALONE
int MPINU_tag_match( int pattern, int tag ) {
  int ret_val;

  assert( tag != MPI_ANY_TAG );
  if ( MPI_ANY_TAG == pattern ) {
    ret_val = ( 1 == 1 );
  } else {
    ret_val = ( pattern == tag );
  }
  return ret_val;
}
#endif

/*
 * FIX-ME: The wait list should be a doubly linked list so that we can find
 * them by tag and by first come first serve.
 */
void MPINU_waitlist_wakeup( void *wl, int tag, pg_stats_p stats ) {
  assert( tag != MPI_ANY_TAG );
  wait_list_p wlp = (wait_list_p)wl;

  /*
   * Look down the list of tag entries to find a matching tag. It is kept in
   * sorted order.
   *
   * NOTE: Don't match the tag using MPINU_tag_match(), as that is only
   * necessary when matching an MPI_ANY_TAG to a given set of tags. Here, the
   * given tag will be a real one, tag != MPI_ANY_TAG, and the list only
   * contains real tags. So a straight comparison works.
   */
  tag_entry_p tep = (tag_entry_p)cdll_head( &wlp->head );
  while ( tep != (tag_entry_p)&wlp->head && tep->tag != tag ) {
    assert( tep->tag != MPI_ANY_TAG );
    tep = (tag_entry_p)cdll_next( &tep->ce );
  }
  if ( tep != (tag_entry_p)&wlp->head
      && ( tep->r.waiters + tep->p.waiters ) > 0 ) {
    // There are waiters for this tag.

    // Wake up probe waiters first to give them all a chance to possibly see
    // the message before the receive waiters consume it.
    int woke = wait_broadcast( &tep->p );
    if ( wait_signal( &tep->r ) ) {
      woke = TRUE;
    }
    if ( woke ) {
      stats->match_wakeup++;
    } else {
      stats->waiters_gtr_events++;
    }
  } else {
    /*
     * NOTE: We might not match this ANY_TAG against the given tag as it could
     * be a tag used for internal communication.
     */
    if ( MPINU_tag_match( MPI_ANY_TAG, tag )
	&& ( wlp->any.r.waiters + wlp->any.p.waiters ) > 0 ) {
      /*
       * Wake any ANY_TAG waiters.
       */
      // Wake up probe waiters first to give them all a chance to possibly see
      // the message before the receive waiters consume it.
      int woke = wait_broadcast( &wlp->any.p );
      if ( wait_signal( &wlp->any.r ) ) {
	woke = TRUE;
      }
      if ( woke ) {
	stats->match_anytag_wakeup++;
      } else {
	stats->waiters_gtr_events_anytag++;
      }
    } else if ( tep != (tag_entry_p)&wlp->head && tep->awaiters > 0 ) {
      stats->match_anysrc_wakeup++;
      assert( 0 == pthread_mutex_lock( tep->amu ) );
      /*
       * Find some ANY_SOURCE waiter to wake-up. Note that we wake up any
       * targetted readers first, since an ANY_SOURCE reader might not end up
       * actually consuming the newly entered matching tag.
       */
      MPINU_waitlist_wakeup( tep->awl, tag, tep->astats );
      assert( 0 == pthread_mutex_unlock( tep->amu ) );
    } else if ( wlp->any.awaiters > 0 && MPINU_tag_match( MPI_ANY_TAG, tag ) ) {
      stats->match_anysrc_anytag_wakeup++;
      /*
       * Wake any ANY_TAG ANY_SOURCE waiters.
       *
       * NOTE: We might not match this ANY_TAG against the given tag as it could
       * be a tag used for internal communication.
       */
      assert( 0 == pthread_mutex_lock( wlp->any.amu ) );
      MPINU_waitlist_wakeup( wlp->any.awl, tag, wlp->any.astats );
      assert( 0 == pthread_mutex_unlock( wlp->any.amu ) );
    } else {
      stats->wakeup_noop++;
    }
  }
}

void MPINU_waitlist_wakeup_reader( void *wl, pg_stats_p stats ) {
  wait_list_p	wlp = (wait_list_p)wl;
  int		waiters;

  if ( !cdll_empty( &wlp->head ) ) {
    /*
     * First, wake a reader from the first tag with registered readers.
     *
     * We select a non-ANY_TAG reader to wake first because they might stay
     * and service the socket longer than an ANY_TAG reader.
     */
    tag_entry_p tep = (tag_entry_p)cdll_head( &wlp->head );
    int woke_one = FALSE;
    while ( tep != (tag_entry_p)&wlp->head && !woke_one ) {
      waiters = tep->p.waiters + tep->r.waiters;
      if ( waiters > 0 ) {
	woke_one = wait_signal( &tep->p );
	if ( !woke_one ) {
	  woke_one = wait_signal( &tep->r );
	}
	if ( woke_one ) {
	  stats->reader_wakeup++;
	} else {
	  stats->reader_waiters_gtr_events++;
	}
	// We did not have to perform a wakeup as threads have already been
	// woken up.
	return;
      } else {
	tep = (tag_entry_p)cdll_next( &tep->ce );
      }
    }
  }

  if ( ( wlp->any.p.waiters + wlp->any.r.waiters ) > 0 ) {
    /*
     * If we didn't find a thread to become the reader above, try waking an
     * ANY_TAG reader.
     */
    int woke_one = wait_signal( &wlp->any.p );
    if ( !woke_one ) {
      woke_one = wait_signal( &wlp->any.r );
    }
    if ( woke_one ) {
      stats->reader_anytag_wakeup++;
    } else {
      stats->reader_waiters_gtr_events_anytag++;
    }
    return;
  }

  if ( !cdll_empty( &wlp->head ) ) {
    /*
     * We could not find a targeted reader to wake up, so try to find some
     * ANY_SOURCE reader to wake. We look for a non-ANY_TAG waiter first to
     * raise the likely hood their stay around.
     */
    tag_entry_p tep = (tag_entry_p)cdll_head( &wlp->head );
    while ( tep != (tag_entry_p)&wlp->head ) {
      assert( 0 == ( tep->p.waiters + tep->r.waiters ) );
      if ( tep->awaiters > 0 ) {
	stats->reader_anysrc_wakeup++;
	assert( 0 == pthread_mutex_lock( tep->amu ) );
	MPINU_waitlist_wakeup_reader( tep->awl, tep->astats );
	assert( 0 == pthread_mutex_unlock( tep->amu ) );
	return;
      } else {
	tep = (tag_entry_p)cdll_next( &tep->ce );
      }
    }
  }

  if ( wlp->any.awaiters > 0 ) {
    stats->reader_anysrc_anytag_wakeup++;
    /*
     * Wake up an ANY_TAG / ANY_SOURCE reader to service the socket. Not
     * ideal, but better than nobody.
     */
    assert( 0 == pthread_mutex_lock( wlp->any.amu ) );
    MPINU_waitlist_wakeup_reader( wlp->any.awl, wlp->any.astats );
    assert( 0 == pthread_mutex_unlock( wlp->any.amu ) );
  } else {
    stats->reader_wakeup_noop++;
  }
}

#ifdef STANDALONE

# define OP_DISJOINT	1
# define OP_JOINT	2

typedef struct testWaitLists_tag {
  void			**list;
  pthread_mutex_t	*mu;
  pg_stats_p		stats;
  int			max;
} testWaitLists_t, *testWaitLists_p;

typedef struct op_tag {
  int			tag;
  int			kind;
  int			block;
  void			*wl;
  pthread_mutex_t	*mu;
  int			predicate;
  pg_stats_p		stats;
  testWaitLists_p	testp;
} op_t, *op_p;

static void *blocker( void *arg ) {
  op_p	op = (op_p)arg;
  assert( !op->testp );

  if ( op->kind == OP_DISJOINT ) {
    assert( 0 == pthread_mutex_lock( op->mu ) );
    while ( op->predicate != 0 ) {
      MPINU_waitlist_block_init( op->wl, op->tag, op->block, op->stats );
      assert( 0 == pthread_mutex_unlock( op->mu ) );

      sched_yield();

      assert( 0 == pthread_mutex_lock( op->mu ) );
      MPINU_waitlist_block_wait( op->wl, op->mu, op->tag, op->block, op->stats );
      assert( 0 == pthread_mutex_unlock( op->mu ) );

      sched_yield();

      assert( 0 == pthread_mutex_lock( op->mu ) );
      MPINU_waitlist_block_destroy( op->wl, op->tag, op->block, op->stats );
    }
    assert( 0 == pthread_mutex_unlock( op->mu ) );
  } else {
    assert( 0 == pthread_mutex_lock( op->mu ) );
    while ( op->predicate != 0 ) {
      MPINU_waitlist_block( op->wl, op->mu, op->tag, op->block, op->stats );
    }
    assert( 0 == pthread_mutex_unlock( op->mu ) );
  }

  return arg;
}

static void test01( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
  pthread_t thdBlocker;

  printf( "%s: Create one blocker on MPI_ANY_TAG\n", __func__ );
  op_p op = (op_p)calloc( 1, sizeof(op_t) );
  assert( op );
  op->tag = MPI_ANY_TAG;
  op->kind = kind;
  op->block = MPINU_RECV;
  op->wl = wl;
  op->mu = mu;
  op->predicate = 1;
  op->stats = stats;

  assert( 0 == pthread_create( &thdBlocker, NULL, blocker, (void *)op ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );
  op->predicate = 0;
  MPINU_waitlist_wakeup( wl, 0, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  assert( 0 == pthread_join( thdBlocker, NULL ) );
  free( op );
}

static void test01p( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test01p_RBLOCKERS	2
  // Add new blockers
  pthread_t rblockers[test01p_RBLOCKERS];
#define test01p_PBLOCKERS	4
  // Add new blockers
  pthread_t pblockers[test01p_PBLOCKERS];

  // Create a bunch of blockers
  op_p rop = (op_p)calloc( test01p_RBLOCKERS, sizeof(op_t) );
  assert( rop );
  int i;
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    rop[i].tag = MPI_ANY_TAG;
    rop[i].kind = kind;
    rop[i].block = MPINU_RECV;
    rop[i].wl = wl;
    rop[i].mu = mu;
    rop[i].predicate = 1;
    rop[i].stats = stats;
  }
  op_p pop = (op_p)calloc( test01p_PBLOCKERS, sizeof(op_t) );
  assert( pop );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    pop[i].tag = MPI_ANY_TAG;
    pop[i].kind = kind;
    pop[i].block = MPINU_PROBE;
    pop[i].wl = wl;
    pop[i].mu = mu;
    pop[i].predicate = 1;
    pop[i].stats = stats;
  }

  printf( "%s: Create %d receive blockers on MPI_ANY_TAG\n", __func__,
      test01p_RBLOCKERS );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == pthread_create( &rblockers[i], NULL, blocker, (void *)&rop[i] ) );
    sched_yield();
  }
  printf( "%s: Create %d probe blockers on MPI_ANY_TAG\n", __func__,
      test01p_PBLOCKERS );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    assert( 0 == pthread_create( &pblockers[i], NULL, blocker, (void *)&pop[i] ) );
    sched_yield();
  }

  assert( 0 == pthread_mutex_lock( mu ) );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    rop[i].predicate = 0;
  }
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    pop[i].predicate = 0;
  }
  MPINU_waitlist_wakeup( wl, 0, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  printf( "%s: Pull in probe blockers\n", __func__ );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    assert( 0 == pthread_join( pblockers[i], NULL ) );
  }

  assert( 0 == pthread_mutex_lock( mu ) );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == rop[i].predicate );
  }
  MPINU_waitlist_wakeup( wl, 0, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  printf( "%s: Pull in receive blockers\n", __func__ );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == pthread_join( rblockers[i], NULL ) );
  }

  free( rop );
  free( pop );
}

static void test02( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
  pthread_t thdBlocker;

  printf( "%s: Create one blocker on TAG 1\n", __func__ );
  op_p op = (op_p)calloc( 1, sizeof(op_t) );
  assert( op );
  op->tag = 1;
  op->kind = kind;
  op->block = MPINU_RECV;
  op->wl = wl;
  op->mu = mu;
  op->predicate = 1;
  op->stats = stats;

  assert( 0 == pthread_create( &thdBlocker, NULL, blocker, (void *)op ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );
  op->predicate = 0;
  MPINU_waitlist_wakeup( wl, 1, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  assert( 0 == pthread_join( thdBlocker, NULL ) );
  free( op );
}

static void test02p( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test01p_RBLOCKERS	2
  // Add new blockers
  pthread_t rblockers[test01p_RBLOCKERS];
#define test01p_PBLOCKERS	4
  // Add new blockers
  pthread_t pblockers[test01p_PBLOCKERS];

  // Create a bunch of blockers
  op_p rop = (op_p)calloc( test01p_RBLOCKERS, sizeof(op_t) );
  assert( rop );
  int i;
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    rop[i].tag = 1;
    rop[i].kind = kind;
    rop[i].block = MPINU_RECV;
    rop[i].wl = wl;
    rop[i].mu = mu;
    rop[i].predicate = 1;
    rop[i].stats = stats;
  }
  op_p pop = (op_p)calloc( test01p_PBLOCKERS, sizeof(op_t) );
  assert( pop );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    pop[i].tag = 1;
    pop[i].kind = kind;
    pop[i].block = MPINU_PROBE;
    pop[i].wl = wl;
    pop[i].mu = mu;
    pop[i].predicate = 1;
    pop[i].stats = stats;
  }

  printf( "%s: Create %d receive blockers on tag 1\n", __func__,
      test01p_RBLOCKERS );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == pthread_create( &rblockers[i], NULL, blocker, (void *)&rop[i] ) );
    sched_yield();
  }
  printf( "%s: Create %d probe blockers on tag 1\n", __func__,
      test01p_PBLOCKERS );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    assert( 0 == pthread_create( &pblockers[i], NULL, blocker, (void *)&pop[i] ) );
    sched_yield();
  }

  assert( 0 == pthread_mutex_lock( mu ) );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    rop[i].predicate = 0;
  }
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    pop[i].predicate = 0;
  }
  MPINU_waitlist_wakeup( wl, 1, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  printf( "%s: Pull in probe blockers\n", __func__ );
  for ( i = 0; i < test01p_PBLOCKERS; i++ ) {
    assert( 0 == pthread_join( pblockers[i], NULL ) );
  }

  assert( 0 == pthread_mutex_lock( mu ) );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == rop[i].predicate );
  }
  MPINU_waitlist_wakeup( wl, 1, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  printf( "%s: Pull in receive blockers\n", __func__ );
  for ( i = 0; i < test01p_RBLOCKERS; i++ ) {
    assert( 0 == pthread_join( rblockers[i], NULL ) );
  }

  free( rop );
  free( pop );
}

static void test03( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test03_BLOCKERS	4
  // Add new blockers
  pthread_t blockers[test03_BLOCKERS];

  // Create a bunch of blockers
  op_p op = (op_p)calloc( test03_BLOCKERS, sizeof(op_t) );
  assert( op );
  int i;
  for ( i = 0; i < test03_BLOCKERS; i++ ) {
    op[i].tag = i;
    op[i].kind = kind;
    op[i].block = MPINU_RECV;
    op[i].wl = wl;
    op[i].mu = mu;
    op[i].predicate = 1;
    op[i].stats = stats;
  }

  printf( "%s: Create one blocker on TAG 1\n", __func__ );
  assert( 0 == pthread_create( &blockers[1], NULL, blocker, (void *)&op[1] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 3\n", __func__ );
  assert( 0 == pthread_create( &blockers[3], NULL, blocker, (void *)&op[3] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 0\n", __func__ );
  assert( 0 == pthread_create( &blockers[0], NULL, blocker, (void *)&op[0] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 2\n", __func__ );
  assert( 0 == pthread_create( &blockers[2], NULL, blocker, (void *)&op[2] ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );

  printf( "%s: wakeup %d\n", __func__, 1 );
  op[1].predicate = 0;
  MPINU_waitlist_wakeup( wl, 1, stats );

  printf( "%s: wakeup %d\n", __func__, 2 );
  op[2].predicate = 0;
  MPINU_waitlist_wakeup( wl, 2, stats );

  printf( "%s: wakeup %d\n", __func__, 0 );
  op[0].predicate = 0;
  MPINU_waitlist_wakeup( wl, 0, stats );

  printf( "%s: wakeup %d\n", __func__, 3 );
  op[3].predicate = 0;
  MPINU_waitlist_wakeup( wl, 3, stats );

  assert( 0 == pthread_mutex_unlock( mu ) );

  for ( i = 0; i < test03_BLOCKERS; i++ ) {
    assert( 0 == pthread_join( blockers[i], NULL ) );
  }
  free( op );
}

static void test04( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test04_BLOCKERS	5
  // Add new blockers
  pthread_t blockers[test04_BLOCKERS];

  // Create a bunch of blockers
  op_p op = (op_p)calloc( test04_BLOCKERS, sizeof(op_t) );
  assert( op );
  int i;
  for ( i = 0; i < test04_BLOCKERS; i++ ) {
    op[i].tag = i;
    op[i].kind = kind;
    op[i].block = MPINU_RECV;
    op[i].wl = wl;
    op[i].mu = mu;
    op[i].predicate = 1;
    op[i].stats = stats;
  }

  printf( "%s: Create one blocker on TAG 1\n", __func__ );
  assert( 0 == pthread_create( &blockers[1], NULL, blocker, (void *)&op[1] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 3\n", __func__ );
  assert( 0 == pthread_create( &blockers[3], NULL, blocker, (void *)&op[3] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 0\n", __func__ );
  assert( 0 == pthread_create( &blockers[0], NULL, blocker, (void *)&op[0] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 2\n", __func__ );
  assert( 0 == pthread_create( &blockers[2], NULL, blocker, (void *)&op[2] ) );
  sched_yield();

  printf( "%s: Try a wakeup for a non-existant tag\n", __func__ );
  assert( 0 == pthread_mutex_lock( mu ) );
  MPINU_waitlist_wakeup( wl, 0xbadf00d, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  printf( "%s: Create one more blocker on TAG ANY\n", __func__ );
  op[4].tag = MPI_ANY_TAG;
  assert( 0 == pthread_create( &blockers[4], NULL, blocker, (void *)&op[4] ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );

  op[4].predicate = 0;
  MPINU_waitlist_wakeup( wl, 0xbadf100d, stats );

  op[1].predicate = 0;
  MPINU_waitlist_wakeup( wl, 1, stats );

  op[2].predicate = 0;
  MPINU_waitlist_wakeup( wl, 2, stats );

  op[0].predicate = 0;
  MPINU_waitlist_wakeup( wl, 0, stats );

  op[3].predicate = 0;
  MPINU_waitlist_wakeup( wl, 3, stats );

  assert( 0 == pthread_mutex_unlock( mu ) );

  for ( i = 0; i < test04_BLOCKERS; i++ ) {
    assert( 0 == pthread_join( blockers[i], NULL ) );
  }
  free( op );
}

static void test05( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test05_BLOCKERS	5
  // Add new blockers
  pthread_t blockers[test05_BLOCKERS];

  // Create a bunch of blockers
  op_p op = (op_p)calloc( test05_BLOCKERS, sizeof(op_t) );
  assert( op );
  int i;
  for ( i = 0; i < test05_BLOCKERS; i++ ) {
    op[i].tag = i;
    op[i].kind = kind;
    op[i].block = MPINU_RECV;
    op[i].wl = wl;
    op[i].mu = mu;
    op[i].predicate = 1;
    op[i].stats = stats;
  }

  printf( "%s: Create one blocker on TAG 1\n", __func__ );
  assert( 0 == pthread_create( &blockers[1], NULL, blocker, (void *)&op[1] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 3\n", __func__ );
  assert( 0 == pthread_create( &blockers[3], NULL, blocker, (void *)&op[3] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 0\n", __func__ );
  assert( 0 == pthread_create( &blockers[0], NULL, blocker, (void *)&op[0] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 2\n", __func__ );
  assert( 0 == pthread_create( &blockers[2], NULL, blocker, (void *)&op[2] ) );
  sched_yield();
  printf( "%s: Create one more blocker on TAG ANY\n", __func__ );
  op[4].tag = MPI_ANY_TAG;
  assert( 0 == pthread_create( &blockers[4], NULL, blocker, (void *)&op[4] ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );

  op[0].predicate = 0;
  MPINU_waitlist_wakeup_reader( wl, stats );

  assert( 0 == pthread_mutex_unlock( mu ) );

  assert( 0 == pthread_join( blockers[0], NULL ) );

  assert( 0 == pthread_mutex_lock( mu ) );
  op[4].predicate = 0;
  MPINU_waitlist_wakeup( wl, 0x400, stats );
  op[1].predicate = 0;
  MPINU_waitlist_wakeup( wl, 1, stats );
  op[2].predicate = 0;
  MPINU_waitlist_wakeup( wl, 2, stats );
  op[3].predicate = 0;
  MPINU_waitlist_wakeup( wl, 3, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  for ( i = 1; i < test05_BLOCKERS; i++ ) {
    assert( 0 == pthread_join( blockers[i], NULL ) );
  }
  free( op );
}

static void test06( int kind, void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
#define test06_BLOCKERS	5
  // Add new blockers
  pthread_t blockers[test06_BLOCKERS];

  // Create a bunch of blockers
  op_p op = (op_p)calloc( test06_BLOCKERS, sizeof(op_t) );
  assert( op );
  int i;
  for ( i = 0; i < test06_BLOCKERS; i++ ) {
    op[i].tag = i;
    op[i].kind = kind;
    op[i].block = MPINU_RECV;
    op[i].wl = wl;
    op[i].mu = mu;
    op[i].predicate = 1;
    op[i].stats = stats;
  }

  printf( "%s: Create one blocker on TAG 1\n", __func__ );
  assert( 0 == pthread_create( &blockers[1], NULL, blocker, (void *)&op[1] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 3\n", __func__ );
  assert( 0 == pthread_create( &blockers[3], NULL, blocker, (void *)&op[3] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 0\n", __func__ );
  assert( 0 == pthread_create( &blockers[0], NULL, blocker, (void *)&op[0] ) );
  sched_yield();
  printf( "%s: Create one blocker on TAG 2\n", __func__ );
  assert( 0 == pthread_create( &blockers[2], NULL, blocker, (void *)&op[2] ) );
  sched_yield();
  printf( "%s: Create one more blocker on TAG ANY\n", __func__ );
  op[4].tag = MPI_ANY_TAG;
  assert( 0 == pthread_create( &blockers[4], NULL, blocker, (void *)&op[4] ) );
  sched_yield();

  assert( 0 == pthread_mutex_lock( mu ) );

  op[0].predicate = 0;
  MPINU_waitlist_wakeup_reader( wl, stats );

  assert( 0 == pthread_mutex_unlock( mu ) );

  assert( 0 == pthread_join( blockers[0], NULL ) );

  assert( 0 == pthread_mutex_lock( mu ) );
  op[1].predicate = 0;
  MPINU_waitlist_wakeup( wl, 1, stats );
  op[2].predicate = 0;
  MPINU_waitlist_wakeup( wl, 2, stats );
  op[3].predicate = 0;
  MPINU_waitlist_wakeup( wl, 3, stats );

  // Re-use a tag that has no waiters on it to make sure only tags with
  // waiters are selected. this tag should end up waking up a the only thread
  // waiting for MPI_ANY_TAG.
  op[4].predicate = 0;
  MPINU_waitlist_wakeup( wl, 0, stats );
  assert( 0 == pthread_mutex_unlock( mu ) );

  for ( i = 1; i < test06_BLOCKERS; i++ ) {
    assert( 0 == pthread_join( blockers[i], NULL ) );
  }
  free( op );
}

static void *ablocker( void *arg ) {
  op_p op = (op_p)arg;
  assert( op->testp );
  testWaitLists_p waitlists = op->testp;

  assert( 0 == pthread_mutex_lock( op->mu ) );
  while ( op->predicate != 0 ) {
    MPINU_waitlist_block_init( op->wl, op->tag, op->block, op->stats );
    assert( 0 == pthread_mutex_unlock( op->mu ) );

    sched_yield();

    // register at all other waitlists
    int i;
    for ( i = 0; i < waitlists->max; i++ ) {
      assert( 0 == pthread_mutex_lock( &waitlists->mu[i] ) );
      MPINU_waitlist_register( waitlists->list[i], op->tag, op->wl, op->mu,
	  &waitlists->stats[i], op->stats );
      assert( 0 == pthread_mutex_unlock( &waitlists->mu[i] ) );
    }

    assert( 0 == pthread_mutex_lock( op->mu ) );
    MPINU_waitlist_block_wait( op->wl, op->mu, op->tag, op->block, op->stats );
    assert( 0 == pthread_mutex_unlock( op->mu ) );

    sched_yield();

    // unregister at all other waitlists
    for ( i = 0; i < waitlists->max; i++ ) {
      assert( 0 == pthread_mutex_lock( &waitlists->mu[i] ) );
      MPINU_waitlist_unregister( waitlists->list[i], op->tag, op->wl, op->mu,
	  &waitlists->stats[i], op->stats );
      assert( 0 == pthread_mutex_unlock( &waitlists->mu[i] ) );
    }

    assert( 0 == pthread_mutex_lock( op->mu ) );
    MPINU_waitlist_block_destroy( op->wl, op->tag, op->block, op->stats );

    // Decrement the predicate each time we wake up.
    op->predicate--;
  }
  assert( 0 == pthread_mutex_unlock( op->mu ) );

  return arg;
}

static void test07( void *wl, pthread_mutex_t *mu, pg_stats_p stats ) {
  /*
   * Create the testWaitLists structure that will be passed to each thread of
   * the test.
   */
  testWaitLists_p	testp = (testWaitLists_p)calloc(
      1, sizeof(testWaitLists_t) );
  assert( testp );
  /*
   * Create an array of waitlists, each with their associated mutexes and
   * statistics structures.
   */
# define TEST07_WAITLIST_CNT	5
  testp->max = TEST07_WAITLIST_CNT;
  testp->list = (void**)calloc( TEST07_WAITLIST_CNT, sizeof(void*) );
  testp->mu = (pthread_mutex_t*)calloc(
      TEST07_WAITLIST_CNT, sizeof(pthread_mutex_t) );
  testp->stats = (pg_stats_p)calloc( TEST07_WAITLIST_CNT, sizeof(pg_stats_t) );
  int i;
  for ( i = 0; i < TEST07_WAITLIST_CNT; i++ ) {
    testp->list[i] = MPINU_waitlist_init( NULL );
    assert( NULL != testp->list[i] );
    assert( 0 == pthread_mutex_init( &testp->mu[i], NULL ) );
  }
  /*
   * Now create an array of operation structures, three times as many as the
   * number of test waitlists, and fill it in using the data above. All zero
   * tags are changed to be MPI_ANY_TAGs.
   */
# define TEST07_ABLOCKERS_CNT	3*TEST07_WAITLIST_CNT
  op_p op = (op_p)calloc( TEST07_ABLOCKERS_CNT, sizeof(op_t) );
  assert( op );
  for ( i = 0; i < TEST07_ABLOCKERS_CNT; i++ ) {
    int tag = i%TEST07_WAITLIST_CNT;
    op[i].tag = tag == 0 ? MPI_ANY_TAG : tag;
    op[i].kind = OP_JOINT;
    op[i].block = MPINU_RECV;
    op[i].wl = wl;
    op[i].mu = mu;
# define TEST07_OPERATIONS	5
    op[i].predicate = TEST07_OPERATIONS;
    op[i].stats = stats;
    op[i].testp = testp;
  }
  /*
   * Create a bunch of threads performing ablocker() operations for the
   * different tags, including MPI_ANY_TAG.
   */
  pthread_t ablockers[TEST07_ABLOCKERS_CNT];
  for ( i = 0; i < TEST07_ABLOCKERS_CNT; i++ ) {
    printf( "%s: Create a blocker on TAG %d\n", __func__, op[i].tag );
    assert( 0 == pthread_create( &ablockers[i], NULL, ablocker,
	    (void *)&op[i] ) );
    sched_yield();
  }

  // Wait for everything to settle down
  sleep( 1 );

  // Wake-up an ablocker() thread from each given waitlist above (3 wakeups
  // per blocker thread, TEST07_OPERATIONS times.
  int j;
  for ( j = TEST07_OPERATIONS; j > 0; j-- ) {
    for ( i = 0; i < TEST07_ABLOCKERS_CNT; i++ ) {
      void *twl = testp->list[i%TEST07_WAITLIST_CNT];
      pthread_mutex_t *tmu = &testp->mu[i%TEST07_WAITLIST_CNT];
      pg_stats_p tstats = &testp->stats[i%TEST07_WAITLIST_CNT];
      assert( 0 == pthread_mutex_lock( tmu ) );
      int tag = i%TEST07_WAITLIST_CNT;
      MPINU_waitlist_wakeup( twl, tag, tstats );
      assert( 0 == pthread_mutex_unlock( tmu ) );
    }
    sleep( 1 );
  }

  // Now wait for all the blocker threads to complete as each has been given
  // 100 events to chew on before leaving.
  for ( i = 0; i < TEST07_ABLOCKERS_CNT; i++ ) {
    assert( 0 == pthread_join( ablockers[i], NULL ) );
    sched_yield();
  }

  free( op );
  for ( i = 0; i < TEST07_WAITLIST_CNT; i++ ) {
    testp->list[i] = MPINU_waitlist_destroy( testp->list[i] );
    assert( NULL == testp->list[i] );
    assert( 0 == pthread_mutex_destroy( &testp->mu[i] ) );
  }
  free( testp->list );
  free( testp->mu );
  free( testp->stats );
  free( testp );
}

int main( int argc, char *argv[] ) {
  assert( NULL == MPINU_waitlist_destroy( NULL ) );
  void *wl = MPINU_waitlist_init( NULL );
  assert( NULL != wl );
  assert( wl == MPINU_waitlist_init( wl ) );
  assert( NULL == MPINU_waitlist_destroy( wl ) );

  pg_stats_p		stats = (pg_stats_p)calloc( 1, sizeof(pg_stats_t) );
  pthread_mutex_t	*mu = (pthread_mutex_t *)calloc(
      1, sizeof(pthread_mutex_t) );
  assert( 0 == pthread_mutex_init( mu, NULL ) );

  wl = MPINU_waitlist_init( NULL );
  assert( NULL != wl );

  test01( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test01( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test01p( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test01p( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test02( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test02( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test02p( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test02p( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test03( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test03( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test04( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test04( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test05( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test05( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test06( OP_DISJOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test06( OP_JOINT, wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  test07( wl, mu, stats );

  wl = MPINU_waitlist_init( wl );
  assert( NULL != wl );
  assert( 0 == pthread_mutex_lock( mu ) );
  assert( 0 == pthread_mutex_unlock( mu ) );
  memset( stats, 0, sizeof(*stats) );

  assert( NULL == MPINU_waitlist_destroy( wl ) );
  assert( 0 == pthread_mutex_destroy( mu ) );
  free( stats );
  free( mu );

  return 0;
}
#endif
